# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from .utils.general_utils import link_obj_to_same_collections, set_parent, operator_is_redo, get_addon_preferences
from .utils.helper_functions import check_mesh_ok_for_grid_surface
from .hair_baking.hair_geometry_nodes import move_hsystem_items, HTOOL_OT_TransferHairSystem
from .hair_baking.hair_geometry_nodes_shared import get_hair_mod_by_idx, \
    add_hair_deformer, HD_Track_To_Mesh, HD_Align_Tilt


class HTOOL_OT_CurvesFromsurfaceNew(bpy.types.Operator):
    bl_label = "Curves from grid surface"
    bl_idname = "object.curves_from_grid_new"
    bl_description = "Generate hair curves from grid type mesh object \n" \
                     "For operator to work one border loop must be marked as sharp"
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        active_obj = context.active_object
        return active_obj and active_obj.type == 'MESH' and not active_obj.ht_props.hair_nodes.is_using_gnodes

    def invoke(self, context, event):
        active_obj = context.active_object
        if not active_obj.ht_props.hair_nodes.is_using_gnodes:
            bpy.ops.ed.undo_push()  # ?
            if context.active_object.type != 'MESH':
                self.report({'WARNING'}, 'Select Grid Mesh object first!')
                return {"CANCELLED"}

            use_crease, cancelled = check_mesh_ok_for_grid_surface(self, context, active_obj)
            if cancelled:
                return {"CANCELLED"}

            bpy.ops.object.add_hair_system(mode='GRID_GUIDE')
            add_hair_deformer(active_obj, new_node_type=HD_Align_Tilt)
            active_obj.ht_props.hair_nodes.hair_systems[0].deformer_index = 0  # set back to
            # get generate_strands node from hair system
            # if use_crease:
            #     hair_nodes = active_obj.ht_props.hair_nodes
            #     hair_mod = get_hair_mod_by_idx(hair_nodes.system_index, active_obj)
            #     hair_node_tree = hair_mod.node_group

            move_hsystem_items(self, context)  # init run for transfer_to_object  allow use of dynamic enum 'move_method'
            # bpy.ops.hair_system.transfer_to_object(move_method='EMPTYCURVE')
            HTOOL_OT_TransferHairSystem.transfer_hsys_to_target(context, active_obj, move_method='EMPTYCURVE')

            if get_addon_preferences().draw_grid_as_wire:
                active_obj.display_type = 'WIRE'
                active_obj.show_all_edges = True
            active_obj.visible_camera = False
            active_obj.visible_diffuse = False
            active_obj.visible_glossy = False
            active_obj.visible_transmission = False
            active_obj.visible_volume_scatter = False
            active_obj.visible_shadow = False
        bpy.ops.screen.draw_hsys_popup()
        return {"FINISHED"}

    def execute(self, context):
        bpy.ops.screen.draw_hsys_popup()
        # return context.window_manager.invoke_props_dialog(self, width=240)


class HTOOL_OT_CurvesWithGuideMesh(bpy.types.Operator):
    bl_label = "Hair with Track Mesh"
    bl_idname = "curves.hairsys_with_guide_mesh"
    bl_description = "It will generate Hair System, and automatically add 'Track To Mesh' Deformer\nGuide mesh is created by copying active mesh object, it will be drawn in wireframe mode"
    bl_options = {"REGISTER", "UNDO"}

    face_offset: bpy.props.FloatProperty(name="Initial Offset", default=0.1, min=0.01, soft_max=1.0, description="Initial offset from source surface")

    @classmethod
    def poll(cls, context):  # XXX: may break redo.
        active_obj = context.active_object
        return active_obj and active_obj.type == 'MESH'  # and not active_obj.ht_props.hair_nodes.is_using_gnodes # breaks redoo

    def check(self, context):
        return True

    def draw(self, context):  # XXX: crash if trying to draw DeformersPanel ....
        self.layout.prop(self, 'face_offset')

    def invoke(self, context, event):
        active_obj = context.active_object
        if active_obj and active_obj.ht_props.hair_nodes.is_using_gnodes:
            self.report({'ERROR'}, 'Object is already using Hair system. Cancelling')
            return {'FINISHED'}

        guide_obj = active_obj.copy()
        guide_obj.data = active_obj.data.copy()
        guide_obj.name = active_obj.name + '_guide'
        guide_obj.data.name = active_obj.data.name + '_guide'
        set_parent(guide_obj, active_obj)
        # add track to mesh deform modifier
        link_obj_to_same_collections(active_obj, guide_obj)

        guide_obj.display_type = 'WIRE'  # or as 'BOUNDS' ?
        guide_obj.show_all_edges = True

        guide_obj.visible_camera = False
        guide_obj.visible_diffuse = False
        guide_obj.visible_glossy = False
        guide_obj.visible_transmission = False
        guide_obj.visible_volume_scatter = False
        guide_obj.visible_shadow = False

        with bpy.context.temp_override(object=active_obj, active_object=active_obj, selected_objects=[active_obj], selected_editable_objects=[active_obj], editable_objects=[active_obj]):
            bpy.ops.object.add_hair_system(mode='GENERATOR')

        track_to_mesh_node = add_hair_deformer(active_obj, new_node_type=HD_Track_To_Mesh)
        track_to_mesh_node.inputs['Target'].default_value = guide_obj
        bpy.ops.hair_system.bind_track_object(bind=True)

        align_tilt_node = add_hair_deformer(active_obj, new_node_type=HD_Align_Tilt)
        align_tilt_node.mute = 1  # set to 'Align'
        active_obj.ht_props.hair_nodes.hair_systems[0].deformer_index = 0  # set back to

        guide_obj.shape_key_add(name='Base')
        deform_key = guide_obj.shape_key_add(name='Deformer')
        guide_obj.active_shape_key_index = 1
        deform_key.value = 1
        self.guide_obj_name = guide_obj.name
        guide_obj.select_set(False)
        # lastly check if guide obj has shrinkwrap (received as copy from active_obj). if so remove
        for mod in guide_obj.modifiers:
            if mod.type == 'SHRINKWRAP':
                guide_obj.modifiers.remove(mod)
        return self.execute(context)

    def execute(self, context):
        guide_obj = bpy.data.objects.get(self.guide_obj_name)
        if guide_obj:
            obj_dim = guide_obj.dimensions.length
            # offset faces of deform_key by self.face_offset
            deform_key = guide_obj.data.shape_keys.key_blocks.get('Deformer')
            offset_fac = self.face_offset * obj_dim
            for vert in guide_obj.data.vertices:
                deform_key.data[vert.index].co = vert.co + vert.normal * offset_fac

        return context.window_manager.invoke_props_dialog(self, width=220)
        return {"FINISHED"}
