# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from bpy.props import FloatProperty, BoolProperty, IntProperty
import math
import gpu
from gpu_extras.batch import batch_for_shader
from gpu_extras.presets import draw_circle_2d
from mathutils.kdtree import KDTree
import numpy as np
from bpy_extras import view3d_utils
from bl_math import lerp
import copy
from mathutils import Vector
from mathutils.bvhtree import BVHTree

from .utils.helper_functions import calc_exponent, angle_signed, get_obj_mesh_bvht, get_fallof_fun, gold_random, clamp, smooth_threshold
from .utils.hair_common import HTool_ClassCommon
from .utils.curve_wrapper import Splines, ChainIK, get_len_masked
from .resample2d import parallel_transport_TNB
from .hair_baking.hair_geometry_nodes_utils import align_curves_tilt
from .hair_baking.hair_geometry_nodes_selections import get_selected_strands_mask, get_blender_curve_selection_attrib, get_edit_selected_points_bool


class HTOOL_OT_SelectRootTip(bpy.types.Operator):
    bl_label = "Select tips/roots"
    bl_idname = "curves.select_tips"
    bl_description = "Select tips/roots"
    bl_options = {"REGISTER", "UNDO"}

    limit_sel: bpy.props.EnumProperty(name='Limit Selection', description='Limit selection by some conditions. Helps to randomize selection',
        items=[
            ('DISABLED', 'Disabled', 'Disabled'),
            ('RANDOM', 'Random', '(De)select hair endings randomly'),
            ('LENGTH', 'By length', '(De)select hair endings by length threshold limit'),
        ], default='DISABLED')

    roots: bpy.props.BoolProperty(name="Roots", description="Roots", default=False)
    extend: bpy.props.BoolProperty(name="Extend", description="Extend selection", default=False)
    seed: bpy.props.IntProperty(name="Seed", default=2, min=1, max=500)
    percentage: bpy.props.IntProperty(name="Percentage", description="Selection percent", default=50, min=0,
                                   max=100, subtype='PERCENTAGE')
    len_threshold: bpy.props.FloatProperty(name='Length threshold', description='Select only strands longer than give threshold value', default= 0.5, min=0, max=1)
    margin: bpy.props.FloatProperty(name='Tolerance', description='Smooth out selection transition by increasing length threshold tolerance', default=0.0, min=0, max=1.0)
    invert: bpy.props.BoolProperty(name='Invert', description='', default=False)

    def check(self, context): #DONE: can prop panel be fixed/refreshed when using f6 prop popup
        return True

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'extend', toggle=True)
        layout.label(text='Limit Selection')
        row = layout.row(align=True)
        row.prop(self, 'limit_sel', expand=True)
        col = layout.column(align=True)
        if self.limit_sel == 'RANDOM':
            col.prop(self, 'percentage')
            col.prop(self, 'seed')
        elif self.limit_sel == 'LENGTH':
            row = col.row(align=True)
            row.prop(self, 'len_threshold')
            row.prop(self, 'invert', icon='CLIPUV_DEHLT', icon_only=True)
            col.prop(self, 'margin')
            if self.margin > 0:
                col.prop(self, 'seed')

    def invoke(self, context, event):
        curve_obj = context.active_object
        curve_obj.data.selection_domain == 'POINT'
        self.called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        if self.called_from_ht_wspace:
            self.tip_ids = []
            offset = context.scene.ht_props.select_offset
            active_end = context.scene.ht_props.active_end
            for spl in curve_obj.data.curves:
                pts = spl.points
                last_pid = len(pts)-1
                p_id = clamp(last_pid - offset, 0, last_pid) if active_end == "TIPS" else clamp(offset, 0, last_pid)
                self.tip_ids.append(p_id)

        self.seed = curve_obj.ht_props.uv_seed
        sel_splines = Splines(curve_obj, onlySelection=False)
        euclid_len = np.array([s.euclidean_len for s in sel_splines.splines])
        self.len_norms = euclid_len/np.max(euclid_len)

        return self.execute(context)

    def execute(self, context):
        curve_obj = context.active_object
        if curve_obj.type != 'CURVES':
            self.report({'INFO'}, 'Works only on curves type of object')
            return {"CANCELLED"}

        index = 0 if self.roots else -1
        spl_cnt = len(curve_obj.data.curves)

        sel_attrib = get_blender_curve_selection_attrib(curve_obj, force_create=True)
        if not self.extend: # DESELECT all
            for sel_data  in sel_attrib.data:
                sel_data.value = False

        if self.limit_sel == 'RANDOM':
            sel_ids = []
            np.random.seed(self.seed)
            gold_seed = np.random.rand()  # (0,1) range
            pick_cnt = int(spl_cnt*self.percentage/100)  # pick %
            remaining_splines = [i for i in range(spl_cnt)]
            for i in range(pick_cnt):  # pick half of elements, for blending sel to non_sel strands...
                rd = gold_random(i, gold_seed) * spl_cnt  # low discrepancy sampling
                idx = int(rd//1)
                sel_ids.append(remaining_splines[idx])
                del(remaining_splines[idx])
                spl_cnt -= 1

        elif self.limit_sel == 'LENGTH':
            sel_ids = get_len_masked(self.len_norms, self.len_threshold, self.margin, seed=self.seed)
            if self.invert:
                all_ids = [i for i in range(len(curve_obj.data.curves))]
                sel_ids = list(set(all_ids)-set(sel_ids)) #invert by sub
        else: #disabled
            sel_ids = [i for i in range(len(curve_obj.data.curves))]
        if not hasattr(self, 'called_from_ht_wspace'):
            self.called_from_ht_wspace = False

        for i in sel_ids:
            if self.called_from_ht_wspace:  # use tip pt idx
                index = self.tip_ids[i] # for spl[i] , we have one tip pt
            curve = curve_obj.data.curves[i]
            pt = curve.points[index]
            sel_attrib.data[pt.index].value = True
        if self.called_from_ht_wspace:
            bpy.context.area.tag_redraw()

        curve_obj.data.update_tag()
        context.area.tag_redraw()

        return {"FINISHED"}


class HTOOL_OT_SelectNextPrev(bpy.types.Operator):
    bl_label = "Select netx/previous"
    bl_idname = "curves.ht_select_next"
    bl_description = "Select next/previous point"
    bl_options = {"REGISTER", "UNDO"}

    previous: BoolProperty(name="Previous", description="Select Next/Previous point", default=False)
    extend: BoolProperty(name="Extend", description="Extend selection", default=False)

    def execute(self, context):
        curve_obj = context.active_object
        if curve_obj.type != 'CURVES':
            self.report({'INFO'}, 'Only hair curves are supported. Cancelling!')
            return {"CANCELLED"}

        sel_attrib = get_blender_curve_selection_attrib(curve_obj, force_create=True)
        curveData = curve_obj.data
        for polyline in curveData.curves:  # for strand point
            length = len(polyline.points)
            for i,point in enumerate(polyline.points):
                # if not point.hide and point.select:
                if  sel_attrib.data[point.index].value:
                    next_p = max(min(i+1,length-1),0)
                    prev_p = max(min(i-1,length-1),0)
                    index = prev_p if self.previous else next_p
                    if not self.extend:
                        # point.select = False
                        sel_attrib.data[point.index].value = False
                    # polyline.points[index].select = True
                    pt = polyline.points[index]
                    sel_attrib.data[pt.index].value = True
                    break
        curve_obj.data.update_tag()
        context.area.tag_redraw()
        return {"FINISHED"}


# for custom ui curve mapping panel - wont work in f6 props
def myNodeTree():
    if 'HairToolCurveTmp' not in bpy.data.node_groups:
        ng = bpy.data.node_groups.new('HairToolCurveTmp', 'ShaderNodeTree')
    return bpy.data.node_groups['HairToolCurveTmp'].nodes


def myCurveData(curve_name):
    if curve_name not in myNodeTree().keys():
        cn = myNodeTree().new('ShaderNodeRGBCurve')
        cn.mapping.initialize()
        cn.name = curve_name
    return myNodeTree()[curve_name]

class HTOOL_OT_CurvesTaperRadius(bpy.types.Operator):
    bl_label = "Taper Curve"
    bl_idname = "curves.curve_taper"
    bl_description = "Taper Curve radius over length"
    bl_options = {"REGISTER"}

    strand_width: FloatProperty(name="Profile width", description="Strands profile width", default=0.5, min=0.0, soft_max=10)
    MainRadius: FloatProperty(name="Main Radius", description="Main Radius", default=1, min=0, soft_max=4)
    onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=False)

    def invoke(self, context, event):  # make selection only false, if obj mode
        curve_obj = context.active_object
        if curve_obj.type != 'CURVES':
            self.report({'INFO'}, 'Use operator on hair curves type of object. Cancelling')
            return {"CANCELLED"}

        called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        self.onlySelection = True if curve_obj.mode in {'EDIT', 'SCULPT_CURVES'}  else called_from_ht_wspace
        self.sel_splines = Splines(curve_obj, self.onlySelection) # was using FLAT - for resampling it seems

        curv_node = myCurveData('taper_mapping')
        curv_node.mapping.initialize()
        curv_node.mapping.curves[3].points[0].location = Vector((1, 0))
        curv_node.mapping.curves[3].points[1].location = Vector((0, 1))
        curv_node.mapping.update()
        return context.window_manager.invoke_props_dialog(self)
        # return self.execute(context)

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'strand_width')
        layout.prop(self, 'MainRadius')
        self.layout.template_curve_mapping(bpy.data.node_groups['HairToolCurveTmp'].nodes["taper_mapping"], "mapping")


    def execute(self, context):
        active_obj = context.active_object
        radius_list_to_len = {} # spline len : radius list

        taper_mapping = bpy.data.node_groups['HairToolCurveTmp'].nodes['taper_mapping']
        for polyline in self.sel_splines.splines:  # for strand point
            curveLenght = polyline.length
            if curveLenght not in radius_list_to_len.keys():
                # x = np.linspace(0, 1, curveLenght)
                rad = np.array([taper_mapping.mapping.evaluate(taper_mapping.mapping.curves[3], x/(curveLenght-1)) for x in range(curveLenght)], 'f')
                radius_list_to_len[curveLenght] = self.MainRadius * rad.clip(0, 1)  # apply fallof

            polyline.radii = radius_list_to_len[curveLenght]
            polyline.write_rad_to_blender_spl(active_obj, polyline.orig_spl_id)

        active_obj.data.update_tag()
        context.area.tag_redraw()
        # bpy.context.window_manager.popup_menu(draw_popup, title="Greeting", icon='INFO')
        return context.window_manager.invoke_props_dialog(self)
        # return {"FINISHED"}


class HTOOL_OT_CurvesTiltAlign(bpy.types.Operator):
    bl_label = "Align curve tilt"
    bl_idname = "object.curves_align_tilt"
    bl_description = "Align curve tilt so that it is aligned to target object surface"
    bl_options = {"REGISTER", "UNDO"}

    onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=True)
    resetTilt: BoolProperty(name="Reset Tilt", description="Reset Tilt before aligning it to surface", default=False)
    align_target_bvht = None
    depsgraph = None

    def execute(self, context):
        active_obj = context.active_object
        if not active_obj.type == 'CURVES':
            self.report({'INFO'}, 'Use operator on curve type object. Cancelling')
            return {"CANCELLED"}

        if self.onlySelection:
            sel_strands = get_selected_strands_mask(context, active_obj)
        else:
            sel_strands = [True]*len(active_obj.data.curves)
        align_curves_tilt(active_obj, sel_strands)
        return {"FINISHED"}
        if active_obj.ht_props.target_obj not in bpy.data.objects.keys(): #cos when run from another operator it may skip init..
            self.report({'INFO'}, 'No target to align to')
            return {"CANCELLED"}
        if not self.depsgraph:
            self.depsgraph = context.evaluated_depsgraph_get()
        if self.align_target_bvht is None: #cos sometimes invoke may be skipped??
            align_target = bpy.data.objects[active_obj.ht_props.target_obj]
            self.align_target_bvht = get_obj_mesh_bvht(align_target, self.depsgraph, applyModifiers=True, world_space=True)
        self.align_curve_tilt(context,self.depsgraph,  active_obj, self.align_target_bvht,  self.resetTilt, self.onlySelection)
        # self.execute_slow_ParallelTransport(context)
        return {"FINISHED"}

    @staticmethod
    def find_nearest(p, bvht_tree, searchDistance):
        hitpoint, norm, face_index, distance = bvht_tree.find_nearest(p, searchDistance)  # max_dist = 10
        # hit, hitpoint, norm, face_index = obj.closest_point_on_mesh(p, 10)  # max_dist = 10
        vecP = hitpoint - p
        v = vecP.dot(norm)
        return hitpoint, norm, face_index, distance, v < 0.0  # v<0 = = is inside volume?


    @staticmethod
    def resetTiltFunction(context, curveData, onlySelection = False):
        for polyline in curveData.splines:  # for strand point
            if polyline.type == 'NURBS' or polyline.type == 'POLY':
                if onlySelection:
                    selection_test = [0] * len(polyline.points)
                    polyline.points.foreach_get('select', selection_test)
                    if not any(selection_test):
                        continue  # if not even one pointis selected
                polyline.points.foreach_set('tilt', [0] * len(polyline.points))
            else:
                if onlySelection:
                    selection_test = [0] * len(polyline.bezier_points)
                    polyline.bezier_points.foreach_get('select_control_point', selection_test)
                    if not any(selection_test):
                        continue  # if not even one point is selected
                polyline.bezier_points.foreach_set('tilt', [0] * len(polyline.bezier_points))
        # curveData.splines.update()
        return

    @staticmethod
    def get_tangents(me, index, is_last):
        selected_vert = me.vertices[index]
        if is_last:
            prev = me.vertices[index - 3]
            tangent = selected_vert.co - prev.co
        else:
            nextVert = me.vertices[index + 3]
            tangent = nextVert.co - selected_vert.co
        return tangent.normalized()

    @staticmethod
    def align_curve_tilt(context, depsgraph, curve, align_target_bvht, resetTilt=False, onlySelection=False):
        sel_strands = get_selected_strands_mask(context, curve) if onlySelection else [True]*len(curve.data.curves)
        align_curves_tilt(curve, sel_strands)
        return # TODO: reimplement for HT 3.0 ? Fuck no

    #slower han hack above...
    def execute_slow_ParallelTransport(self, context):
        ''' #not  somewhat worse than hack method...'''
        snapTargetName = context.active_object.ht_props.target_obj
        snapTarget = bpy.data.objects[snapTargetName]
        Curve = context.active_object

        #apply transformation to prevetn 90 deg offset bug
        snapTarget.data.transform(snapTarget.matrix_world)
        if Curve.mode == 'EDIT':  # to make transfrom() work in edit mode
            bpy.ops.object.mode_set(mode="OBJECT")
            Curve.data.transform(Curve.matrix_world)
            bpy.ops.object.mode_set(mode="EDIT")
        else:
            Curve.data.transform(Curve.matrix_world)

        if self.resetTilt:
            self.resetTiltFunction(Curve.data)

        curveData = Curve.data
        snapTarget_BVHTOOL_tree = BVHTree.FromObject(snapTarget, context.evaluated_depsgraph_get())  # render eg use subsurf render settin

        for i, polyline in enumerate(curveData.splines):  # for strand point
            strand_len = len(polyline.bezier_points) if polyline.type == 'BEZIER' else len(polyline.points)
            if polyline.type == 'NURBS' or polyline.type == 'POLY':
                if self.onlySelection:
                    spline_selection_mask = np.zeros(strand_len, dtype=np.bool)
                    polyline.points.foreach_get('select', spline_selection_mask)
                    if not any(spline_selection_mask):
                        continue  # if not even one pointis selected
                points_co = [p.co for p in polyline.points]
                points = polyline.points
            else:
                if self.onlySelection:
                    spline_selection_mask = np.zeros(strand_len, dtype=np.bool)
                    polyline.bezier_points.foreach_get('select_control_point', spline_selection_mask)
                    if not any(spline_selection_mask):
                        continue  # if not even one point is selected
                points_co = [p.co for p in polyline.bezier_points]
                points = polyline.bezier_points
              # helper for tangetnt function (to detect last point on cuve)
            prevAngle = 0
            corrective_angles = np.zeros(strand_len, dtype=np.float16)
            current_tilt = np.zeros(strand_len, dtype=np.float16)
            points.foreach_get('tilt', current_tilt)
            (T, N, B) = parallel_transport_TNB(points_co)
            strand_Ts = [Vector((i)) for i in T]
            strand_Ns = [Vector((i)) for i in N]
            strand_Bs = [Vector((i)) for i in B]
            for j, (point_T, point_N, p_B, curvePoint) in enumerate(zip(strand_Ts, strand_Ns, strand_Bs, points)):  # for strand point
                # pointOnMeshLoc, normSnapTarget, face_index, distance, isInside = self.find_nearest(curvePoint.co.xyz, snapTarget_BVHTOOL_tree, 11)
                biTangentCurve = point_T.cross(point_N)
                pointOnMeshLoc = align_target_bvht.data.splines[i].points[j].co.xyz
                vecSurfaceHit = curvePoint.co.xyz - pointOnMeshLoc
                # if isInside:  # if curve point is outside snapSurface flip it
                #     vecSurfaceHit *= -1
                vectSurfHitProjected90 = point_T.cross(vecSurfaceHit)
                vectSurfHitProjectedToTangent = vectSurfHitProjected90.cross(point_T)
                # angle = biTangentCurve.angle(vectSurfHitProjectedToTangent)  #unsigned angle <0, 180>
                angle = angle_signed(biTangentCurve, vectSurfHitProjectedToTangent, point_T) - math.pi / 2  # a,b, vN  - signed (-180, 180)
                corrective_angles[j] = angle + 1.57  # add 90deg fix
                prevAngle = angle
            if strand_len >= 4:
                corrective_angles[1] = (corrective_angles[2] + corrective_angles[3]) / 2
                corrective_angles[0] = (corrective_angles[1] + corrective_angles[2]) / 2
            elif strand_len == 3:
                corrective_angles[0] = (corrective_angles[1] + corrective_angles[2]) / 2
            if self.onlySelection:
                current_tilt[spline_selection_mask] += corrective_angles[spline_selection_mask]
            else:
                current_tilt = corrective_angles
            points.foreach_set('tilt', current_tilt)  # in radians
            # manually smooth first two points cos they are usually broken

        snapTarget.data.transform(snapTarget.matrix_world.inverted())
        if Curve.mode == 'EDIT':  # to make transfrom() work in edit mode
            bpy.ops.object.mode_set(mode="OBJECT")
            Curve.data.transform(Curve.matrix_world.inverted())
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)
            bpy.ops.object.mode_set(mode="EDIT")
        else:
            Curve.data.transform(Curve.matrix_world.inverted())
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)
        return {"FINISHED"}


class HTOOL_OT_EmbedRoots(bpy.types.Operator):
    bl_label = "Embed Roots"
    bl_idname = "object.embed_roots"
    bl_description = "Embed Roots into target mesh"
    bl_options = {"REGISTER", "UNDO"}

    embed: FloatProperty(name="Embed roots", description="Radius for bezier curve", default=0, min=0, max=10)
    onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=False)

    # diagonal_local = None
    def invoke(self, context, event):  # make selection only false, if obj mode
        obj = context.active_object
        if obj.mode == 'EDIT':
            self.onlySelection = True
        elif obj.mode == 'OBJECT':
            self.onlySelection = False
        # self.diagonal_local =  obj.dimensions.length/obj.matrix_world.to_scale().length
        return self.execute(context)

    def execute(self, context):
        active_obj = context.active_object
        if not active_obj.type == 'CURVES':
            self.report({'INFO'}, 'Use operator on curve type object')
            return {"CANCELLED"}
        # snapTargetName = Curve.ht_props.target_obj

        target_obj = bpy.data.objects.get(active_obj.ht_props.target_obj) if active_obj.ht_props.target_obj else None
        if not target_obj:
            target_obj = active_obj.data.surface
        if not target_obj:
            print("Curve has No target object to embed into (no cure surface or target object set)")
            return None

        target_obj = active_obj.data.surface
        if not target_obj or  not bpy.data.objects.get(target_obj.name):
            self.report({'INFO'}, 'Curve has no source surface to snap to')
            return {"CANCELLED"}
        # snapTargetName = active_obj.ht_props.target_obj
        # snapTarget = bpy.data.objects[snapTargetName]
        #FIXED:  why context.depsgrap.objects is empty on redo? Fixed with 'USE_EVAL_DATA'
        sourceSurface_BVHT = get_obj_mesh_bvht(target_obj, context.evaluated_depsgraph_get())
        # self.diagonal_local = self.diagonal_local if self.diagonal_local else Curve.dimensions.length/Curve.matrix_world.to_scale().length
        sel_strands = get_selected_strands_mask(context, active_obj) if self.onlySelection else [True]*len(active_obj.data.curves)
        self.embed_strands_roots(context, active_obj, sourceSurface_BVHT, self.embed, sel_strands)
        return {"FINISHED"}

    @staticmethod
    def embed_strands_roots(context, Curve, sourceSurface_BVHT, embed, sel_strands):
        # Curve.data.transform(Curve.matrix_world)
        mw = Curve.matrix_world
        sel_strands_ids = [i for i in range(len(Curve.data.curves)) if sel_strands[i]]
        for i, polyline in enumerate(Curve.data.curves):  # for strand point
            if sel_strands[i]:
                points = polyline.points
                snappedPoint, normalChildRoot, rootHitIndex, distance = sourceSurface_BVHT.find_nearest(mw@points[0].position, 100)
                diff = points[0].position - points[1].position
                diff.normalize()
                diff_dot_norm = abs(diff.dot(normalChildRoot))  # the more diff faces the target normal, the smaller this is
                points[0].position += (diff * 0.6 * diff_dot_norm - normalChildRoot * (1 - diff_dot_norm)) * embed  # do childStrandRootNormal to move it more into mesh surface
        # Curve.data.transform(Curve.matrix_world.inverted())


class HTOOL_OT_CurvesTiltRandomize(bpy.types.Operator):
    bl_label = "Randomize tilt"
    bl_idname = "curves.curves_randomize_tilt"
    bl_description = "Change curve tilt, can be used for currly hair"
    bl_options = {"REGISTER", "UNDO"}

    reset_first: BoolProperty(name="Reset on init", description="Clear old tilt before randimizing it", default=True)
    onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=True)
    tilt: FloatProperty(name="Max Tilt Angle", description="Maximum tilt angle applied to strands", default=1.5, min=-100, max=100, step=145, subtype='ANGLE')
    min_tilt: FloatProperty(name="Min Tilt", description="Minimal tilt angle (described as % \of max tilt) applied to strands", default=0.5, min=0, max=1, subtype='PERCENTAGE')
    tilt_transition_fallof: FloatProperty(name="Transition falloff", description="Describes tilt influence strength over strand length", default=0, min=0, max=1, subtype='PERCENTAGE')
    root_influence: FloatProperty(name="Root Influence", description="Strength of tilt shift induced on strand roots", default=0, min=0, max=1, subtype='PERCENTAGE')
    randomDirection: BoolProperty(name="Flip Direction", description="Flip tilt direction angle randomly from CW to CCW", default=False)
    Seed: IntProperty(name="Noise Seed", default=1, min=1, max=500)

    def save_settings(self, target_obj):  # to object  rand_tilt_settings
        for d in self.properties.bl_rna.properties.keys():
            if d == 'rna_type':
                continue
            setattr(target_obj.ht_props.rand_tilt_settings, d, getattr(self.properties, d))

    def load_settings(self, source_obj):  # from  braid_settings
        for d in source_obj.ht_props.rand_tilt_settings.bl_rna.properties.keys():
            if d in {'name', 'rna_type'}:
                continue
            setattr(self.properties, d, getattr(source_obj.ht_props.rand_tilt_settings, d))

    def invoke(self, context, event):  # make selection only false, if obj mode
        obj = context.active_object
        if obj.mode == 'EDIT':
            self.onlySelection = True
        elif obj.mode == 'OBJECT':
            self.onlySelection = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        self.load_settings(context.active_object)  # so load cached settings
        # get set tilt
        tilt_attr = obj.data.attributes.get('tilt')
        if not tilt_attr:
            tilt_attr = obj.data.attributes.new("tilt", "FLOAT", "POINT")
        return self.execute(context)

    def execute(self, context):
        curve_obj = context.active_object
        if not curve_obj.type == 'CURVES':
            self.report({'INFO'}, 'Use operator on curve type object')
            return {"CANCELLED"}

        if self.onlySelection:
            sel_strands = get_selected_strands_mask(context, curve_obj)
            sel_splines = [curve_obj.data.curves[i] for i in range(len(sel_strands)) if sel_strands[i]]
        else:
            sel_splines = curve_obj.data.curves

        np.random.seed(self.Seed*149)
        spline_count = len(curve_obj.data.curves)
        randTiltList = np.random.rand(spline_count)
        cpow = calc_exponent(self.tilt_transition_fallof)
        randDir = np.random.choice([1,-1],spline_count) if self.randomDirection else np.ones(spline_count, dtype=np.int)
        np.random.seed(self.Seed+12)
        # called_from_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        tilt_attr = curve_obj.data.attributes.get('tilt')
        for i, polyline in enumerate(sel_splines):  # for strand point
            points = polyline.points
            pointsNumber = len(points)
            for j, curvePoint in enumerate(points):  # for strand point
                strand_Fallof = math.pow(j / pointsNumber, cpow)
                min_angle_factor  = self.tilt/2 * randDir[i] * randTiltList[i] * (1-self.min_tilt) + 1*self.min_tilt
                randomizedAngle = self.tilt * randDir[i] * (min_angle_factor *  (strand_Fallof * (1 - self.root_influence) + self.root_influence))
                if self.reset_first:
                    tilt_attr.data[curvePoint.index].value = randomizedAngle
                else:
                    tilt_attr.data[curvePoint.index].value += randomizedAngle
        self.save_settings(curve_obj)
        return {"FINISHED"}



class HTOOL_OT_HairStraighten(bpy.types.Operator):
    bl_idname = "curves.hair_straighten"
    bl_label = "Hair Straighten"
    bl_description = "Curve Hair Transform (more, rotate, scale tips) with IK"
    bl_options = {'REGISTER','UNDO'}

    repeat: bpy.props.IntProperty(name='Strength', description='', default= 2, min=1, soft_max=10)
    offset: bpy.props.FloatProperty(name="Transition Offset", description="Moves straighten influence toward the roots of strands", default=0.5, min=0, max=1)
    contrast: bpy.props.FloatProperty(name="Transition Contrast", description="Increase straighten effect transition contrast along strand length. Low values give smoother transition", default=0.5, min=0, soft_max=1)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def invoke(self, context, event):
        curve_obj = context.active_object
        splines = curve_obj.data.curves
        sel_strands = get_selected_strands_mask(context, curve_obj)
        self.orign_ik_chains = []
        sel_spl_ids = []
        for sp_id, is_sel in enumerate(sel_strands):
            if is_sel:
                self.orign_ik_chains.append(ChainIK(splines[sp_id].points))
                sel_spl_ids.append(sp_id)
        self.sel_spl_ids = sel_spl_ids
        return self.execute(context)


    def execute(self, context):
        curve_obj = context.active_object
        def write_pts(pts, moved_pts):
            for p, m_p in zip(pts, moved_pts):
                p.position = m_p
        # clump_falloff = context.scene.ht_props.clump_falloff
        ik_chains = [copy.deepcopy(ik_ch) for ik_ch in self.orign_ik_chains] #makes it work correctly with redoo
        for _ in range(self.repeat):
            for sp_id, ik_chain in zip(self.sel_spl_ids, ik_chains):
                spl = curve_obj.data.curves[sp_id]
                pts = spl.points
                ik_chain.straighten_ik_chain(self.contrast, self.offset)
                moved_pts = ik_chain.get_pts()
                write_pts(pts, moved_pts)
        return {"FINISHED"}


# commented out in UI for now...
class HTOOL_OT_CurveRadiusByLength(bpy.types.Operator, HTool_ClassCommon):
    bl_label = "Radius from length"
    bl_idname = "curve.radius_by_length"
    bl_description = "Radius from strand length"
    bl_options = {"REGISTER", "UNDO"}

    min_radius: bpy.props.FloatProperty(name='Min Radius', description='', default= 0.5, min=0, max=1.0)
    max_radius: bpy.props.FloatProperty(name='Max Radius', description='', default= 1.0, min=0, max=1.0)
    len_threshold: bpy.props.FloatProperty(name='Transition threshold', description='Radius change will affect only strands close to given length threshold', default= 0.5, min=0, max=1)
    margin: bpy.props.FloatProperty(name='Transition Smoothness', description='Smooth out affected strand radius chagne transition', default=1.0, min=0, max=1.0)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def invoke(self, context, event):
        curve_obj = context.active_object
        called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        self.only_selected = True if curve_obj.mode in {'EDIT', 'SCULPT_CURVES'} else called_from_ht_wspace
        s_splines = Splines(curve_obj, onlySelection=self.only_selected)

        euclid_len = np.array([s.euclidean_len for s in s_splines.splines])
        mx = np.max(euclid_len)
        mi = np.min(euclid_len)
        if abs(mx - mi) < 0.02:
            self.len_norms = np.ones(s_splines.length)
        else:
            self.len_norms = (euclid_len - mi)/(mx-mi) # remap <mi, mx> into <0, 1> range
        self.ssplines = s_splines
        self.load_settings(curve_obj.ht_props.radius_from_length_settings)
        return self.execute(context)

    def execute(self, context):
        curve_obj = context.active_object
        len_threshold = smooth_threshold(self.len_norms, 1-self.len_threshold, self.margin/2)
        len_remapped = (self.max_radius - self.min_radius)* len_threshold + self.min_radius  # remap (0, 1) to new (mi, mx)
        for polyline, new_len in zip(self.ssplines.splines, len_remapped):
            polyline.radii = [new_len]*len(polyline.points)
            polyline.write_rad_to_blender_spl(curve_obj, polyline.orig_spl_id)
        self.save_settings(curve_obj.ht_props.radius_from_length_settings)
        # update 3d view
        curve_obj.update_from_editmode()
        context.area.tag_redraw()
        return {"FINISHED"}


class HTOOL_OT_CutWithMesh(bpy.types.Operator):
    bl_label = "Slice using mesh"
    bl_idname = "curves.slice_with_mesh"
    bl_description = "Slice curve using mesh. First select mesh and curves for slicing, then run the operator\nYou have option to change direction of slicing (slice to strands roots or tips)"
    bl_options = {"REGISTER", "UNDO"}

    onlySelection: BoolProperty(name="Only Selected", description="Affect only selected points", default=False)
    flip_dir: bpy.props.BoolProperty(name='Flip Direction', description='', default=False)

    @classmethod
    def poll(cls, context):
        return context.active_object and len(context.selected_objects)>1

    def invoke(self, context, event):  # make selection only false, if obj mode
        sel_curves = [o for o in context.selected_objects if o.type=='CURVES']
        sel_meshes = [o for o in context.selected_objects if o.type=='MESH']
        if not sel_curves:
            self.report({'WARNING'}, 'Add curve object to selection')
            return {"CANCELLED"}
        if not sel_meshes:
            self.report({'WARNING'}, 'Add mesh type of object to selection')
            return {"CANCELLED"}
        called_from_ht_wspace = context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == "hair_tool.hair_transform"
        self.onlySelection = True if context.mode in {'EDIT_CURVES', 'SCULPT_CURVES'} else called_from_ht_wspace

        self.sel_splines = []
        for c in sel_curves:
            # has_shape_key = c.data.shape_keys != None
            # if not has_shape_key:
            self.sel_splines.append(Splines(c, self.onlySelection))
        # if has_shape_key: # for old curves
        #     self.report({'WARNING'}, 'Curves with shape keys will be skipped')

        cutter = sel_meshes[0]
        self.depsgraph = context.evaluated_depsgraph_get()
        self.cutter_bvht = get_obj_mesh_bvht(cutter, self.depsgraph, applyModifiers=True, world_space=True)
        bpy.ops.ed.undo_push()  # to prevent restoring curve from step above on execute()
        return self.execute(context)


    def execute(self, context):
        sel_curves = [o for o in context.selected_objects if o.type=='CURVES'] #NOTE: make sure sel_curve mathes self.sel_splines
        for sel_splines, curve_obj in zip(self.sel_splines, sel_curves):
            mv_inv = curve_obj.matrix_world.inverted()
            sspline = copy.deepcopy(sel_splines)
            for spl_idx, polyline in enumerate(sspline.splines):  # for strand point
                if self.flip_dir:
                    prev_pt = curve_obj.matrix_world @ polyline.points[-1].co
                    for i,pt in reversed(list(enumerate(polyline.points[:-1]))):
                        pt_co = curve_obj.matrix_world @ pt.co
                        dir = pt_co - prev_pt
                        seg_len = dir.length
                        loc, normal, f_index, distance = self.cutter_bvht.ray_cast(prev_pt, dir) # we hit world space
                        if loc:
                            if distance < 0.3 * seg_len:
                                sspline.splines[spl_idx].points = polyline.points[i+1:]
                                sspline.splines[spl_idx].points[0].co = mv_inv @ loc
                                break
                            elif distance < seg_len:
                                sspline.splines[spl_idx].points = polyline.points[i:]
                                sspline.splines[spl_idx].points[0].co = mv_inv @ loc
                                break
                        prev_pt = pt_co
                else:
                    prev_pt = curve_obj.matrix_world @ polyline.points[0].co
                    for i,pt in enumerate(polyline.points[1:]):
                        pt_co = curve_obj.matrix_world @ pt.co
                        dir = pt_co - prev_pt
                        seg_len = dir.length
                        loc, normal, f_index, distance = self.cutter_bvht.ray_cast(prev_pt, dir) # we hit world space
                        if loc:
                            if distance < 0.3 * seg_len:
                                sspline.splines[spl_idx].points = polyline.points[:i+1]
                                sspline.splines[spl_idx].points[-1].co = mv_inv @ loc
                                break
                            elif distance < seg_len:
                                sspline.splines[spl_idx].points = polyline.points[:i+2]
                                sspline.splines[spl_idx].points[-1].co = mv_inv @ loc
                                break

                        prev_pt = pt_co

            sspline.write_splines_to_blender(curve_obj)

        # curve_obj.data.update_tag()
        # context.area.tag_redraw()
        return {'FINISHED'}



shader_2d = gpu.shader.from_builtin('UNIFORM_COLOR')
def draw_line_2d(self, context):
    line_end = [self.mouse_pos[0], self.selection_center_2d[1]]

    batch = batch_for_shader(shader_2d, 'LINES', {"pos": [self.selection_center_2d , line_end]})
    shader_2d.bind()
    shader_2d.uniform_float("color", (1.0, 1.0, 1.0, 0.8))
    gpu.state.line_width_set(2.0)
    gpu.state.blend_set("ALPHA")
    gpu.state.depth_test_set("LESS")

    batch.draw(shader_2d)

    gpu.state.depth_test_set("NONE")
    gpu.state.line_width_set(1.0)
    gpu.state.blend_set("NONE")

    if context.scene.tool_settings.use_proportional_edit:
        color = (0.9, 0.9, 0.9, 0.9)
        # draw_circle_2d(self.mouse_start_pos_2d, color, self.rad_draw_2d, segments=32)
        draw_circle_2d(self.selection_center_2d, color, self.rad_draw_2d, segments=32)



# XXX: undo not work in edit mode - in 3.6 - waiting for fix : https://projects.blender.org/blender/blender/issues/114173
class HTOOL_OT_EditAdjustRadius(bpy.types.Operator):
    bl_idname = "curves.editmode_adjust_tilt_rad"
    bl_label = "Set Tilt/Rad"
    bl_description = "Adjust editmode, selected hair curves points: tilt or radius"
    bl_options = {'REGISTER', 'GRAB_CURSOR', 'BLOCKING', 'UNDO'}

    mode: bpy.props.EnumProperty(name='Mode', description='Smooth Axis', items=[
        ('RADIUS', 'Radius', 'Radius'),
        ('TILT', 'Tilt', 'Tilt')], default='RADIUS')

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES' and context.mode == 'EDIT_CURVES'

    def init_attribs(self, active_obj, default_value):
        attr_type = 'radius' if self.mode == 'RADIUS' else 'tilt'
        affected_attr = active_obj.data.attributes.get(attr_type)
        if affected_attr:
            rad_d = np.empty(len(affected_attr.data))
            affected_attr.data.foreach_get('value', rad_d)
            if attr_type == 'radius' and not np.any(rad_d): # all rad zeroes. Must be bug in blender
                self.attr_cached = np.ones(len(affected_attr.data), dtype=np.float32) # workaround
                affected_attr.data.foreach_set('value', self.attr_cached)
            else:
                self.attr_cached = rad_d
        else:
            affacted_attr = active_obj.data.attributes.new(attr_type, "FLOAT", "POINT")
            self.attr_cached = [default_value]*len(active_obj.data.points)
            for pt_idx in range(len(active_obj.data.points)):
                affacted_attr.data[pt_idx].value = self.attr_cached[pt_idx]
            # affacted_attr.foreach_set('value', self.attr_cached) # shit only work on radius
            bpy.ops.object.mode_set(mode='EDIT')


    def calc_rad_2d(self, context):
        '''For drawing circle'''
        region = context.region
        rv3d = context.region_data
        scn_s = context.scene.ht_props
        center_2d_plus_rad_3d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.selection_center_3d + Vector((scn_s.influence_rad, 0, 0)), default=Vector((0, 0)))
        self.rad_draw_2d = (center_2d_plus_rad_3d - self.selection_center_2d).length


    def euclidean_seg_distances(self, curve_obj):
        ''' Returns  curve.points - min distance to adjacent selected point - as flat array (similar to points attrib)'''
        sel_pts_bool = self.sel_pts_bool

        mw = curve_obj.matrix_world
        spl_pts_dist_to_sel = []
        for spl in curve_obj.data.curves:
            if not any([sel_pts_bool[p.index] for p in spl.points]):
                spl_pts_dist_to_sel.extend([999]*len(spl.points))
                continue # skip
            seg_lens = [] # distance between p[n] and p[n-1] - for p[n] e.g. [0, x1, x2, ..]
            prev_co = mw @ spl.points[0].position
            for p in spl.points:
                current_pt = mw @ p.position
                seg_lens.append((current_pt - prev_co).length)
                prev_co = current_pt

            # calculate distance from current point to first selected point
            per_point_min_dist = [9999] * len(spl.points)
            for i, p in enumerate(spl.points): # first check in
                min_dist = 999
                if sel_pts_bool[p.index]:
                    per_point_min_dist[i] = 0
                    continue

                acc_len_right = 0
                # check for closest selected point to right
                if i < len(spl.points)-1: # if not last point
                    for right_id, sel_p in enumerate(spl.points[i+1:], start=i+1):
                        acc_len_right += seg_lens[right_id]
                        if sel_pts_bool[sel_p.index]:
                            min_dist = acc_len_right
                            break


                # search to left
                acc_len_left = 0
                if i > 0: # if not first point
                    for left_id, sel_p in enumerate(reversed(spl.points[:i])): # handle points to left
                        acc_len_left += seg_lens[left_id+1]
                        if sel_pts_bool[sel_p.index]:
                            if acc_len_left < min_dist:
                                min_dist = acc_len_left
                            break
                        if acc_len_left >= min_dist: # do not check further if we already found closer point in other direction
                            break

                per_point_min_dist[i] = min_dist

            spl_pts_dist_to_sel.extend(per_point_min_dist)

        return spl_pts_dist_to_sel



    def kd_closest_seg_distances(self, curve_obj):
        ''' Returns  curve.points - min distance any selected point - does not have to be adjacent (on same spline)'''
        sel_pts_bool = self.sel_pts_bool

        mw = curve_obj.matrix_world
        spl_pts_dist_to_sel = []
        kdtree = KDTree(sum(sel_pts_bool))
        for point in curve_obj.data.points:
            if sel_pts_bool[point.index]:
                kdtree.insert(mw @ point.position, point.index)
        kdtree.balance()

        # for each non selected point find closest selected point. For selected point distance is 0
        for point in curve_obj.data.points:
            if sel_pts_bool[point.index]:
                spl_pts_dist_to_sel.append(0)
            else:
                co, index, dist = kdtree.find(mw @ point.position)
                spl_pts_dist_to_sel.append(dist)

        return spl_pts_dist_to_sel



    def invoke(self, context, event):
        curve_obj = context.active_object
        if curve_obj.type!= 'CURVES':
            self.report({'ERROR'}, 'Select curve object')
            return {'CANCELLED'}
        sel_pts_bool = get_edit_selected_points_bool(curve_obj)
        if not any(sel_pts_bool):
            self.report({'ERROR'}, 'Select at least one point. Cancelling')
            return {'CANCELLED'}
        self.sel_pts_bool = sel_pts_bool
        # print(f"{self.sel_pts_bool=}")


        # if context.scene.tool_settings.transform_pivot_point == 'CURSOR':
        #     op_center = curve_obj.matrix_world.inverted() @ context.scene.cursor.location
        # else:
        sel_center = Vector((0, 0, 0))
        for p in curve_obj.data.points:
            if sel_pts_bool[p.index]:
                sel_center += p.position
        op_center = sel_center / sum(sel_pts_bool)
        self.selection_center_3d = curve_obj.matrix_world @ op_center

        region = context.region
        rv3d = context.region_data
        self.selection_center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.selection_center_3d, default=None)


        self.mouse_start_pos_2d = Vector((event.mouse_region_x, event.mouse_region_y))
        self.mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
        default_value = 1 if self.mode == 'RADIUS' else 0
        self.init_attribs(curve_obj, default_value) # init missing attributes if needed

        if context.scene.tool_settings.use_proportional_edit:
            if context.tool_settings.use_proportional_connected:
                self.spl_pts_dist_to_sel = self.euclidean_seg_distances(curve_obj)
            else:
                self.spl_pts_dist_to_sel = self.kd_closest_seg_distances(curve_obj)
            # print(f"{self.spl_pts_dist_to_sel=}")

        self.force_zero = False
        self.calc_rad_2d(context)  # for drawing
        args = (self, context)
        self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_line_2d, args, "WINDOW", "POST_PIXEL")
        context.window_manager.modal_handler_add(self)
        return {"RUNNING_MODAL"}


    def update_draw_header(self, context, value):
        txt = 'Radius' if self.mode == 'RADIUS' else 'Tilt'
        scn_s = context.scene.ht_props
        if context.scene.tool_settings.use_proportional_edit:
            context.area.header_text_set(f"{txt}: {value:.2f}   Proportional Size: {scn_s.influence_rad:.2f}")
        else:
            context.area.header_text_set(f"{txt}: {value:.2f}")


    def modal(self, context, event):
        scn_s = context.scene.ht_props
        if event.type == "MOUSEMOVE":
            self.adjust_radilt(context, event)

        if event.type == "LEFTMOUSE":
            self.finish(context)
            return {"FINISHED"}

        if event.type == "ZERO" and event.value == 'RELEASE':
            self.force_zero = not self.force_zero
            self.adjust_radilt(context, event)

        if event.type == "WHEELDOWNMOUSE" and context.scene.tool_settings.use_proportional_edit:
            scn_s.influence_rad *= 1.1
            self.adjust_radilt(context, event)

        if event.type == "WHEELUPMOUSE" and context.scene.tool_settings.use_proportional_edit:
            scn_s.influence_rad = max(0.001, scn_s.influence_rad * 0.9)
            self.adjust_radilt(context, event)

        if event.type in {"RIGHTMOUSE", "ESC"}:
            self.finish(context)
            return {"FINISHED"}

        return {"RUNNING_MODAL"}

    def finish(self, context):
        bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
        context.area.tag_redraw()
        context.area.header_text_set(None)


    def adjust_radilt(self, context, event):
        self.calc_rad_2d(context)  # for drawing
        self.mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
        delta = self.mouse_pos[0] - self.mouse_start_pos_2d[0]
        delta *= 0.02
        self.update_draw_header(context, delta)

        active_obj = context.active_object
        attr_type = 'radius' if self.mode == 'RADIUS' else 'tilt'
        affected_attr = active_obj.data.attributes.get(attr_type)
        sel_pts_bool = get_edit_selected_points_bool(active_obj)

        if self.force_zero:
            delta = 0

        scn_s = context.scene.ht_props
        use_fallof = context.scene.tool_settings.use_proportional_edit
        fallof_fun = get_fallof_fun(context.scene.tool_settings.proportional_edit_falloff)
        for spl in active_obj.data.curves:
            if not any([sel_pts_bool[p.index] for p in spl.points]) and context.tool_settings.use_proportional_connected:
                continue # skip
            for p in spl.points:
                if sel_pts_bool[p.index] or use_fallof:
                    if  self.mode == 'RADIUS':
                        new_val = max(self.attr_cached[p.index] + delta, 0) # base rad + delta
                    else:
                        new_val = self.attr_cached[p.index] + delta
                    if use_fallof:
                        pt_distance = self.spl_pts_dist_to_sel[p.index]
                        dist_fallof = max(1-pt_distance/scn_s.influence_rad, 0)  # map dist - (0, inf), to (1, 0)
                        dist_fallof = fallof_fun(dist_fallof)
                        new_val = lerp(self.attr_cached[p.index], new_val, dist_fallof)
                    affected_attr.data[p.index].value = new_val


        active_obj.update_tag()
        context.view_layer.update()
        context.area.tag_redraw()
        return {'FINISHED'}
