# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import bmesh
import math
import numpy as np
from os import path
from bpy.props import BoolProperty, IntProperty, FloatProperty
from mathutils import Vector, kdtree, Color
from mathutils.geometry import intersect_point_line
from .utils.helper_functions import calc_exponent, get_obj_mesh_bvht, clamp
from .utils.mesh_utils import build_vert_islands, build_edge_islands,sort_edges_by_uv_v,edges_to_strips_by_uv
from .utils.general_utils import set_parent
from .hair_tool_ui import ModifierPopupPanel
from .hair_baking.hair_geometry_nodes_shared import get_node_group, get_profile_mod
from mathutils.interpolate import poly_3d_calc
from .jiggle_chain import OBJECT_OT_Assign_Bones_Jiggle
from .jiggle_chain_b4 import OBJECT_OT_Assign_Bones_JiggleB4
from itertools import cycle
from .utils.curve_wrapper import SplineFlat
import gpu
from gpu_extras.batch import batch_for_shader


shader = gpu.shader.from_builtin('UNIFORM_COLOR')
# TODO: origin at hair obj, or at vec(0,0,0)?
# TODO: better bone gen for  cylinder profile
# DONE: offset to tip of bone
# DONE: bone for all sel islands mode

def build_strand_islands(bm, all_verts=False):
    """
       Builds list of vert rows for each strand island. Only for selected verts.
    Args:
        bm ():

    Returns:
        islands (): list of vert rows for each strand island
        active_idx (): index of active island

    """
    islands = []
    island = []  #  [[verts_row1], [verts_row2], ...]
    island_curl_verts = []  # list of  [curl_verts]
    def add_to_island(vert_row):
        cnt = len(vert_row)
        if cnt > 1: # avoid adding single vert rows - caused by uv curls
            island.append(vert_row)
        elif cnt == 1:
            island_curl_verts.append(vert_row[0]) # single vert anyway

    vert_row_ids = [] # island is list of rows of vert ids
    prev_loop = bm.verts[0].link_loops[0]
    uv_lay = bm.loops.layers.uv['HairTool_UV']
    # go through all verts and build islands - based on UV[1]. Curl verts are added separately
    for v in bm.verts: # DONE: deal with curly uvs... - ignore len(vert_row_ids) == 1 ?
        loop = v.link_loops[0] # all should have same uv
        diff = loop[uv_lay].uv[1] - prev_loop[uv_lay].uv[1] # fixed in 3.5.1 :,
        # if bpy.app.version < (3, 5, 0):
        #     diff = loop[uv_lay].uv[1] - prev_loop[uv_lay].uv[1]
        # else:
        #     diff = prev_loop[uv_lay].uv[1] - loop[uv_lay].uv[1]
        if diff <= 0.00001 and diff >=0 :   # NOTE: same row
            vert_row_ids.append(v.index) # vert is invalidated after edit_bone mode, thus store index
        elif diff > 0.00001:  # NOTE: moved to next row
            add_to_island(vert_row_ids)
            vert_row_ids = [v.index] # start filling new row
        else:             # NOTE: negative - we moved to new island
            add_to_island(vert_row_ids)
            if bm.verts[vert_row_ids[-1]].select or all_verts:
                islands.append(island)
            island = []
            island_curl_verts = []
            vert_row_ids = [v.index]
        prev_loop = v.link_loops[0]
    add_to_island(vert_row_ids)
    if bm.verts[vert_row_ids[-1]].select or all_verts:
        islands.append(island) #no next island, thus conditions from loop above wont be triggered
    # print(islands)
    return islands

g_islands_data = {} # inslads data for HTOOL_OT_GenerateBones
g_islands_data['all_islands'] = None
g_islands_data['island_vert_strips_co'] = None
g_islands_data['longest_island_len'] = None
g_islands_data['islands_len'] = None
g_islands_data['vert_cnt'] = 0

def store_to_g_islands(all_islands, island_vert_strips_co, islands_len, longest_island_len, vert_face_cnt):
    # obj['all_islands'] = all_islands # this sucks - invalid on operator redo... use global instead...
    # obj['all_islands_curl_verts'] = all_islands_curl_verts
    # obj['longest_island_len'] = longest_island_len
    # obj['islands_len'] = islands_len
    global g_islands_data
    g_islands_data['all_islands'] = all_islands
    g_islands_data['island_vert_strips_co'] = island_vert_strips_co
    g_islands_data['longest_island_len'] = longest_island_len
    g_islands_data['islands_len'] = islands_len
    g_islands_data['vert_cnt'] = vert_face_cnt



def precalc_island_cache(bm, uv_threshold=0.2, smooth_factor=0.5, reverse=True):
    '''
    returns:
        obj['all_islands'] = all_islands
        obj['all_islands_curl_verts'] = all_islands_curl_verts
        obj['longest_island_len'] = longest_island_len
        obj['islands_len'] = islands_len
    '''
    # NOTE: build mesh islands based on HairTool_UV uvs
    # Take all islands - for bone count normalization by longest island...
    bm.faces.ensure_lookup_table()
    bm.verts.ensure_lookup_table()

    # if self.normalize_by_length:
    # all_islands = build_strand_islands(bm, all_verts=True) # sorted verts (assuming cards comewith ordered verts)
    all_islands = build_vert_islands(bm) # list of lists of tuples (vert, uv_v) - not sorted
    all_edge_islands = build_edge_islands(bm) # list of tuples (edge, uv_v) - not sorted

    #  XXX: issue - vert islands vere made from verts idx, now from verts.. -they may get invalidated after operator mode chages...
    islands_as_vert_ids = [[v.index for v in island] for island in all_islands]

    # gen vert strips for all islands
    # sort_vertices_by_uv_v: return list of tuples (vert, uv_v)
    island_vert_strips_pos = []
    for e_island in all_edge_islands:
        sorted_edges = sort_edges_by_uv_v(bm, e_island, reverse=reverse)
        out_verts_pos = edges_to_strips_by_uv(sorted_edges, uv_threshold=uv_threshold, smooth_factor=smooth_factor) # TODO: make it a property
        island_vert_strips_pos.append(out_verts_pos)

    # get only subset of selected islands for actuall bone creation
    longest_island_len = 0.0
    islands_len = []
    for vert_strip in island_vert_strips_pos:
        # NOTE: calc len for all islands.
        prev_co = vert_strip[0] # use first vert in first row
        acc_len = 0
        for v_co in vert_strip[1:]:
            acc_len += (prev_co - v_co).length
            prev_co = v_co
        if acc_len > longest_island_len: # globally longest
            longest_island_len = acc_len
        islands_len.append(acc_len)

    # get island with longest length (for normalization)
    # return all_islands, longest_island_len,islands_len
    store_to_g_islands(islands_as_vert_ids, island_vert_strips_pos, islands_len, longest_island_len, (len(bm.verts),len(bm.faces)))
    # return all_islands, island_vert_strips_co,longest_island_len,islands_len
    return island_vert_strips_pos



class HTOOL_OT_PrecalcIslandsCache(bpy.types.Operator):
    bl_label = "Generate Haircards Cache"
    bl_idname = "object.precalc_islands_cache"
    bl_description = "Generate cache for bone generation. Use Ctrl+MMB Scroll to adjust simplification level (Shift+MMB for smoothness). ESC to cancel"
    bl_options = {"REGISTER", "UNDO"}

    uv_threshold: FloatProperty(name="Simplification", description='Based on mesh UVs strip detection', default=0.2, min=0.0, soft_min=0.0, soft_max=1.0)
    smooth_factor: FloatProperty(name="Smooth Factor", description='Smooth factor for UV strip detection', default=0.9, min=0.0, max=1.0)

    def modal(self, context, event):
        if event.type in {'RET','ESC', 'RIGHTMOUSE'}:
            self.cancel(context)
            return {'CANCELLED'}

        elif event.type == 'WHEELUPMOUSE':
            # if shift - smooth
            if event.shift:
                self.smooth_factor = min(1.0, self.smooth_factor + 0.05)
            elif event.ctrl:
                self.uv_threshold = min(1.0, self.uv_threshold + 0.05)
            else:
                return {'PASS_THROUGH'}
            self.island_strips = precalc_island_cache(self.bm, uv_threshold=self.uv_threshold, smooth_factor=self.smooth_factor)
            context.area.header_text_set(f"UV Threshold: [{self.uv_threshold:.2f}]         Smooth Factor: [{self.smooth_factor:.2f}]")
            context.area.tag_redraw()

        elif event.type == 'WHEELDOWNMOUSE':
            if event.shift:
                self.smooth_factor = max(0.0, self.smooth_factor - 0.05)
            elif event.ctrl:
                self.uv_threshold = max(0.0, self.uv_threshold - 0.05)
            else:
                return {'PASS_THROUGH'}
            self.island_strips = precalc_island_cache(self.bm,  uv_threshold=self.uv_threshold, smooth_factor=self.smooth_factor)
            context.area.header_text_set(f"UV Threshold: [{self.uv_threshold:.2f}]         Smooth Factor: [{self.smooth_factor:.2f}]")
            context.area.tag_redraw()

        # allow mmb navigation
        if event.type in {'MIDDLEMOUSE'}: #   , 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        obj = context.active_object
        obj_bone_cache = obj.get('ht_mesh_to_bones_cache')
        if obj_bone_cache:
            self.uv_threshold, self.smooth_factor = obj_bone_cache[0], obj_bone_cache[1]

        # check if mesh has uv if not - wanr user it is required to determine cards  orientation
        if len(obj.data.uv_layers) < 1:
            self.report({'ERROR'}, 'No UV found. They are required to determine cards orientation! Cancelling')
            return {"CANCELLED"}

        # init bmesh depending on mode
        self.free_bmesh = False # free if in obj mode

        if obj.mode == 'OBJECT':
            bm = bmesh.new()
            bm.from_mesh(obj.data)
            self.free_bmesh = True
        elif obj.mode == 'EDIT':
            bm = bmesh.from_edit_mesh(obj.data)

        self.bm = bm
        self.island_strips = precalc_island_cache(self.bm,  uv_threshold=self.uv_threshold, smooth_factor=self.smooth_factor)
        context.window_manager.modal_handler_add(self)

        #  draw self.uv_threshold in both status bar and header
        context.area.header_text_set(f"UV Threshold: [{self.uv_threshold:.2f}]         Smooth Factor: [{self.smooth_factor:.2f}]")
        context.workspace.status_text_set('[Ctrl+MMB Scroll] - Simplification       [Shift+MMB Scroll] - Smoothing        [ENTER] - Confirm')


        args = (self, context)
        self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_callback, args, "WINDOW", "POST_VIEW")

        context.area.tag_redraw()
        return {'RUNNING_MODAL'}

    def cancel(self, context):
        bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
        context.area.header_text_set(None)
        context.workspace.status_text_set(None)
        # store self.uv_threshold  and self.smooth_factor in obj props
        context.active_object['ht_mesh_to_bones_cache'] = (self.uv_threshold, self.smooth_factor)
        if self.free_bmesh:
            self.bm.free()
        self.report({'INFO'}, "Mesh cache generation, for bone generation, done")
        context.area.tag_redraw()


    @staticmethod
    def draw_callback(self, context):
        gpu.state.blend_set("ALPHA")
        gpu.state.line_width_set(2.0)
        gpu.state.depth_test_set('NONE')

        with gpu.matrix.push_pop():
            gpu.matrix.multiply_matrix(context.active_object.matrix_world)
            for strip in self.island_strips:
                if len(strip) > 1:
                    batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": strip})
                    shader.bind()
                    shader.uniform_float("color", (1.0, 1.0, 1.0, 0.5))  # Set color with alpha opacity (0.5)
                    batch.draw(shader)

                    gpu.state.point_size_set(3)
                    # also draw points for each strip
                    batch = batch_for_shader(shader, 'POINTS', {"pos": strip[:-1]})
                    shader.uniform_float("color", (1.0, 1.0, 1.0, 1.0))  # Set color with alpha opacity (0.5)
                    batch.draw(shader)

                    # drw last pt with red
                    gpu.state.point_size_set(4)
                    batch = batch_for_shader(shader, 'POINTS', {"pos": [strip[-1]]})
                    shader.uniform_float("color", (0.5, 1.0, 0.2, 1.0))  # Set color with alpha opacity (0.5)
                    batch.draw(shader)

        # Re-enable the depth test after drawing
        gpu.state.depth_test_set("NONE")
        gpu.state.blend_set("NONE")
        gpu.state.point_size_set(1)
        gpu.state.line_width_set(1.0)


class HTOOL_OT_GenerateBones(bpy.types.Operator):
    bl_label = "Generate Bones"
    bl_idname = "object.ribbon_to_bones"
    bl_description = "Create bones from selected mesh hair strands.\nStrand with active mesh element will be used to generate bone chain.\nOther selected strand will be automatically skinned to generated bone chain\nWorks only on strands generated with hair tool - cylinders or strips geometry"
    bl_options = {"REGISTER", "UNDO"}

    absolete_start: bpy.props.FloatProperty(name="Absolute Start", description='Absolute offset from strand root location', default=0.0, min=0.0, soft_max=10.0, subtype='DISTANCE')
    relative_start: bpy.props.IntProperty(name="Chain relative start", description='Place first bone at some start distance from strand root location', default=60, min=0, max=90, subtype='PERCENTAGE')
    to_tip_bias: bpy.props.IntProperty(name="Bias toward Tip", description='Clump generated bones more toward strand tip', default=0, min=0, max=100, subtype='PERCENTAGE')
    assign_jiggle: BoolProperty(name="Assign Jiggle", description="Assign jiggle physics to bones", default=True)
    generation_mode: bpy.props.EnumProperty(name='Bone generation mode', description='Descibes how new bones will be generated',
        items=[
            ('SELECTED_ISLANDS', 'For selected stripes', 'Generate bone chain for all selected strans (mesh islands)'),
            ('ACTIVE_ISLAND', 'For active stripe only', 'Generate bone chain only for active stripe (island with active face/edge/vert)')
        ], default='ACTIVE_ISLAND')
    normalize_by_length: bpy.props.BoolProperty(name="Normalize by length", default=False, description="Bones count will depend on haircard length - longer strands will receive more bones in chain - up to 'Max Bones Count' for longest strand")
    bone_count: bpy.props.IntProperty(name="Bones count", description='', default=2, min=1, soft_max=8)
    max_bone_count: bpy.props.IntProperty(name="Max bones count", description='Maximum bones count for longest strand in normalized algorithm version', default=6, min=1, soft_max=16)

    # @classmethod
    # def poll(cls, context):
    #     obj = context.active_object
    #     return obj and obj.type == 'MESH' and obj.mode == 'EDIT' and obj.get('all_islands')

    def get_active_sel_islands(self, bm, all_islands):
        ''' returns: sel_islands_ids, active_island_idx - to pick from initialized: self.all_islands  '''

        active_element = bm.select_history.active
        active_vert = None

        # NOTE: Always find active island by active element (since resutls are cached and calculated only once)
        active_vert = active_element
        if isinstance(active_element, bmesh.types.BMFace):
            active_vert = active_element.verts[0]
        elif isinstance(active_element, bmesh.types.BMEdge):
            active_vert = active_element.verts[0]
        if not active_vert and self.generation_mode == 'ACTIVE_ISLAND':
            self.report({'WARNING'}, 'No active mesh element (vert, edge, face) found. Switching to \'All Selected Stripes\' mode')
            self.generation_mode = 'SELECTED_ISLANDS'

        # NOTE: from all islands, gete sel_islands (and other sel data)
        active_island_idx = None
        sel_islands_ids = []
        for i,island in enumerate(all_islands): # these from cache
            if bm.verts[island[0]].select: # island selected - if any vert in it is selected
                sel_islands_ids.append(i)

                # NOTE: find active island - by finding active elem
                if active_vert and not active_island_idx:
                    if any([v_id == active_vert.index for v_id in island]):
                        active_island_idx = i

        return sel_islands_ids, active_island_idx

    def invoke(self, context, event):
        obj = context.active_object
        if len(obj.data.uv_layers) < 2: # user should pick uv.. to island processing
            self.report({'WARNING'}, 'No second UV channel found! Cancelling')
            return {"CANCELLED"}

        self.prev_bones_names = [] # if operator redo - remove previously created bones
        self.load_settings(obj.ht_props.bone_gen_props)  #else from target mesh

        mesh = obj.data
        bm = bmesh.from_edit_mesh(mesh)

        # check if bm has some non manifold vertices or edges without adjacent faces
        for v in bm.verts:
            if not v.link_faces:
                self.report({'ERROR'}, 'Mesh has some non manifold vertices! Cancelling')
                return {"CANCELLED"}
        for e in bm.edges:
            if not e.link_faces:
                self.report({'ERROR'}, 'Mesh has some non manifold edges! Cancelling')
                return {"CANCELLED"}

        ht_uvs = bm.loops.layers.uv.get('HairTool_UV')
        if not ht_uvs:
            self.report({'ERROR'}, 'No \'HairTool UV\' layer found! Cancelling')
            return {"CANCELLED"}


        # NOTE : check if global cache exists, and if it is not outdated
        vert_face_cnt = (len(bm.verts),len(bm.faces))
        if not g_islands_data.get('all_islands') or g_islands_data['vert_cnt'] != vert_face_cnt:
            obj_bone_cache = obj.get('ht_mesh_to_bones_cache')
            uv_threshold, smooth_factor = 0.2, 0.8
            if obj_bone_cache:
                uv_threshold, smooth_factor = obj_bone_cache[0], obj_bone_cache[1]
            precalc_island_cache(bm,  uv_threshold=uv_threshold, smooth_factor=smooth_factor)
            self.report({'WARNING'}, "HairCards cache regenerated.")

        # NOTE: load data from global cache
        self.all_islands = g_islands_data['all_islands']
        self.longest_island_len = g_islands_data['longest_island_len']
        self.islands_len = g_islands_data['islands_len']
        island_vert_strips_co = g_islands_data['island_vert_strips_co']

        (self.sel_islands_ids, self.active_island_idx) = self.get_active_sel_islands(bm, self.all_islands)

        self.sel_strip_as_splines = {}
        for i in self.sel_islands_ids:
            vert_strip = island_vert_strips_co[i]
            pts_spline = SplineFlat(vert_strip)
            pts_spline.resample_pts(10, uniform=True) # TODO: maybe resampling should be customizable or something..
            pts_spline.cache_coefficients()
            self.sel_strip_as_splines[i] = pts_spline

        # bm.free()  # free and prevent further access
        bpy.ops.ed.undo_push() # damm - obj['all_islands'] is reverted anyway...
        return self.execute(context)

    def save_settings(self, save_props):
        for d in self.properties.bl_rna.properties.keys():
            if d == 'rna_type':
                continue
            setattr(save_props, d, getattr(self.properties, d))

    def load_settings(self, get_props):
        # what exception could occur here??
        for d in get_props.bl_rna.properties.keys():
            if d in {'name', 'rna_type'}:
                continue
            if hasattr(self.properties, d):
                setattr(self.properties, d, getattr(get_props, d))

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'generation_mode')
        col = layout.column(align=True)
        col.prop(self, 'absolete_start')
        col.prop(self, 'relative_start')
        layout.prop(self, 'to_tip_bias')
        layout.prop(self, 'normalize_by_length', toggle=True)
        if self.normalize_by_length:
            layout.prop(self, 'max_bone_count')
        else:
            layout.prop(self, 'bone_count')
        layout.prop(self, 'assign_jiggle', toggle=True)

    def get_set_rig(self, context, ht_props):
        active_obj = self.active_obj
        if ht_props.target_rig:
            rig = ht_props.target_rig
        else:
            rig_name = active_obj.name + '_Rig'
            rig = bpy.data.objects.get(rig_name)
            ht_props.target_rig = rig
        if rig:
            armature = rig.data
        else:
            armature = bpy.data.armatures.new(active_obj.name + "_Bones")
            rig = bpy.data.objects.new(active_obj.name + '_Rig', armature)
            rig.matrix_world = active_obj.matrix_world
            context.collection.objects.link(rig)
            #   also add armature modifier to active_obj
            # arma_mods = [mod for mod in active_obj.modifiers if mod.type == 'ARMATURE']
            mod_armature = active_obj.modifiers.new('Hair Armature', type='ARMATURE')
            mod_armature.object = rig
            # mayeb should parent too?
            set_parent(active_obj, rig)

        # rig.show_in_front = True
        is_in_scene = context.scene.objects.get(rig.name)
        if not is_in_scene:
            context.scene.collection.objects.link(rig)
        return rig

    def create_edit_bones(self, pin_bone_name, bone_pts_local):
        created_bones = []
        created_bones_names = []
        to_rig_space = self.rig.matrix_world.inverted() @ self.active_obj.matrix_world
        prev_pt = bone_pts_local[0]
        prev_bone = None
        for i,bone_pt in enumerate(bone_pts_local[1:]):
            bone = self.rig.data.edit_bones.new('HT_Bone')
            bone.envelope_distance = 0.0001
            bone.head_radius = 0.002
            bone.tail_radius = 0.002
            if i == 0 and pin_bone_name :
                root_bone = self.rig.data.edit_bones.get(pin_bone_name)
                if root_bone:
                    bone.parent = root_bone
            created_bones.append(bone)
            created_bones_names.append(bone.name)
            bone.head = to_rig_space @ prev_pt # from active obj local, to rig local
            bone.tail = to_rig_space @ bone_pt # from active obj local, to rig local
            if prev_bone:
                bone.parent = prev_bone
                bone.use_connect = True
            prev_pt = bone_pt
            prev_bone = bone
        self.prev_bones_names.extend(created_bones_names) # backup - we remove new bones on operator redo (cos blender cant perform it properly)
        return created_bones, created_bones_names

    def gen_bone_chain_pos(self, bm, pts_spline, bone_cnt, start_absolute_offset):
        ''' Returns:
            bone chain points.co,
            bones_co_kd: boones kd_tree for weights gen'''
        bones_co_kd = kdtree.KDTree(bone_cnt+1) # holds bone points

        ids = np.linspace(0, 1, bone_cnt+1)
        if self.to_tip_bias:
            exponent = self.to_tip_bias/100*1.5+1 # remap (0,1) to (1,3)
            ids = 1-np.power(-ids+1, exponent)

        mi = start_absolute_offset
        ids = ids*(1-mi) + mi # remap from (0,1) to (self.absolute_start, 1)
        mi = self.relative_start/100
        ids = ids*(1-mi) + mi # remap from (0,1) to (self.relative_start/100, 1)

        island_bone_pts_local = [Vector(pt) for pt in pts_spline.evaluate(ids)]

        for i, pt in enumerate(island_bone_pts_local):
            bones_co_kd.insert(pt, i)

        bones_co_kd.balance()
        return island_bone_pts_local, bones_co_kd

    def generate_weights(self, bm, created_bones, pin_bone, islands, bones_co_kd):
        active_obj = self.active_obj
        deform = bm.verts.layers.deform.active

        to_rig_space_mat = self.rig.matrix_world.inverted() @ active_obj.matrix_world
        to_active_space = to_rig_space_mat.inverted() # back from rig to active object space
        pin_group = active_obj.vertex_groups.get(pin_bone)
        if not pin_group:
            pin_group = active_obj.vertex_groups.new(name=pin_bone)

        # NOTE: if vertex group already exists, for created bones, remove it
        for bone in created_bones:
            bgroup = active_obj.vertex_groups.get(bone.name)
            if bgroup:
                for v in bm.verts:
                    for group_id in v[deform].keys():
                        if group_id == bgroup.index:
                            del v[deform][group_id]
                            v[deform][pin_group.index] = 1


        bone_cnt = len(created_bones)
        for island in islands:
            for v_id in island:
                v = bm.verts[v_id]
                co, bone_idx, dist = bones_co_kd.find(v.co)
                bones_to_check = {max(0,bone_idx-1), min(bone_idx, bone_cnt-1)} # check two closest bones to  pt
                for b_idx in bones_to_check:
                    bone = created_bones[b_idx]
                    ## factor = as from first pt
                    pt, fac = intersect_point_line(v.co, to_active_space @ bone.head, to_active_space @ bone.tail)
                    bgroup = active_obj.vertex_groups.get(bone.name)
                    if not bgroup:
                        bgroup = active_obj.vertex_groups.new(name=bone.name)

                    # remap fac from  (0,1) - so that 0.5 is at max influcence
                    if b_idx == bone_cnt-1 and fac > 0.5:   # we are 50% toward tip
                        weight = 1.0  # the tip verts always get 1
                    else:
                        x = abs(fac-0.5)  #  remap from 1 at head, to 0 at center - 0.5 at both ends

                        weight = -abs(x)+1  # flip from 0 center, -0.5 at both ends, to 1 center, and +0.5 at both ends
                        # weight = 0.5*(math.cos(fac * 3.14)+1) #smooth fallof
                        # weight = clamp(weight, -1, 1)
                    # vert_groups = v[deform].keys() # current vert groups
                    v[deform][bgroup.index] = weight
                    if b_idx == 0 and fac <= 0.5:  # assign pin_bone vert_group to first half, of first bone
                        v[deform][pin_group.index] = 1-weight

    def set_lock_weight(self, all_new_bones_names, lock):
        '''Lock so that normalize wont affect generated bones weights'''
        for bones_chain in all_new_bones_names:
            vg = self.active_obj.vertex_groups.get(bones_chain[0])
            if vg:
                vg.lock_weight = lock # so that normalize wont affect it

    def execute(self, context):
        active_obj = context.active_object
        self.active_obj = active_obj
        ht_props = active_obj.ht_props

        try:    # prevent poll error
            bpy.ops.object.vertex_group_remove_from(use_all_groups=True)
        except:
            pass
        mesh = active_obj.data
        bm = bmesh.from_edit_mesh(mesh)
        bm.faces.ensure_lookup_table() # gives error due same function used in invoke?
        bm.verts.ensure_lookup_table()

        # load values calculated in invoke
        active_island_idx =self.active_island_idx

        if active_island_idx is None:
            self.report({'WARNING'}, "No active mesh element (vert, edge, face) found. Switching to 'All Selected Stripes' mode")
            self.generation_mode = 'SELECTED_ISLANDS'

        if not self.sel_islands_ids:
            return {'CANCELLED'}

        if self.generation_mode == 'ACTIVE_ISLAND' and active_island_idx is not None:
            active_island_ids = [active_island_idx]
        else:
            active_island_ids = self.sel_islands_ids

        # NOTE: calc bones count based on absolute_start, relative_start, normalize_by_length
        rel_shorten_fac = 1-self.relative_start/100


        # NOTE: detect existing or create new rig for hair. And setup armature modifier on active_obj
        self.rig  = self.get_set_rig(context, ht_props )

        # NOTE: generate bone (head.co, tail.co) by traveling over island row verts.
        is_bone_pts_co = []  # list of lists of tuples (head.co, tail.co)
        is_bones_co_kd = []
        for island_idx in active_island_ids:
            ilen = self.islands_len[island_idx]
            bone_cnt_norm_fac = max(ilen-self.absolete_start, 0)/self.longest_island_len*rel_shorten_fac

            if self.normalize_by_length:
                bone_cnt_norm = max(int(self.max_bone_count*bone_cnt_norm_fac+.6), 0)
            else:   # skip very short bone chains, even if no normalize
                bone_cnt_norm = self.bone_count if bone_cnt_norm_fac > 0.1 else 0

            bones_absolute_offset = min(self.absolete_start/ilen, 1.)

            if bone_cnt_norm < 1:
                continue
            island_bone_pts_co, island_bones_co_kd = self.gen_bone_chain_pos(bm, self.sel_strip_as_splines[island_idx], bone_cnt_norm, bones_absolute_offset)
            is_bone_pts_co.append(island_bone_pts_co)
            is_bones_co_kd.append(island_bones_co_kd)

        # NOTE: create actual bones
        context.view_layer.objects.active = self.rig
        bpy.ops.object.mode_set(mode='EDIT')

        # NOTE: if operator redo - remove previously created bones
        if self.prev_bones_names:
            for bname in reversed(self.prev_bones_names):
                pbone = self.rig.pose.bones.get(bname)
                if pbone:
                    pbone.ht_props.jiggle_index = -1
                    pbone.bone_group = None
                ebone = self.rig.data.edit_bones.get(bname)
                if ebone:
                    self.rig.data.edit_bones.remove(ebone)
            self.prev_bones_names.clear()

        # NOTE: create bones from bones pts
        all_new_bones_names = []
        islands_new_bones = []
        for island_bone_pts_co in is_bone_pts_co:
            island_new_bones, created_bones_names = self.create_edit_bones(ht_props.pin_bone, island_bone_pts_co) # fills self.prev_bones_names; also destroys bm (by switch to edit mode)
            all_new_bones_names.append(created_bones_names)
            islands_new_bones.append(island_new_bones)
            self.prev_bones_names.extend(created_bones_names) # backup - we remove new bones on operator redo (cos blender cant perform it properly)
        # bpy.ops.object.editmode_toggle() #reqyured fir bones to show in view
        context.view_layer.objects.active = self.active_obj
        # bmesh.update_edit_mesh(mesh, True)


        # NOTE: assign weights to all hair strips
        if all_new_bones_names: # at least one new bone created
            # Ensure custom data exists.
            bm.verts.layers.deform.verify()
            # bm.verts.ensure_lookup_table()
            pin_group_name = ht_props.pin_bone if ht_props.pin_bone else 'ht_root'
            for active_island_idx, island_bones_co_kd, island_new_bones in zip(active_island_ids, is_bones_co_kd, islands_new_bones):
                if self.generation_mode == 'ACTIVE_ISLAND': #   skin - all selected to active bone chain
                    islands_to_skin = [self.all_islands[idx] for idx in self.sel_islands_ids]
                else: # each bone chain has its own island
                    islands_to_skin = [self.all_islands[active_island_idx]]

                self.generate_weights(bm, island_new_bones, pin_group_name, islands_to_skin, island_bones_co_kd)
                    # bm.to_mesh(mesh) # for obj mode
            bmesh.update_edit_mesh(mesh, destructive=True) #save weights
            # bm.free()  # free and prevent further access Note for editmesh.


            # NOTTE: finally smooth ver groups, and assign jiggle physics
            self.rig.select_set(True)
            bpy.ops.object.mode_set(mode='WEIGHT_PAINT')
            bpy.context.object.data.use_paint_mask_vertex = True
            bpy.ops.object.vertex_group_smooth(group_select_mode='ALL', factor=0.4, repeat=1)
            # active_obj.vertex_groups.active = active_obj.vertex_groups[created_bones_names[0].name]

            # NOTE: normalize without affecting root of bone chains
            self.set_lock_weight(all_new_bones_names, True)
            try:
                bpy.ops.object.vertex_group_normalize_all(lock_active=False)
            except Exception as e:
                print(e)
            self.set_lock_weight(all_new_bones_names, False)

            bpy.ops.object.mode_set(mode='OBJECT')

            context.view_layer.objects.active = self.rig
            bpy.ops.object.mode_set(mode='OBJECT')
            context.view_layer.objects.active = active_obj

            active_obj.update_tag()
            self.rig.data.edit_bones.update()
            self.rig.pose.bones.update()
            # context.view_layer.update()
            self.rig.pose.use_auto_ik = True
            active_obj.ht_props.target_rig = self.rig

            self.save_settings(active_obj.ht_props.bone_gen_props)  #else from target mesh
            # save island cache, since operator redoo will clear it if inside invoke...
            # save_island_data_to_obj(active_obj, self.all_islands, self.all_islands_curl_verts, self.islands_len, self.longest_island_len)
            bpy.ops.object.mode_set(mode='EDIT')

            # NOTE: add springs
            if self.assign_jiggle:
                pbones = [self.rig.pose.bones.get(bname) for bone_chain in all_new_bones_names for bname in bone_chain]
                if bpy.app.version < (4, 0, 0):
                    OBJECT_OT_Assign_Bones_Jiggle.assign_jiggle_to_bones(self.rig, pbones)
                else: # for B.4
                    OBJECT_OT_Assign_Bones_JiggleB4.assign_jiggle_to_bones(self.rig, pbones)
        else:
            self.report({'WARNING'}, 'No new bones created')
        return {"FINISHED"}


class HTOOL_OT_TextureUVGradient_Bake(bpy.types.Operator):
    bl_idname = "object.uvgradient_bake"
    bl_label = "Generate texture Gradient "
    bl_description = "Generate texture Gradient "
    bl_options = {"REGISTER","UNDO"}

    # image resolution enum
    image_res: bpy.props.EnumProperty(
        name="Image Resolution",
        items=(
            ('256', "256", "256"),
            ('512', "512", "512"),
            ('1024', "1024", "1024"),
            ('2048', "2048", "2048"),
            ),
        default="1024"
    )
    out_img_name: bpy.props.StringProperty(name="Output Image Name", default="UVGradient")

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH'

    @staticmethod
    def import_uv_hair_mat(mat_name, force_update=False):
        old_mat = bpy.data.materials.get(mat_name)
        hair_baking_file = path.dirname(__file__)+"\hair_baking\hair_bake_3.0.blend"
        if not old_mat or force_update:
            with bpy.data.libraries.load(hair_baking_file) as (data_from, data_to):
                data_to.materials = [mat for mat in data_from.materials if mat == mat_name]

            if old_mat:
                old_mat.user_remap(data_to.materials[0])
                # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
                # we have to remove data block that holds the name. Then use that name, on new data block...
                bpy.data.materials.remove(old_mat)
                data_to.materials[0].name = mat_name

            return data_to.materials[0]
        else:
            return old_mat

    def execute(self, context):
        obj = context.active_object
        mat_name = 'UVGradient'
        uv_grad_mat = bpy.data.materials.get(mat_name)
        if not uv_grad_mat:
            uv_grad_mat = self.import_uv_hair_mat(mat_name)
            # create new output image

        # backup material slots of obj
        old_mats = [slot.material for slot in obj.material_slots if slot.material]
        # replace obj materials with uv_grad_mat
        obj.data.materials.clear()
        obj.data.materials.append(uv_grad_mat)

        img_res = int(self.image_res)
        out_img = bpy.data.images.new(name=self.out_img_name, width=img_res, height=img_res, alpha=True, float_buffer=False)
        # add image node to uv_grad_mat
        texture_node = uv_grad_mat.node_tree.nodes.get("Image Texture")
        # set texture node iamge to out_img
        texture_node.image = out_img
        uv_grad_mat.node_tree.nodes.active = texture_node

        # bpy.ops.object.bake(type='EMIT')
        prev_render_engine = context.scene.render.engine
        context.scene.render.engine = "CYCLES"
        padding_size = int(img_res/64)
        print("padding_size", padding_size)

        bpy.ops.object.bake(type='EMIT', filepath="", pass_filter={'NONE'},
                                        width=img_res, height=img_res, margin=padding_size,
                                        use_selected_to_active=False,
                                        save_mode='INTERNAL', use_clear=True, use_cage=False,
                                        use_split_materials=False, use_automatic_name=False)

        context.scene.render.engine = prev_render_engine
        # restore old_mats
        obj.data.materials.clear()
        for mat in old_mats:
            obj.data.materials.append(mat)
        return {"FINISHED"}


class HTOOL_OT_GenerateWeight(bpy.types.Operator):
    bl_label = "Add Weight Gradient"
    bl_idname = "object.ribbon_weight"
    bl_description = "Generate vertex weight gradient from root to tip of each hair card"
    bl_options = {"REGISTER", "UNDO"}

    createNewVC: BoolProperty(name="As New Weight Group", description="Adds new vertex weight layer instead of overriding active one", default=False)
    gradientFalloff: FloatProperty(name="Gradient Falloff", description="Gradient Falloff", default=0,
                                    min=-1, max=1, subtype='PERCENTAGE')
    inverted: BoolProperty(name="Invert", description="Invert gradient", default=False)

    def execute(self, context):
        ribbonMesh = context.active_object
        if len(ribbonMesh.data.uv_layers)<2:
            self.report({'INFO'}, 'No second UV channel found!')
            return {"CANCELLED"}
        mesh = ribbonMesh.data
        if not ribbonMesh.vertex_groups or self.createNewVC:
            ribbonMesh.vertex_groups.new(name = 'gradient')
        if self.createNewVC:
            vg = ribbonMesh.vertex_groups[-1]
        else:
            vg = ribbonMesh.vertex_groups.active
        vertUV = {}
        cpow = calc_exponent(self.gradientFalloff)
        for face in mesh.polygons:
            for vert_idx, loop_idx in zip(face.vertices, face.loop_indices):
                uv_coords = mesh.uv_layers['HairTool_UV'].data[loop_idx].uv
                vertUV[vert_idx]= (uv_coords.x, uv_coords.y)
        selected_verts = [v for v in mesh.vertices if v.select]
        verts = selected_verts if mesh.use_paint_mask_vertex else mesh.vertices

        bm = bmesh.new()
        bm.from_mesh(mesh)
        # Select all loose vertices
        valid_verts = []
        bm.verts.ensure_lookup_table()
        for vert in verts:
            bm_vert = bm.verts[vert.index]
            if len(bm_vert.link_edges) != 0 and len(bm_vert.link_faces) != 0:
                valid_verts.append(vert)
        bm.free()

        for vert in valid_verts:
            x,y = vertUV[vert.index]
            weight = abs(max(min(1-y, 1), 0))  # d over box height
            if self.inverted:
                weight = 1-weight
            w2 = math.pow(weight, cpow)
            vg.add([vert.index],w2 , "REPLACE")
        if self.createNewVC:
            ribbonMesh.vertex_groups.active = ribbonMesh.vertex_groups[-1]
        bpy.ops.object.mode_set(mode='WEIGHT_PAINT')
        return {"FINISHED"}

class HTOOL_OT_GenerateVerColorGradient(bpy.types.Operator):
    bl_label = "Generate vertext gradient"
    bl_idname = "object.ribbon_vert_color_grad"
    bl_description = "Generate vertex color gradient from root to tip of each hair card"
    bl_options = {"REGISTER", "UNDO"}

    createNewVC: BoolProperty(name="Add new VColor", description="Adds new vertex color layer instead of overriding active one", default=False)
    gradientFalloff: FloatProperty(name="Gradient Falloff", description="Gradient Falloff", default=0,
                                  min=-1, max=1, subtype='PERCENTAGE')

    def execute(self, context):
        ribbonMesh = context.active_object
        if len(ribbonMesh.data.uv_layers)<2:
            self.report({'INFO'}, 'No second UV channel found!')
            return {"CANCELLED"}
        mesh = ribbonMesh.data
        if not mesh.vertex_colors or self.createNewVC:
            mesh.vertex_colors.new(name="Gradient")

        # color_layer = mesh.vertex_colors["Col"]
        # or you could avoid using the color_layer name
        if self.createNewVC:
            color_layer = mesh.vertex_colors[-1]
        else:
            color_layer = mesh.vertex_colors.active
        cpow = calc_exponent(self.gradientFalloff)
        selected_faces = [f for f in mesh.polygons if f.select]
        faces = selected_faces if mesh.use_paint_mask_vertex else mesh.polygons
        for face in faces:
            for loop_idx in face.loop_indices:
                x,y = mesh.uv_layers['HairTool_UV'].data[loop_idx].uv
                col = abs(max(min(y, 1), 0))
                col2 = math.pow(col, cpow)
                color_layer.data[loop_idx].color = [col2,col2,col2,1]
        if self.createNewVC:
            mesh.vertex_colors.active = mesh.vertex_colors[-1]
        bpy.ops.object.mode_set(mode='VERTEX_PAINT')
        ribbonMesh.update_tag()
        return {"FINISHED"}

class HTOOL_OT_GenerateVerColorRandom(bpy.types.Operator):
    bl_label = "Generate random vertex colors"
    bl_idname = "object.ribbon_vert_color_random"
    bl_description = "Makes each ribbon vertex color unique"
    bl_options = {"REGISTER", "UNDO"}

    Seed: IntProperty(name="Noise Seed", default=1, min=1, max=1000)
    mergeSimilar: BoolProperty(name="Merge Similar", description="Merge Simila", default=False)
    createNewVC: BoolProperty(name="Add new VColor", description="Adds new vertex color layer instead of overriding active one", default=False)


    def execute(self, context):
        np.random.seed(self.Seed)
        obj = bpy.context.active_object
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)  # Go to edit mode
        bpy.ops.mesh.select_all(action="DESELECT")  # unselect everything

        bm = bmesh.from_edit_mesh(obj.data)  # load mesh
        bm.faces.ensure_lookup_table()

        if self.createNewVC or not obj.data.vertex_colors:
            obj.data.vertex_colors.new(name="Random Color")

        if self.createNewVC:
            color_layer =bm.loops.layers.color[-1]
        else:
            color_layer =bm.loops.layers.color.active

        faces_lislands = []
        faces = bm.faces

        while faces:
            faces[0].select = True  # select 1st face
            bpy.ops.mesh.select_linked()  # select all linked faces makes a full loop
            faces_lislands.append([f for f in faces if f.select])
            bpy.ops.mesh.hide(unselected=False)  # hide the detected loop
            faces = [f for f in bm.faces if not f.hide]  # update faces

        bpy.ops.mesh.reveal()  # unhide all faces
        islandColor = {} #dictt will contain faceCount : Color pairs
        for facesIsland in faces_lislands: # face islands list
            if self.mergeSimilar:
                islandFaceCount = len(facesIsland)
                if islandFaceCount in islandColor.keys():
                    random_color = islandColor[islandFaceCount]
                else:
                    random_color = np.random.random_sample(3).tolist()
                    islandColor[islandFaceCount] = random_color  #asign faceCount : Color"
            else:
                random_color = np.random.random_sample(3).tolist()
            for face in facesIsland: #island faces
                for loop in face.loops:
                        loop[color_layer] = random_color+[1]

        bm.free()  # free and prevent further access
        if self.createNewVC:
            obj.data.vertex_colors.active = obj.data.vertex_colors[-1]
        self.report({'INFO'}, 'Vertex colors have been generated')
        bpy.ops.object.mode_set(mode='VERTEX_PAINT')
        return {"FINISHED"}



# operator to create edge strips from mesh islands
class HTOOL_OT_CreateEdgeStrips(bpy.types.Operator):
    bl_label = "Haircards to Mesh loops"
    bl_idname = "object.create_edge_strips"
    bl_description = "Create edge loop strips from mesh islands. Can be converted to curves (W > Convert To > Curve and then W > Convert To > Curves). After that profile, can be added."
    bl_options = {'REGISTER', 'UNDO'}

    uv_name: bpy.props.StringProperty(name="UV name", default="UVMap", description="UV map to use for detecting of haircard orientation")
    uv_threshold: bpy.props.FloatProperty(name="Simplification", description='Merge points with UV.y difference below this value. Higher value - stronger simplification', default=0.01, min=0.0, max=1.0, precision=3)
    smooth_factor: bpy.props.FloatProperty(name="Smooth Factor", default=0.5, description="Smooth generated strips", min=0, max=1)
    reverse: bpy.props.BoolProperty(name="Reverse Direction", default=False)

    def draw(self, context):
        layout = self.layout
        # layout.prop_search(self, "uv_name", context.active_object.data, "uv_layers", text="UV Map")
        layout.prop(self, "uv_threshold")
        layout.prop(self, "smooth_factor")
        layout.prop(self, "reverse", toggle=True)

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    @staticmethod
    def create_edge_strip_positions_list(bm, out_verts_pos):
        new_verts = []
        # Create new vertices from the list of vertices
        for v_co in out_verts_pos:
                new_vert = bm.verts.new(v_co)
                new_verts.append(new_vert)

            # Create edges between the new vertices to form a strip
        for i in range(len(new_verts) - 1):
                bm.edges.new((new_verts[i], new_verts[i+1]))
        # bmesh.update_edit_mesh(me)

    def execute(self, context):
        sel_objs = context.selected_objects[:]
        if not sel_objs:
            self.report({'WARNING'}, "Select at least one mesh haircard object for conversion. Cancelling")
            return {'CANCELLED'}

        bpy.ops.object.select_all(action='DESELECT')

        convert_cnt = 0
        for obj in sel_objs:
            if obj.type != 'MESH':
                self.report({'WARNING'}, f"{obj.name} object is not mesh - skipping")
                continue

            bm = bmesh.new()
            bm.from_mesh(obj.data)

            out_bm = bmesh.new()

            island_strips = precalc_island_cache(bm, uv_threshold=self.uv_threshold, smooth_factor=self.smooth_factor, reverse=self.reverse)
            for e_island in island_strips:
                self.create_edge_strip_positions_list(out_bm, e_island)

            new_name = obj.name + "_strip"
            me = bpy.data.meshes.new(new_name)
            out_bm.to_mesh(me)
            out_bm.free()
            bm.free()

            strip_obj = bpy.data.objects.new(new_name, me)
            context.collection.objects.link(strip_obj)
            context.view_layer.objects.active = strip_obj

            strip_obj.select_set(True)
            strip_obj.matrix_world = obj.matrix_world
            strip_obj.show_in_front = True

            convert_cnt += 1

        self.report({'INFO'}, f"Converted {convert_cnt} objects")
        return {'FINISHED'}



def three_d_bincount_add(out, e_k, weights):
    out[:, 0] += np.bincount(e_k, weights[:, 0], out.shape[0])
    out[:, 1] += np.bincount(e_k, weights[:, 1], out.shape[0])
    out[:, 2] += np.bincount(e_k, weights[:, 2], out.shape[0])


def new_bincount(out, e_k, weights=None):
    out += np.bincount(e_k, weights, out.shape[0])


def close_bmesh(context,  bm, source):
    bm.normal_update()
    if context.object.mode == "EDIT":
        bmesh.update_edit_mesh(source, tessface=False, destructive=False)
    else:
        bm.to_mesh(source)
        bm.free()
        # source.update()


def get_bmesh(context, mesh):
    bm = bmesh.new()
    if context.active_object.mode == 'OBJECT':
        bm.from_mesh(mesh)
    elif context.active_object.mode == 'EDIT':
        bm = bmesh.from_edit_mesh(mesh)
    return bm


def smooth_mesh_data(context, bm, data, iteration_amount):
    # bm.verts.ensure_lookup_table()
    # bm.edges.ensure_lookup_table()

    # o_co = np.array([vert.co for vert in bm.verts])
    o_co = data
    vert_count = len(bm.verts)

    edge_keys = np.array([[edge.verts[0].index, edge.verts[1].index] for edge in bm.edges])
    # freeze_verts = np.array([vert.is_boundary or not vert.is_manifold for vert in bm.verts])  # manifold for pocket sewing edges

    p_co = np.copy(o_co)  # current new vertex position

    #sum edges that have vert X  and either end of edge (Va, Vb)
    v_connections = np.zeros(vert_count, dtype=np.float32)
    new_bincount(v_connections, edge_keys[:, 1])
    new_bincount(v_connections, edge_keys[:, 0])

    for x in range(iteration_amount):
        q_co = np.copy(p_co)  # prevoius iteration vert position
        p_co.fill(0)
        #calc average position weighted...
        if len(p_co.shape) == 1:
            new_bincount(p_co, edge_keys[:, 1], q_co[edge_keys[:, 0]])
            new_bincount(p_co, edge_keys[:, 0], q_co[edge_keys[:, 1]])

            p_co = p_co / v_connections  # new aver posision
        else:
            three_d_bincount_add(p_co, edge_keys[:, 1], q_co[edge_keys[:, 0]])
            three_d_bincount_add(p_co, edge_keys[:, 0], q_co[edge_keys[:, 1]])

            p_co = p_co / v_connections[:, None]  # new aver posision

        # p_co[freeze_verts] = q_co[freeze_verts]  # not selected ver - reset to original pos o_co
    # source.vertices.foreach_set('co', p_co.ravel())
    # for v in bm.verts:
    #     v.co = p_co[v.index]
    return p_co

def remap_01(arr):
    return list(np.interp(arr, (np.nanmin(arr), np.nanmax(arr)), (0, 1)))

class HTOOL_OT_AoToVcol(ModifierPopupPanel, bpy.types.Operator):
    bl_idname = "hsystem.add_ao_mod"
    bl_label = "Generate AO (modifier)"
    bl_description = "Generate vertex colors Ambient Occlusion attribute, using geometry nodes modifier.\nToo see its effect, you have to connect 'AO' attribute node, in your hair shader"
    bl_options = {"REGISTER"}

    # offset: bpy.props.FloatProperty(name="Surface Offset", default=0.01, min=0.0, max=1.0 )
    # dist: bpy.props.FloatProperty(name="AO Distance", default=1.0, min=0.0)
    # smooth_iter: bpy.props.IntProperty(name='Smooth', description='', default=1, min=0, max=12)
    # bounces: bpy.props.IntProperty(name='Bounces', description='', default=1, min=0, max=5)
    # contrast: bpy.props.FloatProperty(name='Contrast', description='', default=0.3, min=0, max=1)
    # side: bpy.props.EnumProperty(name='Side', description='Shoot light rays from Front of faces, back or both',
    #     items=[
    #         ('FRONT', 'Front', ''),
    #         ('BACK', 'Back', ''),
    #         ('BOTH', 'Both', '')
    #     ], default='FRONT')

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and (get_profile_mod(obj) or obj.type == 'MESH')

    def draw(self, context):
        ModifierPopupPanel.modifier_name = self.modifier_name
        ModifierPopupPanel.draw(self, context)

    def invoke(self, context, event):
        # self.depsgraph = context.evaluated_depsgraph_get()
        # self.bvh = get_obj_mesh_bvht(context.active_object, self.depsgraph, applyModifiers=True, world_space=True) #fastest
        # self.other_bvh = [get_obj_mesh_bvht(obj, self.depsgraph, applyModifiers=True, world_space=True) for obj in context.selected_objects if obj.type == 'MESH' and obj != context.active_object]
        return self.execute(context)

    def execute(self, context):
        obj = context.active_object
        ao_node_group_name = 'HT_Mesh_AO'
        gen_strands_ng = get_node_group(ao_node_group_name)
        # detectl if old version of mod is present
        for mod in obj.modifiers:
            if mod.type == 'NODES' and mod.node_group.name == 'HT_Mesh_AO':
                self.modifier_name = mod.name
                break
        else:
            short_hair_mod = obj.modifiers.new(name='Mesh AO with Bounces', type='NODES')
            short_hair_mod.node_group = gen_strands_ng
            self.modifier_name = short_hair_mod.name
        return context.window_manager.invoke_props_dialog(self, width=220)
        # return {'FINISHED'}
        # NOTE: old 'manual' ao version
        me = obj.data
        bm = bmesh.new()   # create an empty BMesh
        bm.from_mesh(me)   # fill it in from a Mesh

        if 'AO' in bm.loops.layers.color.keys():
            color_layer = bm.loops.layers.color['AO']
        else:
            color_layer = bm.loops.layers.color.new('AO')

        per_face_ao = []
        faces_hit_ids = []
        ray_cast_count = 18 if self.side == 'BOTH' else 9
        mat = obj.matrix_world
        for f in bm.faces:
            #convert to pt by .to_4d().  Then to 3d
            ray_casts = self.simple_sample((mat@f.calc_center_median().to_4d()).xyz, (mat@Vector(f.normal[:]+(0,))).xyz)
            # total distance loc, norm, idx, dist
            base_ao = sum(dist / self.dist if loc and dist < self.dist else 1 for (loc, norm, idx, dist, is_active_obj) in ray_casts)  # ! fix 1.0 to self.dist
            faces_hit_ids.append([(idx, is_active_obj) if loc and dist < self.dist else (-1, is_active_obj) for (loc, norm, idx, dist, is_active_obj) in ray_casts])
            per_face_ao.append(base_ao/ray_cast_count)

        import math
        prev_ao = remap_01(per_face_ao)
        #* average current color + other faces that were ray_hit from current vert
        for n in range(self.bounces):
            next_bounce_ao = []
            for src_f_id, f_hits in enumerate(faces_hit_ids):  # sum per_face ao, from raycast hits
                sum_aos = 0
                miss_count = 0
                for (f_id, is_active_obj) in f_hits:
                    if is_active_obj: #use light cache
                        sum_aos += prev_ao[f_id] if f_id > -1 else 1
                    else:  # no light cache, so just 0 or 1
                        sum_aos += 0 if f_id > -1 else 1 # add zero - hited solid face...

                bounce_ao = sum_aos / ray_cast_count + prev_ao[src_f_id]/2
                # next_bounce_ao.append(bounce_ao)  # max to prevent darkening
                next_bounce_ao.append(max(bounce_ao, prev_ao[src_f_id]))  # max to prevent darkening

            # prev_ao = next_bounce_ao.copy()
            prev_ao = remap_01(next_bounce_ao)

        ao_per_vert = [0.0]*len(bm.verts)
        for i,f in enumerate(bm.faces):
            for v in f.verts:
                ao_per_vert[v.index] += prev_ao[i]

        for v_id, v in enumerate(bm.verts):
            f_cnt = len(v.link_faces)
            if f_cnt > 0:
                ao_per_vert[v_id] /= f_cnt
            else:
                ao_per_vert[v_id] = 1.0

        smoothed_ao = smooth_mesh_data(context, bm, np.array(ao_per_vert), self.smooth_iter)
        smoothed_ao -= np.nanmin(smoothed_ao)  #normalize
        smoothed_ao /= np.nanmax(smoothed_ao)  # normalize cd
        smoothed_ao = np.power(smoothed_ao, self.contrast*1.5 + 0.5)
        d3_ao = np.repeat(smoothed_ao[:, None], [3], axis=1)  # repeats 3x, row element
        for v, sm_ao in zip(bm.verts, list(smoothed_ao)):
            for loop in v.link_loops:
                loop[color_layer] = [sm_ao, sm_ao, sm_ao, 1]  # rgb + a
        bm.to_mesh(me)
        bm.free()  # free and prevent further access
        obj.data.update()
        return {"FINISHED"}

    def raycast_objs(self, ray_orig, ray_dir):
        # loc, norm, idx, dist, + (active_obj)
        rayc = self.bvh.ray_cast(ray_orig, ray_dir)
        if not rayc[0] and self.other_bvh:  # test other selected objs
            for sel_obj_bvj in self.other_bvh:
                rayc = sel_obj_bvj.ray_cast(ray_orig, ray_dir)
                if rayc[0]:
                    return rayc + (False,)
        return rayc + (True,)  # loc, norm, idx, dist, active_obj?

    def simple_sample(self, center, normal):
        component_A = normal.cross(normal.yxz).normalized() if normal.yxz != normal else normal.cross(normal.zyx).normalized()
        component_B = normal.cross(component_A).normalized()
        above_offset = center+self.offset * normal
        below_offset = center-self.offset * normal
        assert abs(normal.cross(component_A).dot(normal)) < 0.001
        casts = []
        if self.side != 'BACK': #so in case of both or front
            casts.append(self.raycast_objs(above_offset, normal))
            casts.append(self.raycast_objs(above_offset, 2*normal + component_A))
            casts.append(self.raycast_objs(above_offset, 2*normal - component_A))
            casts.append(self.raycast_objs(above_offset, 2*normal + component_B))
            casts.append(self.raycast_objs(above_offset, 2*normal - component_B))

            casts.append(self.raycast_objs(above_offset, normal + component_A + component_B))
            casts.append(self.raycast_objs(above_offset, normal + component_A - component_B))
            casts.append(self.raycast_objs(above_offset, normal - component_A + component_B))
            casts.append(self.raycast_objs(above_offset, normal - component_A - component_B))
        if self.side != 'FRONT':
            casts.append(self.raycast_objs(below_offset, -1*normal))
            casts.append(self.raycast_objs(below_offset, -2*normal + component_A))
            casts.append(self.raycast_objs(below_offset, -2*normal - component_A))
            casts.append(self.raycast_objs(below_offset, -2*normal + component_B))
            casts.append(self.raycast_objs(below_offset, -2*normal - component_B))

            casts.append(self.raycast_objs(below_offset, -1*normal + component_A + component_B))
            casts.append(self.raycast_objs(below_offset, -1*normal + component_A - component_B))
            casts.append(self.raycast_objs(below_offset, -1*normal - component_A + component_B))
            casts.append(self.raycast_objs(below_offset, -1*normal - component_A - component_B))
        return casts  # loc, norm, idx, dist, active_obj?


def get_linked_faces(f):
    link_faces = [f]
    f.tag = True
    new_faces = [f]
    while True:
        new_edges = [e for new_face in new_faces for e in new_face.edges if len(e.link_faces) == 2]
        new_faces = []
        for e in new_edges:
            for f in e.link_faces:
                if not f.tag:
                    new_faces.append(f)
                    f.tag = True
        if not new_faces:
            break #break while
        else:
            link_faces.extend(new_faces)
    return link_faces


def get_face_islands(bm):
    ''' Damm it slow'''
    next_face = bm.faces[0]
    islands = []
    face_count = len(bm.faces)
    scanned_faces_count = 0
    while True:
        islands.append(get_linked_faces(next_face))
        scanned_faces_count += len(islands[-1])
        if scanned_faces_count >= face_count:
            break
        non_tag_faces = [f for f in bm.faces if not f.tag]
        if not non_tag_faces:
            break
        else:
            next_face = non_tag_faces[0]
    return islands

UVS = []
def get_uv_layers(self, context):
    active_obj = bpy.context.active_object
    target = bpy.data.objects[active_obj.ht_props.target_obj]
    uv_layers = []
    for i,uv_layer in enumerate(target.data.uv_layers):
        uv_layers.append((uv_layer.name, uv_layer.name, ""))
    global UVS
    if len(UVS) != len(uv_layers):
        UVS = uv_layers
    return uv_layers

class HTOOL_OT_UVSampling(bpy.types.Operator):
    bl_idname = "object.uv_sample_from_target"
    bl_label = "Sample UV from target"
    bl_description = "Sample UV from target object and write it to new UV channel\nSamples are taken at each hair-card roots"
    bl_options = {"REGISTER", "UNDO"}

    method: bpy.props.EnumProperty(name='Sampling Method', description='Use hair card root Center or Profile vertices for picking UVs',
        items=[
            ('CENTER', 'Root center', 'Sample UVs for each strand at root center. Less accurate but works great for strands placed at seams edges'),
            ('PROFILE', 'Root Profile', 'Sample UVs at each strand root row, then propagate it to other vert rows. More accurate, but may give ugly artefeacts on seams'),
            ('PROJECTED', 'Projected', 'Sample UVs by projecting hair to target geometry. Wrapper fro Blender build-in Data Transer operator')
        ],  default='CENTER')

    collapse_threshold: bpy.props.FloatProperty(name='Collapse threshold', description='Colapse strand profile to point if picked UV size is too big (usually long UVs are generated for strands taht are placed over seams', default = 0.1, min=0, max=1)
    #dynamic enum property
    uv_layer: bpy.props.EnumProperty(name='UV To Sample', description='UV layer to write sampled UVs to',
        items=get_uv_layers)

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'method')
        if self.method == 'PROFILE':
            layout.prop(self, 'collapse_threshold')
        layout.prop(self, 'uv_layer')

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.target_obj

    @staticmethod
    def get_face_3d_to_uv(face, co, uv_lay):
        '''Get uv position corresponding to 3d loc co, on face, using uv '''
        face3d_barycentric_w = poly_3d_calc([v.co for v in face.verts], co)  # get barycentric weights
        # faceUV_barycentric_w = poly_3d_calc([loop[uv_lay].uv for loop in face.loops], co)  # get barycentric weights
        # for face in bm.faces:
        avg_uv_co = Vector((0, 0))
        for bary_weigth, loop in zip(face3d_barycentric_w, face.loops):
            uv = loop[uv_lay].uv
            avg_uv_co += uv * bary_weigth
        # for i, vert in enumerate(face.verts):
        #     avg_weight += vert[w_lay][source_id] * face3d_barycentric_w[i] if source_id in vert[w_lay].keys() else 0
        return avg_uv_co

    def execute(self, context):
        mesh_ribbons = context.active_object
        if not mesh_ribbons.ht_props.target_obj:
            self.report({'ERROR'}, f'No target object specified. Cancelling')
            return {'CANCELLED'}

        target = bpy.data.objects[mesh_ribbons.ht_props.target_obj]
        picked_uv_name = target.name+ '_picked_uv'
        if self.method == 'PROJECTED': # TODO: make sure it works - maybe use mod ?
            picked_uvs = mesh_ribbons.data.uv_layers.get(picked_uv_name)
            if not picked_uvs:
                picked_uvs = mesh_ribbons.data.uv_layers.new(name=picked_uv_name)
            project_uv_mod = mesh_ribbons.modifiers.new(name='Project UV', type='DATA_TRANSFER')
            project_uv_mod.object = target
            project_uv_mod.use_loop_data = True
            project_uv_mod.use_poly_data = True
            project_uv_mod.data_types_loops = {'UV'}
            project_uv_mod.layers_uv_select_src = self.uv_layer
            project_uv_mod.layers_uv_select_dst = picked_uv_name
            project_uv_mod.poly_mapping = 'POLYINTERP_PNORPROJ'
            project_uv_mod.loop_mapping = 'POLYINTERP_NEAREST'

            bpy.ops.object.modifier_apply(modifier=project_uv_mod.name)

            # overrides newer work...
            # override = {'active_object': target, 'selected_objects': [target, mesh_ribbons],
            #     'selected_editable_objects': [target, mesh_ribbons], 'editable_objects': [target, mesh_ribbons],
            #     'visible_objects': [target, mesh_ribbons]}
            # context.view_layer.objects.active = target
            # mesh_ribbons.data.uv_layers.active = picked_uvs
            # bpy.ops.object.data_transfer(data_type='UV', loop_mapping='POLYINTERP_NEAREST', poly_mapping='NORMAL', use_auto_transform=False, use_object_transform=True,
                                         # layers_select_src='ACTIVE', layers_select_dst='ACTIVE')
            return {'FINISHED'}


        helper_vvs = mesh_ribbons.data.uv_layers.get('HairTool_UV')
        if not helper_vvs:
            self.report({'ERROR'}, f'No required: HairTool_UV layer found on hair object. Cancelling')
            return {'CANCELLED'}

        target_mwi = target.matrix_world.inverted()

        target_BVHT = get_obj_mesh_bvht(target, context.evaluated_depsgraph_get(), applyModifiers=False, world_space=True)

        ribbon_bm = bmesh.new()   # create an empty BMesh
        ribbon_bm.from_mesh(mesh_ribbons.data)   # fill it in from a Mesh
        target_bm = bmesh.new()   # create an empty BMesh
        target_bm.from_mesh(target.data)   # fill it in from a Mesh

        target_bm.faces.ensure_lookup_table()
        ribbon_bm.faces.ensure_lookup_table()

        target_uv = target_bm.loops.layers.uv[self.uv_layer]
        hair_square_uv = ribbon_bm.loops.layers.uv['HairTool_UV']
        projected_uv = ribbon_bm.loops.layers.uv.get(picked_uv_name)
        if not projected_uv:
            projected_uv = ribbon_bm.loops.layers.uv.new(picked_uv_name)

        ribbon_bm.faces.ensure_lookup_table()
        ribbon_bm.verts.ensure_lookup_table()

        # faces_lislands = get_face_islands(ribbon_bm) # old sloooww
        # NOTE: build mesh islands based on HairTool_UV uvs
        islands = build_strand_islands(ribbon_bm, all_verts=True)

        for island_idx,island in enumerate(islands):
            first_row_verts = [ribbon_bm.verts[i] for i in island[0]]
            strip_width = len(first_row_verts)
            first_vert = first_row_verts[0]
            last_vert = first_row_verts[-1]


            def get_row_center():
                nonlocal first_vert, last_vert
                avg_co = (first_vert.co + last_vert.co)/2
                is_cylinder = strip_width != 2 and len([e for e in first_vert.link_edges if e.other_vert(first_vert) == last_vert]) > 0
                if is_cylinder: # cylinder profile
                    strip_center_row_idx = round(strip_width/2)
                    is_even = strip_width % 2 == 0
                    if is_even:
                        last_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (first_vert.co + last_vert.co)/2
                    else:
                        diagonal_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (avg_co + diagonal_vert.co)/2
                return avg_co

            root_co = get_row_center()
            hit_co, _, f_id, _ = target_BVHT.find_nearest(mesh_ribbons.matrix_world@root_co)

            center_hit_uv = self.get_face_3d_to_uv(target_bm.faces[f_id], target_mwi@hit_co, target_uv)

            if self.method == 'CENTER': #collapse uv in root center
                for vert_row_ids in island:  # island faces
                    for v_idx in vert_row_ids:
                        vert = ribbon_bm.verts[v_idx]
                        for loop in vert.link_loops:
                            loop[projected_uv].uv = center_hit_uv[:]
            else:  # use first row piced uvs for whole stripe
                picked_root_uvs = [] #* store each root picked UV
                for root_vert in first_row_verts:
                    vert_hit_co, _, f_id, _ = target_BVHT.find_nearest(mesh_ribbons.matrix_world @ root_vert.co)
                    hit_uv = self.get_face_3d_to_uv(target_bm.faces[f_id], target_mwi@vert_hit_co, target_uv)
                    picked_root_uvs.append(hit_uv)

                #* check uv are not too big, if so use center_hit_uv
                for i, p_uv in enumerate(picked_root_uvs):
                    if (p_uv-center_hit_uv).length > self.collapse_threshold:
                        picked_root_uvs[i] = center_hit_uv # collapse partially uv = True

                #* write UV for each hair card / island
                for row_vert_ids in island:
                    for v_idx, picked_uv in zip(row_vert_ids, cycle(picked_root_uvs)):
                        # print(f"{row_vert_ids=}  uvs: {picked_uv=}")
                        row_vert = ribbon_bm.verts[v_idx]
                        for loop in row_vert.link_loops:
                            loop[projected_uv].uv = picked_uv[:]


        ribbon_bm.to_mesh(mesh_ribbons.data)
        ribbon_bm.free()  # free and prevent further access
        self.report({'INFO'}, 'New uv channel \'projected_uv\' generated successfully')
        mesh_ribbons.ht_props.used_uv_pick = True  # just helper on convert
        return {"FINISHED"}


VCOLS = []
def get_vcol_layers(self, context):
    active_obj = bpy.context.active_object
    target = bpy.data.objects[active_obj.ht_props.target_obj]
    vcol_layers = []
    for i,vcol_layer in enumerate(target.data.vertex_colors):
        vcol_layers.append((vcol_layer.name, vcol_layer.name, ""))
    global VCOLS
    if len(VCOLS) != len(vcol_layers):
        VCOLS = vcol_layers
    return vcol_layers

class HTOOL_OT_VColorSampling(bpy.types.Operator):
    bl_idname = "object.color_sample_from_target"
    bl_label = "Sample Vert Color from target"
    bl_description = "Sample vertex colors from target object; sampled in place of each hair card root position"
    bl_options = {"REGISTER", "UNDO"}

    method: bpy.props.EnumProperty(name='Sampling Method', description='Use hair card root Center or Profile vertices for picking UVs',
                                   items=[
            ('CENTER', 'Root center', 'Sample weights for each strand at root center. Less accurate but fast'),
            ('PROFILE', 'Root Profile', 'Sample weights at each strand root row, then propagate it to other vert rows. More accurate'),
            ('PROJECTED', 'Projected', 'Sample weights by projecting hair to target geometry. Wrapper fro Blender build-in Data Transer operator')
           ],  default='CENTER')
    collapse_threshold: bpy.props.FloatProperty(name='Collapse threshold', description='Colapse strand profile to point if picked UV size is too big (usually long UVs are generated for strands taht are placed over seams', default = 0.1, min=0, max=1)
    vcol_layer: bpy.props.EnumProperty(name='Vert Color To Sample', description='Vertex color layer to sample colors form',
        items=get_vcol_layers)


    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.target_obj

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'method')
        if self.method == 'PROFILE':
            layout.prop(self, 'collapse_threshold')
        layout.prop(self, 'vcol_layer')

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)
        # return self.execute(context)

    @staticmethod
    def get_face_3d_to_col(face, co, vcol_lay):
        '''Get uv position corresponding to 3d loc co, on face, using uv '''
        face3d_barycentric_w = poly_3d_calc([v.co for v in face.verts], co)  # get barycentric weights
        # faceUV_barycentric_w = poly_3d_calc([loop[uv_lay].uv for loop in face.loops], co)  # get barycentric weights
        # for face in bm.faces:
        avg_vcol = Vector((0, 0, 0, 0))
        for bary_weigth, loop in zip(face3d_barycentric_w, face.loops):
            avg_vcol += loop[vcol_lay] * bary_weigth
        # for i, vert in enumerate(face.verts):
        #     avg_weight += vert[w_lay][source_id] * face3d_barycentric_w[i] if source_id in vert[w_lay].keys() else 0
        return avg_vcol


    def execute(self, context):
        mesh_ribbons = context.active_object
        if not mesh_ribbons.ht_props.target_obj:
            self.report({'ERROR'}, 'No target object specified. Cancelling')
            return {'CANCELLED'}
        target = bpy.data.objects[mesh_ribbons.ht_props.target_obj]
        if len(target.data.vertex_colors) == 0:
            self.report({'ERROR'}, 'No vertex groups found on target mesh. Cancelling')
            return {'CANCELLED'}

        picked_vert_col_name = target.name + '_' + self.vcol_layer
        if self.method == 'PROJECTED': # DONE: make sure it works - maybe use mod ?
            sample_type = 'VCOL' if bpy.app.version >= (3,4,0) else 'COLOR_CORNER'
            project_vcol_mod = mesh_ribbons.modifiers.new(name='Project UV', type='DATA_TRANSFER')
            project_vcol_mod.object = target
            project_vcol_mod.use_vert_data = False
            # project_vcol_mod.data_types_verts = {'VCOL'}

            project_vcol_mod.use_loop_data = True
            project_vcol_mod.data_types_loops = {sample_type}
            # project_vcol_mod.data_types_loops = {'VCOL'}

            project_vcol_mod.use_poly_data = False

            project_vcol_mod.vert_mapping = 'POLYINTERP_NEAREST'
            project_vcol_mod.loop_mapping = 'NEAREST_POLY'


            project_vcol_mod.layers_vcol_loop_select_src = self.vcol_layer
            project_vcol_mod.layers_vcol_loop_select_dst = 'NAME'
            # generate vcol on vcolor receiver
            # target.data.vertex_colors.active = target.data.vertex_colors[self.vcol_layer]
            bpy.ops.object.datalayout_transfer(modifier=project_vcol_mod.name, data_type=sample_type, layers_select_src='ACTIVE', layers_select_dst='NAME')
            # bpy.ops.geometry.color_attribute_add(name=picked_vert_col_name, domain='CORNER', data_type='BYTE_COLOR')

            # project_uv_mod.loop_mapping = 'POLYINTERP_NEAREST'
            bpy.ops.object.modifier_apply(modifier=project_vcol_mod.name)
            return {'FINISHED'}

        helper_vvs = mesh_ribbons.data.uv_layers.get('HairTool_UV')
        if not helper_vvs:
            self.report({'ERROR'}, 'No required: HairTool_UV layer found on hair object. Cancelling')
            return {'CANCELLED'}

        # mesh_ribbons.data.update()
        target_BVHT = get_obj_mesh_bvht(target, context.evaluated_depsgraph_get(), applyModifiers=False, world_space=True)

        ribbon_bm = bmesh.new()   # create an empty BMesh
        ribbon_bm.from_mesh(mesh_ribbons.data)   # fill it in from a Mesh
        ribbon_bm.faces.ensure_lookup_table()
        ribbon_bm.verts.ensure_lookup_table()

        target_bm = bmesh.new()   # create an empty BMesh
        target_bm.from_mesh(target.data)   # fill it in from a Mesh
        target_bm.faces.ensure_lookup_table()
        target_bm.verts.ensure_lookup_table()


        if picked_vert_col_name in ribbon_bm.loops.layers.uv.keys():
            projected_col = ribbon_bm.loops.layers.color[picked_vert_col_name]
        else:
            projected_col = ribbon_bm.loops.layers.color.new(picked_vert_col_name)

        target_mwi = target.matrix_world.inverted()
        target_vcol = target_bm.loops.layers.color[self.vcol_layer]
        # w_lay_ribb = ribbon_bm.verts.layers.deform.verify()  # create or retur existin one layer
        #* vert[w_lay] is dict - group.ids: g.weights

        # NOTE: build mesh islands based on HairTool_UV uvs
        islands = build_strand_islands(ribbon_bm, all_verts=True)

        for island_idx,island in enumerate(islands):
            first_row_verts = [ribbon_bm.verts[i] for i in island[0]]
            strip_width = len(first_row_verts)
            first_vert = first_row_verts[0]
            last_vert = first_row_verts[-1]


            def get_row_center():
                nonlocal first_vert, last_vert
                avg_co = (first_vert.co + last_vert.co)/2
                is_cylinder = strip_width != 2 and len([e for e in first_vert.link_edges if e.other_vert(first_vert) == last_vert]) > 0
                if is_cylinder: # cylinder profile
                    strip_center_row_idx = round(strip_width/2)
                    is_even = strip_width % 2 == 0
                    if is_even:
                        last_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (first_vert.co + last_vert.co)/2
                    else:
                        diagonal_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (avg_co + diagonal_vert.co)/2
                return avg_co

            root_co = get_row_center()
            hit_co, _, f_id, _ = target_BVHT.find_nearest(mesh_ribbons.matrix_world@root_co)

            center_hit_col = self.get_face_3d_to_col(target_bm.faces[f_id], target_mwi@hit_co, target_vcol)

            if self.method == 'CENTER': #collapse uv in root center
                for vert_row_ids in island:  # island faces
                    for v_idx in vert_row_ids:
                        vert = ribbon_bm.verts[v_idx]
                        for loop in vert.link_loops:
                            loop[projected_col] = center_hit_col[:]
            else:  # use first row piced uvs for whole stripe
                picked_root_cols = [] #* store each root picked UV
                for root_vert in first_row_verts:
                    vert_hit_co, _, f_id, _ = target_BVHT.find_nearest(mesh_ribbons.matrix_world @ root_vert.co)
                    hit_uv = self.get_face_3d_to_col(target_bm.faces[f_id], target_mwi@vert_hit_co, target_vcol)
                    picked_root_cols.append(hit_uv)

                #* check uv are not too big, if so use center_hit_uv
                for i, p_uv in enumerate(picked_root_cols):
                    if (p_uv-center_hit_col).length > self.collapse_threshold:
                        picked_root_cols[i] = center_hit_col # collapse partially uv = True

                #* write UV for each hair card / island
                for row_vert_ids in island:
                    for v_idx, picked_uv in zip(row_vert_ids, cycle(picked_root_cols)):
                        # print(f"{row_vert_ids=}  uvs: {picked_uv=}")
                        row_vert = ribbon_bm.verts[v_idx]
                        for loop in row_vert.link_loops:
                            loop[projected_col] = picked_uv[:]


        ribbon_bm.to_mesh(mesh_ribbons.data)
        ribbon_bm.free()  # free and prevent further access
        target_bm.free()  # free and prevent further access
        self.report({'INFO'}, f"Generated Vertex Color '{picked_vert_col_name}'")
        # mesh_ribbons.ht_props.used_weight_pick = True  # TODO:  just helper on convert - implement for vcol?
        # bpy.ops.object.mode_set(mode='VERTEX_PAINT')
        context.area.tag_redraw()
        return {"FINISHED"}


class HTOOL_OT_WeightSampling(bpy.types.Operator):
    bl_idname = "object.weight_sample_from_target"
    bl_label = "Sample Weights from target"
    bl_description = "Sample vertex weights from target object; sampled in place of each hair card root position"
    bl_options = {"REGISTER", "UNDO"}

    method: bpy.props.EnumProperty(name='Sampling Method', description='Use hair card root Center or Profile vertices for picking UVs',
                                   items=[
            ('CENTER', 'Root center', 'Sample weights for each strand at root center. Less accurate but fast'),
            ('PROFILE', 'Root Profile', 'Sample weights at each strand root row, then propagate it to other vert rows. More accurate'),
            ('PROJECTED', 'Projected', 'Sample weights by projecting hair to target geometry. Wrapper fro Blender build-in Data Transer operator')
           ],  default='CENTER')

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.target_obj

    @staticmethod
    def get_avg_weight_all(face, co, w_lay):
        # this will be output groups ids
        face_barycentric_w = poly_3d_calc([v.co for v in face.verts], co)  # get barycentric weights

        face_groups_li = {g for vert in face.verts for g in vert[w_lay].keys()}
        avg_weight = {g_idx: 0 for g_idx in face_groups_li}

        for i, vert in enumerate(face.verts):
            for g_idx in face_groups_li:
                avg_weight[g_idx] += vert[w_lay][g_idx] * face_barycentric_w[i] if g_idx in vert[w_lay].keys() else 0
        return avg_weight

    @staticmethod
    def get_face_3d_to_uv(face, co, uv_lay):
        '''Get uv position corresponding to 3d loc co, on face, using uv '''
        face3d_barycentric_w = poly_3d_calc([v.co for v in face.verts], co)  # get barycentric weights
        # faceUV_barycentric_w = poly_3d_calc([loop[uv_lay].uv for loop in face.loops], co)  # get barycentric weights
        # for face in bm.faces:
        avg_uv_co = Vector((0, 0))
        for bary_weigth, loop in zip(face3d_barycentric_w, face.loops):
            uv = loop[uv_lay].uv
            avg_uv_co += uv * bary_weigth
        # for i, vert in enumerate(face.verts):
        #     avg_weight += vert[w_lay][source_id] * face3d_barycentric_w[i] if source_id in vert[w_lay].keys() else 0
        return avg_uv_co

    def execute(self, context):
        mesh_ribbons = context.active_object
        if not mesh_ribbons.ht_props.target_obj:
            self.report({'ERROR'}, f'No target object specified. Cancelling')
            return {'CANCELLED'}
        target = bpy.data.objects[mesh_ribbons.ht_props.target_obj]
        if len(target.vertex_groups) == 0:
            self.report({'ERROR'}, f'No vertex groups found on target mesh. Cancelling')
            return {'CANCELLED'}

        if self.method == 'PROJECTED': # TODO: make sure it works - maybe use mod ?
            project_uv_mod = mesh_ribbons.modifiers.new(name='Project UV', type='DATA_TRANSFER')
            project_uv_mod.object = target
            project_uv_mod.use_vert_data = True
            project_uv_mod.use_loop_data = False
            project_uv_mod.use_poly_data = False
            project_uv_mod.data_types_verts = {'VGROUP_WEIGHTS'}
            project_uv_mod.vert_mapping = 'POLYINTERP_NEAREST'

            project_uv_mod.layers_vgroup_select_src = 'ALL'
            project_uv_mod.layers_vgroup_select_dst = 'NAME'
            bpy.ops.object.datalayout_transfer(modifier=project_uv_mod.name, data_type='VGROUP_WEIGHTS', layers_select_src='ALL', layers_select_dst='NAME')
            # project_uv_mod.loop_mapping = 'POLYINTERP_NEAREST'
            bpy.ops.object.modifier_apply(modifier=project_uv_mod.name)
            return {'FINISHED'}

        helper_vvs = mesh_ribbons.data.uv_layers.get('HairTool_UV')
        if not helper_vvs:
            self.report({'ERROR'}, f'No required: HairTool_UV layer found on hair object. Cancelling')
            return {'CANCELLED'}
        #* create vgroups first, and map target vg ids, to new vgroup ids
        target_to_new_vg_id = {}
        for t_vg in target.vertex_groups:
            mesh_rib_vg = mesh_ribbons.vertex_groups.get(t_vg.name)
            if not mesh_rib_vg:
                mesh_rib_vg = mesh_ribbons.vertex_groups.new(name=t_vg.name)  # changes active idx
            target_to_new_vg_id[t_vg.index] = mesh_rib_vg.index  # map

        mesh_ribbons.data.update()

        target_BVHT = get_obj_mesh_bvht(target, context.evaluated_depsgraph_get(), applyModifiers=False, world_space=True)

        ribbon_bm = bmesh.new()   # create an empty BMesh
        ribbon_bm.from_mesh(mesh_ribbons.data)   # fill it in from a Mesh
        ribbon_bm.faces.ensure_lookup_table()
        ribbon_bm.verts.ensure_lookup_table()

        target_bm = bmesh.new()   # create an empty BMesh
        target_bm.from_mesh(target.data)   # fill it in from a Mesh
        target_bm.faces.ensure_lookup_table()
        target_bm.verts.ensure_lookup_table()

        w_lay_trg = target_bm.verts.layers.deform.active  # vertex groups
        w_lay_ribb = ribbon_bm.verts.layers.deform.verify()  # create or retur existin one layer
        #* vert[w_lay] is dict - group.ids: g.weights

        # NOTE: build mesh islands based on HairTool_UV uvs
        islands = build_strand_islands(ribbon_bm, all_verts=True)
        target_mwi = target.matrix_world.inverted()
        for _,island in enumerate(islands):
            first_row_verts = [ribbon_bm.verts[i] for i in island[0]]
            strip_width = len(first_row_verts)
            first_vert = first_row_verts[0]
            last_vert = first_row_verts[-1]

            def get_row_center():
                nonlocal first_vert, last_vert
                avg_co = (first_vert.co + last_vert.co)/2
                is_cylinder = strip_width != 2 and len([e for e in first_vert.link_edges if e.other_vert(first_vert) == last_vert]) > 0
                if is_cylinder: # cylinder profile
                    strip_center_row_idx = round(strip_width/2)
                    is_even = strip_width % 2 == 0
                    if is_even:
                        last_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (first_vert.co + last_vert.co)/2
                    else:
                        diagonal_vert = ribbon_bm.verts[vert_row_ids[strip_center_row_idx]] # opposite vert
                        avg_co = (avg_co + diagonal_vert.co)/2
                return avg_co

            root_co = get_row_center()
            hit_co, _, f_id, _ = target_BVHT.find_nearest(mesh_ribbons.matrix_world@root_co)

            if self.method == 'CENTER': #collapse uv in root center
                avg_weight_all = self.get_avg_weight_all(target_bm.faces[f_id], target_mwi@hit_co, w_lay_trg)

                for vert_row_ids in island:  # island faces
                    for v_idx in vert_row_ids:
                        vert = ribbon_bm.verts[v_idx]
                        vert[w_lay_ribb].clear()
                        for g_id, weight in avg_weight_all.items():
                            ribbon_vg_id = target_to_new_vg_id[g_id]
                            vert[w_lay_ribb][ribbon_vg_id] = weight

            else:  # use first row piced uvs for whole stripe
                picked_row_weights = []  # * store  each root picked UV
                for root_vert in first_row_verts:
                    hit_co, normal, f_id, distance = target_BVHT.find_nearest(mesh_ribbons.matrix_world@root_vert.co)  # ? maybe times target.mwi ?
                    avg_weight_all = self.get_avg_weight_all(target_bm.faces[f_id], target_mwi@hit_co, w_lay_trg)
                    picked_row_weights.append(avg_weight_all)

                #* write UV for each hair card / island
                for row_vert_ids in island:
                    for v_idx, picked_weights in zip(row_vert_ids, cycle(picked_row_weights)):
                        # print(f"{row_vert_ids=}  uvs: {picked_uv=}")
                        row_vert = ribbon_bm.verts[v_idx]
                        row_vert[w_lay_ribb].clear()
                        for g_id, weight in picked_weights.items():
                            ribbon_vg_id = target_to_new_vg_id[g_id]
                            row_vert[w_lay_ribb][ribbon_vg_id] = weight


        ribbon_bm.to_mesh(mesh_ribbons.data)
        ribbon_bm.free()  # free and prevent further access
        target_bm.free()  # free and prevent further access
        self.report({'INFO'}, f'Vertex groups picked successfully')
        mesh_ribbons.ht_props.used_weight_pick = True  # just helper on convert
        return {"FINISHED"}
