# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with self program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2020 JOSECONSCO
# Created by JOSECONSCO

from copy import deepcopy
from math import cos, ceil, floor, sin, degrees
from bl_math import lerp

import bpy
import gpu
import numpy as np
from bpy_extras import view3d_utils
from gpu_extras.batch import batch_for_shader
from gpu_extras.presets import draw_circle_2d
from mathutils import Euler, Matrix, Quaternion, Vector

from .utils.curve_wrapper import ChainIK, Splines
from .utils.gpu_helpers import depth_offset_matrix_method
from .hair_baking.hair_geometry_nodes_selections import get_tips_pts, get_selected_strands_mask, select_blender_curves, get_edit_selected_points_bool
from .utils.helper_functions import clamp, get_fallof_fun, get_obj_mesh_bvht
from .hair_baking.hair_geometry_nodes_utils import del_sel_strands, align_curves_tilt, duplicate_sel, apply_mod_force


from .hair_baking.hair_geometry_nodes_shared import auto_update_all_curves_uv_regions

HT_WORKSPACE_SELECT_TIP = "ht_workspace_select_tip"

def get_set_ht_select_attr(curve_obj):
    ht_workspace_select = curve_obj.data.attributes.get(HT_WORKSPACE_SELECT_TIP)
    if not ht_workspace_select: # add select attribute
        ht_workspace_select = curve_obj.data.attributes.new(HT_WORKSPACE_SELECT_TIP, 'BOOLEAN','POINT')
    return ht_workspace_select

def workspace_de_select_all(context, curve_obj, mode = 'DESELECT'):
    ''' Mode - DESELECT, SELECT, INVERT'''
    sel_tip_pts = []
    not_sel_tip_pts = []
    sel_splines_ids = [False]*len(curve_obj.data.curves) # for .selection
    offset = context.scene.ht_props.select_offset
    active_end = context.scene.ht_props.active_end
    ht_workspace_select = get_set_ht_select_attr(curve_obj)
    for spl in curve_obj.data.curves:
        last_pid = len(spl.points)-1
        p_id = clamp(last_pid - offset, 0, last_pid) if active_end == "TIPS" else clamp(offset, 0, last_pid)
        # pt_sel_state = spl.points[p_id].select if spl.type in {'NURBS', 'POLY'} else spl.points[p_id].select_control_point
        pt_sel_state = ht_workspace_select.data[spl.points[p_id].index].value

        if pt_sel_state:
            sel_tip_pts.append(spl.points[p_id])
            sel_splines_ids[spl.index] = True
        else:
            not_sel_tip_pts.append(spl.points[p_id])

    ht_workspace_select.data.foreach_set('value', [False]*len(ht_workspace_select.data))

    if mode != "DESELECT": #we already deselected
        pts_to_set = sel_tip_pts+not_sel_tip_pts if mode == "SELECT" else not_sel_tip_pts
        for pt in pts_to_set:
            ht_workspace_select.data[pt.index].value = True

    # handle .select
    if mode == "DESELECT":
        select_blender_curves(curve_obj,[], deselect=True)
    elif mode == "SELECT":
        select_blender_curves(curve_obj,[i for i in range(len(curve_obj.data.curves))] , deselect=True)
    else: # INVERT
        spl_ids = [i for i, sel in enumerate(sel_splines_ids) if not sel]
        select_blender_curves(curve_obj,spl_ids, deselect=True)


def workspace_select_by_spl_ids(context, curve_obj, sel_splines_ids, deselect=True):
    is_ht_on = context.workspace.tools.from_space_view3d_mode(mode=bpy.context.mode).idname == HTOOL_OT_ToolModelHair.bl_idname
    if context.mode in ('OBJECT','SCULPT_CURVES') and not is_ht_on: # switch to using HT worksapce tool - will/should create HT_WORKSPACE_SELECT_TIP attrib
        bpy.ops.wm.tool_set_by_id(name="hair_tool.hair_transform")

    ht_workspace_select = get_set_ht_select_attr(curve_obj)
    _, pts, _ = get_tips_pts(context, curve_obj, world_space=False, only_sel=False)
    if deselect:
        ht_workspace_select.data.foreach_set('value', [False]*len(ht_workspace_select.data))
    for spl_id in sel_splines_ids:
        pt = pts[spl_id]
        ht_workspace_select.data[pt.index].value = True
    select_blender_curves(curve_obj, sel_splines_ids, deselect=deselect)


shader_3d_uniform = gpu.shader.from_builtin('UNIFORM_COLOR')
shader_2d_flat = gpu.shader.from_builtin('UNIFORM_COLOR')  # 2D_FLAT_COLOR
shader_3d_smooth = gpu.shader.from_builtin('SMOOTH_COLOR')

sel_col = (0.3, 1., 0.1)
adj_verts_colors_sel1 = [sel_col+(0.0,), sel_col+(0.8,), sel_col+(0.0,)]
adj_verts_colors_sel2 = [sel_col+(0.0,), sel_col+(0.5,), sel_col+(0.8,), sel_col+(0.5,), sel_col+(0.0,)]

not_sel_col = (0.9, 0.7, 0.2)
adj_verts_colors_not_sel1 = [not_sel_col+(0.0,), not_sel_col+(0.8,),  not_sel_col+(0.0,)]
adj_verts_colors_not_sel2 = [not_sel_col+(0.0,), not_sel_col+(0.5,), not_sel_col+(0.8,),  not_sel_col+(0.5,), not_sel_col+(0.0,)]

def draw_tips_pts(self, context):
    obj = context.active_object
    if context.space_data.overlay.show_overlays and obj and obj.type == 'CURVES' and obj.mode in ('OBJECT','SCULPT_CURVES') and obj.visible_get():
        higlight_bias = context.scene.ht_props.higlight_bias
        higlight_range = context.scene.ht_props.higlight_range
        theme = bpy.context.preferences.themes['Default']
        g_vertex_size = theme.view_3d.vertex_size + 1
        _, pts, pts_co, adj_points = get_tips_pts(context, obj, world_space=True, only_sel=False, with_adj_pts=True)
        if not pts:
            return
        center = Vector((0, 0, 0))
        for p_co in pts_co:
            center += p_co
        center /= len(pts_co)

        #* bias toward camera offset
        g_rv3d = context.region_data
        camera_pos = g_rv3d.view_matrix.inverted().translation
        camDistance = camera_pos - obj.matrix_world@center
        bias = higlight_bias / camDistance.length_squared
        dynamic_depth_off = 0.01* higlight_bias / camDistance.length

        # g_vertex_color = theme.view_3d.vertex
        gpu.state.blend_set("ALPHA")
        gpu.state.depth_test_set("LESS")
        gpu.state.line_width_set(2.0)
        ht_workspace_select = get_set_ht_select_attr(obj)

        # bgl.glDepthRange(0, 0.9999) - replacement
        # winmat = gpu.matrix.get_projection_matrix()
        # winmat[3][2] -= 0.1*bias  # bias toward camera offset
        with gpu.matrix.push_pop(): # or else obj.mw may be applied twice, if other draw handler is active with matrix.multiply ...

            depth_offset_matrix_method(obj.matrix_world, dynamic_depth_off)

            # TODO: find replacement for glDepthRange(zNear, zFar) in gpu module.
            # with gpu.matrix.push_pop(): # XXX: there was crash (too many matrix push - or something)
            for p, p_co, connected_pts in zip(pts, pts_co, adj_points):
                sel = ht_workspace_select.data[p.index].value
                gpu.state.point_size_set(g_vertex_size + 1 if sel else g_vertex_size)
                batch_pts = batch_for_shader(shader_3d_uniform, 'POINTS', {"pos": [p_co]})
                shader_3d_uniform.bind()
                shader_3d_uniform.uniform_float("color", sel_col+(1,) if sel else not_sel_col+(1,))
                batch_pts.draw(shader_3d_uniform)

                if higlight_range:
                    if higlight_range == 1:
                        adj_col = adj_verts_colors_sel1 if sel else adj_verts_colors_not_sel1
                    else:
                        adj_col = adj_verts_colors_sel2 if sel else adj_verts_colors_not_sel2
                    batch_edges = batch_for_shader(shader_3d_smooth, 'LINE_STRIP', {"pos": connected_pts, "color": adj_col})
                    shader_3d_smooth.bind()
                    batch_edges.draw(shader_3d_smooth)

            gpu.matrix.pop_projection() # not needed with push_pop...jik

        gpu.state.depth_test_set("NONE")
        gpu.state.blend_set("NONE")
        gpu.state.point_size_set(1)
        gpu.state.line_width_set(1.0)


def draw_2d_circle(self, context):
    color = (0.9, 0.9, 0.9, 0.9)
    draw_circle_2d(self.center_2d, color, self.rad_draw_2d, segments=32)
    pass


def delayed_refresh():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            area.tag_redraw()

def toggle_blender_brush_key(active=True):
    '''Disable sculpt_curves.brush_stroke - since it conflicts with ctrl+lmb lasso select'''
    user_kc = bpy.context.window_manager.keyconfigs.user
    view_3d_keymap = user_kc.keymaps.get('3D View')  # Get hotkeys from the currently used keyconfig
    if view_3d_keymap is None:
        return
    sculpt_curve_mode_keymap = user_kc.keymaps.get('Sculpt Curves')
    if not active:
        ### XXX: crashes when re-enabling brush hotkey (eg when switching blend file)
        # thus only leave disabled
        for keymap_item in sculpt_curve_mode_keymap.keymap_items:
            if keymap_item.idname == 'sculpt_curves.brush_stroke' and keymap_item.properties.mode == 'INVERT':
                keymap_item.active = active



handle_SpaceView3D = None
class HT_TransformGiGroup(bpy.types.GizmoGroup):
    bl_idname = "HT_TransformGiGroup"  # this shit has to be same as cls name!
    bl_label = "Hair Tool Curve Tips Points"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'
    bl_options = {'3D'}  # , 'PERSISTENT'

    def __init__(self):  # enabled once on worktool ON
        # print('iinit')
        global handle_SpaceView3D
        is_ht_on = bpy.context.workspace.tools.from_space_view3d_mode(mode=bpy.context.mode).idname == HTOOL_OT_ToolModelHair.bl_idname
        toggle_blender_brush_key(active=False)
        if handle_SpaceView3D is None and is_ht_on:
            # print('Addind draw')
            #! not greatest sollutions it seems.
            args = (None, bpy.context)  # u can pass arbitrary class as first param  Instead of (self, context)
            handle_SpaceView3D = bpy.types.SpaceView3D.draw_handler_add(draw_tips_pts, args, 'WINDOW', 'POST_VIEW')
            bpy.context.area.tag_redraw()

    def __del__(self):  # enabled once on worktool OFF
        # print('del')
        global handle_SpaceView3D
        toggle_blender_brush_key(active=True)
        if handle_SpaceView3D is not None:
            # print('rem draw')
            #! not greatest sollutions it seems. Unbinds on operator each operator redoo.....
            bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
            handle_SpaceView3D = None

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'
        # if bpy.context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == MyTool.bl_idname:
        #     return True
        # else:
        #     context.window_manager.gizmo_group_type_unlink_delayed(cls.bl_idname)
        #     return False

    def invoke_prepare(self, context, gizmo):
        # print('Inovke')
        pass

    def setup(self, context):
        # print('setup')
        curve_obj = context.active_object
        get_set_ht_select_attr(curve_obj)
        bpy.app.timers.register(delayed_refresh, first_interval=0.05)

    def refresh(self, context):
        global handle_SpaceView3D
        is_ht_on = bpy.context.workspace.tools.from_space_view3d_mode(mode=bpy.context.mode).idname == HTOOL_OT_ToolModelHair.bl_idname
        if not is_ht_on and handle_SpaceView3D is not None:
            # print('rem draw')
            #! not greatest sollutions it seems. Unbinds on operator each operator redoo.....
            bpy.types.SpaceView3D.draw_handler_remove(handle_SpaceView3D, 'WINDOW')
            handle_SpaceView3D = None
        # print('GizmoGroup refresh')
        pass
        # TODO: put cache update here?


class HT_PT_WorkspaceToolShared():
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Tool"
    is_popover = True

    # bl_context = ".ht_hotkeys"  # dot on purpose (access from topbar)
    # bl_ui_units_x = 8
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        active_wspace_tool = context.workspace.tools.from_space_view3d_mode(mode=context.mode)
        return active_wspace_tool and active_wspace_tool.idname == "hair_tool.hair_transform"



class HT_PT_ToolHotkeys(HT_PT_WorkspaceToolShared, bpy.types.Panel):
    bl_idname = "HT_PT_ToolHotkeys"
    bl_label = "Hotkeys"

    def draw(self, context):
        layout = self.layout
        # Active Tool
        # -----------
        # from bl_ui.space_toolsystem_common import ToolSelectPanelHelper
        # tool = ToolSelectPanelHelper.tool_active_from_context(context)
        # props = tool.operator_properties("curves.hair_transform")
        # preferences = bpy.context.preferences.addons[__package__].preferences
        col = layout.column()
        row = col.row()
        row.label(text="Ctrl+MWheel")
        row.label(text="Offset selection")
        row = col.row()
        row.label(text="Ctrl+RMB")
        row.label(text="Select object")
        row = col.row()
        row.label(text="(Shift)+RMB")
        row.label(text="Select Points")
        row = col.row()
        row.label(text="(Alt)+A")
        row.label(text="(de)Select All")
        row = col.row()
        row.label(text="Ctrl+I")
        row.label(text="Invert Selection")
        row = col.row()
        row.label(text="(Alt)+B+LMB")
        row.label(text="Box (de)Select")
        row = col.row()
        row.label(text="G, R, S")
        row.label(text="Transform spline")
        row = col.row()
        row.label(text="C+LMB(MMB)")
        row.label(text="Circle (de)Select")
        row = col.row()
        row.label(text="Ctr+(Shift)+LMB")
        row.label(text="Lasso (de)Select")
        row = col.row()
        row.label(text="X, Del")
        row.label(text="Delete Selected")
        row = col.row()
        row.label(text="Ctrl+D")
        row.label(text="Linear Deformer")
        row = col.row()
        row.label(text="K")
        row.label(text="Cut Tool")
        row = col.row()
        row.label(text="Shift+D")
        row.label(text="Duplicate Strand")
        row = col.row()
        row.label(text="Alt+S")
        row.label(text="Adjust Radius")
        row = col.row()
        row.label(text="Alt+R")
        row.label(text="Reset Radius")
        row = col.row()
        row.label(text="Ctrl+T")
        row.label(text="Adjust Tilt")
        row = col.row()
        row.label(text="Alt+T")
        row.label(text="Reset Tilt")
        row = col.row()
        row.label(text="E")
        row.label(text="Extend Curve")
        row = col.row()
        row.label(text="I")
        row.label(text="Straighten Curve")
        row = col.row()
        row.label(text="Ctrl+J")
        row.label(text="Join Strands")
        row = col.row()
        row.label(text="P")
        row.label(text="Separate Strands")

class HT_PT_DrawSettings(HT_PT_WorkspaceToolShared, bpy.types.Panel):
    bl_idname = "HT_PT_DrawSettings"
    bl_label = "Overlay Settings"

    def draw(self, context):
        layout = self.layout
        ht_scn_props = context.scene.ht_props
        layout.prop(ht_scn_props, "higlight_bias")
        layout.prop(ht_scn_props, "higlight_range")

# popup for use collisions
class HTOOL_PT_UseCollisionsMenu(HT_PT_WorkspaceToolShared, bpy.types.Panel):
    bl_idname = "HTOOL_PT_UseCollisionsMenu"
    bl_label = "Collisions"

    def draw(self, context):
        layout = self.layout
        ht_scn_props = context.scene.ht_props
        layout.prop(ht_scn_props, "use_collisions", toggle=True, icon = 'MOD_PHYSICS')
        if ht_scn_props.use_collisions:
            layout.prop(ht_scn_props, "collison_offset")
            layout.prop(ht_scn_props, "use_custom_collider", toggle=True, icon = 'MODIFIER')
            if ht_scn_props.use_custom_collider:
                layout.prop_search(ht_scn_props, "custom_collider", bpy.data, "objects", text="")



# panel for proportional_edit_falloff
class HT_PT_ProportionalEditMenu(HT_PT_WorkspaceToolShared, bpy.types.Panel):
    bl_idname = "HT_PT_ProportionalEditMenu"
    bl_label = "Proportional Edit"

    def draw(self, context):
        ht_scn_props = context.scene.ht_props
        layout = self.layout
        layout.props_enum(ht_scn_props, "proportional_edit_falloff")



class HTOOL_OT_ToolBase(bpy.types.WorkSpaceTool):
    bl_space_type = 'VIEW_3D'
    # The prefix of the idname should be your add-on name.
    bl_idname = "hair_tool.hair_transform"
    bl_label = "Hair Modeling"
    bl_description = "Select, transform, bend, extend, delete, cut etc. curves using Hair Tool modeling tools"
    bl_icon = "brush.particle.smooth"
    bl_widget = "HT_TransformGiGroup"
    bl_keymap = (
                ("curves.join_strands", {"type": 'J', "value": 'ANY', 'ctrl': True}, None),
                ("curve.deform_curve", {"type": 'D', "value": 'ANY', 'ctrl': True}, None),
                ("curves.separate_selected", {"type": 'P', "value": 'ANY'}, None),
                ("hair.knife_cut_modal", {"type": 'K', "value": 'ANY'}, None),
                ("curve.curve_lenghten", {"type": 'E', "value": 'PRESS'}, None),
                ("curves.hair_straighten", {"type": 'I', "value": 'PRESS'}, None),
                ("wm.call_menu", {"type": 'DEL', "value": 'PRESS'}, {"properties": [('name', 'HTOOL_MT_DeleteMenu')]}),
                ("wm.call_menu", {"type": 'X', "value": 'PRESS'}, {"properties": [('name', 'HTOOL_MT_DeleteMenu')]}),
                # ("curve.hair_delete", {"type": 'DEL', "value": 'PRESS'}, None),
                # ("curve.hair_delete", {"type": 'X', "value": 'PRESS'}, None),
                ("curves.hair_duplicate", {"type": 'D', "value": 'ANY', 'shift': True}, None),
                ("curve.hair_radilt_fallof", {"type": 'S', "value": 'ANY', 'alt': True},  {"properties": [('mode', 'RADIUS')]}),
                ("curve.hair_radilt_reset", {"type": 'R', "value": 'ANY', 'alt': True}, {"properties": [('mode', 'RADIUS')]}),
                ("curve.hair_radilt_fallof", {"type": 'T', "value": 'ANY', 'ctrl': True},  {"properties": [('mode', 'TILT')]}),
                ("curve.hair_radilt_reset", {"type": 'T', "value": 'ANY', 'alt': True}, {"properties": [('mode', 'TILT')]}),
                ("hair.offset_selection", {"type": 'WHEELUPMOUSE', "value": 'ANY', 'ctrl': True}, {"properties": [("increase", True)]}),
                ("hair.offset_selection", {"type": 'WHEELDOWNMOUSE', "value": 'ANY', 'ctrl': True}, {"properties": [("increase", False)]}),
                ("curves.hair_transform", {"type": 'G', "value": 'PRESS'}, {"properties": [("transform_mode", 'TRANSLATE')]}),
                ("curves.hair_transform", {"type": 'R', "value": 'PRESS'}, {"properties": [("transform_mode", 'ROTATE')]}),
                ("curves.hair_transform", {"type": 'S', "value": 'PRESS'}, {"properties": [("transform_mode", 'SCALE')]}),
                ("view3d.select", {"type": 'RIGHTMOUSE', "value": 'PRESS', "ctrl": True}, None),
                ("view3d.select", {"type": 'RIGHTMOUSE', "value": 'PRESS', "ctrl": True, "shift": True}, {"properties": [("toggle", True)]}),
                ("curve.hair_select", {"type": 'RIGHTMOUSE', "value": 'PRESS'}, {"properties": [("extend", False)]}),
                ("curve.hair_select", {"type": 'RIGHTMOUSE', "value": 'PRESS', 'shift': True}, {"properties": [("extend", True)]}),
                ("curve.hair_select_set", {"type": 'A', "value": 'PRESS'}, {"properties": [("mode", "SELECT")]}),
                ("curve.hair_select_set", {"type": 'A', "value": 'PRESS', "alt": True}, {"properties": [("mode", "DESELECT")]}),
                ("curve.hair_select_set", {"type": 'A', "value": 'DOUBLE_CLICK'}, {"properties": [("mode", "DESELECT")]}),
                ("curve.hair_select_set", {"type": 'I', "value": 'PRESS', "ctrl": True}, {"properties": [("mode", "INVERT")]}),
                ("hair.select_region_modal", {"type": 'B', "value": 'PRESS'}, {"properties": [("sel_type", 'BOX'), ('mode', 'ADD')]}),
                ("hair.select_region_modal", {"type": 'B', "value": 'PRESS', "alt": True}, {"properties": [("sel_type", 'BOX'), ('mode', 'SUB')]}),
                ("hair.select_region_modal", {"type": 'C', "value": 'PRESS'}, {"properties": [("sel_type", 'CIRCLE'), ('mode', 'ADD')]}),
                ("hair.select_region_modal", {"type": 'C', "value": 'PRESS', "alt": True}, {"properties": [("sel_type", 'CIRCLE'), ('mode', 'SUB')]}),
                ("hair.select_region_modal", {"type": 'LEFTMOUSE', "value": 'CLICK_DRAG', "ctrl": True}, {"properties": [("sel_type", 'LASSO'), ('mode', 'ADD')]}),
                ("hair.select_region_modal", {"type": 'LEFTMOUSE', "value": 'CLICK_DRAG', "ctrl": True, "shift": True}, {"properties": [("sel_type", 'LASSO'), ('mode', 'SUB')]}),

                # ("hair.select_region_modal", {"type": 'EVT_TWEAK_L', "value": 'ANY', "ctrl": True}, {"properties": [("sel_type", 'LASSO'), ('mode', 'ADD')]}),
                # ("hair.select_region_modal", {"type": 'EVT_TWEAK_L', "value": 'ANY', "ctrl": True, "shift": True}, {"properties": [("sel_type", 'LASSO'), ('mode', 'SUB')]}),
    )
    # ,"key_modifier": 'Ctrl' ; EVT_TWEAK_L

    def draw_settings(context, layout, tool):
        transform_op = tool.operator_properties("curves.hair_transform")
        ht_scn_props = context.scene.ht_props
        row = layout.row(align=True)
        row.prop(ht_scn_props, "active_end", expand=True)
        ic = 'PINNED' if transform_op.pin_end else 'UNPINNED'
        row.prop(transform_op, "pin_end", text='', icon=ic)
        # layout.prop(ht_scn_props, "select_offset")
        # layout.label(text=f'Offset {ht_scn_props.select_offset}')
        split = layout.split(factor=0.3, align=True)
        split.label(text='Mode')
        split.prop(ht_scn_props, "move_mode", text='')
        if ht_scn_props.move_mode in ('IK','ROOT'): # ROOT also uses IK for translation
            layout.popover(panel="HTOOL_PT_UseCollisionsMenu")

        layout.popover(panel="HT_PT_DrawSettings")

        # fallof settings
        col = layout.column()
        # col.active = ht_scn_props.move_mode in ('FK','ROOT')
        prop_ic = 'PROP_ON' if ht_scn_props.use_proportional_edit else 'PROP_OFF'
        row = col.row(align=True)
        row.prop(ht_scn_props,'use_proportional_edit', text='', icon=prop_ic, toggle=True)

        # get icon from  active ht_scn_props.proportional_edit_falloff
        ic = ht_scn_props.proportional_edit_falloff.upper()+"CURVE"
        if ht_scn_props.proportional_edit_falloff == 'CONSTANT': # 2 special cases
            ic = 'NOCURVE'
        elif ht_scn_props.proportional_edit_falloff == 'LINEAR':
            ic = 'LINCURVE'
        row.popover(panel="HT_PT_ProportionalEditMenu", text='', icon=ic)

        layout.prop(ht_scn_props, "align_tilt", icon="SNAP_NORMAL")
        reg = context.region.type
        if reg == 'TOOL_HEADER':
            # popover_kw = {"space_type": 'VIEW_3D', "region_type": 'UI', "category": "Tool"}
            # layout.popover_group(context=".ht_hotkeys", **popover_kw)
            layout.popover(panel="HT_PT_ToolHotkeys")
        # if context.area.type == 'PROPERTIES':

class HTOOL_OT_ToolModelHair(HTOOL_OT_ToolBase):
    bl_context_mode = 'OBJECT'


class HTOOL_OT_ToolModelHairSculptMode(HTOOL_OT_ToolBase):
    bl_context_mode = 'SCULPT_CURVES'


class HTOOL_OT_HairTransform(bpy.types.Operator):  # ! revrite to this operator
    bl_idname = "curves.hair_transform"
    bl_label = "Hair Transform"
    bl_description = "Curve Hair Transform (more, rotate, scale tips) with IK"
    bl_options = {'REGISTER', 'UNDO'}

    pin_end: bpy.props.BoolProperty(name="Pinned", description="Keep other spline ending in place", default=True)
    transform_mode: bpy.props.EnumProperty(name='Transform mode',
                                           items=(
                                               ('TRANSLATE', 'Translate', ''),
                                               ('ROTATE', 'Rotate', ''),
                                               ('SCALE', 'Scale', '')), default='TRANSLATE')

    @staticmethod
    def draw_debug_3d(self, context):
        gpu.state.blend_set("ALPHA") # if bpy.app.version >= (3, 5, 0) else bgl.glEnable(bgl.GL_BLEND)
        gpu.state.line_width_set(1.0) #if bpy.app.version >= (3, 5, 0) else bgl.glLineWidth(1)
        gpu.state.depth_test_set("LESS") # if bpy.app.version >= (3,5,0) else bgl.glEnable(bgl.GL_DEPTH_TEST)

        center_w = self.center_3dw
        if self.x:
            line_pts = [Vector((-100, 0, 0))+center_w, Vector((100, 0, 0))+center_w]
            batch_pts = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": line_pts})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", (1, .1, .1, 0.8))
            batch_pts.draw(shader_3d_uniform)
        elif self.y:
            line_pts = [Vector((0, -100, 0))+center_w, Vector((0, 100, 0))+center_w]
            batch_pts = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": line_pts})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", (0, 1, 0, 0.8))
            batch_pts.draw(shader_3d_uniform)
        elif self.z:
            line_pts = [Vector((0, 0, -100))+center_w, Vector((0, 0, 100))+center_w]
            batch_pts = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": line_pts})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", (0.2, 0.2, 1, 0.8))
            batch_pts.draw(shader_3d_uniform)

        gpu.state.depth_test_set("NONE") # if bpy.app.version >= (3,5,0) else bgl.glDisable(bgl.GL_DEPTH_TEST)
        gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)


    @staticmethod
    def draw_ui_2d(self, context):
        gpu.state.blend_set("ALPHA")
        gpu.state.line_width_set(2)
        gpu.state.depth_test_set("LESS")

        # draw line from center w to mouse
        region = context.region
        rv3d = context.region_data
        # move center 3d to 2d
        scn_s = context.scene.ht_props
        if scn_s.move_mode in ('FK','IK'):
            center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw)
            line_pts = [self.mouse_co_2d, center_2d]
        else: #'ROOT'
            if self.transform_mode == 'SCALE':
                line_pts = [self.first_click, Vector((self.mouse_co_2d.x, self.first_click.y))]
            elif self.transform_mode == 'ROTATE':
                center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw)
                line_pts = [self.mouse_co_2d, center_2d]
            else: # do not draw when translate
                return


        batch_pts = batch_for_shader(shader_2d_flat, 'LINES', {"pos": line_pts})
        shader_2d_flat.bind()
        shader_2d_flat.uniform_float("color", (0.0, 0.0, 0.0, 0.8))
        batch_pts.draw(shader_2d_flat)

        gpu.state.blend_set("NONE")
        gpu.state.depth_test_set("NONE")



    def update_draw_header(self, context):
        scn_s = context.scene.ht_props
        append = ''
        if scn_s.move_mode == 'ROOT' and self.transform_mode == 'SCALE':
            append = '        [Shift] - Unify Strands Length'
        context.area.header_text_set(f"Proportional Size: {scn_s.influence_rad: .2f}  Fallof: {scn_s.proportional_edit_falloff} {append}")

    def calc_rad_2d(self, context):
        '''For drawing circle'''
        region = context.region
        rv3d = context.region_data
        scn_s = context.scene.ht_props
        center_2d_plus_rad_3d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw + Vector((scn_s.influence_rad, 0, 0)), default=Vector((0, 0)))
        self.rad_draw_2d = (center_2d_plus_rad_3d - self.center_2d).length

    def move_tip(self, context, event):  #
        ''' returns (hit, hit_co_world, hit_obj)
        If nothing hit, get output location based on region_2d_to_location_3d
        '''
        def write_pts(pts, moved_pts):
            for p, m_p in zip(pts, moved_pts):
                p.position = m_p

        scn_ht = context.scene.ht_props
        active_roots = scn_ht.active_end == "ROOTS"
        # clump_falloff = scn_ht.clump_falloff
        move_mode = scn_ht.move_mode
        splines = self.curve_obj.data.curves
        mw = self.curve_obj.matrix_world
        m_inv = mw.inverted()

        if move_mode != 'IK':
            self.calc_rad_2d(context)  # for drawing
            self.update_draw_header(context)

        def move_the_chains(spl, target_co):
            pts = spl.points[:]
            pid_offset = scn_ht.select_offset
            last_pid = len(pts)-1
            if last_pid < 1:
                return
            bvh_collider = self.target_BVHT if scn_ht.use_collisions else None
            coll_offset = scn_ht.collison_offset
            p_id = pid_offset if active_roots else last_pid - pid_offset
            p_id = clamp(p_id, 0, last_pid)
            other_end_r = p_id == 0 and not active_roots  # we scrolled to other end - root
            other_end_t = p_id == last_pid and active_roots  # we scrolled to other end - tip
            if p_id == 0 or p_id == last_pid:
                flip = active_roots
                if other_end_r or other_end_t:
                    flip = not flip
                ik_chain = ChainIK(pts[::-1]) if flip else ChainIK(pts)
                ik_chain.move_ik_chain(target_co, self.pin_end, bvh_collider=bvh_collider, offset=coll_offset)
                moved_pts = ik_chain.get_pts()[::-1] if flip else ik_chain.get_pts()

                write_pts(pts, moved_pts)
            else:
                if active_roots:  # when pull by roots, use pts order
                    sub_chain_a = pts[p_id:]
                    sub_chain_b = pts[:p_id+1]  # !check if in (0, len) range
                    ik_chain = ChainIK(sub_chain_a[::-1])  # rev
                    ik_chain.move_ik_chain(target_co, self.pin_end, bvh_collider=bvh_collider, offset=coll_offset)
                    moved_pts_a = ik_chain.get_pts()[::-1]  # rev

                    ik_chain_b = ChainIK(sub_chain_b)
                    ik_chain_b.move_ik_chain(moved_pts_a[0], pin_end=False, bvh_collider=bvh_collider, offset=coll_offset)
                    moved_pts_b = ik_chain_b.get_pts()
                else:  # when pull by tips, flip pts order
                    sub_chain_a = pts[:p_id+1]
                    sub_chain_b = pts[p_id:]  # ! cos we use seame point p[p_id] in both chains...
                    if len(sub_chain_a) > 1:
                        ik_chain = ChainIK(sub_chain_a)
                        ik_chain.move_ik_chain(target_co, self.pin_end, bvh_collider=bvh_collider, offset=coll_offset)
                        moved_pts_a = ik_chain.get_pts()
                    else:
                        moved_pts_a = [target_co]

                    ik_chain_b = ChainIK(sub_chain_b[::-1])
                    ik_chain_b.move_ik_chain(moved_pts_a[-1], pin_end=False, bvh_collider=bvh_collider, offset=coll_offset)
                    moved_pts_b = ik_chain_b.get_pts()[::-1]
                write_pts(sub_chain_a, moved_pts_a)
                write_pts(sub_chain_b, moved_pts_b)


        region = context.region
        rv3d = context.region_data
        mouse_co_2d = Vector((event.mouse_region_x, event.mouse_region_y))
        mw = self.curve_obj.matrix_world
        mw_inv = self.curve_obj.matrix_world.inverted()

        # *********************** TRANSLATE  **************************************************************************************
        if self.transform_mode == 'TRANSLATE':
            first_click_3d = view3d_utils.region_2d_to_location_3d(region, rv3d, self.first_click, depth_location=self.center_3dw)
            mouse_co_3d = view3d_utils.region_2d_to_location_3d(region, rv3d, mouse_co_2d, depth_location=self.center_3dw)
            offset = mouse_co_3d - first_click_3d
            offset = mw_inv @ Vector((offset.xyz[:] + (0,)))  # to obj space - as vector  (it is diff.co after all)
            offset = offset.to_3d()
            if self.x:
                offset = offset.project(Vector((1, 0, 0)))
            elif self.y:
                offset = offset.project(Vector((0, 1, 0)))
            elif self.z:
                offset = offset.project(Vector((0, 0, 1)))

            if move_mode in ('IK', 'ROOT'):
                for sp_id, orig_tip_co in zip(self.sel_spl_ids, self.original_spl_tips_co):
                    move_the_chains(splines[sp_id], target_co=orig_tip_co+offset)
            elif move_mode == 'FK':
                # spl_copy = deepcopy(self.sel_splines)
                for orig_spl in self.sel_splines.splines:  # with original init positions
                    bspl = splines[orig_spl.orig_spl_id]
                    out_pts = bspl.points
                    fallof_fun = get_fallof_fun(scn_ht.proportional_edit_falloff)
                    for out_p, p in zip(out_pts, orig_spl.points):
                        if scn_ht.use_proportional_edit:
                            dist_fallof = max(1-p.dist_w/scn_ht.influence_rad, 0)  # map dist - (0, inf), to (1, 0)
                            dist_fallof = fallof_fun(dist_fallof)
                        else:
                            dist_fallof = p.dist_w < 0.0001
                        out_p.position = p.co.lerp((p.co + offset), dist_fallof)  # lerp old pos, to new  target_co

                # spl_copy.write_splines_to_blender(self.curve_obj)
            else:
                pass


        # ************************* ROTATE  **************************************************************************************
        if self.transform_mode == 'ROTATE':

            def rot_from_mouse(center_3dw):
                center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, center_3dw, default=None)
                cen2d_to_fclick = self.first_click - center_2d
                cen_sec_click = mouse_co_2d - center_2d
                rot_diff_ang = cen2d_to_fclick.angle_signed(cen_sec_click, 0)  # for first click to current mouse pos

                view_dir = view3d_utils.region_2d_to_vector_3d(region, rv3d, center_2d)  # ? or: view_dir = rv3d.view_rotation * Vector((0.0, 0.0, -1.0))
                rot_dir = view_dir
                if self.x:
                    rot_dir = rot_dir.project(Vector((1, 0, 0)))
                elif self.y:
                    rot_dir = rot_dir.project(Vector((0, 1, 0)))
                elif self.z:
                    rot_dir = rot_dir.project(Vector((0, 0, 1)))

                # * rot in local space:
                # rot_mat_loc = mw_inv @ Quaternion(rot_dir, rot_diff_ang).to_matrix().to_4x4()  # get 3d rot, form screen rot
                rot_quat_local = Quaternion(rot_dir, rot_diff_ang)
                rot_quat_local.rotate(mw_inv)
                return rot_quat_local

            # Helpers
            no_rot = Quaternion((0.0, 1.0, 0.0), 0) # helper for blendering by fallof mask
            fallof_fun = get_fallof_fun(scn_ht.proportional_edit_falloff) # for 'FK' and 'ROOT' modes


            def rot_with_fallof(orig_spl, out_bspl, center, rot_q):
                '''Write rotated pts  to out_bspl '''
                pts_cnt = len(orig_spl.points)-1
                for i,(out_p, p) in enumerate(zip(out_bspl.points, orig_spl.points)):
                    center_to_p_rot = p.co - center
                    if move_mode == 'FK':
                        if scn_ht.use_proportional_edit:
                            dist_fallof = max(1-p.dist_w/scn_ht.influence_rad, 0)  # map dist - (0, inf), to (1, 0)
                            dist_fallof = fallof_fun(dist_fallof)
                        else:
                            dist_fallof = p.dist_w < 0.0001
                    else: # use p index for 'ROOT'
                        if scn_ht.use_proportional_edit:
                            dist_fallof = i/pts_cnt
                            dist_fallof = fallof_fun(dist_fallof)
                        else:
                            dist_fallof = 1


                    center_to_p_rot.rotate(no_rot.slerp(rot_q, dist_fallof))  # rotate orig vec...
                    out_p.position = center + center_to_p_rot  # lerp old pos, to new  target_co


            if move_mode == 'IK':
                rot_quat_l = rot_from_mouse(self.center_3dw)
                for sp_id, orig_tip_co in zip(self.sel_spl_ids, self.original_spl_tips_co):
                    center_to_pt_rot = orig_tip_co - self.center_l
                    center_to_pt_rot.rotate(rot_quat_l)  # rotate orig vec...
                    move_the_chains(splines[sp_id], target_co=self.center_l+center_to_pt_rot)

            elif move_mode == 'FK':
                rot_quat_l = rot_from_mouse(self.center_3dw)
                for orig_spl in self.sel_splines.splines:  # with original init positions
                    bspl = splines[orig_spl.orig_spl_id]
                    rot_with_fallof(orig_spl, bspl, center=self.center_l, rot_q = rot_quat_l)

            else: # 'ROOT' mode - rot aruond each spl root
                for orig_spl in self.sel_splines.splines:  # with original init positions
                    bspl = splines[orig_spl.orig_spl_id]
                    first_pt_co = bspl.points[0].position

                    rot_quat_l = rot_from_mouse(mw@first_pt_co)
                    rot_with_fallof(orig_spl, bspl, center=first_pt_co, rot_q = rot_quat_l)

        # ************************* SCALE  **************************************************************************************
        if self.transform_mode == 'SCALE':
            x = Vector((1, 0, 0))
            y = Vector((0, 1, 0))
            z = Vector((0, 0, 1))

            def calc_scale(center):
                '''Scale calc in world space'''
                center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, center, default=None)
                base_scale = max((self.first_click - center_2d).length, 0.01)
                new_dist = (mouse_co_2d - center_2d).length if not self.force_zero else 0.0
                return new_dist / base_scale

            fallof_fun = get_fallof_fun(scn_ht.proportional_edit_falloff)

            def scale_strands(orig_spl, out_bspl, center, scale_fac):
                ''' For FK and ROOT modes. Write scaled pts  to out_bspl '''
                # pts_cnt = len(orig_spl.points)-1
                for i, (out_p, p) in enumerate(zip(out_bspl.points, orig_spl.points)):
                    center_to_pt = p.co - center
                    if self.x:
                        scale_f = p.co + (scale_fac-1)*center_to_pt.project(x)
                    elif self.y:
                        scale_f = p.co + (scale_fac-1)*center_to_pt.project(y)
                    elif self.z:
                        scale_f = p.co + (scale_fac-1)*center_to_pt.project(z)
                    else:
                        scale_f = center + scale_fac*center_to_pt

                    if move_mode == 'FK':
                        if scn_ht.use_proportional_edit:
                            dist_fallof = max(1-p.dist_w/scn_ht.influence_rad, 0)  # map dist - (0, inf), to (1, 0)
                            dist_fallof = fallof_fun(dist_fallof)
                        else:
                            dist_fallof = p.dist_w < 0.0001
                    else: # use p index for 'ROOT'
                        dist_fallof = 1

                    out_p.position = p.co.lerp(scale_f, dist_fallof)  # lerp old pos, to new  target_co


            scale_fac = calc_scale(self.center_3dw)
            if move_mode == 'IK':
                for sp_id, orig_tip_co in zip(self.sel_spl_ids, self.original_spl_tips_co):
                    center_to_pt = orig_tip_co - self.center_l
                    if self.x:
                        scale_move = orig_tip_co + (scale_fac-1)*center_to_pt.project(x)
                    elif self.y:
                        scale_move = orig_tip_co + (scale_fac-1)*center_to_pt.project(y)
                    elif self.z:
                        scale_move = orig_tip_co + (scale_fac-1)*center_to_pt.project(z)
                    else:
                        scale_move = self.center_l + scale_fac*center_to_pt
                    move_the_chains(splines[sp_id], target_co=scale_move)

            elif move_mode == 'FK':
                for orig_spl in self.sel_splines.splines:  # with original init positions
                    bspl = splines[orig_spl.orig_spl_id]
                    out_pts = bspl.points
                    scale_strands(orig_spl, bspl, center=self.center_l, scale_fac=scale_fac)

            else: # 'ROOT' mode
                # scale = calc_scale(mw@first_pt_co)
                scale = self.first_click.x - event.mouse_region_x
                scale /= -50 # scale from (0, 200) to (0, 1)

                for spl_id, orig_spl in enumerate(self.sel_splines.splines):   # with original init positions
                    bspl = splines[orig_spl.orig_spl_id]
                    first_pt_co = bspl.points[0].position

                    if scale > 0:  # scale shorter strands toward longerst
                        if self.scale_unify:
                            spl_len_norm = self.root_len['max_lens'][spl_id] # shortest - 0, longest - 1
                        else:
                            spl_len_norm = 0.5
                        fac = min(scale, 1)
                        scale_by_strand = lerp(1, 1/spl_len_norm, fac)
                    else:  # scale longer strands toward shortest
                        if self.scale_unify:
                            spl_len_norm = self.root_len['min_lens'][spl_id] # shortest - 0, longest - 1
                        else:
                            spl_len_norm = 2
                        fac = min(-1*scale,1)
                        scale_by_strand = lerp(1, 1/spl_len_norm, fac)

                    scale_strands(orig_spl, bspl, center=first_pt_co, scale_fac=scale_by_strand)


    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES' and context.space_data.type == 'VIEW_3D'

    # backup hair curves tilt and posiitons attributes - for undo (since blender build in undo crashes on ESC)
    def backup_curves(self, obj):
        pos_attr = obj.data.attributes.get('position')
        pos_back = np.empty((len(pos_attr.data), 3), dtype=np.float32)
        pos_attr.data.foreach_get('vector', pos_back.ravel())

        tilt_attr = obj.data.attributes.get('tilt')
        if tilt_attr:
            tilt_back = np.empty(len(tilt_attr.data), dtype=np.float32)
            tilt_attr.data.foreach_get('value', tilt_back.ravel())
        else:
            tilt_back = None
        self.backup_data = {'position': pos_back, 'tilt': tilt_back}

    # restore backup (for undo)
    def restore_curves(self, obj):
        pos_attr = obj.data.attributes.get('position')
        pos_attr.data.foreach_set('vector', self.backup_data['position'].ravel())


        if self.backup_data['tilt'] is not None:
            tilt_attr = obj.data.attributes.get('tilt')
            tilt_attr.data.foreach_set('value', self.backup_data['tilt'].ravel())
        else:
            if obj.data.attributes.get('tilt'):
                obj.data.attributes.remove('tilt')


    def invoke(self, context, event):
        self.sel_spl_ids = []
        self.curve_obj = curve_obj = context.active_object
        self.x, self.y, self.z = False, False, False
        self.first_click = Vector((event.mouse_region_x, event.mouse_region_y))
        self.mouse_co_2d = Vector((event.mouse_region_x, event.mouse_region_y))

        self.scale_unify = False # for 'ROOT' mode to scale all strands to same length

        self.backup_curves(curve_obj)

        #* for tilt align
        scn_ht = context.scene.ht_props
        self.snap_surface = bpy.data.objects.get(curve_obj.ht_props.target_obj) if curve_obj.ht_props.target_obj and scn_ht.align_tilt else None
        if self.snap_surface:
            #* fix strand jump on tilt aligning, if obj.rot != 0
            if curve_obj.rotation_euler != Euler((0.0, 0.0, 0.0), 'XYZ') or min(curve_obj.scale[:]) < 0:  # fix rot if curve rot is non 0,0,0, or scele x,y,z <0
                override = {'selected_editable_objects': [curve_obj]}
                if curve_obj.data.users > 1:
                    curve_obj.data = curve_obj.data.copy()
                bpy.ops.object.transform_apply(override, location=False, rotation=True, scale=False)  # ! apply  scale breaks /changes pts radius
                # context.view_layer.update()

            if self.snap_surface:
                sel_strands = get_selected_strands_mask(context, self.curve_obj)
                self.align_mod = align_curves_tilt(self.curve_obj, sel_strands, apply_mod=False)


        if scn_ht.move_mode in ('FK','ROOT'):
            self.sel_splines = Splines(curve_obj, onlySelection=True)
            self.sel_splines.euclidean_seg_distances(context, curve_obj.matrix_world)  # ads p.dist_w - from sel point to each strand

        self.target_BVHT = None
        if scn_ht.move_mode in ('IK','ROOT') and scn_ht.use_collisions:
            if scn_ht.use_custom_collider:
                target = scn_ht.custom_collider
            else:
                target = curve_obj.data.surface # TODO: can try to fallback to hair system Setup modifier..
            if target:
                # in target mw space, and then back to curve space
                self.target_BVHT = get_obj_mesh_bvht(target, context.evaluated_depsgraph_get(), applyModifiers=False, world_space=True, m_inv=curve_obj.matrix_world.inverted())


        # if 'ROOT' mode and scale - calculate euclidean_len for each strand
        if scn_ht.move_mode == 'ROOT' and self.transform_mode == 'SCALE':
            splines = self.sel_splines.splines
            # normalize - to (0, 1) range
            self.spl_lens = [spl.euclidean_len for spl in splines]
            # calc standard deviation. If low, then all strands are close to same length - se we will scale uniform
            self.root_len = {}

            max_len = max(self.spl_lens)
            min_len = min(self.spl_lens)
            # self.spl_lens = [(l-min_len)/(max_len-min_len) for l in self.spl_lens]
            self.root_len['max_lens'] = [max(l/max_len, 0.01) for l in self.spl_lens]
            if min_len < 0.001:
                self.root_len['min_lens'] = [max(l/0.01, 0.01) for l in self.spl_lens]
            else:
                self.root_len['min_lens'] = [max(l/min_len, 0.01) for l in self.spl_lens]


        self.sel_spl_ids, _, pts_co = get_tips_pts(context, curve_obj, world_space=False, only_sel=True)
        if not self.sel_spl_ids:
            if self.transform_mode == "TRANSLATE":
                bpy.ops.transform.translate('INVOKE_DEFAULT', release_confirm = False)
            elif self.transform_mode == "SCALE":
                bpy.ops.transform.resize('INVOKE_DEFAULT', release_confirm = False)
            elif self.transform_mode == "ROTATE":
                bpy.ops.transform.rotate('INVOKE_DEFAULT', release_confirm = False)
            return {'FINISHED'}

        #* operation origing / center
        if scn_ht.move_mode in ('FK','IK'):
            if context.scene.tool_settings.transform_pivot_point == 'CURSOR':
                self.center_l = curve_obj.matrix_world.inverted() @ context.scene.cursor.location
            else:
                sel_center = Vector((0, 0, 0))
                for p_co in pts_co:
                    sel_center += p_co
                self.center_l = sel_center / len(pts_co)
        else: # 'ROOT' get average root pos
            roots = [spl.points[0].co for spl in self.sel_splines.splines]
            aver_co = Vector((0, 0, 0))
            for co in roots:
                aver_co += co
            self.center_l = aver_co / len(roots)

        self.center_3dw = curve_obj.matrix_world @ self.center_l
        self.original_spl_tips_co = pts_co

        #* cor circle drawing
        region = context.region
        rv3d = context.region_data
        self.center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw, default=None)

        self.force_zero = False
        self.calc_rad_2d(context)
        args = (self, context)
        self._handle_deb = bpy.types.SpaceView3D.draw_handler_add(self.draw_debug_3d, args, "WINDOW", "POST_VIEW")
        self._handle_2d_lines = bpy.types.SpaceView3D.draw_handler_add(self.draw_ui_2d, args, "WINDOW", "POST_PIXEL")
        if scn_ht.use_proportional_edit:
            self._handle_2d_circ = bpy.types.SpaceView3D.draw_handler_add(draw_2d_circle, args, "WINDOW", "POST_PIXEL") if scn_ht.move_mode == 'FK' else None
        else:
            self._handle_2d_circ = None
        context.window_manager.modal_handler_add(self)
        return {"RUNNING_MODAL"}

    def finish(self, context):
        if self._handle_deb:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_deb, "WINDOW")
        if self._handle_2d_lines:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_2d_lines, "WINDOW")
        # force_update(context.active_object) # looks like addin 'Surface Derom' modifier, will update curves ok
        if self._handle_2d_circ:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_2d_circ, "WINDOW")
        if self.snap_surface:
            node_tree = self.align_mod.node_group
            apply_mod_force(self.curve_obj, self.align_mod)
            bpy.data.node_groups.remove(node_tree)


        context.area.header_text_set(None)
        context.area.tag_redraw()

    def modal(self, context, event):
        scn_s = context.scene.ht_props
        self.mouse_co_2d = Vector((event.mouse_region_x, event.mouse_region_y))
        if event.type in {"LEFTMOUSE", 'RET'}:
            self.finish(context)
            return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            self.finish(context)
            self.restore_curves(self.curve_obj)
            context.active_object.data.update_tag()
            context.area.tag_redraw()
            return {'CANCELLED'}

        elif event.type == 'X' and event.value == 'PRESS':
            self.x, self.y, self.z = not self.x, False, False

        elif event.type == 'Y' and event.value == 'PRESS':
            self.x, self.y, self.z = False, not self.y, False

        elif event.type == 'Z' and event.value == 'PRESS':
            self.x, self.y, self.z = False, False, not self.z

        elif event.type == 'MOUSEMOVE':
            if event.shift:
                self.scale_unify = True
            else:
                self.scale_unify = False
            self.move_tip(context, event)
            context.area.tag_redraw()
            # return {'PASS_THROUGH'}
        elif event.type == "ZERO" and event.value == 'RELEASE':
            self.force_zero = not self.force_zero
            self.move_tip(context, event)
            context.area.tag_redraw()

        # NOTE: FK movement aka. 'normal'
        elif scn_s.move_mode == 'FK' and event.type == "WHEELDOWNMOUSE":
            scn_s.influence_rad *= 1.1
            self.move_tip(context, event)
        elif scn_s.move_mode == 'FK' and event.type == "WHEELUPMOUSE":
            scn_s.influence_rad = max(0.001, scn_s.influence_rad * 0.9)
            self.move_tip(context, event)

        # NOTE: IK movement
        # elif scn_s.move_mode == 'IK' and event.type == "WHEELUPMOUSE":
        #     context.scene.ht_props.clump_falloff *= 1.1
        #     self.move_tip(context, event)
        # elif scn_s.move_mode == 'IK' and event.type == "WHEELDOWNMOUSE":
        #     context.scene.ht_props.clump_falloff *= 0.9
        #     self.move_tip(context, event)

        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        return {"RUNNING_MODAL"}


class OBJECT_OT_SelectOffset(bpy.types.Operator):
    bl_idname = "hair.offset_selection"
    bl_label = "Select Offset"
    bl_description = "Select Offset"
    bl_options = {"REGISTER", "UNDO_GROUPED"}

    increase: bpy.props.BoolProperty(name='Increase', description='', default=True)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type in ('CURVES')

    def execute(self, context):
        offset_prev = context.scene.ht_props.select_offset
        ht_props = context.scene.ht_props
        obj = context.active_object
        next_offset = ht_props.select_offset + 1 if self.increase else ht_props.select_offset - 1
        longest_pid = 0
        if obj.type == 'CURVE':
            self.report({'WARNING'}, f'Old Curve type of objects are no longer supported')

        ht_workspace_select = get_set_ht_select_attr(obj) # Domain: Point
        for i, spl in enumerate(obj.data.curves):
            pts = spl.points
            spl_len = len(pts)
            last_pid = spl_len-1
            longest_pid = max(longest_pid, last_pid)

            if ht_props.active_end == "TIPS":
                p_id_current = clamp(last_pid - ht_props.select_offset, 0, last_pid)
                p_id_next = clamp(last_pid - next_offset, 0, last_pid)
            else:
                p_id_current = clamp(ht_props.select_offset, 0, last_pid)
                p_id_next = clamp(next_offset, 0, last_pid)
            sel_mask = [False]*spl_len
            if p_id_current != p_id_next:  # deselect old, select new one
                sel_mask[p_id_next] = ht_workspace_select.data[pts[p_id_current].index].value
                # pts.foreach_set('select', sel_mask)
                for pt, sel in zip(pts, sel_mask):
                    ht_workspace_select.data[pt.index].value = sel

            else:
                pass
                # if spl.type in {'NURBS', 'POLY'}:
                #     sel_mask[p_id_next] = True
                #     pts.foreach_set('select', sel_mask)
                # else:
                #     sel_mask[p_id_next] = True
                #     pts.foreach_set('select_control_point', sel_mask)
        obj.data.update_tag()
        ht_props.select_offset = clamp(next_offset, 0, longest_pid)
        # bpy.ops.ed.undo_push()
        bpy.context.area.tag_redraw()
        return {"FINISHED"}


class HTOOL_OT_HairSelectAll(bpy.types.Operator):
    bl_idname = "curve.hair_select_set"
    bl_label = "(de)Select hair alll"
    bl_description = "(de)Select hair alll"
    bl_options = {'REGISTER', "UNDO"}

    # select: bpy.props.BoolProperty(name='Select', description='Select All', default=True)
    mode: bpy.props.EnumProperty(name='Mode',
        items=[
            ('SELECT', 'Select', 'Select'),
            ('DESELECT', 'Deselect', 'Deselect'),
            ('INVERT', 'Invert', 'Invert selection')
        ], default='SELECT')


    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def execute(self, context):
        curve_obj = context.active_object
        workspace_de_select_all(context, curve_obj, self.mode)
        context.area.tag_redraw()
        return {'FINISHED'}


class HTOOL_OT_LmbHairSelect(bpy.types.Operator):
    bl_idname = "curve.hair_select"
    bl_label = "Select hair tip"
    bl_description = "Select Hair tips (MTool modeling)"
    bl_options = {'REGISTER', "UNDO"}

    extend: bpy.props.BoolProperty(name='Extend', description='', default=False)

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def mouse_click_select(self, event):
        return (event.type == 'LEFTMOUSE' and self.left_click_select) or (event.type == 'RIGHTMOUSE' and not self.left_click_select)

    def invoke(self, context, event):
        wm = bpy.context.window_manager
        km = wm.keyconfigs.user.keymaps.get('3D View Tool: Object, Hair Modeling')
        self.left_click_select = True
        if km:
            kmi = km.keymap_items['curve.hair_select']
            self.left_click_select = True if kmi.type == 'LEFTMOUSE' else False

        curve_obj = context.active_object
        if self.mouse_click_select(event) and event.value == 'PRESS':
            sel_spl_ids, sel_pts, sel_pts_co = get_tips_pts(context, curve_obj, world_space=True, only_sel=True)

            # handle blender .select attr
            spl_sel_mask = [False]*len(curve_obj.data.curves)

            if self.extend:  # deselect all pts first
                for sp_id in sel_spl_ids:
                    spl_sel_mask[sp_id] = True

            mouse_co = Vector((event.mouse_region_x, event.mouse_region_y))
            region = context.region
            rv3d = context.region_data
            best_dist_2d = 20
            overlapping_pts = [] #deal with selecting overlapping pts
            best_pt = None
            _, all_pts, pts_co = get_tips_pts(context, curve_obj, world_space=True, only_sel=False)
            mw = curve_obj.matrix_world
            for pt, p_co in zip(all_pts, pts_co):
                tip_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, mw @ p_co, default=None)
                if not tip_2d:
                    continue
                current_dist = (tip_2d-mouse_co).length
                if current_dist < best_dist_2d:
                    overlapping_pts.clear()
                    overlapping_pts.append(pt)
                    best_dist_2d = current_dist
                elif current_dist == best_dist_2d: # must overlap
                    overlapping_pts.append(pt)

            overlap_pts_cnt = len(overlapping_pts)

            if overlap_pts_cnt == 0: # deselect all (like blender does)
                if not self.extend:
                    workspace_select_by_spl_ids(context, curve_obj, [], deselect=True)
                    context.area.tag_redraw()
                return {'FINISHED'}

            elif overlap_pts_cnt == 1: #deal tips that are in same coordinates...
                best_pt = overlapping_pts[0]
            else: # more than one point under cursor
                sel_overlapping_pts = [op for op in overlapping_pts if op in sel_pts]
                # all points under cursor are selected - pick first unselected or
                if len(sel_overlapping_pts) == 0 or len(sel_overlapping_pts) == overlap_pts_cnt: # no point under cursor selected - just pick first
                    best_pt = overlapping_pts[0]
                else: # at least one point under cursor is selected already..., pick next one
                    prev_pt = overlapping_pts[-1]
                    i = 0
                    while True:
                        candidate = overlapping_pts[i % overlap_pts_cnt]
                        #not sel point, but with prev_pt selected
                        if candidate not in sel_overlapping_pts and prev_pt in sel_overlapping_pts:
                            best_pt = candidate
                            break
                        else:
                            prev_pt = candidate
                        i += 1
                        if i >= overlap_pts_cnt:
                            break

            if best_pt:
                found = False
                for i,spl in enumerate(curve_obj.data.curves):
                    for p in spl.points:
                        if p == best_pt:
                            spl_sel_mask[i] = True
                            found = True
                            break
                    if found:
                        break
                spl_sel_ids = [i for i, sel in enumerate(spl_sel_mask) if sel]
                workspace_select_by_spl_ids(context, curve_obj, spl_sel_ids, deselect=True)
            else: # no points under cursor - deselect all (like blender does)
                workspace_select_by_spl_ids(context, curve_obj, [], deselect=True)
            context.area.tag_redraw()

        return {'FINISHED'}


class HTOOL_MT_DeleteMenu(bpy.types.Menu):
    bl_idname = "HTOOL_MT_DeleteMenu"
    bl_label = "Hair Tool Menu Delete"
    @classmethod
    def poll(cls, context):
        return context.mode in ('OBJECT','SCULPT_CURVES')

    def draw(self, context):
        layout = self.layout
        layout.operator_enum("curve.hair_delete", "mode")


# fill dynamic enum mode
DEL_TYPE  = []
def uv_box_items(self, context):
    obj = context.active_object if context.active_object else None
    if not obj:
        return []
    global DEL_TYPE
    if obj.type == 'CURVES':
        if len(DEL_TYPE) != 3:
            DEL_TYPE = [
                     ('OBJECT', 'Object', 'Delete selected object'),
                     ('SPLINE', 'Spline', 'Delete whole strand (spline)'),
                     ('SLICE', 'Cut', 'Cut spline at selection point')
            ]
    elif obj.type == 'MESH':
        if len(DEL_TYPE) != 1:
            DEL_TYPE = [
                     ('OBJECT', 'Object', 'Delete selected object'),
            ]
    return DEL_TYPE

class HTOOL_OT_HairDelete(bpy.types.Operator):
    bl_idname = "curve.hair_delete"
    bl_label = "Hair Delete"
    bl_description = "Delete Hair curve"
    bl_options = {'REGISTER'}

    mode: bpy.props.EnumProperty(name='Mode', description='Delete mode',
                                 items=uv_box_items)

    @classmethod
    def poll(cls, context):
        return context.active_object

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "mode", expand=True)

    def invoke(self, context, event):
        # bpy.ops.ed.undo_push()  # to prevent restoring curve from step above on execute()
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):
        curve_obj = context.active_object
        if self.mode == 'OBJECT':
            if context.mode == 'SCULPT_CURVES':
                bpy.ops.curves.sculptmode_toggle()
            bpy.ops.object.delete(use_global=False)
            return {'FINISHED'}

        sel_spl_ids, sel_pts, _ = get_tips_pts(context, curve_obj, world_space=False, only_sel=True)
        if not sel_spl_ids:
            self.report({'WARNING'}, 'No curve endings selected! Cancelling')
            return {'CANCELLED'}
        splines = curve_obj.data.curves
        if self.mode == 'SPLINE':
            # for sp_id in reversed(sel_spl_ids):
            #     splines.remove(splines[sp_id])
            sel_strands = [i in sel_spl_ids for i in range(len(splines))]
            del_sel_strands(curve_obj, sel_strands)
        else:
            sel_splines = Splines(context.active_object, onlySelection=True)
            for spl in sel_splines.splines:
                for i, p in enumerate(spl.points):
                    if p.select:
                        del spl.points[i+1:]
                        spl.points[i].select = True
                        break
            context.scene.ht_props.select_offset = 0
            sel_splines.write_splines_to_blender(curve_obj)  # override old spl
        bpy.ops.ed.undo_push()
        return {'FINISHED'}


class HTOOL_OT_ResetRadiusTilt(bpy.types.Operator):
    bl_idname = "curve.hair_radilt_reset"
    bl_label = "Hair Radius/Tilt Reset"
    bl_description = "Hair Radius/Tilt Reset"
    bl_options = {'REGISTER', "UNDO"}

    mode: bpy.props.EnumProperty(name='Mode', description='Mode', items=[('RADIUS', 'Radius', 'Radius'), ('TILT', 'Tilt', 'Tilt')], default='RADIUS')

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def invoke(self, context, event):
        curve_obj = context.active_object

        attr_name = 'radius' if self.mode == 'RADIUS' else 'tilt'
        attr = curve_obj.data.attributes.get(attr_name)
        if not attr: # no attr, no need to reset
            return {'CANCELLED'}

        if context.mode in {'EDIT_CURVES', 'SCULPT_CURVES'}:
            if self.mode == 'RADIUS':
                bpy.ops.curves.set_attribute('INVOKE_DEFAULT',attr_name=attr_name, default_value=1, value=1)
            else:
                bpy.ops.curves.set_attribute('INVOKE_DEFAULT', attr_name=attr_name, default_value=0, value=0)

        else:  # mostly for hair workspace tool
            spl_ids, _, _ = get_tips_pts(context, curve_obj, only_sel=True)

            val = 1 if self.mode == 'RADIUS' else 0
            for spl_id in spl_ids:
                spl = curve_obj.data.curves[spl_id]
                for p in spl.points:
                    attr.data[p.index].value = val

        return {'FINISHED'}


class HTOOL_OT_AdjustRadTitl(bpy.types.Operator):
    bl_idname = "curve.hair_radilt_fallof"
    bl_label = "Radius/Tilt Adjustment"
    bl_description = "Proportianl Radius/Tilt Adjustment, that uses proportional editing fallof to adjust nearby points radius/tilt"
    bl_options = {'REGISTER', 'GRAB_CURSOR', 'BLOCKING', 'UNDO'}

    mode: bpy.props.EnumProperty(name='Mode', description='Smooth Axis', items=[
        ('RADIUS', 'Radius', 'Radius'),
        ('TILT', 'Tilt', 'Tilt')], default='RADIUS')

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def calc_rad_2d(self, context):
        '''For drawing circle'''
        region = context.region
        rv3d = context.region_data
        scn_s = context.scene.ht_props
        center_2d_plus_rad_3d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw + Vector((scn_s.influence_rad, 0, 0)), default=Vector((0, 0)))
        self.rad_draw_2d = (center_2d_plus_rad_3d - self.center_2d).length

    def invoke(self, context, event):
        curve_obj = context.active_object
        if len(curve_obj.data.curves) == 0:
            self.report({'ERROR'}, "There are no curves in active object. You can use Ctrl+Shift+H > 'Bake..' to convert procedural strands into 'real' geometry. Cancelling" )
            return {'CANCELLED'}
        self.sel_splines = Splines(curve_obj, onlySelection=True)
        self.sel_splines.euclidean_seg_distances(context, curve_obj.matrix_world) # ? for what
        self.first_click = Vector((event.mouse_region_x, event.mouse_region_y))
        self.scale_fac = 1  # radius scale factor

        self.sel_spl_ids, _, pts_co = get_tips_pts(context, curve_obj, world_space=False, only_sel=True)
        if not self.sel_spl_ids:
            self.sel_spl_ids, _, pts_co = get_tips_pts(context, curve_obj, world_space=False, only_sel=False)


        if context.scene.tool_settings.transform_pivot_point == 'CURSOR':
            op_center = curve_obj.matrix_world.inverted() @ context.scene.cursor.location
        else:
            sel_center = Vector((0, 0, 0))
            for p_co in pts_co:
                sel_center += p_co
            op_center = sel_center / len(pts_co)
        self.center_3dw = curve_obj.matrix_world @ op_center

        region = context.region
        rv3d = context.region_data
        self.center_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.center_3dw, default=None)

        self.force_zero = False
        self.calc_rad_2d(context)  # for drawing
        scn_ht = context.scene.ht_props
        if scn_ht.use_proportional_edit:
            args = (self, context)
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_2d_circle, args, "WINDOW", "POST_PIXEL")
        else:
            self._handle = None
        context.window_manager.modal_handler_add(self)
        return {"RUNNING_MODAL"}

    def update_draw_header(self, context):
        scn_s = context.scene.ht_props
        if self.mode == 'RADIUS':
            txt = 'Radius'
            val = self.scale_fac
        else:
            txt = 'Tilt'
            val = degrees(self.scale_fac)
        context.area.header_text_set(f"{txt}: {val:.2f}   Proportional Size: {scn_s.influence_rad:.2f}")

    def modal(self, context, event):
        scn_s = context.scene.ht_props
        if event.type == "MOUSEMOVE":
            self.adjust_radilt(context, event)

        if event.type == "LEFTMOUSE":
            return self.finish(context)

        if event.type == "ZERO" and event.value == 'RELEASE':
            self.force_zero = not self.force_zero
            self.adjust_radilt(context, event)

        if event.type == "WHEELDOWNMOUSE":
            scn_s.influence_rad *= 1.1
            self.adjust_radilt(context, event)

        if event.type == "WHEELUPMOUSE":
            scn_s.influence_rad = max(0.001, scn_s.influence_rad * 0.9)
            self.adjust_radilt(context, event)

        if event.type in {"RIGHTMOUSE", "ESC"}:
            return self.cancelled(context)

        return {"RUNNING_MODAL"}

    def finish(self, context):
        if self._handle:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
        context.area.tag_redraw()
        context.area.header_text_set(None)
        return {"FINISHED"}

    def cancelled(self, context):
        if self._handle:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
        context.area.tag_redraw()
        self.sel_splines.write_splines_to_blender(context.active_object)
        context.area.header_text_set(None)
        return {"CANCELLED"}

    def adjust_radilt(self, context, event):
        self.calc_rad_2d(context)  # for drawing
        self.update_draw_header(context)
        obj = context.active_object
        mode_rad = self.mode == 'RADIUS'
        mouse_co = Vector((event.mouse_region_x, event.mouse_region_y))
        base_scale = abs(max(self.first_click.x - self.center_2d.x, 0.01))
        if self.force_zero:
            new_dist = 0
        else:
            new_dist = mouse_co.x - self.center_2d.x

        if mode_rad:
            scale_fac = abs(new_dist) / base_scale  # ==1 for default radius 1. For thilt compenstae later to 0
        else:
            scale_fac = (new_dist - base_scale) / 200  # ==0 for default tilt 0. For thilt compenstae later to 0

        self.scale_fac = scale_fac
        self.update_draw_header(context)
        spl_copy = deepcopy(self.sel_splines)
        scn_ht = context.scene.ht_props
        fallof_fun = get_fallof_fun(scn_ht.proportional_edit_falloff)
        for spl in spl_copy.splines:
            for i,p in enumerate(spl.points):
                if scn_ht.use_proportional_edit:
                    dist_fallof = max(1-p.dist_w/scn_ht.influence_rad, 0)  # map dist - (0, inf), to (1, 0)
                    dist_fallof = fallof_fun(dist_fallof)
                else:
                    dist_fallof = p.dist_w < 0.0001

                if mode_rad:
                    spl.radii[i] = lerp(spl.radii[i],spl.radii[i] * scale_fac, dist_fallof) # basically lerp old rad, to new rad aka. scale_fac
                else:  # (scale_fac-1) - cos reset tilt is 0, not 1
                    spl.tilts[i] = lerp(spl.tilts[i], spl.tilts[i] + scale_fac, dist_fallof)  # basically lerp old rad, to new rad aka. scale_fac

        # spl_copy.write_splines_to_blender(context.active_object)  # override old spl
        if mode_rad:
            [sp.write_rad_to_blender_spl(obj) for sp in spl_copy.splines]
        else:
            [sp.write_tilt_to_blender_spl(obj) for sp in spl_copy.splines]
        obj.update_tag()
        context.view_layer.update()
        context.area.tag_redraw()
        return {'FINISHED'}


class HTOOL_OT_HairDuplicate(bpy.types.Operator):
    bl_idname = "curves.hair_duplicate"
    bl_label = "Hair Duplicate"
    bl_description = "Duplicate Hair curve(s)"
    bl_options = {'REGISTER', "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def execute(self, context):
        curve_obj = context.active_object
        init_strands_cnt = len(curve_obj.data.curves)
        if context.mode == 'EDIT_CURVES':
            sel_mask = get_edit_selected_points_bool(curve_obj) # based on blender selection '.select' atrib
            point_mode = True
        else:
            sel_mask = get_selected_strands_mask(context, curve_obj) # based on blender selection '.select' atrib
            point_mode = False
        if sum(sel_mask) == 0:
            # print("Nothing to Duplicate. No strands selected. Cancelling")
            # bpy.ops.object.duplicate()
            bpy.ops.object.duplicate()
            bpy.ops.transform.translate('INVOKE_DEFAULT')
            return {'FINISHED'}
        if sum(sel_mask) == 0:
            self.report({'WARNING'}, "Nothing selected")
            return {'CANCELLED'}
        else:
            duplicate_sel(curve_obj, sel_mask, point_mode=point_mode)
            self.report({'INFO'}, f"Duplicated {sum(sel_mask)} {'points' if point_mode else 'strands'}")
        # deselect original curves, and select duplis (dupli if   spl_id > init_strands_cnt)

        spl_ids, pts, _ = get_tips_pts(context, curve_obj, world_space=True, only_sel=False)
        ht_workspace_select = get_set_ht_select_attr(curve_obj)
        for spl_id, p in zip(spl_ids, pts):
            ht_workspace_select.data[p.index].value = spl_id >= init_strands_cnt

        if curve_obj.ht_props.use_auto_uv:
            auto_update_all_curves_uv_regions(curve_obj)
        if context.mode == 'EDIT_CURVES': # use default blender transform
            bpy.ops.transform.translate('INVOKE_DEFAULT')
        else: # for Obj mode, or Sculpt mode use HT Transform
            bpy.ops.curves.hair_transform('INVOKE_DEFAULT', transform_mode='TRANSLATE')
        return {'FINISHED'}


class HTOOL_OT_HairRegionSelect(bpy.types.Operator):
    bl_idname = "hair.select_region_modal"
    bl_label = "Select Hair Region"
    bl_options = {"REGISTER", "UNDO"}

    sel_type: bpy.props.EnumProperty(name='selection type',
                                     items=(
                                         ('BOX', 'Box', ''),
                                         ('LASSO', 'Lasso', ''),
                                         ('CIRCLE', 'Circle', '')), default='BOX')
    mode: bpy.props.EnumProperty(name='mode',
                                 items=(
                                     ('ADD', 'Add', ''),
                                     ('SUB', 'Substract', ''),
                                     ('SET', 'Set', '')), default='ADD')

    clicked = False

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    @staticmethod
    def draw_line_loop_2d(self, context):
        if self.mouse_pts:
            gpu.state.blend_set("ALPHA") # if bpy.app.version >= (3, 5, 0) else bgl.glEnable(bgl.GL_BLEND)
            batch_edges = batch_for_shader(shader_2d_flat, 'LINE_LOOP', {"pos": self.mouse_pts})
            shader_2d_flat.bind()
            shader_2d_flat.uniform_float("color", (1, 1, 1, 1))
            batch_edges.draw(shader_2d_flat)
            gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)


    def select_circle(self, context, curve_obj, center2d):
        region = context.region
        rv3d = context.region_data
        _, _, pts_co = get_tips_pts(context, curve_obj, world_space=True, only_sel=False)
        pre_sel_spl_ids, _, _ = get_tips_pts(context, curve_obj, world_space=False, only_sel=True)
        mw = curve_obj.matrix_world

        # for blender .select  attr
        spl_sel_mask = [False]*len(curve_obj.data.curves)
        for sel_id in pre_sel_spl_ids:
            spl_sel_mask[sel_id] = True

        for idx, p_co in enumerate(pts_co):
            tip_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, mw@p_co, default=None)
            if (tip_2d-center2d).length < context.scene.ht_props.circle_sel_rad2d:
                spl_sel_mask[idx] = self.mode == 'ADD'
        spl_sel_ids = [i for i, sel in enumerate(spl_sel_mask) if sel]
        workspace_select_by_spl_ids(context, curve_obj, spl_sel_ids, deselect=True)

    def select_box(self, context, curve_obj):
        ''' Not used anymore'''
        region = context.region
        rv3d = context.region_data
        min_x, max_x = min(self.mouse_pts[0].x, self.mouse_pts[2].x), max(self.mouse_pts[0].x, self.mouse_pts[2].x)
        min_y, max_y = min(self.mouse_pts[0].y, self.mouse_pts[2].y), max(self.mouse_pts[0].y, self.mouse_pts[2].y)
        for curve in curve_obj.data.splines:
            tip_world = curve_obj.matrix_world @ curve.points[-1].co.xyz
            tip_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, tip_world, default=None)
            if min_x < tip_2d.x < max_x and min_y < tip_2d.y < max_y:
                curve.points[-1].select = True if self.mode == 'ADD' else False

    def is_point_inside_winding(self, context, curve_obj):
        if not self.mouse_pts:
            return
        mouse_pts = self.mouse_pts
        region = context.region
        rv3d = context.region_data
        _, pts, pts_co = get_tips_pts(context, curve_obj, world_space=True, only_sel=False)
        pre_sel_spl_ids, _, _ = get_tips_pts(context, curve_obj, world_space=False, only_sel=True)
        mw = curve_obj.matrix_world

        # for blender .select  attr
        spl_sel_mask = [False]*len(curve_obj.data.curves)
        for sel_id in pre_sel_spl_ids:
            spl_sel_mask[sel_id] = True

        for idx, p_co in enumerate(pts_co):
            rad = 0
            tip_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, mw@p_co, default=None)
            prev_V = mouse_pts[-1]
            for i in range(len(mouse_pts)):
                next_V = mouse_pts[i]
                rad += (prev_V-tip_2d).angle_signed((next_V-tip_2d))
                prev_V = mouse_pts[i]
            if abs(abs(rad) - 6.283) < 0.1:  # is it close to 2pi= 6.283 ?
                spl_sel_mask[idx] = self.mode == 'ADD'
        spl_sel_ids = [i for i, sel in enumerate(spl_sel_mask) if sel]
        workspace_select_by_spl_ids(context, curve_obj, spl_sel_ids, deselect=True)

    def mouse_click_select(self, event):
        return (event.type == 'LEFTMOUSE' and self.left_click_select) or (event.type == 'RIGHTMOUSE' and not self.left_click_select)

    def modal(self, context, event):
        if self.sel_type == 'CIRCLE':
            scn_s = context.scene.ht_props
            if event.type == "WHEELUPMOUSE":
                scn_s.circle_sel_rad2d = ceil(scn_s.circle_sel_rad2d * 1.1)
                self.rad_draw_2d = scn_s.circle_sel_rad2d

            elif event.type == "WHEELDOWNMOUSE":
                scn_s.circle_sel_rad2d = floor(max(0.001, scn_s.circle_sel_rad2d * 0.9))
                self.rad_draw_2d = scn_s.circle_sel_rad2d

            elif self.mouse_click_select(event) and event.value == 'PRESS':
                self.clicked = True
                self.mode = 'ADD'

            elif event.type == "MIDDLEMOUSE" and event.value == 'PRESS':
                self.clicked = True
                self.mode = 'SUB'

            elif event.type == "MOUSEMOVE":
                self.center_2d = Vector((event.mouse_region_x, event.mouse_region_y))
                if self.clicked:
                    self.select_circle(context, self.curve_obj, center2d=self.center_2d)
                context.area.tag_redraw()

            elif self.mouse_click_select(event) and event.value == 'RELEASE':
                self.clicked = False

            elif event.type in {"RIGHTMOUSE", "ESC", "RET"}:
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
                context.area.tag_redraw()
                return {"FINISHED"}


        elif self.sel_type == 'BOX':
            if self.mouse_click_select(event) and event.value == 'PRESS':
                self.first_co = Vector((event.mouse_region_x, event.mouse_region_y))
                self.clicked = True

            elif event.type == "MOUSEMOVE":
                if self.clicked:
                    a = self.first_co
                    b = Vector((event.mouse_region_x, self.first_co.y))
                    c = Vector((event.mouse_region_x, event.mouse_region_y))
                    d = Vector((self.first_co.x, event.mouse_region_y))
                    self.mouse_pts = [a, b, c, d]
                context.area.tag_redraw()

            elif self.mouse_click_select(event) and event.value == 'RELEASE':
                self.is_point_inside_winding(context, self.curve_obj)
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
                context.area.tag_redraw()
                return {"FINISHED"}

        elif self.sel_type == 'LASSO':
            if event.type == "MOUSEMOVE":
                self.mouse_pts.append(Vector((event.mouse_region_x, event.mouse_region_y)))
                context.area.tag_redraw()

            elif self.mouse_click_select(event) and event.value == 'RELEASE':
                self.is_point_inside_winding(context, self.curve_obj)
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
                context.area.tag_redraw()
                return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
            context.area.tag_redraw()
            return {"CANCELLED"}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        self.curve_obj = context.active_object
        self.mouse_pts = []

        self.center_2d = event.mouse_region_x, event.mouse_region_y #for 'CIRCLE'
        self.rad_draw_2d = context.scene.ht_props.circle_sel_rad2d  # for 'CIRCLE'

        #get left click select, or right click
        if event.type in {"RIGHTMOUSE"}: # had  {"RIGHTMOUSE", "EVT_TWEAK_R"}  in 3.1 and before
            self.left_click_select = False
            self.first_co = Vector((event.mouse_region_x, event.mouse_region_y))
            self.clicked = True
        elif event.type in {"LEFTMOUSE"}: # had  {"RIGHTMOUSE", "EVT_TWEAK_L"}  in 3.1 and before
            self.left_click_select = True
            self.first_co = Vector((event.mouse_region_x, event.mouse_region_y))
            self.clicked = True
        else:
            self.left_click_select = True
            # wm = bpy.context.window_manager
            # km = wm.keyconfigs.user.keymaps.get('3D View Tool: Object, Hair Modeling')
            # if km:
            #     if self.sel_type == 'LASSO':
            #         for kmi in km.keymap_items:
            #             if kmi.idname == 'hair.select_region_modal' and kmi.properties.sel_type == 'LASSO' and kmi.properties.mode == self.mode: # kmi.type == 'EVT_TWEAK_L'
            #                 self.left_click_select = True if kmi.type == 'EVT_TWEAK_L' else False
            #                 break
            #     else:
            #         kmi = km.keymap_items['curve.hair_select']
            #         self.left_click_select = True if kmi.type == 'LEFTMOUSE' else False

        args = (self, context)
        if self.sel_type in {'LASSO','BOX'}:
            self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_line_loop_2d, args, "WINDOW", "POST_PIXEL")
        else: #* circle sel
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_2d_circle, args, "WINDOW", "POST_PIXEL")
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}


class HTOOL_OT_CurvesAdjustLen(bpy.types.Operator):
    bl_label = "Hair lengthen"
    bl_idname = "curve.curve_lenghten"
    bl_description = "Extend or shorten selected strands"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'CURVES'

    def invoke(self, context, event):
        self.scale_strand = 100
        self.sel_splines = Splines(context.active_object, onlySelection=True) # XXX: was using - since Flat will auto resample tilt, radius
        self.pt_cnt = [s.length for s in self.sel_splines.splines]
        self.euclid_len = [s.euclidean_len for s in self.sel_splines.splines]  # reduce growth on long strands, or else they blow too fast
        for s in self.sel_splines.splines:
            s.circle_from_3points() # calculate circle from spl last 3 points
        self.t_spat = []
        # for spl in self.sel_splines.splines:
        #     seg_lens = np.apply_along_axis(np.linalg.norm, 1, spl.points_co[:-1, :]-spl.points_co[1:, :])  # do an - an-1 and get length
        #     t_spat = np.insert(seg_lens, 0, 0).cumsum()  # add zero and sum
        #     self.t_spat.append(t_spat/t_spat[-1])  # it will be now proportional to segments distances - bigger point spacing - more time.
        for spl in self.sel_splines.splines:
            prev_p = spl.points[0].co
            t_spat = [0]
            for p in spl.points[1:]:
                t_spat.append(t_spat[-1] + (p.co-prev_p).length)
                prev_p = p.co
            self.t_spat.append(np.array(t_spat)/t_spat[-1])


        self.init_cox_2d = event.mouse_region_x

        bpy.ops.ed.undo_push()  # to prevent restoring curve from step above on execute()
        context.window_manager.modal_handler_add(self)
        # return self.execute(context)
        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        if event.type == "MOUSEMOVE":
            scale = clamp(1 + (event.mouse_region_x - self.init_cox_2d)/600, 0.2, 3)
            context.area.header_text_set(f"Factor: {scale:.2f}")
            self.extend_shorten_spl(context, scale)
            context.area.tag_redraw()

        elif event.type in {"LEFTMOUSE", "RET"}:
            context.area.header_text_set(None)
            context.area.tag_redraw()
            return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            context.area.header_text_set(None)
            bpy.ops.ed.undo()
            context.area.tag_redraw()
            return {"CANCELLED"}

        return {'RUNNING_MODAL'}

    def circle_from_3points(p1, p2, p3):
        p12 = (p2 - p1).normalized()
        p13 = (p3 - p1).normalized()
        circle_norm = p12.cross(p13) # XXX: if cross prouct is close to 0 - we have parallel lines..
        A = np.array(p1)
        B = np.array(p2)
        C = np.array(p3)
        a = np.linalg.norm(C - B)
        b = np.linalg.norm(C - A)
        c = np.linalg.norm(B - A)
        s = (a + b + c) / 2
        rad = a*b*c / 4 / np.sqrt(s * (s - a) * (s - b) * (s - c))
        b1 = a*a * (b*b + c*c - a*a)
        b2 = b*b * (a*a + c*c - b*b)
        b3 = c*c * (a*a + b*b - c*c)
        center = np.column_stack((A, B, C)).dot(np.hstack((b1, b2, b3)))
        center /= b1 + b2 + b3
        cent_vec = Vector(center)
        circle_ox = p3 - cent_vec # x arm of circle
        circle_oy = circle_norm.cross(circle_ox).normalized() # y arm of circle
        return circle_ox, circle_oy, rad, center

    def eval_circle(t, ox, oy, center, rad):
        # at t=0, we get ox axis
        x = rad * cos(t)*ox + center
        y = rad * sin(t)*oy + center
        return x + y

    def extend_shorten_spl(self, context, scale):
        splines = deepcopy(self.sel_splines)
        max_e_len = max(self.euclid_len)
        for e_len, t_spat, pts_cnt, spl in zip(self.euclid_len, self.t_spat, self.pt_cnt, splines.splines):
            t_range = pts_cnt - 1.01
            tn_new = scale*t_range  # adjust tn to get even spacing
            if scale > 100: # make long hair glow slower
                extra_t = tn_new - t_range
                grow_mul = max_e_len/e_len
                tn_new = t_range + extra_t * grow_mul
                cut_t = grow_mul*(scale-1)+1
                new_tns = np.linspace(0, cut_t, pts_cnt, endpoint=True)

            elif scale < 100:
                t_cut = scale
                new_tns = np.linspace(0, t_cut, pts_cnt, endpoint=True)
            spl.resample_pts(pts_cnt, uniform=False, t_custom=new_tns)

        splines.write_splines_to_blender(context.active_object)
        return {"FINISHED"}



class HTOOL_OT_HairModalKnife(bpy.types.Operator):
    bl_idname = "hair.knife_cut_modal"
    bl_label = "Hair Knife Cut"
    bl_options = {"REGISTER", "UNDO"}

    @staticmethod
    def draw_line_loop_2d(self, context):
        gpu.state.point_size_set(4)
        draw_pts = [self.mouse_hover_co]
        if self.mouse_pts:
            draw_pts.extend(self.mouse_pts)
        batch_pts = batch_for_shader(shader_2d_flat, 'POINTS', {"pos": draw_pts})
        shader_2d_flat.bind()
        shader_2d_flat.uniform_float("color", (0.6, 1, 0.4, 1))
        batch_pts.draw(shader_2d_flat)
        gpu.state.point_size_set(1)

        if self.mouse_pts:
            batch_edges = batch_for_shader(shader_2d_flat, 'LINE_STRIP', {"pos": self.mouse_pts+[self.mouse_hover_co]})
            shader_2d_flat.bind()
            shader_2d_flat.uniform_float("color", (0.6, 1, 0.4, 1))
            batch_edges.draw(shader_2d_flat)


    def modal(self, context, event):
        if event.type == "LEFTMOUSE" and event.value == 'RELEASE':
            self.mouse_pts.append(Vector((event.mouse_region_x, event.mouse_region_y)))
            if len(self.mouse_pts) >= 2:
                self.b_splines.splines_line_intersect(context, self.curve_obj, self.mouse_pts[-2], self.mouse_pts[-1]) #cut in screen space..
                # reselect cut strands
                spl_sel_ids = [spl.orig_spl_id for spl in self.b_splines.splines]
                workspace_select_by_spl_ids(context,  self.curve_obj, spl_sel_ids, deselect=True)
            context.area.tag_redraw()

        elif event.type == 'MOUSEMOVE':
            self.mouse_hover_co = Vector((event.mouse_region_x, event.mouse_region_y))
            context.area.tag_redraw()

        elif event.type in {"RET", "SPACE"}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
            context.area.tag_redraw()
            return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, "WINDOW")
            bpy.ops.ed.undo()
            context.active_object.data.update_tag()
            context.area.tag_redraw()
            return {"CANCELLED"}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        bpy.ops.ed.undo_push()
        self.curve_obj = context.active_object
        self.mouse_hover_co = Vector((event.mouse_region_x, event.mouse_region_y))
        self.mouse_pts = []
        self.b_splines = Splines(self.curve_obj, onlySelection=True)

        args = (self, context)
        self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_line_loop_2d, args, "WINDOW", "POST_PIXEL")
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}
