# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from mathutils import  Matrix, Euler
from mathutils.geometry import intersect_point_line
from bpy.app.handlers import persistent
from .jiggle_chain_b4 import SpringChains

import gpu
from gpu_extras.batch import batch_for_shader


shader = gpu.shader.from_builtin('UNIFORM_COLOR')


# define signed distance field for rounded cone
# https://iquilezles.org/www/articles/distfunctions/distfunctions.htm
# def sdRoundCone(pbone, point):
#     # sampling independent computations (only depend on shape)
#     r1 = pbone.bone.head_radius
#     r2 = pbone.bone.tail_radius
#     head_tail_vec = pbone.tail - pbone.head
#     l2 = head_tail_vec.dot(head_tail_vec)
#     rad_diff = r1 - r2
#     a2 = l2 - rad_diff*rad_diff
#
#     # sampling dependant computations
#     head_p_vec = point - pbone.head
#     y = head_p_vec.dot(head_tail_vec)
#     z = y - l2
#     help = head_p_vec*l2 - head_tail_vec*y
#     x2 =  help.dot(help)
#     y2 = y*y*l2
#     z2 = z*z*l2
#
#     k =  copysign(rad_diff*rad_diff*x2, rad_diff) # copysign(a,b) =  a * sign(b)
#     if copysign(a2*z2, z) > k:
#         return  sqrt(x2 + z2) /l2 - r2
#     if copysign(a2*y2, y) < k:
#         return  sqrt(x2 + y2) /l2 - r1
#     return (sqrt(x2*a2/l2)+y*rad_diff)/l2 - r1
def rem_copy_rot_constraint(arma, pbone):
    copy_rot = pbone.constraints.get('Set Rotation')
    if copy_rot:
        world_rot_from_constr = Euler((copy_rot.min_x, copy_rot.min_y, copy_rot.min_z)).to_matrix().to_4x4()
        rot_from_constr = arma.convert_space(pose_bone=pbone, matrix=world_rot_from_constr, from_space='WORLD', to_space='LOCAL')
        pbone.rotation_quaternion = rot_from_constr.to_quaternion()
        pbone.constraints.remove(copy_rot)


def bone_hit_simple(pbone, point):
    ''' Simple bbox check - sphere vs point'''
    center = (pbone.head+pbone.tail)/2
    rad = (pbone.head-pbone.tail).length
    rad += max(pbone.bone.head_radius , pbone.bone.tail_radius ) + 2*pbone.bone.envelope_distance
    rad *= 0.5
    return False if (point-center).length > rad else True


def bone_hit_detect_aprox(pbone, point):
    ''' aproximated with cone. Less accurate version
    (we hit straight perpendicular to |head tail| - rather than at angle caused by r1-r2) but should be good enough
    return outside - bool, normal, hit_co '''
    line_hit, a_fac = intersect_point_line(point, pbone.head, pbone.tail) # intersect_point_line(p, a, b)

    #NOTE: we have to clamp line to segment hit
    if a_fac < 0:
        line_hit = pbone.head
        rad = pbone.bone.head_radius
    elif a_fac > 1:
        line_hit = pbone.tail
        rad = pbone.bone.tail_radius
    else:
        rad = pbone.bone.head_radius*(1-a_fac) + pbone.bone.tail_radius*a_fac # lerp from ra to rb

    colision_offset = pbone.bone.envelope_distance
    if (line_hit - point).length > rad+1.1*colision_offset: # fast check if we are outside
        return False, None, None

    normal = (point - line_hit).normalized()
    surf_hit = line_hit +  (rad+colision_offset)*normal

    surf_hit_to_p = point - surf_hit # vec from inside line to capsule surface hit
    is_inside = surf_hit_to_p.dot(normal) < 0
    #dist = (surf_hit-point).length
    return is_inside, normal, surf_hit


def bone_hit_detect_exact(pbone, point):
    ''' cone hit detection - more accuartate but not 100% exact
    (we hit at angle defined by r1-r2 difference
    offset_fac = 1.02 => move 2% outside of surface
    return outside - bool, normal, hit_co '''
    r_head = pbone.bone.head_radius
    r_tail = pbone.bone.tail_radius
    colision_offset = pbone.bone.envelope_distance
    a_p = point - pbone.head
    a_b = pbone.tail - pbone.head
    ab_len = a_b.length

    ap_projected_ab = a_p.project(a_b)
    ap_perp = a_p - ap_projected_ab
    cone_slope = (r_head-r_tail)/ ab_len
    ap_fix = cone_slope * ap_perp.length # dist from P_projected' to  P2' - point projected on a_b based on dr/ab_len slope

    line_hit = pbone.head + ap_projected_ab - ap_fix * (a_b/ab_len)

    a_to_hit = line_hit-pbone.head

    # NOTE: we hit cone line a_b -> now clamp and  move line_hit outside to surface of cone
    sign = -1 if a_to_hit.dot(a_b) < 0 else 1
    a_fac = sign * a_to_hit.length/ab_len # a_fac - distance from the first point of a_b line as a percentage of the length of the line

    # NOTE: clamp line to segment hit
    if a_fac < 0:
        line_hit = pbone.head
        rad = r_head
    elif a_fac > 1:
        line_hit = pbone.tail
        rad = r_tail
    else:
        rad = r_head*(1-a_fac) + r_tail*a_fac # lerp from ra to rb

    line_hit_to_p = point - line_hit

    normal = line_hit_to_p.normalized()
    surf_hit = line_hit +  (rad+colision_offset)*normal

    surf_hit_to_p_vec = point - surf_hit # vec from inside line to capsule surface hit
    is_inside = surf_hit_to_p_vec.dot(normal) < 0
    return is_inside, normal, surf_hit

class OBJECT_OT_Select_Jiggle_Group(bpy.types.Operator):
    bl_idname = "object.select_jiggle_group"
    bl_label = "Select bones in group"
    bl_description = "Select jiggle bones assigned to active group"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE' and obj.mode == 'POSE' and len(obj.ht_props.jiggle_props.physics_props) > 0

    def execute(self, context):
        obj = context.active_object
        jiggle_props = obj.ht_props.jiggle_props

        # NOTE: setup jiggle bones color - using bone_groups
        physics_props_cnt = len(jiggle_props.physics_props)
        if jiggle_props.active_index > physics_props_cnt-1:
            self.report({'WARNING'}, f'Active jiggle group index is out of range.')
            return {'CANCELLED'}

        for pbone in obj.pose.bones:
            pbone.bone.select = False # deselect first all
            if pbone.ht_props.jiggle_index == jiggle_props.active_index:
                pbone.bone.select = True

        return {"FINISHED"}

class OBJECT_OT_Assign_Bones_Jiggle(bpy.types.Operator):
    bl_idname = "object.assign_jiggle"
    bl_label = "Assign Jiggle to bones"
    bl_description = "Assign jiggle physics to selected pose bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return context.selected_pose_bones and obj.ht_props.jiggle_props.active_index < len(obj.ht_props.jiggle_props.physics_props)

    @staticmethod
    def assign_jiggle_to_bones(obj, pbones):
        jiggle_props = obj.ht_props.jiggle_props

        # NOTE: setup jiggle bones color - using bone_groups
        physics_props_cnt = len(jiggle_props.physics_props)
        if physics_props_cnt == 0:
            OBJECT_OT_Add_Jiggle_Settings.add_jiggle_settings(obj)
            jiggle_props.active_index = 0
        if jiggle_props.active_index < physics_props_cnt:
            jiggle_props.active_index = len(jiggle_props.physics_props)-1

        jiggle_bone_group = obj.pose.bone_groups[jiggle_props.active_index]
        for pbone in pbones:
            pbone.bone_group = jiggle_bone_group
            pbone.ht_props.jiggle_index = jiggle_props.active_index

    def execute(self, context):
        obj = context.active_object
        self.assign_jiggle_to_bones(obj, context.selected_pose_bones)
        return {"FINISHED"}


class OBJECT_OT_Remove_Bones_Jiggle(bpy.types.Operator):
    bl_idname = "object.remove_bones_jiggle"
    bl_label = "Remove Jiggle from bones"
    bl_description = "Remove jiggle physics from selected pose bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.selected_pose_bones

    def execute(self, context):
        for pbone in context.selected_pose_bones:
            pbone.bone_group = None
            pbone.ht_props.jiggle_index = -1
            rem_copy_rot_constraint(context.active_object, pbone)
        return {"FINISHED"}


class OBJECT_OT_Add_Jiggle_Settings(bpy.types.Operator):
    bl_idname = "object.add_jiggle_settings"
    bl_label = "New Jiggle Settings"
    bl_description = "Create new jiggle physics settings"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'ARMATURE'

    @staticmethod
    def add_jiggle_settings(obj):
        jiggle_props = obj.ht_props.jiggle_props
        new_props = jiggle_props.physics_props.add()
        physics_props_cnt = len(jiggle_props.physics_props)
        new_props.name = f'Jiggle_{physics_props_cnt}'

        # NOTE: setup jiggle bones color - using bone_groups
        # jiggle_bone_group = obj.pose.bone_groups.get(f'Jiggle_{idx_str}')
        bone_groups_cnt = len(obj.pose.bone_groups)
        if bone_groups_cnt < physics_props_cnt:
            jiggle_bone_group = obj.pose.bone_groups.new(name=f'Jiggle')
            idx_str = f'{bone_groups_cnt+1:02d}' # with leading zeros
            jiggle_bone_group.color_set = f'THEME{idx_str}'

        jiggle_props.active_index = physics_props_cnt-1

    def execute(self, context):
        obj = context.active_object
        self.add_jiggle_settings(obj)
        if context.selected_pose_bones:
            bpy.ops.object.assign_jiggle()
        return {"FINISHED"}


class OBJECT_OT_Remove_Jiggle_Settings(bpy.types.Operator):
    bl_idname = "object.remove_jiggle_settings"
    bl_label = "Remove chain"
    bl_description = "Remove jiggle chain"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE'

    def execute(self, context):
        obj = context.active_object
        jiggle_props = obj.ht_props.jiggle_props

        physics_props_cnt = len(jiggle_props.physics_props)
        if jiggle_props.active_index > physics_props_cnt - 1:
            jiggle_props.active_index = physics_props_cnt - 1

        jiggle_props.physics_props.remove(jiggle_props.active_index)

        bone_groups_cnt = len(obj.pose.bone_groups)
        if jiggle_props.active_index < bone_groups_cnt:
            sel_bone_group = obj.pose.bone_groups[jiggle_props.active_index]
            obj.pose.bone_groups.remove(sel_bone_group)

        for i in range(jiggle_props.active_index, physics_props_cnt):
            if i == jiggle_props.active_index:
                for bone in obj.pose.bones:
                    if bone.ht_props.jiggle_index == jiggle_props.active_index:
                        bone.ht_props.jiggle_index = -1
                        bone.bone_group = None
            else:
                for bone in obj.pose.bones:
                    if bone.ht_props.jiggle_index == i:
                        bone.ht_props.jiggle_index = i - 1
                        bone.bone_group = obj.pose.bone_groups[i-1]


        last_idx = len(jiggle_props.physics_props)-1
        if jiggle_props.active_index > last_idx:
            jiggle_props.active_index = last_idx

        return {"FINISHED"}

#Select coliision bones operator
class OBJECT_OT_Select_Collision_Bones(bpy.types.Operator):
    bl_idname = "object.select_collision_bones"
    bl_label = "Select Collision Bones"
    bl_description = "Select collision bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE'

    def execute(self, context):
        obj = context.active_object
        obj.data.display_type = 'ENVELOPE'
        for pbone in obj.pose.bones:
            if pbone.ht_props.jiggle_collision:
                pbone.bone.hide = False
                pbone.bone.select = True
        return {"FINISHED"}


class HT_MT_Jiggle_BonesMenu(bpy.types.Menu):
    bl_idname = "HT_MT_Jiggle_BonesMenu"
    bl_label = "Jiggle Bones"

    def draw(self, context):
        layout = self.layout
        layout.operator("object.assign_jiggle")
        layout.operator("object.remove_bones_jiggle")
        layout.operator("object.select_jiggle_group")


