# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from math import degrees, sin, copysign, sqrt
from mathutils import Vector, Matrix, Quaternion, Euler
from mathutils.geometry import intersect_point_line
from bpy.app.handlers import persistent
# import random
from mathutils import noise
import gpu
from gpu_extras.batch import batch_for_shader


shader = gpu.shader.from_builtin('UNIFORM_COLOR')

# define signed distance field for rounded cone
# https://iquilezles.org/www/articles/distfunctions/distfunctions.htm
# def sdRoundCone(pbone, point):
#     # sampling independent computations (only depend on shape)
#     r1 = pbone.bone.head_radius
#     r2 = pbone.bone.tail_radius
#     head_tail_vec = pbone.tail - pbone.head
#     l2 = head_tail_vec.dot(head_tail_vec)
#     rad_diff = r1 - r2
#     a2 = l2 - rad_diff*rad_diff
#
#     # sampling dependant computations
#     head_p_vec = point - pbone.head
#     y = head_p_vec.dot(head_tail_vec)
#     z = y - l2
#     help = head_p_vec*l2 - head_tail_vec*y
#     x2 =  help.dot(help)
#     y2 = y*y*l2
#     z2 = z*z*l2
#
#     k =  copysign(rad_diff*rad_diff*x2, rad_diff) # copysign(a,b) =  a * sign(b)
#     if copysign(a2*z2, z) > k:
#         return  sqrt(x2 + z2) /l2 - r2
#     if copysign(a2*y2, y) < k:
#         return  sqrt(x2 + y2) /l2 - r1
#     return (sqrt(x2*a2/l2)+y*rad_diff)/l2 - r1

def get_jiggle_colls(obj):
    return [c for c in  obj.data.collections if c.physics_props.is_jiggle]

def jiggle_cnt(obj):
    return len(get_jiggle_colls(obj))

def get_pbone_physics(pbone):
    for col in pbone.bone.collections:
        if col.physics_props.is_jiggle:
            return col.physics_props
    return None

def is_pbone_physics(pbone):
    return pbone.bone.collections and any([c for c in pbone.bone.collections if c.physics_props.is_jiggle])

def rem_copy_rot_constraint(arma, pbone):
    copy_rot = pbone.constraints.get('Set Rotation')
    if copy_rot:
        world_rot_from_constr = Euler((copy_rot.min_x, copy_rot.min_y, copy_rot.min_z)).to_matrix().to_4x4()
        rot_from_constr = arma.convert_space(pose_bone=pbone, matrix=world_rot_from_constr, from_space='WORLD', to_space='LOCAL')
        pbone.rotation_quaternion = rot_from_constr.to_quaternion()
        pbone.constraints.remove(copy_rot)


def set_bone_col(pbone, color):
    # # Set theme color:
    # bone.color.palette = 'THEME09'
    pbone.bone.color.palette = 'CUSTOM'
    pbone.bone.color.custom.normal = color
    pbone.bone.color.custom.select = (1.0, 0.9, 0.3)
    pbone.bone.color.custom.active = (1.0, 1.0, 1.0)
    pbone.color.palette = 'CUSTOM'
    pbone.color.custom.normal = color
    pbone.color.custom.select = (1.0, 0.9, 0.3)
    pbone.color.custom.active = (1.0, 1.0, 1.0)


def reset_bone_col(pbone):
    pbone.bone.color.palette = 'DEFAULT'
    pbone.color.palette = 'DEFAULT'

def bone_hit_simple(pbone, point):
    ''' Simple bbox check - sphere vs point'''
    center = (pbone.head+pbone.tail)/2
    rad = (pbone.head-pbone.tail).length
    rad += max(pbone.bone.head_radius , pbone.bone.tail_radius ) + 2*pbone.bone.envelope_distance
    rad *= 0.5
    return False if (point-center).length > rad else True


def bone_hit_detect_aprox(pbone, point):
    ''' aproximated with cone. Less accurate version
    (we hit straight perpendicular to |head tail| - rather than at angle caused by r1-r2) but should be good enough
    return outside - bool, normal, hit_co '''
    line_hit, a_fac = intersect_point_line(point, pbone.head, pbone.tail) # intersect_point_line(p, a, b)

    #NOTE: we have to clamp line to segment hit
    if a_fac < 0:
        line_hit = pbone.head
        rad = pbone.bone.head_radius
    elif a_fac > 1:
        line_hit = pbone.tail
        rad = pbone.bone.tail_radius
    else:
        rad = pbone.bone.head_radius*(1-a_fac) + pbone.bone.tail_radius*a_fac # lerp from ra to rb

    colision_offset = pbone.bone.envelope_distance
    if (line_hit - point).length > rad+1.1*colision_offset: # fast check if we are outside
        return False, None, None

    normal = (point - line_hit).normalized()
    surf_hit = line_hit +  (rad+colision_offset)*normal

    surf_hit_to_p = point - surf_hit # vec from inside line to capsule surface hit
    is_inside = surf_hit_to_p.dot(normal) < 0
    #dist = (surf_hit-point).length
    return is_inside, normal, surf_hit


def bone_hit_detect_exact(pbone, point):
    ''' cone hit detection - more accuartate but not 100% exact
    (we hit at angle defined by r1-r2 difference
    offset_fac = 1.02 => move 2% outside of surface
    return outside - bool, normal, hit_co '''
    r_head = pbone.bone.head_radius
    r_tail = pbone.bone.tail_radius
    colision_offset = pbone.bone.envelope_distance
    a_p = point - pbone.head
    a_b = pbone.tail - pbone.head
    ab_len = a_b.length

    ap_projected_ab = a_p.project(a_b)
    ap_perp = a_p - ap_projected_ab
    cone_slope = (r_head-r_tail)/ ab_len
    ap_fix = cone_slope * ap_perp.length # dist from P_projected' to  P2' - point projected on a_b based on dr/ab_len slope

    line_hit = pbone.head + ap_projected_ab - ap_fix * (a_b/ab_len)

    a_to_hit = line_hit-pbone.head

    # NOTE: we hit cone line a_b -> now clamp and  move line_hit outside to surface of cone
    sign = -1 if a_to_hit.dot(a_b) < 0 else 1
    a_fac = sign * a_to_hit.length/ab_len # a_fac - distance from the first point of a_b line as a percentage of the length of the line

    # NOTE: clamp line to segment hit
    if a_fac < 0:
        line_hit = pbone.head
        rad = r_head
    elif a_fac > 1:
        line_hit = pbone.tail
        rad = r_tail
    else:
        rad = r_head*(1-a_fac) + r_tail*a_fac # lerp from ra to rb

    line_hit_to_p = point - line_hit

    normal = line_hit_to_p.normalized()
    surf_hit = line_hit +  (rad+colision_offset)*normal

    surf_hit_to_p_vec = point - surf_hit # vec from inside line to capsule surface hit
    is_inside = surf_hit_to_p_vec.dot(normal) < 0
    return is_inside, normal, surf_hit

class OBJECT_OT_Select_Jiggle_Group_B4(bpy.types.Operator):
    bl_idname = "object.select_jiggle_group_b4"
    bl_label = "Select bones in group"
    bl_description = "Select jiggle bones assigned to active group"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE' and obj.mode == 'POSE' # and len(obj.ht_props.jiggle_props.physics_props) > 0

    def execute(self, context):
        obj = context.active_object
        active_col = obj.data.collections.active
        if not active_col.physics_props.is_jiggle:
            print('Active bone collection is not jiggle collection')
            return
        bpy.ops.armature.collection_select()

        # for pbone in obj.pose.bones:
        #     pbone.bone.select = False # deselect first all
        #     if pbone.ht_props.jiggle_index == jiggle_props.active_index:
        #         pbone.bone.select = True

        return {"FINISHED"}

class OBJECT_OT_Assign_Bones_JiggleB4(bpy.types.Operator):
    bl_idname = "object.assign_jiggle_b4"
    bl_label = "Assign Jiggle to bones"
    bl_description = "Assign jiggle physics to selected pose bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.selected_pose_bones

    @staticmethod
    def assign_jiggle_to_bones(obj, pbones):
        jig_colls = get_jiggle_colls(obj)
        if len(jig_colls) == 0:
            OBJECT_OT_Add_Jiggle_Settings.add_jiggle_settings(obj)

        active_col = obj.data.collections.active
        if not active_col.physics_props.is_jiggle:
            print('Active bone collection is not jiggle collection')
            return

        color = active_col.physics_props.bone_color[:]
        for pbone in pbones:
            pbone.bone.collections.clear() # remove from other if assigned
            active_col.assign(pbone)

            # Set custom color:
            set_bone_col(pbone, color)


    def execute(self, context):
        obj = context.active_object
        self.assign_jiggle_to_bones(obj, context.selected_pose_bones)

        return {"FINISHED"}


class OBJECT_OT_Remove_Bones_Jiggle(bpy.types.Operator):
    bl_idname = "object.remove_bones_jiggle_b4"
    bl_label = "Remove Jiggle from bones"
    bl_description = "Remove jiggle physics from selected pose bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.selected_pose_bones

    def execute(self, context):
        obj = context.active_object
        jiggle_colls = get_jiggle_colls(obj)
        if not jiggle_colls:
            self.report({'WARNING'}, 'No jiggle collections found. Cancelling')
            return

        for pbone in context.selected_pose_bones:
            for col in pbone.bone.collections:
                if col in jiggle_colls:
                    col.unassign(pbone)
                    reset_bone_col(pbone)
                    rem_copy_rot_constraint(obj, pbone)

        return {"FINISHED"}


class OBJECT_OT_Add_Jiggle_Settings(bpy.types.Operator):
    bl_idname = "object.add_jiggle_settings_b4"
    bl_label = "Add Settings"
    bl_description = "Create new jiggle physics settings"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'ARMATURE'

    @staticmethod
    def add_jiggle_settings(obj):
        physics_props_cnt = jiggle_cnt(obj)
        jiggel_bcol_name = f'Jiggle_{physics_props_cnt}' # bone collection name
        new_bcoll = obj.data.collections.new(name=jiggel_bcol_name)
        new_bcoll.physics_props.is_jiggle = True

        obj.data.collections.active = new_bcoll

        col = 0.5*noise.random_unit_vector() + Vector((0.5, 0.5, 0.5))
        new_bcoll.physics_props.bone_color = col[:]
        # jiggle_bone_group.color_set = f'THEME{idx_str}'
        return new_bcoll

    def execute(self, context):
        obj = context.active_object
        self.add_jiggle_settings(obj)
        if context.selected_pose_bones:
            bpy.ops.object.assign_jiggle_b4()
        return {"FINISHED"}


class OBJECT_OT_Remove_Jiggle_Settings(bpy.types.Operator):
    bl_idname = "object.remove_jiggle_settings_b4"
    bl_label = "Remove Settings"
    bl_description = "Remove Jiggle Settings"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE'


    def execute(self, context):
        obj = context.active_object
        active_col = obj.data.collections.active
        active_idx = obj.data.collections.active_index
        if not active_col.physics_props.is_jiggle:
            self.report({'Warning'}, 'Active bone collection is not jiggle collection')
            return

        # reset color of bones in group
        for pbone in obj.pose.bones:
            if active_col.bones.get(pbone.name):
                reset_bone_col(pbone)
                rem_copy_rot_constraint(obj, pbone)


        obj.data.collections.remove(active_col)
        # bone.color.palette = 'THEME09'
        # bpy.context.object.data.bones["Bone.002"].color.palette = 'DEFAULT'

        remaining_phys_colls = get_jiggle_colls(obj)
        if remaining_phys_colls:
            prev_idx = max(active_idx-1, 0)
            obj.data.collections.active = remaining_phys_colls[min(prev_idx, len(remaining_phys_colls)-1)]

        return {"FINISHED"}

#Select coliision bones operator
class OBJECT_OT_Select_Collision_Bones(bpy.types.Operator):
    bl_idname = "object.select_collision_bones_b4"
    bl_label = "Select Collision Bones"
    bl_description = "Select collision bones"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'ARMATURE'

    def execute(self, context):
        obj = context.active_object
        obj.data.display_type = 'ENVELOPE'
        for pbone in obj.pose.bones:
            if pbone.ht_props.jiggle_collision:
                pbone.bone.hide = False
                pbone.bone.select = True
        return {"FINISHED"}



class JBONES_UL_props_B4(bpy.types.UIList):
    VGROUP_EMPTY = 1 << 0

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        obj = data.id_data
        bcoll = item
        # item_idx = int(item.path_from_id('mass')[-7])  # wont work if  jiggle_propx idx is 2 digitspan
        if bcoll:
            layout.prop(bcoll, 'name', text='', emboss=False)
            row = layout.row(align=True)
            row.prop(bcoll.physics_props, "bone_color", text="")
        else:
            layout.label(text="", translate=False, icon_value=icon)


    def filter_items_coll(self, context, bcolls):
        return {c.name:c.physics_props.is_jiggle  for c in bcolls}

    def filter_items(self, context, data, propname):
        bone_colls = getattr(data, propname)

        # Default return values.
        flt_flags = []
        flt_neworder = []

        # Pre-compute of vgroups data, CPU-intensive. :/
        vec_pts_attribs = self.filter_items_coll(context, bone_colls)

        flt_flags = [self.bitflag_filter_item] * len(bone_colls) # special bitmask. First item for removing unwanted items

        # Filter by curve tag attribs (spline type, bool)
        for idx, attrib in enumerate(bone_colls):
            if not vec_pts_attribs[attrib.name]:  # non Tag type - add to ignore
                flt_flags[idx] |= self.VGROUP_EMPTY
                # if self.ignore_neg:  # always filter
                flt_flags[idx] &= ~self.bitflag_filter_item

        # Reorder by name - or no reorder....
        # flt_neworder = helper_funcs.sort_items_by_name(attributes, "name")
        # print(flt_flags, flt_neworder)
        return flt_flags, flt_neworder


class JBONES_UL_props(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        obj = data.id_data
        active_physics_prop = item
        # jiggle_props = data
        # item_idx = int(item.path_from_id('mass')[-7])  # wont work if  jiggle_propx idx is 2 digits
        item_idx = [i for i, ph in enumerate(obj.ht_props.jiggle_props.physics_props) if ph == active_physics_prop][0]
        if active_physics_prop:
            layout.prop(active_physics_prop, 'name', text='', emboss=False)
            if item_idx < len(obj.pose.bone_groups):
                group = obj.pose.bone_groups[item_idx]
                row = layout.row(align=True)
                row.enabled = group.is_custom_color_set  # only custom colors are editable
                row.prop(group.colors, "normal", text="")
        else:
            layout.label(text="", translate=False, icon_value=icon)


class  HTOOL_PT_VIEW3D_PT_JIGGLE(bpy.types.Panel):
    bl_label = "Bones and Jiggle Physics"
    bl_idname = "HTOOL_PT_VIEW3D_PT_JIGGLE"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'
    # bl_context = 'objectmode'
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def draw_b36(self, context, obj):
        layout = self.layout
        if  obj and obj.type == 'ARMATURE' and obj.mode == 'POSE':
            obj_jiggle_props = obj.ht_props.jiggle_props
            if context.active_pose_bone:
                col = layout.column(align=True)
                row = col.row(align=True)
                txt = 'Add Collision' if not context.active_pose_bone.ht_props.jiggle_collision else 'Remove Collision'
                row.prop(context.active_pose_bone.ht_props, 'jiggle_collision', icon='MOD_PHYSICS', text=txt)
                row.operator("object.select_collision_bones", icon="BONE_DATA", text='')
                if context.active_pose_bone.ht_props.jiggle_collision:
                    col.prop(context.active_pose_bone.ht_props, 'friction')

            if obj:
                layout.label(text='Assign Jiggle to selection:')
                row = layout.row()
                row.template_list("JBONES_UL_props", "", obj_jiggle_props, "physics_props", obj_jiggle_props, "active_index", rows=3)
                col = row.column(align=True)
                col.operator("object.add_jiggle_settings", icon="ADD", text='')
                col.operator("object.remove_jiggle_settings", icon="REMOVE", text='')
                col.menu("HT_MT_Jiggle_BonesMenu", text='', icon='DOWNARROW_HLT')

                chain_idx = obj_jiggle_props.active_index
                physics_props_cnt = len(obj_jiggle_props.physics_props)
                if physics_props_cnt and chain_idx < physics_props_cnt:
                    chain_props = obj_jiggle_props.physics_props[chain_idx]
                    col = layout.column(align=True)
                    col.prop(chain_props, 'mass')
                    col.prop(chain_props, 'stiffness')
                    col.prop(chain_props, 'damping')
                    col.prop(chain_props, 'gravity')


    def draw_b4(self, context, obj):
        # almost same but use b4 version of operators
        layout = self.layout
        if  obj and obj.type == 'ARMATURE' and obj.mode == 'POSE':
            if context.active_pose_bone:
                col = layout.column(align=True)
                row = col.row(align=True)
                txt = 'Add Collision' if not context.active_pose_bone.ht_props.jiggle_collision else 'Remove Collision'
                row.prop(context.active_pose_bone.ht_props, 'jiggle_collision', icon='MOD_PHYSICS', text=txt)
                row.operator("object.select_collision_bones_b4", icon="BONE_DATA", text='')
                if context.active_pose_bone.ht_props.jiggle_collision:
                    col.prop(context.active_pose_bone.ht_props, 'friction')

            if obj:
                layout.label(text='Assign Jiggle to selection:')
                row = layout.row()
                row.template_list("JBONES_UL_props_B4", "", obj.data, "collections", obj.data.collections, "active_index", rows=3)
                col = row.column(align=True)
                col.operator("object.add_jiggle_settings_b4", icon="ADD", text='')
                col.operator("object.remove_jiggle_settings_b4", icon="REMOVE", text='')
                col.menu("HT_MT_Jiggle_BonesMenu_B4", text='', icon='DOWNARROW_HLT')

                active_bcol = obj.data.collections.active
                if active_bcol and active_bcol.physics_props.is_jiggle:
                    col = layout.column(align=True)
                    col.prop(active_bcol.physics_props, 'mass')
                    col.prop(active_bcol.physics_props, 'stiffness')
                    col.prop(active_bcol.physics_props, 'damping')
                    col.prop(active_bcol.physics_props, 'gravity')

    def draw(self, context):
        layout = self.layout
        obj = context.active_object
        if obj and obj.type == 'MESH' and context.mode == 'EDIT_MESH':
            col = layout.column(align=True)
            col.operator("object.ribbon_to_bones", icon="BONE_DATA")
            col.prop(obj.ht_props, 'target_rig')
            if obj.ht_props.target_rig and obj.ht_props.target_rig.type == 'ARMATURE':
                col.prop_search(obj.ht_props, 'pin_bone', obj.ht_props.target_rig.data, 'bones')

        if context.mode == 'EDIT_MESH':
            return

        scn_jiggle_props = context.scene.ht_props.jiggle_props
        text = 'Run Spring Physics' if not scn_jiggle_props.sim_is_running else 'Simulation enabled'
        if scn_jiggle_props.sim_is_running:      # disable with bool prop
            layout.prop(scn_jiggle_props, 'sim_is_running', text=text, icon="GP_MULTIFRAME_EDITING")
        else: # if disabled then give option to enable
            layout.operator("object.enable_spring_physics", icon="GP_MULTIFRAME_EDITING", text=text, depress=scn_jiggle_props.sim_is_running )
            layout.prop(scn_jiggle_props, 'sim_mode')

        if bpy.app.version < (4, 0, 0):
            self.draw_b36(context, obj)
        else:
            self.draw_b4(context, obj)



class HT_MT_Jiggle_BonesMenu_B4(bpy.types.Menu):
    bl_idname = "HT_MT_Jiggle_BonesMenu_B4"
    bl_label = "Jiggle Bones"

    def draw(self, context):
        layout = self.layout
        layout.operator("object.assign_jiggle_b4")
        layout.operator("object.remove_bones_jiggle_b4")
        layout.operator("object.select_jiggle_group_b4")


class ChainSegment():
    def __init__(self, root_data, tip_data):
        '''
            root_data: loc, exp_map
        '''
        self.mass = 1.0
        self.damping = 2.0
        self.stiffness = 10.
        self.gravity = Vector((0, 0, -9.81))

        root_co = root_data['loc']
        self.prev_root_co = root_co.copy() # needed to handle node translation
        self.root_co = root_co.copy() # needed to handle node translation
        self.tip_co = tip_data['loc'].copy()

        self.mag = (self.tip_co - self.root_co).length
        # self.track_axis = Vector((0, 1, 0))*self.mag

        self.init_dir = root_data['rest_rot'].copy() # stores initial chain rot
        self.rest_dir = root_data['rest_rot'].copy() # target for jiggle
        self.prev_dir = root_data['curr_rot'].copy()
        self.current_dir = root_data['curr_rot'].copy()
        self.velocity = Vector((0.0, 0.0, 0.0))

        self.ang_vel = Vector((0, 0, 0))

        self.init_mat = root_data['rest_mat'].copy()
        self.rest_mat = root_data['rest_mat'].copy()
        self.current_quat = root_data['rest_mat'].to_quaternion()
        self.m_p_inv = root_data['m_p_inv']
        self.next = None

        self.pbone = None
        self.phys_prop = None

    def update_physics_props(self):
        self.mass = self.phys_props.mass
        self.damping = self.phys_props.damping
        self.stiffness = self.phys_props.stiffness
        self.gravity = Vector((0, 0, -self.phys_props.gravity))

    def get_current_rotation(self):
        return self.root_co, self.current_quat # quat(dir*axis)


class SubChain():
    def __init__(self, chain_data):
        '''
            chain_data is a list of tuples (co, rotation - axis*angle)
        '''
        self.dt = 0.04
        self.reached_rest_state = False # TODO: add per each subchain
        self.segments = []
        self.pbones = []
        self.empties = []
        self.collision_bones = []

        # NOTE: create chain [(d1,d2), (d2,d3), (d3, ..)]
        prev_seg_data = chain_data[0]
        prev_seg = None
        segments = []
        for seg_data in chain_data[1:]:  # 4 objs give 3 segs
            new_seg = ChainSegment(prev_seg_data,seg_data )
            if prev_seg:
                prev_seg.next = new_seg
            segments.append(new_seg)
            prev_seg_data = seg_data
            prev_seg = new_seg
        self.segments = segments

    def set_target_bones(self, child_bones):
        ''' they will receive the sim output '''
        parent_bone = child_bones[0].parent
        self.pbones = child_bones
        # NOTE: root local matrix required to to properly propagate movement from root.parent bone
        # NOTE: get root.pbone.matrix_local by diffing  prev_bone.mw-root.mw
        if parent_bone:
            self.root_pbone_mat_local = parent_bone.matrix.inverted() @ child_bones[0].matrix @ child_bones[0].matrix_basis.inverted()
        else:
            self.root_pbone_mat_local = child_bones[0].matrix @ child_bones[0].matrix_basis.inverted()


    def solve_semi_euler_euler(self, seg):
        '''  stable it dt < 1/4 * T
        vel[n+2] = v[n] - h * omgea^2 * x[n]
        x[n+1] = x[n] + h * vel[n+2]
        omega^2 = k/m     #angular freq -  controls rate of oscilation
        2*dr*omega = c/m  # dr - damping ratio - dimensionless and controls amount of oscilation (0, 1) (no_damp, perfect_decay)


        w = w0 + a * dt
        q[n+1] = q[n] + 0.5 wq * q[n]
        '''
        pass

    def integrate_euler_loc(self, seg):
        ''' had trouble - damping increasing with time - slower and slower after 4 sec'''
        dt = self.dt
        # seg.set_physics(phys_props.mass, phys_props.damping, phys_props.stiffness, phys_props.gravity)

        if seg.next:
            if seg.init_mat != seg.rest_mat: # NOTE: if prev chain rest_dir changed, propagete this change up the chain
                seg.next.rest_mat = seg.rest_mat @ seg.next.m_p_inv  # aka  parent.mw @ child.mp_inv
                seg.next.rest_dir = seg.next.rest_mat.col[1].xyz  # track with  y
            else:
                seg.next.rest_mat = seg.next.init_mat.copy()
                seg.next.rest_dir = seg.next.rest_mat.col[1].xyz  # track with  y

        if seg.prev_root_co != seg.root_co: # NOTE: handles root_tranlation
            prev_root_to_tip = seg.tip_co - seg.prev_root_co # from prev_root to next_seg.root
            root_to_tip = seg.tip_co - seg.root_co # from current_root to next_seg.root
            q_diff = prev_root_to_tip.rotation_difference(root_to_tip)
            seg.current_dir.rotate(q_diff)
            seg.prev_root_co = seg.root_co.copy()


        # T =R x F  ;  a = T/I  = T/ (m R^2) = F R sin() / m R^2  = g * sin() / R
        grav = Vector((0,0, -seg.phys_props.gravity)) # longer lever arm, bigger gravity acc
        # grav_tangent = grav - grav.project(seg.rest_dir) # tangetn to rest_dir (current_dir wobbles)
        # grav_tangent *= seg.mag

        # global DRAW_PTS
        # DRAW_PTS = [seg.current_dir, seg.rest_dir]

        x =  seg.current_dir - seg.rest_dir
        # print(f"{x.length=:.3f}")
        spring_force = - seg.phys_props.stiffness*x   # F = -kx - bv - g*m  - vel here. causes it to groww too much with time..
        # print(f"{spring_force.length=:.3f}")
        acc = spring_force / seg.phys_props.mass + grav  # in theory  /= m R^2 => I
        # get acc perpendicualr to cur_dir
        # acc = acc - acc.project(seg.current_dir) # not sure if better
        # print(f"{acc.length=:.3f}")
        seg.velocity += acc * dt - seg.phys_props.damping * seg.velocity # FIXED: problem with growing vel.length .... causes huge damping with time. Fixed?
        seg.velocity = seg.velocity - seg.velocity.project(seg.current_dir) # project vel to current_dir

        # print(f"{seg.velocity.length=:.3f}")

        for collision_bone in self.collision_bones:
            if collision_bone.name == seg.pbone.name: # DONE: avoid self collision
                continue

            is_inside, normal, surf_hit = bone_hit_detect_aprox(collision_bone, seg.tip_co)
            if is_inside:
                root_to_tip = seg.current_dir # why using current_dir gives jitter?
                root_to_hit = surf_hit - seg.root_co # from current_root to next_seg.root
                q_diff = root_to_tip.rotation_difference(root_to_hit)
                seg.current_dir.rotate(q_diff)
                seg.velocity -= collision_bone.ht_props.friction * seg.velocity * dt  #  Frict = f * m * a,   vf = f * v * dt

        # NOTE: is this slower than quat?
        seg.current_dir = seg.current_dir + seg.velocity * dt
        seg.current_dir.normalize()


        # NOTE: does not reuire normalization... quat is slower?
        # new_dir = seg.current_dir + seg.velocity * dt
        # quat_curr_to_new = seg.current_dir.rotation_difference(new_dir)
        # seg.current_dir.rotate(quat_curr_to_new)
        seg.tip_co = seg.root_co + seg.mag * seg.current_dir


        rd_prev_curr = seg.rest_dir.rotation_difference(seg.current_dir)
        seg.current_quat = seg.rest_mat.to_quaternion()
        seg.current_quat.rotate(rd_prev_curr)

        if seg.next:  # NOTE: seg.tip.co  -> seg[n+1].root.co
            seg.next.root_co = seg.tip_co.copy()
        else:  # extra check if sim reached rest state
            if seg.velocity.length < 0.0005:
                self.reached_rest_state = True


        # seg.prev_dir = seg.current_dir.copy()


    def integrate_pbd(self, seg):
        '''Looks ok - no mass, sfiffness ok in 0.01 to 0.8 range, ... what about damping?
         with stiffness == 1 > perfacty stiff body - no jiggle basically
         Does not jiggele as nice os euler above... need more substeps?
        '''
        dt = self.dt
        # seg.set_physics(phys_props.mass, phys_props.damping, phys_props.stiffness, phys_props.gravity)

        if seg.next:
            if seg.init_mat != seg.rest_mat: # NOTE: if prev chain rest_dir changed, propagete this change up the chain
                seg.next.rest_mat = seg.rest_mat @ seg.next.m_p_inv  # aka  parent.mw @ child.mp_inv
                seg.next.rest_dir = seg.next.rest_mat.col[1].xyz  # track with  y
            else:
                seg.next.rest_mat = seg.next.init_mat.copy()
                seg.next.rest_dir = seg.next.rest_mat.col[1].xyz  # track with  y

        if seg.prev_root_co != seg.root_co: # NOTE: handles root_tranlation
            prev_root_to_tip = seg.tip_co - seg.prev_root_co # from prev_root to next_seg.root
            root_to_tip = seg.tip_co - seg.root_co # from current_root to next_seg.root
            q_diff = prev_root_to_tip.rotation_difference(root_to_tip)
            seg.current_dir.rotate(q_diff)
            seg.prev_root_co = seg.root_co.copy()


        # T =R x F  ;  a = T/I  = T/ (m R^2) = F R sin() / m R^2  = g * sin() / R
        g_const = Vector((0,0, -seg.phys_props.gravity)) # longer lever arm, bigger gravity acc
        grav_tangent = g_const - g_const.project(seg.rest_dir) # tangetn to rest_dir (current_dir wobbles)
        grav_tangent *= seg.mag

        # global DRAW_PTS
        # DRAW_PTS = [seg.current_dir, seg.rest_dir]
        vel = seg.velocity + g_const * dt
        prev_dir = seg.current_dir.copy()

        seg.current_dir = seg.current_dir + vel * dt

        # do the contraints
        x =  seg.current_dir - seg.rest_dir
        # print(f"{x.length=:.3f}")
        print(f"{seg.current_dir.length=:.3f}")
        print(f"{seg.rest_dir.length=:.3f}")
        rot_diff = seg.current_dir.rotation_difference(seg.rest_dir)
        # scale rot by stiffness...
        spring_force = - seg.phys_props.stiffness*x #  - seg.phys_props.damping * seg.velocity  # F = -kx - bv - g*m
        acc = spring_force / seg.phys_props.mass # + grav_tangent  # in theory  /= m R^2 => I
        # seg.velocity += acc * dt + seg.velocity
        # print(f"{seg.velocity.length=}")

        for collision_bone in self.collision_bones:
            if collision_bone.name == seg.pbone.name: # DONE: avoid self collision
                continue

            is_inside, normal, surf_hit = bone_hit_detect_aprox(collision_bone, seg.tip_co)
            if is_inside:
                root_to_tip = seg.current_dir # why using current_dir gives jitter?
                root_to_hit = surf_hit - seg.root_co # from current_root to next_seg.root
                q_diff = root_to_tip.rotation_difference(root_to_hit)
                seg.current_dir.rotate(q_diff)
                seg.velocity -= collision_bone.ht_props.friction * seg.velocity * dt  #  Frict = f * m * a,   vf = f * v * dt

        target_rot = seg.current_dir.copy()
        target_rot.rotate(rot_diff)

        seg.current_dir = seg.current_dir.lerp(target_rot, seg.phys_props.stiffness)

        # NOTE: is this slower than quat?
        # seg.current_dir = seg.current_dir + seg.velocity * dt
        seg.current_dir.normalize()

        seg.velocity = (1 - seg.phys_props.damping) * (seg.current_dir - prev_dir) / dt
        print(f"{seg.velocity.length=:.3f}")


        # NOTE: does not reuire normalization... quat is slower?
        # new_dir = seg.current_dir + seg.velocity * dt
        # quat_curr_to_new = seg.current_dir.rotation_difference(new_dir)
        # seg.current_dir.rotate(quat_curr_to_new)
        seg.tip_co = seg.root_co + seg.mag * seg.current_dir


        rd_prev_curr = seg.rest_dir.rotation_difference(seg.current_dir)
        seg.current_quat = seg.rest_mat.to_quaternion()
        seg.current_quat.rotate(rd_prev_curr)

        if seg.next:  # NOTE: seg.tip.co  -> seg[n+1].root.co
            seg.next.root_co = seg.tip_co.copy()
        else:  # extra check if sim reached rest state
            vel_mag = seg.ang_vel.length
            if vel_mag + acc.length < 0.001:
                self.reached_rest_state = True


        # seg.prev_dir = seg.current_dir.copy()


    def solve_euler_axis_angle(self, seg):
        ''' This sheisse works
        stable it dt < 1/4 * T
        vel[n+1] = v[n] - h * omgea^2 * x[n]
        x[n+1] = x[n] + h * vel[n+1]
        omega^2 = k/m     #angular freq -  controls rate of oscilation
        2*dr*omega = c/m  # dr - damping ratio - dimensionless and controls amount of oscilation (0, 1) (no_damp, perfect_decay)

        w = w0 + a * dt
        q[n+1] = q[n] + 0.5 wq * q[n]
        '''
        dt = self.dt

        if seg.next:
            if seg.init_dir != seg.rest_dir: # NOTE: if prev chain rest_dir changed, propagete this change up the chain
                diff = seg.init_dir - seg.rest_dir
                seg.next.rest_dir = seg.next.init_dir - diff
            else:  # FIXed: eg. reseting root bone rot wont  reset rest of chaing
                seg.next.rest_dir = seg.next.init_dir

        if seg.prev_root_co != seg.root_co: # NOTE: handles root_tranlation
            prev_root_to_tip = seg.tip_co - seg.prev_root_co # from prev_root to next_seg.root
            root_to_tip = seg.tip_co - seg.root_co # from current_root to next_seg.root
            q_diff = prev_root_to_tip.rotation_difference(root_to_tip)
            seg.current_dir += q_diff.to_exponential_map()
            seg.prev_root_co = seg.root_co.copy()


        diff = seg.current_dir - seg.rest_dir
        spring_force = - seg.stiffness*diff - seg.damping * seg.ang_vel # F = -kx - bv - g*m
        acc = spring_force / seg.mass #+ g*grav_dir.to_exponential_map()
        seg.ang_vel += acc * dt
        seg.current_dir += seg.ang_vel * dt

        track_curr = seg.track_axis.copy()
        curr_quat = Quaternion(seg.current_dir)
        track_curr.rotate(curr_quat)

        if True:
            grav = Vector((0, 0, -0.5))*seg.mag
            grav_tan = grav - grav.project(track_curr)
            grav_dir = track_curr.rotation_difference(track_curr+grav_tan)
            track_curr.rotate(grav_dir)
            seg.curr_with_grav_dir = (grav_dir @ curr_quat).to_exponential_map()

        seg.tip_co = seg.root_co + track_curr

        if seg.next:  # NOTE: seg.tip.co  -> seg[n+1].root.co
            seg.next.root_co = seg.tip_co.copy()
        else:  # extra check if sim reached rest state
            vel_mag = seg.ang_vel.length
            if vel_mag + acc.length < 0.1:
                self.reached_rest_state = True

    def solve_verlet_integration_euler(self, seg):
        ''' does not behave well with double current_dir. Problems with maintaining dir
         args:  seg - processed segment
        '''
        dt = self.dt
        if seg.next:
            if seg.init_euler != seg.rest_euler: # NOTE: if prev chain rest_euler changed, propagete this change up the chain
                diff = seg.init_euler - seg.rest_euler
                new_rest =  seg.next.init_euler - diff
                total_diff = new_rest - seg.next.rest_euler
                seg.next.rest_euler = new_rest
                stiff_01 = seg.stiffness/60 # (40  max stiffnes) -  remap from (0, 40) - (1, 0.5) => (max_wiggle, 50% wiggle)
                seg.current_euler += total_diff * stiff_01 # NOTE: update both current and prev euler - so that we wont build velocity in diff dir
                seg.prev_euler += total_diff * stiff_01
            else:  # FIX: eg. reseting root bone rot wont  reset rest of chaing
                seg.next.rest_euler = seg.next.init_euler

        if seg.prev_root_co != seg.root_co: # NOTE: handles root_tranlation
            prev_root_to_tip = seg.tip_co - seg.prev_root_co # from prev_root to next_seg.root
            root_to_tip = seg.tip_co - seg.root_co # from current_root to next_seg.root
            q_diff = prev_root_to_tip.rotation_difference(root_to_tip)
            q_diff_evec = Vector(q_diff.to_euler()[:])
            stiff_01 = 1 - seg.stiffness/60 # (40  max stiffnes) -  remap from (0, 40) - (1, 0.5) => (max_wiggle, 50% wiggle)
            seg.current_euler += q_diff_evec * stiff_01 # NOTE: update both current and prev euler - so that we wont build velocity in diff dir
            seg.prev_euler += q_diff_evec * stiff_01
            seg.prev_root_co = seg.root_co

        if True: # gravity causes jiggle on gimball?
            current_dir = seg.track_axis.copy()
            current_dir.rotate(Euler(seg.current_euler).to_quaternion())
            gravity = Vector((0,0,-0.1))
            g_diff = gravity.rotation_difference(current_dir)

            angle = g_diff.angle
            g_dirr = Vector(g_diff.to_euler()[:])*sin(angle)


        velocity = (seg.current_euler - seg.prev_euler)/dt    # vel =  curr - prev dir
        x = seg.current_euler - seg.rest_euler
        spring_force = - seg.stiffness*x -  seg.damping * velocity - 10*g_dirr*seg.mass # F = -kx - bv - g*m
        acceleration = spring_force / seg.mass

        next_dir = 2 * seg.current_euler - seg.prev_euler + acceleration * dt * dt #  v = 2v(n) - v(n-1) + at^2
        seg.prev_euler = seg.current_euler
        seg.current_euler = next_dir

        track_cpy = seg.track_axis.copy()
        track_cpy.rotate(Euler(seg.current_euler)) # XXX: do we need convert to quat?
        seg.tip_co = seg.root_co + track_cpy

        if seg.next:  # NOTE: seg.tip.co  -> seg[n+1].root.co
            seg.next.root_co = seg.tip_co.copy()
        else:  # extra check if sim reached rest state
            vel_mag = velocity.length
            accel_mag = acceleration.length
            if vel_mag + accel_mag < 0.1:
                self.reached_rest_state = True

    def step(self):
        ''' run euler motion integration for each element in chain '''
        for seg in self.segments:
            # self.solve_verlet_integration_euler(seg) # number 1)
            self.integrate_euler_loc(seg) # number 1) # had trouble - damping increasing with time - slower and slower after 4 sec
            # self.integrate_pbd(seg) # number 1)


class SpringChains():
    def __init__(self, armature, write_loc = True, write_rot = False, can_rest=True):
        ''' chain_data[n]['loc'] - segment root loc, curr_rot, etc, '''
        self.sub_chains = []
        self.write_loc = write_loc #write translation to chain root_objs
        self.write_rot = write_rot #write rotation to chain root_objs
        self.armature = armature
        self.can_rest = can_rest # can chain rest if movement is small?

    def add_sub_chain(self, chain_data):
        sc = SubChain(chain_data)
        self.sub_chains.append(sc)
        return sc

    def solve_chains(self):
        ''' run euler motion integration for each element in chain '''
        for chain in self.sub_chains:
            pbones = chain.pbones
            if not (self.can_rest and chain.reached_rest_state):
                # NOTE: write transformation from chain.parent, to first root.seg (chaing[0])
                chain_root_pb = pbones[0]
                root_mat = chain.root_pbone_mat_local
                if chain_root_pb.parent:
                    root_mat = chain_root_pb.parent.matrix @ root_mat

                if chain_root_pb.bone.use_inherit_rotation: # or else chain wont align its direction, to new parent dir.
                    chain.segments[0].rest_dir = root_mat.col[1].xyz # track with y
                    chain.segments[0].rest_mat = root_mat
                chain.segments[0].root_co = root_mat.col[3].xyz # propagate change from pbone to root empty
                chain.step()

                # NOTE: write chain.seg coordinates
                for i, (pbone, sim_seg) in enumerate(zip(pbones, chain.segments)):
                    loc, current_quat = sim_seg.get_current_rotation()
                    rot_constraint = pbone.constraints.get('Set Rotation')
                    if not rot_constraint:
                        rot_constraint = pbone.constraints.new('LIMIT_ROTATION')
                        rot_constraint.use_limit_x = rot_constraint.use_limit_y = rot_constraint.use_limit_z = True
                        rot_constraint.owner_space = 'WORLD'
                        rot_constraint.influence = 1.0
                        rot_constraint.name = 'Set Rotation'


                    rot_eu = current_quat.to_euler()
                    rot_constraint.min_x = rot_constraint.max_x = rot_eu.x
                    rot_constraint.min_y = rot_constraint.max_y = rot_eu.y
                    rot_constraint.min_z = rot_constraint.max_z = rot_eu.z

                    if chain.empties:
                        empty = chain.empties[i]
                        empty.location = loc
                        empty.rotation_quaternion = current_quat

    def finish_chains(self):
        for sub_chain in self.sub_chains: # force chains movement on scene change
            for pbone in reversed(sub_chain.pbones):
                copy_rot = pbone.constraints.get('Set Rotation')
                if copy_rot:
                    world_rot_from_constr = Euler((copy_rot.min_x, copy_rot.min_y, copy_rot.min_z)).to_matrix().to_4x4()
                    rot_from_constr = self.armature.convert_space(pose_bone=pbone, matrix=world_rot_from_constr, from_space='WORLD', to_space='LOCAL')
                    pbone.rotation_quaternion = rot_from_constr.to_quaternion()
                    pbone.constraints.remove(copy_rot)


def matrix_world(armature_ob, pbone):
    ''' Gives pbone.matrix -> but calculated manually'''
    local = pbone.bone.matrix_local
    basis = pbone.matrix_basis
    if pbone.parent:
        parent_local = pbone.parent.bone.matrix_local
        return matrix_world(armature_ob, pbone.parent) @ (parent_local.inverted() @ local) @ basis
    else:
        return  local @ basis


def get_child_chains_b4(pbone, out_chain):
    ''' get armature childs bones as flat child list of branches '''
    for i,child in enumerate(pbone.children):
        if is_pbone_physics(child):
            if i > 0 or not child.bone.use_connect or not is_pbone_physics(child.parent):  # more than 1 child or not connected, or parent was not jiggle
                out_chain.append([]) # generate new subchain
            out_chain[-1].append(child)
        get_child_chains_b4(child, out_chain)


def get_child_chains(pbone, out_chain):
    ''' get armature childs bones as flat child list of branches  old - for b 3.6'''
    for i,child in enumerate(pbone.children):
        if child.ht_props.jiggle_index > -1:
            if i > 0 or not child.bone.use_connect or child.parent.ht_props.jiggle_index == -1:  # more than 1 child or not connected, or parent was not jiggle
                out_chain.append([]) # generate new subchain
            out_chain[-1].append(child)
        get_child_chains(child, out_chain)

def get_root_bones(arma):
    ''' get all root bones for rig. In theory only one'''
    root_bones = []
    for pbone in arma.pose.bones:
        if not pbone.parent:
            root_bones.append(pbone)
    return root_bones

def setup_empty(context,name, mat):
    empty = bpy.data.objects.get(name)
    if not empty:
        empty = bpy.data.objects.new(name, None)
        empty.rotation_mode = 'QUATERNION'
        empty.empty_display_type = 'ARROWS'
        empty.empty_display_size = 0.02
        context.scene.collection.objects.link(empty)
    empty.matrix_world = mat
    return empty

def init_arma_spring_chains(context):
    armature_springs = []
    arma_objs = [obj for obj in context.visible_objects if obj.type == 'ARMATURE']
    dt = 1/context.scene.render.fps
    scn_jiggle_props = context.scene.ht_props.jiggle_props
    can_rest = scn_jiggle_props.sim_mode == 'REALTIME' # NOTE: enable 'chain freeze' only for 'REALTIME' mode
    # physics_colls = get_jiggle_colls(context.active_object)
    for arma_obj in arma_objs:
        obj_springs = SpringChains(arma_obj, write_loc = True, write_rot = True, can_rest = can_rest)

        root_pbones = get_root_bones(arma_obj)
        root_pbones = [rb for rb in root_pbones if rb]

        #get collision pbones
        collision_bones = []
        for pbone in arma_obj.pose.bones:
            if pbone.ht_props.jiggle_collision:
                collision_bones.append(pbone)

        for root_pbone in root_pbones:
            if bpy.app.version >= (4, 0, 0):
                in_pbone_chains = [[root_pbone]] if is_pbone_physics(root_pbone) else [[]]
                get_child_chains_b4(root_pbone, in_pbone_chains)  # scan child branches for chains if jiggle bones
            else: # old blender 3.6
                in_pbone_chains = [[root_pbone]] if root_pbone.ht_props.jiggle_index > -1 else [[]]
                get_child_chains(root_pbone, in_pbone_chains)  # scan child branches for chains if jiggle bones

            # import pprint
            # pprint.pprint(in_pbone_chains)
            in_pbone_chains = [chain for chain in in_pbone_chains if chain]
            for pbone_chain in in_pbone_chains:
                chain_init_data = []
                empties = []

                parent = pbone_chain[0].parent
                parent_mat = Matrix.Identity(4)
                if parent:  # get parent_mat - without local transfroms of all pbone_chain[0] parents
                    parent_mat = arma_obj.convert_space(pose_bone=parent, matrix=parent_mat, from_space='LOCAL_WITH_PARENT', to_space='WORLD')

                for pbone in pbone_chain:
                    # pbone.bone.use_inherit_rotation = True
                    pbone.rotation_mode = 'QUATERNION'
                    world_matrix = pbone.matrix
                    rest_mat_world = arma_obj.convert_space(pose_bone=pbone, matrix=Matrix.Identity(4), from_space='LOCAL_WITH_PARENT', to_space='WORLD')
                    # rest_mat_world = world_matrix @ pbone.matrix_basis.inverted() # without matrix_basis - but icludes all parent local transf mat_basis...
                    link_data = {}
                    link_data['loc'] = world_matrix.col[3].xyz
                    link_data['curr_rot'] = world_matrix.col[1].xyz
                    link_data['rest_rot'] = rest_mat_world.col[1].xyz
                    link_data['rest_mat'] = rest_mat_world
                    link_data['m_p_inv']  = parent_mat.inverted() @ rest_mat_world
                    chain_init_data.append(link_data)
                    # empties.append(setup_empty(context,pbone.name, world_matrix)) # NOTE: for debugging
                    parent_mat = rest_mat_world

                # NOTE: helper - tail point - required for tip_co - root_co - in sim step
                link_data = {}
                link_data['loc'] = pbone_chain[-1].tail
                link_data['curr_rot'] = world_matrix.col[1].xyz
                link_data['rest_rot'] = rest_mat_world.col[1].xyz
                link_data['rest_mat'] = rest_mat_world
                link_data['m_p_inv']  = parent_mat.inverted() @ rest_mat_world
                chain_init_data.append(link_data)

                spring_sub_chain = obj_springs.add_sub_chain(chain_init_data)
                for pbone, seg in zip(pbone_chain, spring_sub_chain.segments):
                    seg.pbone = pbone
                    # XXX: what in pbone is in multiple colls?
                    if bpy.app.version >= (4, 0, 0):
                        seg.phys_props = get_pbone_physics(pbone) # idx should be > -1
                    else:
                        obj_jiggle_props = arma_obj.ht_props.jiggle_props
                        seg.phys_props = obj_jiggle_props.physics_props[pbone.ht_props.jiggle_index] # idx should be > -1
                    # seg.set_physics(phys_props.mass, phys_props.damping, phys_props.stiffness, phys_props.gravity)

                if empties:
                    spring_sub_chain.empties = empties
                spring_sub_chain.collision_bones = collision_bones
                spring_sub_chain.dt = dt
                spring_sub_chain.set_target_bones(pbone_chain)
        if obj_springs.sub_chains: # does armature has any jiggle bone chains
            armature_springs.append(obj_springs)
    return armature_springs

SPRING_OBJS = []
@persistent
def spring_bone_frame_change(scn):
    # print('Frame change')
    # bpy.context.scene.ht_props.jiggle_props.jiggle_frame_step = True
    # NOTE: for this SpringChains.can_rest = False - sim is always running
    for obj_springs in SPRING_OBJS:
        obj_springs.solve_chains() # jiggle them always - even in no scene update (wait for chain.reached_rest_state)


class HT_OT_SpringBones(bpy.types.Operator):
    bl_idname = "object.enable_spring_physics"
    bl_label = "Enable Spring Sim"
    bl_description = "Enable Spring Simulation on bones. Assign jiggle physics to bones first (in pose mode).\nSimulation timesteps is based on scene animation fps\n[ESC] - to finish the sim"
    bl_options = {'REGISTER','UNDO_GROUPED'}

    _timer = None
    _handle = None

    @staticmethod
    def draw_callback_px(self, context):
        gpu.state.point_size_set(6)
        gpu.state.depth_test_set("NONE") # if bpy.app.version >= (3,5,0) else bgl.glDisable(bgl.GL_DEPTH_TEST)
        gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)

        batch = batch_for_shader(shader, 'POINTS', {"pos": DRAW_PTS})
        shader.uniform_float("color", (1.0, 1.0, 0.2, 1.0))
        shader.bind()
        batch.draw(shader)

        gpu.state.point_size_set(1)

    def invoke(self, context, event):
        scn_jiggle_props = context.scene.ht_props.jiggle_props
        scn_jiggle_props.sim_is_running = True
        spring_chains = init_arma_spring_chains(context)
        if len(spring_chains) == 0:
            self.report({'ERROR'}, f'No spring bones set on visible armatures. Cancelling!')
            return {'CANCELLED'}
        self.spring_objs = spring_chains
        self.dep = context.evaluated_depsgraph_get()
        if scn_jiggle_props.sim_mode == 'FRAME_CHANGE': # modal only takes care of finishing the sim
            global SPRING_OBJS
            SPRING_OBJS = spring_chains
            bpy.app.handlers.frame_change_post.append(spring_bone_frame_change)
        else: # 'REALTIME'
            dt = 1/context.scene.render.fps
            self._timer = context.window_manager.event_timer_add(dt, window=context.window)
        context.window_manager.modal_handler_add(self)

        # args = (self, context)
        # self._handle = bpy.types.SpaceView3D.draw_handler_add(self.draw_callback_px, args, 'WINDOW', 'POST_VIEW')
        return {"RUNNING_MODAL"}


    def modal(self, context, event):
        scn_jiggle_props = context.scene.ht_props.jiggle_props
        if event.type == 'TIMER':
            # print('Timer')
            # NOTE: force chains movement on scene change
            if len(self.dep.updates) > 0 or context.screen.is_animation_playing: # XXX: disabled or else sim is paused on slow animation
                # for update in self.dep.updates:
                #     if update.id.original == context.active_object
                for obj_springs in self.spring_objs:
                    for sub_chain in obj_springs.sub_chains:
                        sub_chain.reached_rest_state = False #it is possible we need update arma

            # for sub_chain in self.spring_objs.sub_chains: # bones was deleted => finish sim
            #     if not sub_chain.pbones[0].bone:
            #         return self.finish(context)
            for obj_springs in self.spring_objs:
                obj_springs.solve_chains() # jiggle them always - even in no scene update (wait for chain.reached_rest_state)

        # on undo event finish spring sim
        elif event.type == 'Z' and event.value == 'PRESS' and event.ctrl:
            return self.finish(context)

        elif event.type in {"ESC"} and event.value == 'PRESS':
            return self.finish(context)
        elif not scn_jiggle_props.sim_is_running:
            return self.finish(context)

        return {"PASS_THROUGH"}
        # return {"RUNNING_MODAL"}

    def finish(self, context):
        #apply pose bones constraints
        if self._handle:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
        context.scene.ht_props.jiggle_props.sim_is_running = False
        for obj_springs in self.spring_objs:
            obj_springs.finish_chains()

        if context.scene.ht_props.jiggle_props.sim_mode == 'FRAME_CHANGE':
            bpy.app.handlers.frame_change_post.remove(spring_bone_frame_change)
        else:
            wm = context.window_manager
            wm.event_timer_remove(self._timer)
            self._timer = None
        return {"FINISHED"}
