# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

import bpy
from bpy_extras import view3d_utils
from mathutils.geometry import intersect_line_plane
from mathutils import Vector, Matrix
from ..hair_baking.hair_geometry_nodes_selections import get_tips_pts, get_selected_strands_mask, select_blender_curves, get_edit_selected_points_bool
from . import mi_utils_base as ut_base
from . import mi_widget_linear_deform as l_widget

HT_WORKSPACE_SELECT_TIP = "ht_workspace_select_tip"

pass_keys = ['NUMPAD_0', 'NUMPAD_1', 'NUMPAD_3', 'NUMPAD_4',
             'NUMPAD_5', 'NUMPAD_6', 'NUMPAD_7', 'NUMPAD_8',
             'NUMPAD_9', 'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE',
             'MOUSEMOVE']

# get mouse on a plane


def get_mouse_on_plane(context, plane_pos, plane_dir, mouse_coords):
    region = context.region
    rv3d = context.region_data
    final_dir = rv3d.view_rotation @ Vector((0.0, 0.0, -1.0)) if plane_dir is None else plane_dir

    mouse_pos = view3d_utils.region_2d_to_origin_3d(region, rv3d, mouse_coords)
    mouse_dir = view3d_utils.region_2d_to_vector_3d(region, rv3d, mouse_coords)
    new_pos = intersect_line_plane(mouse_pos, mouse_pos + (mouse_dir * 10000.0), plane_pos, final_dir, False)
    if new_pos:
        return new_pos
    return None


class HTOOL_OT_Linear_Deformer(bpy.types.Operator):
    """Draw a line with the mouse"""
    bl_idname = "curve.deform_curve"
    bl_label = "Deform Curve"
    bl_description = "Deformer Curve using gizmo"
    bl_options = {'REGISTER', 'UNDO'}

    # curve tool mode
    tool_modes = ('IDLE', 'MOVE_POINT', 'DRAW_TOOL', 'SCALE_ALL', 'SCALE_FRONT', 'MOVE_ALL', 'TWIST', 'TAPE', 'ROTATE_ALL', 'BEND_ALL', 'BEND_SPIRAL')
    tool_mode = 'IDLE'
    manipulator = None

    do_update = None
    lw_tool = None
    lw_tool_axis = None
    active_lw_point = None
    deform_mouse_pos = None
    deform_vec_pos = None

    bend_scale_len = None

    start_work_center = None
    pointsCurv = None
    apply_tool_verts = None

    # history
    h_undo = None
    h_redo = None

    def add_history(self):
        self.h_redo.clear()
        obj = bpy.context.active_object
        verts =[obj.data.points[vert_data[0]] for vert_data in self.apply_tool_verts] # (vert.index, weights, original_pos)
        self.h_undo.append([vert.position.copy() for vert in verts])


    def undo_history(self):
        if self.h_undo:
            pre_history = self.h_undo[-1]
            if len(self.h_undo) > 1:  # 0 index is always original verts
                self.h_undo.remove(pre_history)
                self.h_redo.append(pre_history)

            obj = bpy.context.active_object
            verts = [obj.data.points[vert_data[0]] for vert_data in self.apply_tool_verts] # (vert.index, weights, original_pos)
            for pt, h_vert in zip(verts, self.h_undo[-1]):
                pt.position = h_vert.copy()


    def redo_history(self):
        if self.h_redo:
            obj = bpy.context.active_object
            verts =[obj.data.points[vert_data[0]] for vert_data in self.apply_tool_verts] # (vert.index, weights, original_pos)
            for v, h_vert in zip(verts, self.h_redo[-1]):
                v.position = h_vert.copy()
            self.h_undo.append(self.h_redo[-1])
            del self.h_redo[-1]



    @staticmethod
    def get_selected_points(context, active_obj):
        curveData = active_obj.data
        points = []
        points_factor = []
        # DONE: make it work only on selected tips in OBj mode?
        select = active_obj.data.attributes.get(HT_WORKSPACE_SELECT_TIP)
        if context.mode == 'EDIT_CURVES':
            sel_mask = get_edit_selected_points_bool(active_obj) # based on blender selection '.select' atrib
            for spl in active_obj.data.curves:
                if not any([sel_mask[p.index] for p in spl.points]):
                    continue # skip
                spl_len = spl.points_length # for new curves
                for i, p in enumerate(spl.points):
                    if sel_mask[p.index]:
                        points.append(p)
                        points_factor.append(i/spl_len)


        elif context.mode in ('OBJECT','SCULPT_CURVES'):  # for workspace tool mostly
            spl_ids, _, _ = get_tips_pts(context, active_obj, world_space=False, only_sel=True)
            for spl_id in spl_ids:  # for strand point
                polyline = curveData.curves[spl_id]
                points += [p for p in polyline.points]
                points_factor += [i/polyline.points_length for i in range(polyline.points_length)]
        return points, points_factor

    @staticmethod
    def get_not_hidden_points(active_obj):
        points = []
        points_factor = []
        for polyline in active_obj.data.curves:  # for strand point
            points += [p for p in polyline.points] #  if p.hide is False
            points_factor += [i/polyline.points_length for i in range(polyline.points_length)]
        return points, points_factor

    def invoke(self, context, event):
        reset_params(self)
        if context.area.type == 'VIEW_3D':
            # the arguments we pass the the callbackection
            active_obj = context.active_object
            if len(active_obj.data.curves):

                pre_work_points, points_factor = self.get_selected_points(context, active_obj)
                if not pre_work_points:
                    pre_work_points, points_factor = self.get_not_hidden_points(active_obj)
                self.pointsCurv = pre_work_points
                self.points_factor = points_factor

                if pre_work_points:
                    self.start_work_center = ut_base.get_vertices_center(pre_work_points, active_obj, False)  # works ok on curve it seems

                    # add original vert to history
                    self.init_history = True

                    # change manipulator
                    self.manipulator = context.space_data.show_gizmo
                    context.space_data.show_gizmo = False

                    # tooltip
                    tooltip_text = "I:Invert, Z:Z-Constraint, X:X-Constraint, S:Scale, Shift-S:ScaleForward, G:Move, R:Rotate, B:Bend, Shift-B:BendSpiral, T:Tape, Shift-T:Twist, Ctrl+Z:Undo, Ctrl+Shift+Z:Redo"
                    context.area.header_text_set(tooltip_text)

                    # Add the region OpenGL drawing callback
                    args = (self, context)
                    self.lin_deform_handle_2d = bpy.types.SpaceView3D.draw_handler_add(lin_def_draw_2d, args, 'WINDOW', 'POST_PIXEL')
                    context.window_manager.modal_handler_add(self)

                    return {'RUNNING_MODAL'}

                else:
                    self.report({'WARNING'}, "No verts!!")
                    return {'CANCELLED'}

            else:
                self.report({'WARNING'}, "Curve has not splines!")
                return {'CANCELLED'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}

    def modal(self, context, event):
        region = context.region
        rv3d = context.region_data
        m_coords = event.mouse_region_x, event.mouse_region_y
        active_obj = context.active_object
        # update check
        self.do_update = True

        keys_pass = event.type in pass_keys

        mw_i = active_obj.matrix_world.inverted()

        # key pressed
        if self.tool_mode == 'IDLE' and event.value == 'PRESS' and keys_pass is False:

            # init here second time or points.indices will be  messed up
            pre_work_points, points_factor = self.get_selected_points(context, active_obj)
            if not pre_work_points:
                pre_work_points, points_factor = self.get_not_hidden_points(active_obj)
            self.pointsCurv = pre_work_points
            self.points_factor = points_factor

            if event.type in {'LEFTMOUSE', 'SELECTMOUSE'}:
                if self.lw_tool:
                    # pick linear widget point
                    picked_point = l_widget.pick_lw_point(context, m_coords, self.lw_tool)
                    if picked_point:
                        self.deform_mouse_pos = Vector(m_coords)
                        self.active_lw_point = picked_point
                        self.tool_mode = 'MOVE_POINT'
                else:
                    picked_point = get_mouse_on_plane(context, self.start_work_center, None, m_coords)
                    if picked_point:
                        self.lw_tool = l_widget.MI_Linear_Widget()

                        self.lw_tool.start_point = l_widget.MI_LW_Point(picked_point.copy())
                        self.lw_tool.middle_point = l_widget.MI_LW_Point(picked_point.copy())
                        self.lw_tool.end_point = l_widget.MI_LW_Point(picked_point)

                        self.active_lw_point = self.lw_tool.end_point
                        self.tool_mode = 'MOVE_POINT'

            elif event.type in {'S', 'G', 'R', 'B', 'T'}:
                # set tool type
                if event.type == 'S':
                    self.tool_mode = 'SCALE_FRONT' if event.shift else 'SCALE_ALL'
                elif event.type == 'R':
                    self.tool_mode = 'ROTATE_ALL'
                elif event.type == 'G':
                    self.tool_mode = 'MOVE_ALL'
                elif event.type == 'B':
                    self.tool_mode = 'BEND_SPIRAL' if event.shift else 'BEND_ALL'
                elif event.type == 'T':
                    self.tool_mode = 'TWIST' if event.shift else 'TAPE'

                # get tool verts
                if self.tool_mode in {'SCALE_FRONT', 'TAPE'}:
                    # do not clamp for SCALE_FRONT mode
                    self.apply_tool_verts = l_widget.get_tool_verts_curv(self.lw_tool, self.pointsCurv, active_obj, False, True, self.points_factor)
                else:
                    self.apply_tool_verts = l_widget.get_tool_verts_curv(self.lw_tool, self.pointsCurv, active_obj, True, True, self.points_factor)

                # add to undo history
                if self.init_history:
                    self.add_history()
                    self.init_history = False


                # set some settings for tools
                if self.tool_mode in {'SCALE_ALL', 'SCALE_FRONT', 'TAPE'}:
                    self.deform_mouse_pos = Vector(m_coords)

                elif self.tool_mode == 'MOVE_ALL':
                    mouse_pos_3d = get_mouse_on_plane(context, self.lw_tool.start_point.position, None, m_coords)
                    self.deform_vec_pos = mouse_pos_3d  # 3d location

                elif self.tool_mode in {'ROTATE_ALL', 'TWIST', 'BEND_ALL', 'BEND_SPIRAL'}:
                    start_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.lw_tool.start_point.position)
                    self.deform_vec_pos = (Vector(m_coords) - start_2d).normalized()  # 2d direction
                    self.deform_mouse_pos = 0.0  # we will use it as angle counter

                    if self.tool_mode in {'BEND_SPIRAL', 'BEND_ALL'}:
                        self.bend_scale_len = (Vector(m_coords) - start_2d).length

                # return {'RUNNING_MODAL'}

            elif event.type in {'Z', 'X'} and self.lw_tool:
                if event.type == 'Z' and event.ctrl:
                    if event.shift:
                        self.redo_history()
                    else:
                        self.undo_history()

                else:
                    if event.type == 'X':
                        if self.lw_tool_axis:
                            if self.lw_tool_axis == 'X':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'X_Left', 1.0)
                                self.lw_tool_axis = 'X_Left'
                            elif self.lw_tool_axis == 'X_Left':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'X_Right', 1.0)

                                # revert direction
                                #stp = self.lw_tool.start_point.position.copy()
                                #self.lw_tool.start_point.position = self.lw_tool.end_point.position
                                #self.lw_tool.end_point.position = stp

                                self.lw_tool_axis = 'X_Right'
                            elif self.lw_tool_axis == 'X_Right':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'X', 1.0)
                                self.lw_tool_axis = 'X'
                            else:
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'X', 1.0)
                                self.lw_tool_axis = 'X'
                        else:
                            l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'X', 1.0)
                            self.lw_tool_axis = 'X'
                    else:
                        if self.lw_tool_axis:
                            if self.lw_tool_axis == 'Z':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'Z_Top', 1.0)
                                self.lw_tool_axis = 'Z_Top'
                            elif self.lw_tool_axis == 'Z_Top':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'Z_Bottom', 1.0)

                                # revert direction
                                #stp = self.lw_tool.start_point.position.copy()
                                #self.lw_tool.start_point.position = self.lw_tool.end_point.position
                                #self.lw_tool.end_point.position = stp

                                self.lw_tool_axis = 'Z_Bottom'
                            elif self.lw_tool_axis == 'Z_Bottom':
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'Z', 1.0)
                                self.lw_tool_axis = 'Z'
                            else:
                                l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'Z', 1.0)
                                self.lw_tool_axis = 'Z'
                        else:
                            l_widget.setup_lw_tool(rv3d, self.lw_tool, active_obj, self.pointsCurv, 'Z', 1.0)
                            self.lw_tool_axis = 'Z'

            elif event.type == 'I' and self.lw_tool:
                start_copy = self.lw_tool.start_point.position.copy()
                self.lw_tool.start_point.position = self.lw_tool.end_point.position
                self.lw_tool.end_point.position = start_copy

            context.area.tag_redraw()
        # TOOL WORK!
        if self.tool_mode == 'MOVE_POINT':

            if event.value == 'RELEASE':
                self.tool_mode = 'IDLE'
                return {'RUNNING_MODAL'}
            else:
                # move points
                new_point_pos = get_mouse_on_plane(context, self.active_lw_point.position, None, m_coords)
                if self.active_lw_point.position == self.lw_tool.start_point.position or self.active_lw_point.position == self.lw_tool.end_point.position:
                    self.active_lw_point.position = new_point_pos
                    l_widget.update_middle_point(self.lw_tool)
                elif self.active_lw_point.position == self.lw_tool.middle_point.position:
                    self.lw_tool.start_point.position += new_point_pos - self.active_lw_point.position
                    self.lw_tool.end_point.position += new_point_pos - self.active_lw_point.position
                    self.lw_tool.middle_point.position = new_point_pos
                context.area.tag_redraw()
                return {'RUNNING_MODAL'}

        elif self.tool_mode in {'SCALE_ALL', 'SCALE_FRONT', 'TAPE'}:

            if event.value == 'RELEASE' and event.type in {'LEFTMOUSE', 'SELECTMOUSE'}:
                # add to undo history
                self.add_history()
                self.tool_mode = 'IDLE'

            elif self.do_update:
                # move points
                start_point_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.lw_tool.start_point.position)
                if start_point_2d:
                    tool_dist = (start_point_2d - self.deform_mouse_pos).length
                    now_dist = (start_point_2d - Vector(m_coords)).length
                    apply_value = (now_dist - tool_dist) / tool_dist
                    if apply_value != 0.0:
                        tool_orig = mw_i @ self.lw_tool.start_point.position
                        tool_end = mw_i @ self.lw_tool.end_point.position
                        tool_vec = tool_end - tool_orig
                        tool_dir = (tool_end - tool_orig).normalized()
                        for vert_data in self.apply_tool_verts: # (vert.index, weights, original_pos)
                            scale_vec = None
                            scale_value = vert_data[1]

                            if self.tool_mode == 'SCALE_ALL':
                                scale_vec = (vert_data[2] - tool_orig)
                            elif self.tool_mode == 'SCALE_FRONT':
                                scale_vec = (tool_end - tool_orig)
                            else:
                                # TAPE
                                scale_vec = vert_data[2].resized(3) - (tool_orig + (tool_dir * vert_data[1] * (tool_vec).length))
                                scale_value = min(1.0, vert_data[1])
                            out = vert_data[2].resized(3) + (scale_vec * (scale_value * apply_value))
                            out.resize_4d()
                            active_obj.data.points[vert_data[0]].position = out.xyz

                context.area.tag_redraw()
            self.do_update = False
            return {'RUNNING_MODAL'}

        elif self.tool_mode == 'MOVE_ALL':
            if event.value == 'RELEASE' and event.type in {'LEFTMOUSE', 'SELECTMOUSE'}:
                # add to undo history
                self.add_history()
                self.tool_mode = 'IDLE'

            elif self.do_update:
                mouse_pos_3d = get_mouse_on_plane(context, self.lw_tool.start_point.position, None, m_coords)
                mouse_pos_3d = mw_i @ mouse_pos_3d
                start_pos = mw_i @ self.lw_tool.start_point.position
                orig_pos = mw_i @ self.deform_vec_pos
                orig_vec = orig_pos - start_pos
                move_vec = (mouse_pos_3d - start_pos) - orig_vec

                for vert_data in self.apply_tool_verts: # (vert.index, weights, original_pos)
                    move_value = vert_data[1]
                    active_obj.data.points[vert_data[0]].position = vert_data[2] + (move_vec * move_value) # whit this shit is not updating....

                self.do_update = False
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        elif self.tool_mode in {'ROTATE_ALL', 'TWIST', 'BEND_ALL', 'BEND_SPIRAL'}:
            if event.value == 'RELEASE' and event.type in {'LEFTMOUSE', 'SELECTMOUSE'}:
                # add to undo history
                self.add_history()
                self.tool_mode = 'IDLE'

            elif self.do_update:
                m_coords = Vector(m_coords)  # convert into vector for operations
                start_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, self.lw_tool.start_point.position)
                new_vec_dir = (m_coords - start_2d).normalized()
                rot_angle = new_vec_dir.angle(self.deform_vec_pos)

                start_3d = self.lw_tool.start_point.position
                end_3d = self.lw_tool.end_point.position

                if rot_angle != 0.0:
                    # check for left or right direction to rotate
                    vec_check_1 = Vector((new_vec_dir[0], new_vec_dir[1], 0))
                    vec_check_2 = Vector((new_vec_dir[0]-self.deform_vec_pos[0], new_vec_dir[1]-self.deform_vec_pos[1], 0))
                    checker_side_dir = vec_check_1.cross(vec_check_2).normalized()[2]
                    if checker_side_dir > 0.0:
                        rot_angle = -rot_angle

                    start_pos = self.lw_tool.start_point.position
                    rot_dir = None
                    if self.tool_mode == 'ROTATE_FRONT' or self.tool_mode == 'TWIST':
                        # ROTATE_FRONT code
                        rot_dir = (end_3d - start_3d) # .normalized() - no need?
                    else:
                        rot_dir = rv3d.view_rotation @ Vector((0.0, 0.0, -1.0)) # .normalized() - no need?

                    rot_angle += self.deform_mouse_pos  # add rot angle

                    bend_side_dir = None
                    faloff_len = None
                    spiral_value = 0.0  # only for BEND_SPIRAL
                    bend_scale_value = 1.0  # only for BEND_ALL

                    if self.tool_mode == 'BEND_ALL' or self.tool_mode == 'BEND_SPIRAL':
                        bend_side_dir = ((end_3d - start_3d).cross(rot_dir)).normalized()
                        faloff_len = end_3d - start_3d

                        if self.tool_mode == 'BEND_SPIRAL':
                            val_scale = None
                            if rot_angle > 0.0:
                                val_scale = (1.0 - ((m_coords - start_2d).length / self.bend_scale_len))
                            else:
                                val_scale = (((m_coords - start_2d).length / self.bend_scale_len))

                            spiral_value = 1.0 - (faloff_len.length * val_scale)

                        else:
                            val_scale = (((m_coords - start_2d).length / self.bend_scale_len))
                            bend_scale_value = ((val_scale))

                    do_bend = False
                    if self.tool_mode == 'BEND_ALL' or self.tool_mode == 'BEND_SPIRAL':
                        do_bend = True

                    for vert_data in self.apply_tool_verts: # (vert.index, weights, original_pos)
                        weight = vert_data[1]
                        final_apply_value = rot_angle * weight

                        # do rotation
                        if final_apply_value != 0.0:
                            rot_mat = Matrix.Rotation(final_apply_value, 3, rot_dir)
                            vert = active_obj.data.points[vert_data[0]]

                            if do_bend:
                                vert_temp = (active_obj.matrix_world @ vert_data[2]) - ((faloff_len) * weight)
                                back_offset = (((faloff_len).length / (final_apply_value)) + spiral_value) * (weight * bend_scale_value)
                                vert_temp += bend_side_dir * back_offset
                                vert.position = vert_temp
                            else:
                                # set original position
                                vert.position[0] = vert_data[2][0]
                                vert.position[1] = vert_data[2][1]
                                vert.position[2] = vert_data[2][2]

                            # ROTATE VERTS!
                            if do_bend:
                                vert.position = rot_mat @ ((vert.position) - start_pos) + start_pos
                                back_offset = ((faloff_len).length / (final_apply_value)) * (weight * bend_scale_value)
                                vert.position = mw_i @ (vert.position - (bend_side_dir * back_offset))
                            else:
                                vert.position = mw_i @ (rot_mat @ ((active_obj.matrix_world @ vert.position) - start_pos) + start_pos)

                    self.deform_vec_pos = new_vec_dir
                    self.deform_mouse_pos = rot_angle  # set new angle rotation for next step

                self.do_update = False
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        else:
            if event.value == 'RELEASE' and event.type in {'LEFTMOUSE', 'SELECTMOUSE'}:
                self.tool_mode = 'IDLE'
                return {'RUNNING_MODAL'}

        # get keys
        if keys_pass is True:
            return {'PASS_THROUGH'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            context.space_data.show_gizmo = self.manipulator
            bpy.types.SpaceView3D.draw_handler_remove(self.lin_deform_handle_2d, 'WINDOW')
            context.area.header_text_set(None)
            context.area.tag_redraw()
            return {'FINISHED'}

        return {'RUNNING_MODAL'}


def reset_params(self):
    self.tool_mode = 'IDLE'
    self.manipulator = None
    self.deform_mouse_pos = None
    self.deform_vec_pos = None
    self.bend_scale_len = None

    self.do_update = False
    self.lw_tool = None
    self.lw_tool_axis = None
    self.active_lw_point = None

    self.start_work_center = None
    self.work_verts = None
    self.apply_tool_verts = None

    self.h_undo = []
    self.h_redo = []




def lin_def_draw_2d(self, context):
    rv3d = context.region_data
    if self.lw_tool:
        lw_dir = (self.lw_tool.start_point.position - self.lw_tool.end_point.position).normalized()
        cam_view = (rv3d.view_rotation @ Vector((0.0, 0.0, -1.0))).normalized()
        side_dir = lw_dir.cross(cam_view).normalized()
        l_widget.draw_lw(context, self.lw_tool, side_dir, True)
