# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import bmesh
import math
from bpy_extras import view3d_utils
from copy import deepcopy
import numpy as np
from mathutils import Vector
from bpy.props import BoolProperty, IntProperty
from .hair_curve_helpers import get_obj_mesh_bvht
from .resample2d import interpol_Catmull_Rom_splines
from .utils.helper_functions import get_curve_points
from .utils.curve_wrapper import SplineSimple


class HTOOL_OT_ParticleHairToCurves(bpy.types.Operator):
    bl_label = "Particle Hair to Curves"
    bl_idname = "object.hair_to_curves"
    bl_description = "Convert active Hair particle system to curves. "
    bl_options = {"REGISTER", "UNDO"} #fixes operator on redo

    hairType: bpy.props.EnumProperty(name="Hair Type", default="NURBS",
                                      items=(("BEZIER", "Bezier", ""),
                                             ("NURBS", "Nurbs", ""),
                                             ("POLY", "Poly", "")))
    particleHairRes: IntProperty(name="Hair Res", description="How many points output curve will build off", default=2, min=0, soft_max=6)

    def check(self, context):  # DONE: can prop panel be fixed/refreshed when using f6 prop popup
        return True

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        box.label(text="Curve from Particle Hair Settings:")
        box.prop(self, 'particleHairRes')
        box.prop(self, 'hairType')

    def invoke(self, context, event):
        self.active_obj = context.active_object
        self.depsgraph = context.evaluated_depsgraph_get()
        self.obj_eval = context.active_object.evaluated_get(self.depsgraph)
        self.source_surf_bvh_w = get_obj_mesh_bvht(self.obj_eval, self.depsgraph, applyModifiers=True, world_space=True)  # for tilt aliging and other
        return self.execute(context)


    def execute(self, context):
        particleObj = context.active_object
        partsysMod = None
        for mod in particleObj.modifiers:
            if mod.type == 'PARTICLE_SYSTEM':  #  and mod.show_viewport use first visible
                if particleObj.particle_systems.active.name == mod.particle_system.name:
                    if mod.particle_system.settings.type == "HAIR":
                        partsysMod = mod  # use first
                        break
        if partsysMod is None:
            self.report({'INFO'}, 'No active Particle Hair System found')
            return {"CANCELLED"}
        partsysMod.particle_system.settings.display_step = self.particleHairRes
        # partsysMod.particle_system.settings.draw_step = self.particleHairRes

        context.scene.render.hair_subdiv = 1
        bpy.ops.object.modifier_convert(modifier=partsysMod.name)
        generated_hair = context.active_object
        me = generated_hair.data
        bm = bmesh.new()  # create an empty BMesh
        bm.from_mesh(me)
        bm.to_mesh(me)
        bm.free()
        generated_hair.data.transform(particleObj.matrix_world.inverted())  # cos modifier_convert above, applays loc,rot, scale so revert it
        generated_hair.matrix_world = particleObj.matrix_world
        bpy.ops.object.convert(target='CURVE')
        bpy.ops.object.editmode_toggle()
        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.spline_type_set(type=self.hairType)

        if self.hairType == 'NURBS':
            for polyline in context.active_object.data.splines:
                polyline.order_u = 3  # like bezier thing
                polyline.use_endpoint_u = True
        bpy.ops.object.editmode_toggle()

        generated_hair.data.dimensions = '3D'
        generated_hair.data.fill_mode = 'FULL'
        generated_hair.data.bevel_resolution = 2
        generated_hair.data.resolution_u = 3
        # partsysMod.show_viewport = False

        return {"FINISHED"}


def getObjectMassCenter(context, obj):
    local_bbox_center = 0.125 * sum((Vector(b) for b in obj.bound_box), Vector())
    global_bbox_center = obj.matrix_world @ local_bbox_center
    region = context.region
    rv3d = context.space_data.region_3d
    return view3d_utils.location_3d_to_region_2d(region, rv3d, global_bbox_center)  # 3d coords to 2d reg


def particle_hair_override(depsgraph, particleObj, pointsList, idx = -1): #todo maybe add resampling some day. cant update from UI. bad poll - edit.particle_hair
    ''' Redraw last strand. What if different pt count. '''
    # depsgraph = bpy.context.evaluated_depsgraph_get()
    depsgraph.update()
    object_eval = particleObj.evaluated_get(depsgraph)
    particle_sys_eval = object_eval.particle_systems.active
    if len(particle_sys_eval.particles[0].hair_keys) == len(pointsList):
        for i, parp in enumerate(pointsList):
            particle_sys_eval.particles[idx].hair_keys[i].co = parp  # wtf why
        bpy.ops.particle.particle_edit_toggle()
        bpy.ops.particle.particle_edit_toggle()


def particleHairFromPoints(context, particleObj, pointsList, extend=False, stick = True):
    ''' Particle system only works on equal len strands. No api for hair_keys with ++ point, or --'''
    particle_sys = None
    context.view_layer.objects.active = particleObj
    for mod in particleObj.modifiers:
        if mod.type == 'PARTICLE_SYSTEM':  # use first visible
            if mod.particle_system.name == particleObj.particle_systems.active.name:
                mod.show_viewport = True
                particle_sys = mod.particle_system
                if particle_sys.is_global_hair and extend:  # prevent jump in there were exising hairs, if they are not connected
                    bpy.ops.particle.connect_hair(all=False)
                depsgraph = bpy.context.evaluated_depsgraph_get()
                particleObj_eval = particleObj.evaluated_get(depsgraph)
                particle_sys_eval = particleObj_eval.particle_systems.active  # use last
                extendList = []
                if extend and mod.particle_system.settings.type == "HAIR":
                    for strand in particle_sys_eval.particles:  # for strand point
                        extendList.append([hair_key.co for hair_key in strand.hair_keys])
                    if len(extendList) > 0:
                        pointsList = interpol_Catmull_Rom_splines(extendList, len(pointsList[0]), uniform_spacing=False, same_point_count=True) + pointsList
                bpy.ops.particle.edited_clear()  #? toogle editability - without cant set particle count
                break
    if particle_sys is None:  # create new one
        print('No active Particle Hair System found! Adding new one')
        particleObj.modifiers.new("Hair_From_Curves", 'PARTICLE_SYSTEM')
        depsgraph = bpy.context.evaluated_depsgraph_get()
        particle_sys = particleObj.particle_systems[-1]
        particle_sys.name = 'Hair_HTool'

    # Create new settings
    particle_sys.settings.type = 'HAIR'
    particle_sys.settings.emit_from = 'FACE'
    particle_sys.settings.use_strand_primitive = True

    hair_len = len(pointsList[0])-1
    particle_sys.settings.hair_step = hair_len  # == HAIR_KEY NUMB
    particle_sys.settings.count = len(pointsList)
    particle_sys.settings.display_step = max(math.floor(math.log(len(pointsList[0]), 2)), 2) + 1
    # partsysMod.name = 'HairFromCurves'

    if not particle_sys.is_edited:
        bpy.ops.particle.particle_edit_toggle()
        bpy.ops.particle.select_all(action='SELECT')
        x, y = getObjectMassCenter(context, particleObj)
        # bpy.context.scene.tool_settings.particle_edit.tool = 'COMB' #! does not work in Modeling layout?
        bpy.ops.wm.tool_set_by_id(name="builtin_brush.Comb")
        if bpy.app.version > (2, 90):
            bpy.ops.particle.brush_edit(  # do 'empty' stroke to generate comb cache
                stroke=[{"name": "", "location": (0, 0, 0), "mouse": (x, y), "pressure": 0, "size": 0, "pen_flip": False, "time": 0, "is_start": False, "mouse_event": (0, 0), "x_tilt": 0, "y_tilt": 0},
                        {"name": "", "location": (0, 0, 0), "mouse": (x + 1, y + 1), "pressure": 0, "size": 0, "pen_flip": False, "time": 0, "is_start": False, "mouse_event": (0, 0), "x_tilt": 0, "y_tilt": 0}])
        else:
            bpy.ops.particle.brush_edit(  # do 'empty' stroke to generate comb cache
                stroke=[{"name": "", "location": (0, 0, 0), "mouse": (x, y), "pressure": 0, "size": 0, "pen_flip": False, "time": 0, "is_start": False},
                        {"name": "", "location": (0, 0, 0), "mouse": (x + 1, y + 1), "pressure": 0, "size": 0, "pen_flip": False, "time": 0, "is_start": False}])
        weights = [i / hair_len for i in range(hair_len, -1, -1)]
        for particle in particle_sys.particles:
            particle.hair_keys.foreach_set('weight', weights)
        bpy.ops.particle.rekey(keys_number=len(pointsList[0]))  # fixes jumping bug and removes doubles
        bpy.ops.particle.select_all(action='DESELECT')
        #sticks the attached curves to head surface
        bpy.ops.particle.particle_edit_toggle()

    #write hair_keys.co
    depsgraph = bpy.context.evaluated_depsgraph_get()
    particleObj_eval = particleObj.evaluated_get(depsgraph)
    particle_sys_eval = particleObj_eval.particle_systems.active
    for i, points in enumerate(pointsList):  # for strand point
        particle_sys_eval.particles[i].location = points[0]
        for j, point in enumerate(points):  # for strand point
            particle_sys_eval.particles[i].hair_keys[j].co = Vector(point)
    #to update hair co's
    bpy.ops.particle.particle_edit_toggle()
    bpy.ops.particle.particle_edit_toggle()
    if stick:
        if not particleObj_eval.particle_systems.active.is_global_hair:
            bpy.ops.particle.disconnect_hair(all=False)
            bpy.ops.particle.connect_hair(all=False)


class HTOOL_OP_ShortenParticles(bpy.types.Operator):
    bl_idname = "particles.scale_particles"
    bl_label = "Scale Hair"
    bl_description = "Shorten/lengthen particle hair length in uniform way"
    bl_options = {"REGISTER","UNDO"}

    method: bpy.props.EnumProperty(name='Method', description='Method',
        items=[
            ('SCALE', 'Scale', 'Scale'),
            ('GROW', 'Grow', 'Grow')], default='SCALE')

    shorten_fac: bpy.props.IntProperty(name='Scale Hair', description='Change particle hair length (and generated curve ribbons length as an result too)',
                                       default=100, min=10, soft_max=200, subtype='PERCENTAGE')

    @classmethod
    def poll(cls, context):
        return context.active_object

    def invoke(self, context, event):
        particleObj = context.active_object
        self.is_interactive_combing = bpy.context.object.ht_props.modal_hair.draw_strands
        self.old_mode = context.mode

        if context.mode == 'PARTICLE': #exit particle edit mode
            bpy.ops.particle.particle_edit_toggle()

        depsgraph = context.evaluated_depsgraph_get()
        obj_eval = particleObj.evaluated_get(depsgraph)
        if bpy.context.active_object.particle_systems is None:  # create new one
            self.report({'INFO'}, 'No active Particle Hair System found!')
            return {"CANCELLED"}

        psys_active = obj_eval.particle_systems.active
        if psys_active.settings.type != 'HAIR':  # create new one
            self.report({'INFO'}, 'Active Particle System is not Hair type! Cancelling')
            return {"CANCELLED"}

        if len(psys_active.particles) == 0:  # require more that three strands
            self.report({'INFO'}, 'Active Particle System has zero strands! Cancelling')
            return {"CANCELLED"}

        factor = self.shorten_fac/100
        self.hair_splines = [SplineSimple.from_vec_list([hair_key.co for hair_key in particle.hair_keys]) for particle in psys_active.particles]

        self.shorten_fac = 100
        bpy.ops.ed.undo_push()  # to prevent restoring curve from step above on execute()
        # return self.execute(context)
        return context.window_manager.invoke_props_dialog(self)

    # def draw(self, context):
    #     layout = self.layout
    #     layout.prop(self, 'method')
    #     layout.prop(self, 'shorten_fac')

    def execute(self, context):
        particleObj = context.active_object
        factor = self.shorten_fac/100
        splines = deepcopy(self.hair_splines) # fix redo

        if self.method == 'SCALE':
            for strand in splines:
                first_co = strand.points[0].co
                for p in strand.points:
                    p.co = first_co+(p.co-first_co)*factor
        else:
            for strand in splines:
                pt_cnt = strand.length
                custom_tns = np.linspace(0, factor, pt_cnt)
                strand.resample_pts(res=pt_cnt, uniform=False, t_custom=custom_tns)

        #write hair_keys.co
        depsgraph = context.evaluated_depsgraph_get()
        particleObj_eval = particleObj.evaluated_get(depsgraph)
        particle_sys_eval = particleObj_eval.particle_systems.active
        for i, strand in enumerate(splines):  # for strand point
            # particle_sys_eval.particles[i].location = points[0]
            for j, point in enumerate(strand.points):  # for strand point
                particle_sys_eval.particles[i].hair_keys[j].co = point.co
        #to update hair co's
        bpy.ops.particle.particle_edit_toggle()
        if self.old_mode != 'PARTICLE': #to object mode,  if we were not in 'Comb mode'
            bpy.ops.particle.particle_edit_toggle()

        if self.is_interactive_combing:
            bpy.context.object.ht_props.modal_hair.draw_strands = True
        # if stick:
        # if not particleObj_eval.particle_systems.active.is_global_hair:
        #     bpy.ops.particle.disconnect_hair(all=False)
        #     bpy.ops.particle.connect_hair(all=False)

        return {"FINISHED"}

class HTOOL_OT_ParticleHairFromCurves(bpy.types.Operator):
    bl_label = "Particle Hair from Curves"
    bl_idname = "object.hair_from_curves"
    bl_description = "Creates particle hair strands from Curves object. \n" \
                     "Two objects need to be selected: one mesh and one curve type"
    bl_options = {"REGISTER", "UNDO"}

    # extend: BoolProperty(name="Append", description="Appends new curves to already existing Particle hair strands", default=True)
    mode: bpy.props.EnumProperty(name='Mode', description='Choose if curves should be added to active particle system or new particles',
                                items=[
                                    ('REPLACE', 'Replace', 'Replace hair strands from particle system'),
                                    ('EXTEND', 'Extend', 'Add hair strands to particle system'),
                                    # ('NEW', 'Create new Particle system', 'Create new particle system add hair to it')
                                ], default='REPLACE')
    affected_psys: bpy.props.EnumProperty(name='Affected Hair System', description='Affect only active particle system or matching target by name',
                                 items=[
                                     ('BY_NAME', 'Match by name', 'Match selected object name with Particle Hair name. I no match - create new hair system with object name'),
                                     ('ACTIVE', 'Use active particle system', 'Affect only active particle hair system')
                                 ], default='ACTIVE')
    stick_to_target: BoolProperty(name="Stick to target", description="Force floating hairs roots to stick to mesh surface", default=True)
    p_hair_res_mode: bpy.props.EnumProperty(name='Resolution', description='Define points count on output strands of Particle Hair System',
        items=[
            ('AUTO', 'Auto', 'Addon will decide if use Max of Fixed output resolution depending on input curve strands resolution'),
            ('MAX', 'Max', 'If input curve strands have different resolutions, use curve with maximum points count for particle hair resolution'),
            ('MIN', 'Min', 'If input curve strands have different resolutions, use curve with minimum points count for particle hair resolution'),
            ('AVERAGE', 'Average', 'If input curve strands have different resolutions, use average points count for particle hair resolution'),
            ('FIXED', 'Fixed', 'Choose yourself how much points assign to output particle hair system')
        ], default='AUTO')
    points_count: bpy.props.IntProperty(name='Points Count', description='How much points assing to particle hair system', default=5, min=3, soft_max=20)

    def draw(self, context):
        layout = self.layout

        split = layout.split(factor=0.3, align=True)
        split.label(text='Affected Hair System:')
        split.prop(self, 'affected_psys', expand=True)

        col = layout.column()
        col.label(text='Mode:')
        row = col.row()
        row.prop(self, 'mode', expand=True)

        col = layout.column()
        col.label(text='Resolution:')
        row = col.row()
        row.prop(self, 'p_hair_res_mode', expand=True)

        # layout.prop(self, 'stick_to_target')
        if self.p_hair_res_mode == 'FIXED':
            layout.prop(self, 'points_count')

    def invoke(self, context, event):
        particleObj = context.active_object
        if particleObj.type != 'MESH':
            self.report({'INFO'}, 'Select mesh object first')
            return {"CANCELLED"}
        return self.execute(context)


    def execute(self, context):
        tool_settings = context.scene.tool_settings.particle_edit
        back_deflect = tool_settings.use_emitter_deflect #prevent hair stranigten on scaled objs....
        back_edit_type = tool_settings.type
        tool_settings.use_emitter_deflect = False
        tool_settings.type = 'PARTICLES'

        meshObj = [obj for obj in context.selected_objects if obj.type == 'MESH']
        curveObjs = [obj for obj in context.selected_objects if obj.type == 'CURVE']
        if len(meshObj) == 0 or len(curveObjs) == 0:
            self.report({'INFO'}, 'Select one curve and then mesh that will receive particle hair! Cancelling')
            return {"CANCELLED"}

        particleObj = context.active_object
        merged_strands_lens = []  # 'name', 'points', 'len_list' #for affected_psys 'Active
        merged_strands_points = []

        def recalc_pts(pts, strand_lens):
            if self.p_hair_res_mode == 'AUTO':
                if not np.all(strand_lens == strand_lens[0]):
                    max_len = np.amax(strand_lens)
                    # averLen = int((len(pointsList[0]) + len(pointsList[1])) / 2)  # assume at least two splines?
                    pointsList = interpol_Catmull_Rom_splines(pts, max_len, uniform_spacing=False)
            elif self.p_hair_res_mode == 'MAX':
                pointsList = interpol_Catmull_Rom_splines(pts, np.amax(strand_lens), uniform_spacing=False)
            elif self.p_hair_res_mode == 'MIN':
                pointsList = interpol_Catmull_Rom_splines(pts, np.amin(strand_lens), uniform_spacing=False)
            elif self.p_hair_res_mode == 'AVERAGE':
                pointsList = interpol_Catmull_Rom_splines(pts, int(np.average(strand_lens)), uniform_spacing=False)
            else:
                pointsList = interpol_Catmull_Rom_splines(pts, self.points_count, uniform_spacing=False)
            return pointsList

        for curveObj in curveObjs:
            #wtf undo won't work without this .... extend - if disablend, and then enabled breaks appending old hair (old hair co are all null)
            # bpy.context.scene.update()
            curveData = curveObj.data
            curveData.transform(particleObj.matrix_world.inverted()@curveObj.matrix_world) #to make it go into particle obj space
            pointsList = []
            for polyline in curveData.splines:  # for strand point
                pts = get_curve_points(polyline)
                if len(pts) > 1:
                    pointsList.append(pts)
            curveData.transform(curveObj.matrix_world.inverted() @ particleObj.matrix_world)
            all_strand_len = [len(strand) for strand in pointsList]
            if self.affected_psys == 'ACTIVE': #skip next lines
                merged_strands_lens.extend(all_strand_len)
                merged_strands_points.extend(pointsList)
                continue

            #* since we are matching psystem to obj by name:
            matchFound = False
            for i,particleSystem in enumerate(particleObj.particle_systems):
                if particleSystem.name == curveObj.name:
                    particleObj.particle_systems.active_index = i
                    matchFound = True
                    break

            if not matchFound or len(particleObj.particle_systems)==0:  # No matched Particle Hair System found! Adding new one
                particleObj.modifiers.new(curveObj.name, 'PARTICLE_SYSTEM')
                particleObj.particle_systems[-1].name = curveObj.name
                particleObj.particle_systems.active_index = len(particleObj.particle_systems)-1
                particleObj.data.update()
                depsgraph = bpy.context.evaluated_depsgraph_get()
                depsgraph.update()
                # particleObj.particle_systems.update()

            pointsList = recalc_pts(pointsList, all_strand_len)
            particleHairFromPoints(context, particleObj, pointsList, extend=self.mode == 'EXTEND', stick=self.stick_to_target)

        if self.affected_psys == 'ACTIVE':
            pointsList = recalc_pts(merged_strands_points, merged_strands_lens)
            if len(particleObj.particle_systems) == 0:  # No matched Particle Hair System found! Adding new one
                particleObj.modifiers.new('ParticleFromCurves', 'PARTICLE_SYSTEM')
                particleObj.particle_systems[-1].name = 'ParticleFromCurves'
                particleObj.particle_systems.active_index = len(particleObj.particle_systems)-1
                particleObj.data.update()
                depsgraph = bpy.context.evaluated_depsgraph_get()
                depsgraph.update()
                # particleObj.particle_systems.update()
            particleHairFromPoints(context, particleObj, pointsList, extend=self.mode == 'EXTEND', stick=self.stick_to_target)

        tool_settings.use_emitter_deflect = back_deflect
        tool_settings.type = back_edit_type
        return {"FINISHED"}
