# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import bmesh
from .utils.general_utils import link_child_to_collection, link_obj_to_same_collections, unlink_from_scene, link_coll_to_same_collections, unlink_obj_from_scene, set_parent
from .hair_baking.hair_geometry_nodes import HTOOL_OT_TransferHairSystem
from .hair_baking.hair_geometry_nodes_shared import *


def shape_transfer(context, src_obj, target_obj):
    #transfer shapekeys
    if src_obj.data.shape_keys:
        context.view_layer.update()
        context.view_layer.objects.active = target_obj
        if target_obj.data.shape_keys:
            for shape in target_obj.data.shape_keys.key_blocks:
                shape.value = 0

        sourceShapesValuesBackup = []
        for shape in src_obj.data.shape_keys.key_blocks:
            sourceShapesValuesBackup.append(shape.value)
            shape.value = 0

        if not target_obj.data.shape_keys:
            target_obj.shape_key_add(name=src_obj.data.shape_keys.key_blocks[0].name, from_mix=False)

        shape_names = [s.name for s in src_obj.data.shape_keys.key_blocks]
        for i, shape_name in enumerate(shape_names):  # key_blocks['Key 1'].data[0].co
            if i == 0:
                continue
            src_obj.data.shape_keys.key_blocks[shape_name].value = 1

            # target_skey = target_obj.data.shape_keys.key_blocks.get(shape_name)
            if shape_name in target_obj.data.shape_keys.key_blocks.keys():
                target_skey = target_obj.data.shape_keys.key_blocks[shape_name]
            else:
                target_skey = target_obj.shape_key_add(name=shape_name, from_mix=False)

            depsgraph = context.evaluated_depsgraph_get()
            obj_eval = src_obj.evaluated_get(depsgraph)
            mesh_shape = obj_eval.to_mesh()
            if len(mesh_shape.vertices) == len(target_skey.data):
                for s_vert in mesh_shape.vertices:
                    target_skey.data[s_vert.index].co = s_vert.co

            obj_eval.to_mesh_clear()
            src_obj.data.shape_keys.key_blocks[shape_name].value = 0

        for backup, shape in zip(sourceShapesValuesBackup, src_obj.data.shape_keys.key_blocks):
            shape.value = backup


# PROS: handle mesh HSystem well with subinstances
# CONS: does not handle Curve HSystem with subinstances - sub-instance curves get wrong mod applied it seems... - convert to mesh first?
class HTOOL_OT_CustomConvertMesh(bpy.types.Operator):
    bl_label = "ht_tmp Convert To Mesh"
    bl_idname = "ht_tmp.simple_convert"
    bl_description = "Test - Convert selected Hair System to mesh"
    bl_options = {'UNDO', 'REGISTER'}

    apply_modifiers: bpy.props.BoolProperty(name="Apply Modifiers", default=True)
    merge_output: bpy.props.BoolProperty(name="Join Output Geometry", default=True)

    @staticmethod
    def is_object_instance_from_selected(object_instance, base_ob):
        # For instanced objects we check selection of their instancer (more accurately: check
        # selection status of the original object corresponding to the instancer).
        if object_instance.parent:
            return object_instance.parent.original == base_ob
        # For non-instanced objects we check selection state of the original object.
        return object_instance.object.original == base_ob

    @staticmethod
    def get_obj_for_instance(object_instance, apply_modifiers):
        if apply_modifiers:
            return object_instance.object
        else:
            return object_instance.object.original

    @staticmethod
    def copy_mesh_attributes(source_mesh, target_mesh, write=False):
        # Iterate over all attributes in the source mesh
        target_mesh.update()
        for source_attr in source_mesh.attributes:
            # Create a new attribute with the same name, data type, and domain as the source attribute
            if source_attr.name.startswith('.'): # skip hidden attribs
                continue
            new_attr = target_mesh.attributes.new(
                name=source_attr.name,
                type=source_attr.data_type,
                domain=source_attr.domain
            )
        target_mesh.attributes.update()

        for source_attr in source_mesh.attributes:
            new_attr = target_mesh.attributes.get(source_attr.name)
            if not new_attr:
                continue
            # Copy the data from the source attribute to the new attribute
            data_size = len(source_attr.data)
            if source_attr.data_type == 'FLOAT':
                data = np.empty(data_size,'f')
                source_attr.data.foreach_get('value', data)
                new_attr.data.foreach_set('value', data)
            elif source_attr.data_type == 'INT':
                data = np.empty(data_size, 'i')
                source_attr.data.foreach_get('value', data)
                new_attr.data.foreach_set('value', data)
            elif source_attr.data_type == 'BOOLEAN':
                data = np.empty(data_size, '?')
                source_attr.data.foreach_get('value', data)
                new_attr.data.foreach_set('value', data)
            elif source_attr.data_type == 'FLOAT_VECTOR':
                data = np.empty(data_size*3,'f')
                source_attr.data.foreach_get('vector', data)
                new_attr.data.foreach_set('vector', data)
            elif source_attr.data_type == 'FLOAT_COLOR': # RGBA 32 bit
                data = np.empty(data_size*4,'f')
                source_attr.data.foreach_get('color', data)
                new_attr.data.foreach_set('color', data)
            elif source_attr.data_type == 'BYTE_COLOR': # 8 bit float
                data = np.empty(data_size*4, dtype=np.float16 ) #'B')
                source_attr.data.foreach_get('color', data)
                new_attr.data.foreach_set('color', data)
            # elif source_attr.data_type == 'STRING':
            #     new_attr.data.foreach_set('string', source_attr.data[:])
            elif source_attr.data_type == 'FLOAT2':
                data = np.empty(data_size*2,'f')
                source_attr.data.foreach_get('vector', data)
                new_attr.data.foreach_set('vector', data)
            elif source_attr.data_type == 'INT8':
                data = np.empty(data_size, 'b')
                source_attr.data.foreach_get('byte', data)
                new_attr.data.foreach_set('byte', data)

    @staticmethod
    def convert_to_mesh(context, depsgraph, sel_obj, merge_output=True, apply_modifiers=True):
        '''Works in groups too in recursive way. apply_modifiers=False => modes get ignored competly'''
        new_objs = []
        conveted_objs = {} # hold mesh of already converted instances, in cache
        old_seL_obj = None
        if sel_obj.type == 'CURVES': # transfer to mesh first...
            get_set_hair_setup_mod(sel_obj)
            sel_obj.select_set(False)
            old_name = sel_obj.name
            new_obj = HTOOL_OT_TransferHairSystem.transfer_hsys_to_target(context, sel_obj, move_method='EMPTY_MESH') # then run transfer to mesh,
            old_seL_obj = sel_obj
            new_obj.name = old_name
            sel_obj = new_obj
            sel_obj.select_set(True)
        depsgraph = context.evaluated_depsgraph_get()
        for dep_obj_info in depsgraph.object_instances:
            if not HTOOL_OT_CustomConvertMesh.is_object_instance_from_selected(dep_obj_info, sel_obj):
                continue # We only export selected objects

            orig_obj = HTOOL_OT_CustomConvertMesh.get_obj_for_instance(dep_obj_info, apply_modifiers)
            if not orig_obj.data or orig_obj.type!= 'MESH': # skip for non-geometry objects.
                continue
            # print(f"Exporting mesh with {len(mesh.vertices)} vertices "
            #       f"at {object_instance.matrix_world}")
            already_converted_me = conveted_objs.get(orig_obj.data.name)
            if already_converted_me:
                new_mesh = already_converted_me
            else:
                new_mesh = bpy.data.meshes.new_from_object(orig_obj)
                conveted_objs[orig_obj.data.name] = new_mesh

            new_obj=bpy.data.objects.new(name=orig_obj.name, object_data=new_mesh)
            new_obj.matrix_world = dep_obj_info.matrix_world
            new_objs.append(new_obj)
            # self.copy_mesh_attributes(dep_obj_info.object.data, new_mesh)
            #object_instance.to_mesh_clear()
        if old_seL_obj:
            try: # since transfer_hsys_to_target can remove old sel_obj (eg. it transfer from empty curve - it will be auto deleted), we need to check if it still exists
                bpy.data.objects.remove(old_seL_obj) # after modifiers are applied, remove old sel_obj
            except Exception as e:
                print(e)
        if merge_output:
            if len(new_objs) > 1:
                if new_objs[0].data.users > 1: # merge into first
                    new_objs[0].data = new_objs[0].data.copy() # make it sure target is single user
                with bpy.context.temp_override(object=new_objs[0], active_object=new_objs[0], selected_objects=new_objs, selected_editable_objects=new_objs, editable_objects=new_objs):
                    bpy.ops.object.join()
            # remove all modifiers from original sel_obj - since in above step mods are inherited
            for mod in sel_obj.modifiers:
                sel_obj.modifiers.remove(mod)
            if sel_obj.type == 'MESH':
                sel_obj.data = new_objs[0].data # override orig obj
                sel_obj.select_set(True)
                bpy.data.objects.remove(new_objs[0]) # remove merged dupli
            else:
                new_obj = bpy.data.objects.new(name=sel_obj.name, object_data=new_objs[0].data)
                context.collection.objects.link(new_obj)
                new_obj.select_set(True)
                # bpy.data.objects.remove(new_objs[0]) # remove merged dupli
                # return [sel_obj]
            return [sel_obj]

        else:   # no merge - then just delete sel_obj, and replace it with bunch of new objs
            for o in new_objs:
                link_obj_to_same_collections(sel_obj, o)
                o.select_set(True)
            context.view_layer.objects.active = new_objs[0]
            bpy.data.objects.remove(sel_obj)
            return new_objs


    def execute(self, context):
        depsgraph = context.evaluated_depsgraph_get()
        for obj in context.selected_objects:
            self.convert_to_mesh(context, depsgraph, obj, self.merge_output)
        return {'FINISHED'}


class HTOOL_OT_ConvertHairSystemToMesh(bpy.types.Operator):
    bl_idname = "hair_system.convert_to_mesh"
    bl_label = "Convert Hair To Mesh"
    bl_description = "Convert hair object to mesh. You will be able to restore original hair system with 'Restore Hair Object' operator"
    bl_options = {"REGISTER", "UNDO"}

    join_instances: bpy.props.BoolProperty(name="Join Instances", default=True, description="Join meshes generated with 'Instance On Strands' deformer. Helps to goe back from mesh back to Hair System", options={"HIDDEN"})

    # poll breaks finalize ...

    @staticmethod
    def write_col(obj, gn_col, name):
        col_np = np.empty((len(gn_col.data)*4), dtype=np.float32)
        gn_col.data.foreach_get('color', col_np)

        # col_4d = np.ones((len(gn_col.data), 4), dtype=np.float32)
        # col_4d[:, 0:3] = factor_array[:, None]
        mesh_vcol = obj.data.color_attributes.get(name) # was ..data.vertex_colors
        if not mesh_vcol:
            mesh_vcol = obj.data.attributes.new(name=name, type='FLOAT_COLOR', domain='POINT')
            # mesh_vcol = obj.data.vertex_colors.new(name='Factor')
        mesh_vcol.data.foreach_set("color", np.power(col_np, 2.2).ravel())


    @staticmethod
    def write_3d_attrib_to_2d_attrib(obj, src_attr_name, target_attr_name):
        '''More like convert vec3 to vec2'''
        # if src_attr_name == target_attr_name:
        tmp_target_attr_name = target_attr_name + '_TMP_HT' # to avoid collision eg. on 'SampledUV' -> 'SampledUV'

        target_attr = obj.data.attributes.get(tmp_target_attr_name)
        if not target_attr: # should be always true
            target_attr = obj.data.attributes.new(name=tmp_target_attr_name, type='FLOAT2', domain='CORNER')
            obj.data.attributes.update() # or else..
            target_attr = obj.data.attributes.get(tmp_target_attr_name)

        attrib_uv = obj.data.attributes.get(src_attr_name)
        if not attrib_uv:
            print('Cant Read UVMapGN attribute. Make sure shorth hair modifier UVs output is set to "UVMapGN"')
            return

        # write gn_uv to numpy array
        uv_array = np.empty((len(attrib_uv.data)*3), dtype=np.float32)
        attrib_uv.data.foreach_get('vector', uv_array)

        # remove every third element form numpy array
        uv_array = uv_array.reshape((len(attrib_uv.data), 3))
        uv_array = uv_array[:, :2] # only copy x,y  (first 2 elements)

        # write uvs using foreach_set
        target_attr.data.foreach_set('vector', uv_array.ravel())
        # for f in obj.data.polygons:
        #    for loop_idx in f.loop_indices:
        #        default_uv.data[loop_idx].uv = gn_uv.data[loop_idx].vector.xy
        attrib_uv.name = attrib_uv.name + '_OLD'
        obj.data.attributes.remove(attrib_uv)
        target_attr.name = target_attr_name


    @staticmethod
    def write_extra(obj, capture_node):
        # NOTE: vert colors attribs, are auto saved into colr
        # gn_col = obj.data.attributes.get('Factor')
        # if gn_col:
        #     self.write_col(obj, gn_col, 'Gradient')

        # capture_node = profile_mod.node_group.nodes.get(final_profile_capture_node_name)
        sample_input = capture_node.inputs['Sampled UV Name']
        if sample_input and sample_input.default_value: # did user specified UVs to sample...
            HTOOL_OT_ConvertHairSystemToMesh.write_3d_attrib_to_2d_attrib(obj,'SampledUV', 'SampledUV')  # 3D Vector, FCorner,  -> to uv_layer(2D, FCorner)


    @staticmethod
    def write_uv_attrib(obj):
        # XXX: create uvs before accessing UVMapGN or writing won't work
        # XXX: has to be second after uv_lays.get or write error.
        # attrib == uvs in B3.5
        HTOOL_OT_ConvertHairSystemToMesh.write_3d_attrib_to_2d_attrib(obj, 'UVMapGN', 'UVMap') # 3D Vec FCorner  - to uv_layer (2D, FCorner)
        HTOOL_OT_ConvertHairSystemToMesh.write_3d_attrib_to_2d_attrib(obj, 'UVHelperGN', 'HairTool_UV')  # 3D Vec FCorner  - to uv_layer (2D, FCorner)

    @staticmethod
    def del_objs(objs):
        for _ in range(len(objs)):
            bpy.data.objects.remove(objs.pop())

    def hair_system_apply(self, context, target_obj):
        profile_mod = get_profile_mod(target_obj)
        if not profile_mod:
            return
        capture_node = profile_mod.node_group.nodes.get(final_profile_capture_node_name)
        capture_extra = not capture_node.mute

        # check if we have Hair AO mode attached
        ao_attr = False
        for mod in target_obj.modifiers:
            if mod.type == 'NODES' and mod.node_group.name == 'HT_Mesh_AO':
                ao_attr = mod['Input_7']
                break

        # XXX: does not handles Instance on 'instanc eon strand......'
        # objs = [target_obj]
        # wait till https://projects.blender.org/blender/blender/issues/93903 geo_nodes instancing gets fixed
        # with context.temp_override(object=target_obj,
        #                                active_object=target_obj,
        #                                selected_objects=objs,
        #                                selected_editable_objects=objs,
        #                                editable_objects=objs):
        #     bpy.ops.object.duplicates_make_real() #  does not handles Instance on 'instanc eon strand......'

        depsgraph = context.evaluated_depsgraph_get()
        # do not join - hard to manage multiple objs with same uvs, eg UVMap...
        # use my convert_to_mesh - since it handles instancing on strands, etc.
        resulting_objs = HTOOL_OT_CustomConvertMesh.convert_to_mesh(context, depsgraph, target_obj, merge_output=False, apply_modifiers=True)
        # in blender 3.3 - vert color (Byte Color) ==> vertex color
        # in blender 3.5 - Face Corner (2D Float Vec) == uv_map


        converted_data = set() # when using 'Instance on Strand' keep track of already handled instances..
        for result_obj in resulting_objs:
            if result_obj.data.attributes.get('UVMapGN'): # must be hsystem obj, else normal mesh...
                if result_obj.data.name not in converted_data:
                    converted_data.add(result_obj.data.name)
                    result_obj.data.attributes.update()

                    # delete old UVMap if exist ? - automatically sampled from base obj, it seems.
                    old_uvs_map_attr = result_obj.data.attributes.get('UVMap') # it is stored in SampledUV anyway, and just gives problems
                    if old_uvs_map_attr:
                        result_obj.data.attributes.remove(old_uvs_map_attr)

                    self.write_uv_attrib(result_obj) # skip in 3.5?  Nope since name is different 'UVMapGN'=>'UVMap'
                    keep_attribs = ['UVMap','HairTool_UV','SystemColor'] # keep them since from B3.5  2D FCorner attribs are == UV layers
                    if ao_attr:
                        keep_attribs.append(ao_attr)

                    if capture_extra: # capture Sampled UVs
                        if capture_node.inputs['Sampled UV Name'].default_value: # user picked uv to sample
                            self.write_extra(result_obj, capture_node) # 'SampledUV' attrib, to 'SampledUV' uvs (not required in 3.5 and up -> uvs == 2d vec attrib)
                            keep_attribs.append('SampledUV')
                        if capture_node.inputs['Sampled Color Name'].default_value: # user picked uv to sample
                            keep_attribs.append('SampledColor')
                        keep_attribs.extend(['Random','Factor','Normal','Tangent'])
                    keep_attribs.append('material_index') # since 3.4.0 # required for - instance on mesh

                    # Remove ParentPos, ParentTan, ParentNorm, ParentData, ParentID attributes from result_obj
                    # for del_attrib in ['ParentPos', 'ParentTan', 'ParentNorm', 'ParentData', 'ParentID','id','UV_Box_ID','material_index']:
                    for attr in reversed(result_obj.data.attributes):
                        # temp_attrib = resutl_obj.data.attributes.get(attrib)
                        if attr.name not in keep_attribs:
                            try: # may error out in 3.5
                                result_obj.data.attributes.remove(attr)
                            except Exception as e:
                                print(e)

                result_obj.ht_props.hair_nodes.hair_systems.clear()
                result_obj.ht_props.is_curve = False
                result_obj.ht_props.hair_nodes.is_using_gnodes = False
            else: # non hsystem objects (eg. from instance on strands) - cleanup colliding attribs
                # keep_attribs = ['Random','Factor','Normal','Tangent']
                # if bpy.app.version >= (3, 4, 0):
                #     keep_attribs.append('material_index') # since 3.4.0 # required for - instance on mesh

                # Remove ParentPos, ParentTan, ParentNorm, ParentData, ParentID attributes from result_obj
                # for del_attrib in ['ParentPos', 'ParentTan', 'ParentNorm', 'ParentData', 'ParentID','id','UV_Box_ID','material_index']:
                for attr in reversed(result_obj.data.attributes):
                    # temp_attrib = resutl_obj.data.attributes.get(attrib)
                    if attr.name in ('SampledUV','Random','Factor','Normal','Tangent','SampledColor'):
                        try: # may error out in 3.5
                            result_obj.data.attributes.remove(attr)
                        except Exception as e:
                            print(e)

        return resulting_objs


    def execute(self, context):
        for obj in context.selected_objects:
            orig_obj_name = obj.name
            if not get_profile_mod(obj):
                self.report({'WARNING'}, f"Selected object: '{obj.name}' does not have profile, skipping")
                continue
            target_obj = obj.copy()
            target_obj.data = obj.data.copy()
            target_obj.name = orig_obj_name+'_hsys'
            link_obj_to_same_collections(obj, target_obj)
            target_obj.select_set(True)

            result_objs = self.hair_system_apply(context, target_obj)
            if not result_objs:
                self.report({'WARNING'}, f"{obj.name} could not be converted")
                continue

            if self.join_instances: # aka Join with option to go back
                unlink_from_scene(obj)
                obj.use_fake_user = True
                result_objs[0].ht_props.hair_nodes.hair_base_mesh = obj.name # set option to go back to orig obj
                result_objs[0].ht_props.hair_nodes.is_convert_result = True  # mark it for 'mesh_ops' being able to be shown - for converted hair system objs
                # XXX: make sure that active - is always the hsystem obj
                if len(result_objs)>1:
                    with bpy.context.temp_override(object=result_objs[0], active_object=result_objs[0], selected_objects=result_objs, selected_editable_objects=result_objs, editable_objects=result_objs):
                        bpy.ops.object.join()

            else:   # not join == no going back with hair_base_mesh use - for finalize only?
                out_coll = bpy.data.collections.get(orig_obj_name+'_HSys')
                if not out_coll:
                    out_coll = bpy.data.collections.new(orig_obj_name+'_HSys')
                    for c in obj.users_collection:
                        c.children.link(out_coll)
                for result_obj in result_objs:
                    unlink_obj_from_scene(context, result_obj)
                    out_coll.objects.link(result_obj)
                    result_obj.ht_props.hair_nodes.is_convert_result = True  # mark it
                bpy.data.objects.remove(obj)

            context.view_layer.objects.active = result_objs[0]
            result_objs[0].name = orig_obj_name+'_hsys'
            self.report({'INFO'}, f'{orig_obj_name} converted to mesh')

        return {"FINISHED"}


class HTOOL_OT_RestoreHairSystemBaseMesh(bpy.types.Operator):
    bl_idname = "hair_system.restore_base_mesh"
    bl_label = "Restore Hair Object"
    bl_description = "Restore original Hair object"

    @classmethod
    def poll(self, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.hair_nodes.hair_base_mesh

    def execute(self, context):
        hsys_out_mesh = context.active_object
        # bring back base mesh
        base_obj_name = hsys_out_mesh.ht_props.hair_nodes.hair_base_mesh
        base_mesh = bpy.data.objects.get(base_obj_name)
        if not base_mesh:
            return {"CANCELLED"}

        link_obj_to_same_collections(hsys_out_mesh, base_mesh)
        base_mesh.matrix_world = hsys_out_mesh.matrix_world
        base_mesh.use_fake_user = False
        base_mesh.select_set(True)

        context.view_layer.objects.active = base_mesh
        bpy.data.objects.remove(hsys_out_mesh)

        return {'FINISHED'}

# XXX: should I merge ConvertHair system to mesh and CurvesToMeshRibbons? Would work on curves and mesh obj... What bout instances on mesh? Test it...
# DONE: not used anymore almost
class HTOOL_OT_CurvesToMeshRibbons(bpy.types.Operator):
    bl_label = "Curve ribbons to mesh ribbons"
    bl_idname = "curves.curve_to_mesh_ribbon"
    bl_description = 'Convert curve ribbons to mesh hair cards. You can go back to curve mode using "Restore curve ribbons"'
    bl_options = {"REGISTER", "UNDO"}

    # refresh_material: bpy.props.BoolProperty(name='Update mesh material', description='Update mesh ribbon material, or skip this step (sometimes you may just want to update mesh ribbons geometry)', default=True)

    def invoke(self, context, event):
        sel_curves = [obj for obj in context.selected_objects if obj.type == 'CURVES' and get_profile_mod(obj) and not obj.ht_props.hair_nodes.is_using_gnodes]
        if not sel_curves:
            self.report({'INFO'}, 'Select at least one curve hair object, with profile first! Cancelling')
            return {"CANCELLED"}
        return self.execute(context)


    def transfer_vg(self, src_obj, target_data):
        ''' copy vertex group weights from old mesh'''
        if len(src_obj.vertex_groups) == 0: # does old hair_mesh has vgroups? If so transfer them
            return []
        back_vg_names = []
        for vg in src_obj.vertex_groups:
            back_vg_names.append(vg.name)
        bm = bmesh.new()
        bm.from_mesh(target_data)
        bm.faces.ensure_lookup_table()
        dlayer = bm.verts.layers.deform.verify()
        bm.verts.ensure_lookup_table()
        src_data = src_obj.data
        for v_idx in range(len(bm.verts)):
            if v_idx >= len(src_data.vertices):
                break
            for g in src_data.vertices[v_idx].groups:
                bm.verts[v_idx][dlayer][g.group] = g.weight
        if len(bm.verts) != len(src_data.vertices):
            self.report({'WARNING'}, '[Hair Tools] Vertex count was changed. Restored vertex weighs may be corrupted.')
        for v in bm.verts:
            g = v[dlayer]
            print(g.items())
        bm.to_mesh(target_data)
        bm.free()
        return back_vg_names

    def createMeshFromCurve(self, context, parentCurve, mesh_hair_child=None):
        '''Old Not used anymore- will not work on hair system on curve anyway (old curves with bevel worked? )'''
        depsgraph = context.evaluated_depsgraph_get()
        parentCurve_eval = parentCurve.evaluated_get(depsgraph)
        meshFromCurve = bpy.data.meshes.new_from_object(parentCurve_eval) # XXX: errors out on new Hair Curves (no geo data)

        if mesh_hair_child:
            prev_vg_names = self.transfer_vg(mesh_hair_child, meshFromCurve)
            mesh_hair_child.data = meshFromCurve
            for vg_name in prev_vg_names:  # NOTE: vgroups wont show unless we do this
                mesh_hair_child.vertex_groups.new(name=vg_name)

            child_mesh_obj = mesh_hair_child
        else:
            child_mesh_obj = bpy.data.objects.new(parentCurve.name + "_mesh", meshFromCurve)
        link_child_to_collection(parentCurve, child_mesh_obj)
        child_mesh_obj.select_set(True)

        # bpy.context.view_layer.objects.active = meshObjFromCurve
        child_mesh_obj.use_fake_user = False
        child_mesh_obj.ht_props.hair_settings.hair_mesh_parent = parentCurve.name
        parentCurve.ht_props.hair_settings.hair_curve_child = child_mesh_obj.name
        child_mesh_obj.matrix_world = parentCurve.matrix_world

        parentCurve.use_fake_user = True

        # if parentCurve.name == active_obj_name:
        context.view_layer.objects.active = child_mesh_obj

        # apply curve deform mod
        for mod in parentCurve.modifiers:
            if mod.type == 'CURVE':
                guide_curve = mod.object
                bpy.ops.object.modifier_apply(modifier=mod.name)
                unlink_from_scene(guide_curve)

        child_mesh_obj.hide_viewport = False
        shape_transfer(context, parentCurve, child_mesh_obj)
        unlink_from_scene(parentCurve)


    def execute(self, context):
        sel_curves = [obj for obj in context.selected_objects if obj.type == 'CURVES' and get_profile_mod(obj) and not obj.ht_props.hair_nodes.is_using_gnodes]
        if len(sel_curves) == 0:
            self.report({'WARNING'}, 'No hair curves selected. Cancelling')
            return {'CANCELLED'}
        # deselect all, except sel curves
        backup_sel_objs = context.selected_objects[:]
        for obj in context.selected_objects:
            if obj not in sel_curves:
                obj.select_set(False)

        # target_obj.select_set(True)

        with bpy.context.temp_override(object=sel_curves[0], active_object=sel_curves[0], selected_objects=sel_curves, selected_editable_objects=sel_curves, editable_objects=sel_curves):
            bpy.ops.object.duplicate()


        dupli_objs = [obj for obj in context.selected_objects if obj.type == 'CURVES' and get_profile_mod(obj) and not obj.ht_props.hair_nodes.is_using_gnodes] # they break convert to mesh
        for i,mesh_result_obj in enumerate(dupli_objs):
            profile_mod = get_profile_mod(mesh_result_obj)
            if not profile_mod:
                continue
            capture_node = profile_mod.node_group.nodes.get(final_profile_capture_node_name)
            capture_extra = not capture_node.mute

            profile_mat = profile_mod[prof_mat_input_name]

            with bpy.context.temp_override(object=mesh_result_obj, active_object=mesh_result_obj, selected_objects=[mesh_result_obj], selected_editable_objects=[mesh_result_obj], editable_objects=[mesh_result_obj]):
                # object.convert() - works on curves
                # bpy.ops.object.convert - looses InstancesOnStrands deformer on curves. Other that that, works ok on mesh and curves
                bpy.ops.object.convert(target='MESH') # seems it can't be overridden?  https://projects.blender.org/blender/blender/issues/93188
                # bpy.ops.object.simple_convert()  - works only on mesh type of obj (would require transfer hsys, which requires setup mod)


            if mesh_result_obj.data.attributes.get('UVMapGN'): # must be hsystem obj, else normal mesh...
                mesh_result_obj.data.attributes.update()

                # delete old UVMap if exist ? - automatically sampled from base obj, it seems.
                old_uvs_map_attr = mesh_result_obj.data.attributes.get('UVMap') # it is stored in SampledUV anyway, and just gives problems
                if old_uvs_map_attr:
                    mesh_result_obj.data.attributes.remove(old_uvs_map_attr)

                HTOOL_OT_ConvertHairSystemToMesh.write_uv_attrib(mesh_result_obj) # skip in 3.5?  Nope since name is different 'UVMapGN'=>'UVMap'
                keep_attribs = ['UVMap','HairTool_UV'] # keep them since from B3.5  2D FCorner attribs are == UV layers

                if capture_extra: # capture Sampled UVs
                    if capture_node.inputs['Sampled UV Name'].default_value: # user picked uv to sample
                        HTOOL_OT_ConvertHairSystemToMesh.write_extra(mesh_result_obj, capture_node) # 'SampledUV' attrib, to 'SampledUV' uvs (not required in 3.5 and up -> uvs == 2d vec attrib)
                        keep_attribs.append('SampledUV')
                    if capture_node.inputs['Sampled Color Name'].default_value: # user picked uv to sample
                        keep_attribs.append('SampledColor')
                    keep_attribs.extend(['Random','Factor','Normal','Tangent'])
                keep_attribs.append('material_index') # since 3.4.0 # required for - instance on mesh

                # Remove ParentPos, ParentTan, ParentNorm, ParentData, ParentID attributes from result_obj
                # for del_attrib in ['ParentPos', 'ParentTan', 'ParentNorm', 'ParentData', 'ParentID','id','UV_Box_ID','material_index']:
                for attr in reversed(mesh_result_obj.data.attributes):
                    # temp_attrib = resutl_obj.data.attributes.get(attrib)
                    if attr.name not in keep_attribs:
                        try: # may error out in 3.5
                            mesh_result_obj.data.attributes.remove(attr)
                        except Exception as e:
                            print(e)

                mesh_result_obj.ht_props.is_curve = False
                mesh_result_obj.ht_props.hair_nodes.is_using_gnodes = False
                mesh_result_obj.data.materials.clear() # remove all materials from mesh
                mesh_result_obj.data.materials.append(profile_mat) # add profile material to mesh


                orig_hair_curve = sel_curves[i]
                old_mesh_hair = orig_hair_curve.ht_props.hair_settings.hair_curve_child
                if old_mesh_hair:
                    old_mesh_hair = bpy.data.objects.get(old_mesh_hair)

                if old_mesh_hair: # we already have previous result of 'Curve to Mesh conversion'
                    old_mesh_hair.data = mesh_result_obj.data
                    bpy.data.objects.remove(mesh_result_obj)
                    link_obj_to_same_collections(orig_hair_curve, old_mesh_hair)
                    old_mesh_hair.use_fake_user = False
                    context.view_layer.objects.active = old_mesh_hair
                else:
                    mesh_result_obj.use_fake_user = False
                    mesh_result_obj.ht_props.hair_settings.hair_mesh_parent = orig_hair_curve.name
                    mesh_result_obj.name = orig_hair_curve.name + '_mesh'
                    context.view_layer.objects.active = mesh_result_obj
                    orig_hair_curve.ht_props.hair_settings.hair_curve_child = mesh_result_obj.name
                orig_hair_curve.use_fake_user = True
                unlink_from_scene(orig_hair_curve)
        # restore backup_sel_objs, if they are on scene
        context.view_layer.update() # update visible view layer objs state
        for obj in backup_sel_objs:
            if context.view_layer.objects.get(obj.name):
                obj.select_set(True)

        return {"FINISHED"}


        active_obj_name = context.active_object.name if context.active_object else ''
        # duplicate curve objects first, then move to empty mesh
        bpy.ops.object.duplicate()

        sel_curves = [obj for obj in context.selected_objects if obj.type == 'CURVES']

        # TODO: check if we may skip transfer to mesh - and just use blender convert  (if not hair deformers are used, that is)
        # What if we have Setup, and Profile , but not Hair System modes? Is it hair_system, or just treat as curve?
        for sel_curve in sel_curves: # set as active each curve
            # move_hsystem_items(self, context) # init run for transfer_to_object  allow use of dynamic enum 'move_method'
            HTOOL_OT_TransferHairSystem.transfer_hsys_to_target(context, sel_curve, move_method='EMPTY_MESH') # then run transfer to mesh,
        bpy.ops.hair_system.convert_to_mesh()

        return {"FINISHED"}


class HTOOL_OT_MeshRibbonsToCurve(bpy.types.Operator):
    bl_label = "Restore curve ribbons"
    bl_idname = "object.ribbon_convert"
    bl_description = "Restore original curve strands, from which mesh ribbons were created"
    bl_options = {"REGISTER", "UNDO"}

    @staticmethod
    def switch_mesh_to_curve_ribbon(context, selected_mesh_hair, set_active=True):
        if selected_mesh_hair.ht_props.hair_settings.hair_mesh_parent:  # switch to spline hair obj
            curve_hair = bpy.data.objects.get(selected_mesh_hair.ht_props.hair_settings.hair_mesh_parent)
            if curve_hair:
                if curve_hair.name not in context.scene.objects.keys():
                    link_child_to_collection(selected_mesh_hair, curve_hair)
                else:  # possibly already linked because mesh ribbon was duplicated, then duplicate source curve too
                    curve_hair = curve_hair.copy()  # copies dupli_group property(empty), but group property is empty (users_group = None)
                    selected_mesh_hair.ht_props.hair_settings.hair_mesh_parent = curve_hair.name
                    curve_hair.data = curve_hair.data.copy()
                    link_child_to_collection(selected_mesh_hair, curve_hair)
                curve_hair.matrix_world = selected_mesh_hair.matrix_world
                curve_hair.use_fake_user = False
                selected_mesh_hair.use_fake_user = True
                unlink_from_scene(selected_mesh_hair)  # hide mesh object
                if set_active:
                    curve_hair.select_set(True)
                    context.view_layer.objects.active = curve_hair
                curve_hair.ht_props.hair_settings.hair_curve_child = selected_mesh_hair.name  # set child in just in case curve name was changed

                # apply curve deform mod
                for mod in curve_hair.modifiers:
                    if mod.type == 'CURVE':
                        link_child_to_collection(curve_hair, mod.object)
            else:
                return False
        return True


    def execute(self, context):
        sel_curves = [obj for obj in context.selected_objects if obj.type == 'MESH']
        for selected_mesh_hair in sel_curves:
            sucess = self.switch_mesh_to_curve_ribbon(context, selected_mesh_hair)
            if not sucess:
                self.report({'ERROR'}, f"Cant switch to curve ribbons. Target curve {selected_mesh_hair.ht_props.hair_settings.hair_mesh_parent} is missing!")
        return {"FINISHED"}

