# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from os import path
from .utils.general_utils import link_obj_to_same_collections, import_node_group, unlink_from_scene, get_nested_node_groups, remap_node_groups, get_attributes, set_attributes, cache_old_mod_inputs, safe_rename
import numpy as np


shell_hair_node_group = 'ShellHairGN'
shell_hair_mod_name = 'Shell Hair'
# old_dir_path = path.dirname(__file__) # won't on win... - gives empty dir..
dir_path = path.dirname(path.abspath(__file__)) # will work on windows too?
shell_hair_BLEND_FILE = path.join(dir_path, "shell_hair_geo_nodes.blend")
hair_mat_name = 'ShellHairMaterial'
mod_material_input_name = 'Input_16'


def import_shell_hair_mat(mat_name, force_update=False):
    old_mat = bpy.data.materials.get(mat_name)
    if not old_mat or force_update:
        with bpy.data.libraries.load(shell_hair_BLEND_FILE) as (data_from, data_to):
            data_to.materials = [mat for mat in data_from.materials if mat == mat_name]
        if old_mat:
            old_mat.user_remap(data_to.materials[0])
            if old_mat.users == 0:
                bpy.data.materials.remove(old_mat)
                # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
                # we have to remove data block that holds the name. Then use freed name, on new data block...
                # XXX: what is mat_name is taken. Line below wont work
                data_to.materials[0].name = mat_name # rename later, so we can use new name
            else:
                safe_rename(bpy.data.materials, data_to.materials[0], mat_name)

        return data_to.materials[0]
    else:
        return old_mat

def get_node_group(node_name):
    node_group = bpy.data.node_groups.get(node_name)
    if not node_group:
        import_node_group(node_name, lib_file=shell_hair_BLEND_FILE)    # XXX: does it actually works (node.name vs node.node_tree.name)
        node_group = bpy.data.node_groups.get(node_name)
    return node_group


def get_shell_hair_mod(obj):
    guide_curve_mod = [m for m in obj.modifiers if m.type=='NODES' and m.node_group and m.node_group.name == shell_hair_node_group]
    return guide_curve_mod[0] if guide_curve_mod else None

def ensure_shell_hair_mat_setup_ok(obj):
    shell_hair_mod = get_shell_hair_mod(obj)
    shell_hair_mat = shell_hair_mod[mod_material_input_name]
    if not shell_hair_mat:
        if obj.data.materials and obj.data.materials[0]:
            shell_hair_mat = obj.data.materials[0]
        else:
            shell_hair_mat = bpy.data.materials.get(hair_mat_name)
            if not shell_hair_mat:
                shell_hair_mat = import_shell_hair_mat(hair_mat_name)
                shell_hair_mat.use_fake_user = True
        shell_hair_mod[mod_material_input_name] = shell_hair_mat
    # checkc if first obj material is shell_hair_mat. If not set it to shell_hair_mat
    if obj.data.materials and obj.data.materials[0] != shell_hair_mat:
        obj.data.materials[0] = shell_hair_mat
    elif not obj.data.materials:
        obj.data.materials.append(shell_hair_mat)
    return shell_hair_mat


def three_d_bincount_add(out ,e_k, weights):
    out[:,0] += np.bincount(e_k, weights[:,0], out.shape[0])
    out[:,1] += np.bincount(e_k, weights[:,1], out.shape[0])
    out[:,2] += np.bincount(e_k, weights[:,2], out.shape[0])

def new_bincount(out, e_k, weights = None):
    out += np.bincount(e_k, weights, out.shape[0])


class HTOOL_OT_SetupShellHair(bpy.types.Operator):
    bl_idname = "object.setup_shell_hair"
    bl_label = "Setup Shell Hair"
    bl_description = "Great for animal fur but not only\nTo comb generated hair use sculpting tool on generated GUIDE object (it is drawn as bounding box)"
    bl_options = {"REGISTER","UNDO"}

    face_offset: bpy.props.FloatProperty(name="Face Offset", default=0.01, min=0.0, max=1.0, description="Offset from face surface")
    swith_to_material_preview: bpy.props.BoolProperty(name="Switch to Material Preview", default=False, description="Switch to Material Preview")

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and not get_shell_hair_mod(obj)

    def execute(self, context):
        if bpy.app.version < (3, 1, 0):
            self.report({'ERROR'}, "This operation requires Blender 3.1 or higher")
            return {'CANCELLED'}
        obj = context.active_object
        self.base_obj = obj
        obj.select_set(False)

        hair_obj = obj.copy() #  short hair mod will remove base geo, so we need to copy it
        hair_obj.name = obj.name + '_shell_hair'
        hair_obj.data = hair_obj.data # .copy() # use instance this time
        link_obj_to_same_collections(obj, hair_obj)
        context.view_layer.objects.active = hair_obj
        hair_obj.select_set(True)

        guide_obj = hair_obj.copy()
        guide_obj.data = guide_obj.data.copy()  # NOTE: no other way right now to deform guide, withotu affecting base
        seg_guide_ng = bpy.data.node_groups.get(shell_hair_node_group)
        if not seg_guide_ng:
            seg_guide_ng = import_node_group(shell_hair_node_group, shell_hair_BLEND_FILE)
            seg_guide_ng.use_fake_user = True

        shell_hair_mod = hair_obj.modifiers.new(name=shell_hair_mod_name, type='NODES')
        shell_hair_mod.node_group = seg_guide_ng

        guide_obj.name = obj.name + '_hair_guide'
        guide_obj.display_type = 'BOUNDS' # 'WIRE'

        link_obj_to_same_collections(obj, guide_obj)
        shell_hair_mod['Input_6'] = guide_obj # Deformer
        shell_hair_mod['Input_15'] = 12 # 12 shells by default
        # shell_hair_mod['Output_35_attribute_name'] = 'Gradient' # Deformer
        # shell_hair_mod['Output_36_attribute_name'] = 'Density'
        # shell_hair_mod['Output_39_attribute_name'] = 'RandWidth'


        guide_obj.shape_key_add(name='Base')
        deform_key = guide_obj.shape_key_add(name='Deformer')
        guide_obj.active_shape_key_index = 1
        deform_key.value = 1

        # offset faces of deform_key by self.face_offset
        for vert in obj.data.vertices:
                offset = vert.normal * self.face_offset
                deform_key.data[vert.index].co = vert.co + offset

        shell_hair_mat = bpy.data.materials.get(hair_mat_name)
        if not shell_hair_mat:
            shell_hair_mat = import_shell_hair_mat(hair_mat_name)
            shell_hair_mat.use_fake_user = True
        shell_hair_mod[mod_material_input_name] = shell_hair_mat

        # hair_obj.ht_props.shell_hair_props.is_shell_hair = True

        context.view_layer.objects.active = guide_obj
        guide_obj.select_set(True)
        # switch viewport shading to material
        if self.swith_to_material_preview and context.space_data.shading.type != 'MATERIAL':
            context.space_data.shading.type = 'MATERIAL'
        return {'FINISHED'}


class HTOOL_PT_ShellHairPanel(bpy.types.Panel):
    bl_label = "Shell Hair"
    bl_idname = "HTOOL_PT_ShellHairPanel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'
    # bl_context = 'objectmode'
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and context.mode == 'OBJECT'

    def draw(self, context):
        layout = self.layout
        obj = context.active_object

        if not get_shell_hair_mod(obj):
            row = layout.row(align=True)
            row.operator("object.setup_shell_hair")
            row.operator("hair.update_shell_hair_node_group", icon="FILE_REFRESH", text='')
        else:
            shell_hair_mod = get_shell_hair_mod(obj)
            if not shell_hair_mod:
                return
            # guide_obj = shell_hair_mod['Input_6'] # Deformer
            col = layout.column(align=True)
            col.label(text="Modifier Properties:")
            # row = col.row(align=True)
            # row.prop_search(shell_hair_mod, '["Input_6"]', bpy.data, 'objects', text="Guide Object")
            col.prop(shell_hair_mod, '["Input_15"]', text="Layers Count")
            if shell_hair_mod["Input_15"]>2:
                col.prop(shell_hair_mod, '["Input_25"]', text="Bend Factor")
                col.prop(shell_hair_mod, '["Input_26"]', text="Bend Shape")

            def draw_attrib(layout,prop_name, text, kwargs={}):
                row = layout.row(align=True)
                # row.label(text=text)
                if shell_hair_mod[f'{prop_name}_use_attribute']: # eg use vg
                    # row.prop(shell_hair_mod, f'["{prop_name}_attribute_name"]', text=text, **kwargs)
                    row.prop_search(shell_hair_mod, f'["{prop_name}_attribute_name"]', obj, 'vertex_groups', text=text)
                else:
                    row.prop(shell_hair_mod, f'["{prop_name}"]', text=text, **kwargs)
                op = row.operator('object.geometry_nodes_input_attribute_toggle', text='', icon='SPREADSHEET')
                if bpy.app.version < (4, 0, 0):
                    op.prop_path=f'[\"{prop_name}_use_attribute\"]'
                else:
                    op.input_name=f'{prop_name}'

                op.modifier_name=shell_hair_mod.name
                return row

            col = layout.column(align=True)
            draw_attrib(col,'Input_9', 'Placement Mask')
            draw_attrib(col,'Input_40', 'Use Placement Fade', { 'toggle':True}) # 'icon':'CHECKBOX_HLT' if shell_hair_mod["Input_40"] else 'CHECKBOX_DEHLT',
            draw_attrib(col,'Input_42', 'Mesh Border Fade', { 'toggle':True}) # 'icon':'CHECKBOX_HLT' if shell_hair_mod["Input_42"] else 'CHECKBOX_DEHLT',

            # draw_attrib(col,'Input_37', 'Density')
            # draw_attrib(col,'Input_38', 'Random Width')

            col = layout.column(align=True)
            col.prop_search(shell_hair_mod, '["Input_16"]', bpy.data, 'materials', text="Material")
            col.separator()
            shell_mat = shell_hair_mod[mod_material_input_name]
            if shell_mat:
                shell_mat_ntree = shell_mat.node_tree
                shell_shader_node = shell_mat_ntree.nodes.get('Shell_Shader_Node')
                if shell_shader_node:
                    def draw_node_in_out(layout, node_name, in_out, idx,  text=None):
                        node = shell_mat_ntree.nodes.get(node_name)
                        if node:
                            node_attrib = getattr(node, in_out)
                            layout.prop(node_attrib[idx], 'default_value', text=text if text else node_name)

                    col = layout.column(align=False)
                    col.label(text="Material Properties:")

                    col = layout.column(align=False)
                    draw_node_in_out(col, 'Scale', 'outputs', 0, "Hair Scale")

                    col = layout.column(align=True)
                    row = col.row(align=True)
                    row.prop(shell_shader_node.inputs['Base Color'], 'default_value', text='Base Color')
                    color_attribute_node = shell_mat_ntree.nodes.get('Color Attribute')
                    if color_attribute_node:
                        col.prop_search(color_attribute_node, 'layer_name', obj.data, 'color_attributes', text="Vert Color")
                    col.prop(shell_shader_node.inputs['Mix Overlay Factor'], 'default_value', text='Mix VColor')
                    col.prop(shell_shader_node.inputs['AO Strength'], 'default_value', text='AO Strength')

                    col = layout.column(align=True)
                    col.prop(shell_shader_node.inputs['Hair Thickness'], 'default_value', text='Hair Thickness')
                    col.prop(shell_shader_node.inputs['Shell Contrast'], 'default_value', text='Shell Contrast')
                    col.prop(shell_shader_node.inputs['Density'], 'default_value', text='Density')
                    col.prop(shell_shader_node.inputs['Random Width'], 'default_value', text='Random Width')


class HTOOL_OT_ImportShellHairMaterial(bpy.types.Operator):
    bl_idname = "shell_hair.import_default_material"
    bl_label = "Import Shell Hair Material"
    bl_description = "Import Shell Hair material"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        obj = context.active_object
        return obj and obj.type == 'MESH' and get_shell_hair_mod(obj) and len(obj.data.materials)

    def execute(self, context):
        obj = context.active_object
        shell_hair_mod = get_shell_hair_mod(obj)
        shell_hair_mat = import_shell_hair_mat(hair_mat_name, force_update=True)
        shell_hair_mat.use_fake_user = True
        shell_hair_mod[mod_material_input_name] = shell_hair_mat
        if obj.data.materials and obj.data.materials[0] != shell_hair_mat:
            obj.data.materials[0] = shell_hair_mat
        elif not obj.data.materials:
            obj.data.materials.append(shell_hair_mat)
        return {"FINISHED"}



# operator of updating shell_hair_node_group
class HTOOL_OT_UpdateShellHairNodeGroup(bpy.types.Operator):
    bl_idname = "hair.update_shell_hair_node_group"
    bl_label = "Update Shell Hair Node Group"
    bl_description = "Update (or Reset) Shell Hair modifier node group,\nUsefull when you want to grab latest version of Short Hair modifier, after addon update"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        old_mats = {mat.name: mat for mat in bpy.data.materials}
        original_shell_mats = {}
        for obj in bpy.data.objects:
            if obj.type == 'MESH':
                shell_hair_mod = get_shell_hair_mod(obj)
                if shell_hair_mod: # get shell material
                    shell_hair_mat = shell_hair_mod[mod_material_input_name]
                    if shell_hair_mat:
                        original_shell_mats[shell_hair_mat.name] = shell_hair_mat

        original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}
        seg_guide_ng = import_node_group(shell_hair_node_group, shell_hair_BLEND_FILE, force_update=True)
        seg_guide_ng.use_fake_user = True

        # hack to detect new material.., since it may be  hair_mat_name + '.001' or something
        new_mat = None
        for m in bpy.data.materials:
            if m.name not in old_mats:
                new_mat = m
                break

        hair_mat_inner_node_groups = get_nested_node_groups(new_mat.node_tree)
        remap_node_groups(hair_mat_inner_node_groups, original_node_groups)

        # store mat inner nodes cache for old hair mods (like node pos, inputs states etc)
        hair_cache = cache_old_mod_inputs([mat.node_tree for mat in original_shell_mats.values()], attr_over = ['default_value'])

        # remap_node_groups([new_mat], original_shell_mats, data_type='materials') # old generic remap - without input values transfer
        for i, (orig_m_name, old_mat) in enumerate(original_shell_mats.items()): # old name: Twist
            print(f'remapping: {orig_m_name} to new: {new_mat.name}')
            (old_nodes_cache, old_nodes_input_values) = hair_cache[i]

            new_mat_cp = new_mat.copy()
            for old_node in old_mat.node_tree.nodes:
                new_node = new_mat_cp.node_tree.nodes.get(old_node.name)
                if new_node:
                    set_attributes(new_node, old_nodes_cache[old_node.name], ignored_attributes=('node_tree','parent')) # ignore old node_tree may have beeren changed => crash

                    # print(f'Updating {new_node.name} node inputs')
                    for i, old_input_cache in enumerate(old_nodes_input_values[old_node.name]):
                        if i < len(new_node.inputs):
                            set_attributes(new_node.inputs[i],old_input_cache)
                        else:
                            print(f'ERROR: {old_node.name} input {i} not found in {new_node.name} inputs')

            # special case for scale node (since we transfer only inputs above..)
            old_scale_node = old_mat.node_tree.nodes.get('Scale')
            new_scale_node = new_mat_cp.node_tree.nodes.get('Scale')
            if old_scale_node and new_scale_node:
                new_scale_node.outputs[0].default_value = old_scale_node.outputs[0].default_value
            # similar for 'Color Attribute' node
            old_color_attr_node = old_mat.node_tree.nodes.get('Color Attribute')
            new_color_attr_node = new_mat_cp.node_tree.nodes.get('Color Attribute')
            if old_color_attr_node and new_color_attr_node:
                #XXX:  B 4.1  - make sure there is no collision with old name... > add function in general_utils
                new_color_attr_node.layer_name = old_color_attr_node.layer_name

            old_mat.user_remap(new_mat_cp)
            # try:
            #     old_node_gr.user_remap(new_mat)
            # except:
            #     print(f'failed to remap: {orig_m_name} to new: {new_mat.name}')
            # else: # remap went ok

            # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
            # we have to remove data block that holds the name. Then use freed name, on new data block...
            bpy.data.materials.remove(old_mat)
            new_mat.name = orig_m_name


        # handle shell modifier now
            # old_hair_nodegroups = [ng for ng in bpy.data.node_groups if ng.name.startswith(hairsys_mod_group_name) and ng.name != hairsys_mod_group_name]
            #
            # # store nodes cache for old hair mods
            # hair_cache = cache_old_mod_inputs(old_hair_nodegroups, attr_over = ['default_value'])
        # XXX: one big hack to prevent crash (when mod.node_tree  output count changes on ng.users_remap() )
        # Reported: https://developer.blender.org/T95868
        # prof_obj_idx_backup = {} # from update profile node - Hair systemk
        # for obj in bpy.data.objects:
        #     shell_mod = get_shell_hair_mod(obj)
        #     if shell_mod:
        #         # hair_profile_mods_ids = [i for i,m in enumerate(obj.modifiers) if m.type=='NODES' and m.node_group and m.node_group.name.startswith(hair_profile_node_group_name)]
        #         # mod_idx = hair_profile_mods_ids[0]
        #         prof_obj_idx_backup[obj.name] = get_attributes(shell_mod)
        # XXX: hack end

        for obj in bpy.data.objects:
            if obj.type == 'MESH':
                shell_hair_mod = get_shell_hair_mod(obj)
                # obj.ht_props.shell_hair_props.is_shell_hair = True
                obj.modifiers.update()
                if shell_hair_mod:
                    shell_hair_mod.node_group = bpy.data.node_groups[shell_hair_node_group] # forces inputs refresh (eg. add new inputs if any)
                    # shell_hair_mod['Input_6'] = guide_obj # Deformer
                    # shell_hair_mod['Input_15'] = 12 # 12 shells by default
                    # shell_hair_mod['Output_35_attribute_name'] = 'Gradient' # Deformer
                    # shell_hair_mod['Output_36_attribute_name'] = 'Density'
                    # shell_hair_mod['Output_39_attribute_name'] = 'RandWidth'

        return {"FINISHED"}


# oferator wthat will apply modifiers, and write uv attribute to uv
class HTOOL_OT_ConvertShellHairToMesh(bpy.types.Operator):
    bl_idname = "hair.convert_shell_hair_to_mesh"
    bl_label = "Convert Shell Hair To Mesh"
    bl_description = "Apply Shell Hair modifier and write uvs"
    bl_options = {"REGISTER", "UNDO"}

    # @classmethod
    # def poll(self, context):  # make selection only false, if obj mode
    #     obj = context.active_object
    #     return obj and obj.type == 'MESH'

      # free and prevent further access

    @staticmethod
    def write_uv(obj, attrib_name, target_uv_name):
        # XXX: create uvs before accessing UVMapGN or writing wont work
        uv_layer_data = obj.data.uv_layers.get(target_uv_name)
        if not uv_layer_data:
            uv_layer_data = obj.data.uv_layers.new(name=target_uv_name)

        attrib_uv = obj.data.attributes.get(attrib_name)
        if not attrib_uv:
            print(f'Cant Read {attrib_name} attribute. Skipping generation of uvs for it')
            return


        # write gn_uv to numpy array
        uv_array = np.empty((len(attrib_uv.data)*2), dtype=np.float32)
        attrib_uv.data.foreach_get('vector', uv_array)

        # remove every third element form numpy array
        uv_array = uv_array.reshape((len(attrib_uv.data), 2))
        uv_array = uv_array[:, :2] # only copy x,y  (first 2 elements)

        # write uvs using foreach_set
        uv_layer_data.data.foreach_set('uv', uv_array.ravel())
        # for f in obj.data.polygons:
        #    for loop_idx in f.loop_indices:
        #        default_uv.data[loop_idx].uv = gn_uv.data[loop_idx].vector.xy
        attrib_uv.name = attrib_uv.name + '_OLD'
        obj.data.attributes.remove(attrib_uv)

    def write_uv_attrib(self, obj):
        # XXX: create uvs before accessing UVMapGN or writing wont work
        # default_uv = obj.data.uv_layers.get('UVMap')
        # if not default_uv:
        #     default_uv = obj.data.uv_layers.new()
        #
        # # XXX: has to be second after uv_lays.get or write error.
        # gn_uv = obj.data.attributes.get('UVMapGN')
        # if not gn_uv:
        #     print('Cant Read UVMapGN attribute. Make sure shorth hair modifier UVs output is set to "UVMapGN"')
        #     return
        if bpy.app.version < (3, 5, 0): # XXX: from B 3.5 uvs are stored as 2d attrib, so no need to modify,if same name (if they are 2d vec face corner)
            self.write_uv(obj, 'UVMap', 'UVMap')


    def shell_hair_mod_apply(self, context, obj):
        shell_hair_mod = obj.modifiers.get(shell_hair_mod_name)
        if not shell_hair_mod:
            return
        target_obj = obj.copy()
        target_obj.data = obj.data.copy()
        target_obj.name = obj.name+'_shell_hair'
        link_obj_to_same_collections(obj, target_obj)
        target_obj.ht_props.shell_hair_props.is_shell_hair = False
        # target_obj.ht_props.shell_hair_props.is_shell_hair_output = True

        unlink_from_scene(obj)
        obj.use_fake_user = True

        context.view_layer.objects.active = target_obj
        target_obj.select_set(True)
        bpy.ops.object.modifier_apply(modifier=shell_hair_mod_name)

        #NOTE:  GNodes mod creates additional mat slot. Remove first one (empty)
        # bpy.ops.object.material_slot_remove() # XXX:  was it even needed?
        self.write_uv_attrib(target_obj)
        target_obj.ht_props.shell_hair_props.hair_base_mesh = obj.name

        guide_obj = shell_hair_mod['Input_6'] # Deformer
        if guide_obj:
            guide_obj.use_fake_user = True
            unlink_from_scene(guide_obj)


    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH' and get_shell_hair_mod(obj):
                shell_hair_mod = obj.modifiers.get(shell_hair_mod_name)
                if not shell_hair_mod:
                    self.report({'ERROR'}, f'{shell_hair_mod_name} modifier not found. Skipping {obj.name}')
                    continue
                self.shell_hair_mod_apply(context, obj)

            self.report({'INFO'}, f'{obj.name} converted to mesh')

        return {"FINISHED"}


class HTOOL_OT_RestoreShellHairBaseMesh(bpy.types.Operator):
    bl_idname = "object.restore_shell_hair_base_mesh"
    bl_label = "Restore Shell Hair Base Mesh"
    bl_description = "Restore Shell Hair base mesh"

    @classmethod
    def poll(self, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.shell_hair_props.hair_base_mesh

    def execute(self, context):
        shell_hair_converted = context.active_object
        # bring back base mesh
        base_obj_name = shell_hair_converted.ht_props.shell_hair_props.hair_base_mesh
        base_mesh = bpy.data.objects.get(base_obj_name)
        if not base_mesh:
            return {"CANCELLED"}

        link_obj_to_same_collections(shell_hair_converted, base_mesh)
        base_mesh.matrix_world = shell_hair_converted.matrix_world
        base_mesh.use_fake_user = False
        base_mesh.select_set(True)

        context.view_layer.objects.active = base_mesh

        shell_hair_mod = base_mesh.modifiers.get(shell_hair_mod_name)
        if shell_hair_mod:
            guide_obj = shell_hair_mod['Input_6'] # Deformer
            if guide_obj:
                guide_obj.use_fake_user = True

                link_obj_to_same_collections(shell_hair_converted, guide_obj)
                guide_obj.matrix_world = shell_hair_converted.matrix_world
                guide_obj.use_fake_user = False
                guide_obj.select_set(True)

        bpy.data.objects.remove(shell_hair_converted)

        return {'FINISHED'}

class HTOOL_OT_BakeShellTexture(bpy.types.Operator):
    bl_idname = "shell_hair.bake_texture"
    bl_label = "Bake Texture"
    bl_description = "Bake Shell Hair texture"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bake_settings = context.scene.ht_props.hair_bake_settings
        if not bake_settings.hair_bake_path:
            self.report({'ERROR'}, f'Set baking path first. Cancelling')
            return {'CANCELLED'}

        render = context.scene.render

        # proper settings for texture 'bake'
        # render.image_settings.color_management = 'OVERRIDE' # or 'FOLLOW_SCENE'
        # context.scene.view_settings.view_transform = 'Raw'
        # bpy.context.scene.view_settings.look = 'None'

        render.image_settings.file_format = bake_settings.output_format
        render_path = bake_settings.hair_bake_path
        render_file_name = bake_settings.hair_bake_file_name
        # render_quality = bake_settings.render_quality
        render.resolution_x = int(bake_settings.bakeResolution_x)
        render.resolution_y = int(bake_settings.bakeResolution_x)
        render.resolution_percentage = 100
        render.image_settings.file_format = bake_settings.output_format
        if bake_settings.output_format == 'JPEG':  # always bake with alpha, for shader padding, for np color fill
            render.image_settings.color_mode = 'RGB'
        else:
            render.image_settings.color_mode = 'RGBA'

        render.use_file_extension = True
        render.image_settings.color_depth = '8'
        context.scene.cycles.samples = 64
        file_name = render_file_name
        render.filepath = render_path + file_name
        bpy.ops.render.render(write_still=True)  # we cant acess render reuslt
        full_file_name = file_name + '.' + render.image_settings.file_format.lower()

        if full_file_name in bpy.data.images.keys():
            bpy.data.images[full_file_name].reload()
            bpy.data.images[full_file_name].update()

        return {"FINISHED"}


class HTOOL_OT_OpenShellFile(bpy.types.Operator):
    bl_idname = "shell_hair.open_texture_scene"
    bl_label = "Open Shell Texture Scene"
    bl_description = "It will allow you to bake new tiling shell texture"
    bl_options = {"REGISTER"}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        self.layout.label(text='Are you sure you want to open Hair Baking scene?')
        self.layout.label(text='Current file will be closed without saving')

    def execute(self, context):
        bpy.app.handlers.load_post.append(load_shell_scene)
        bpy.ops.wm.read_homefile(app_template="")
        # bpy.ops.wm.open_mainfile(filepath=path.dirname(__file__)+"/hair_bake.blend") # opens baking scene from HTool. But we do not want to override it
        return {"FINISHED"}


from bpy.app.handlers import persistent

def get_shell_texture_file_path():
    return path.dirname(__file__)+"\shell_hair_texture_bake.blend"

@persistent
def load_shell_scene(dummy):
    old_scene = bpy.context.scene
    tmp = bpy.data.scenes.new('tmp')
    bpy.context.window.scene = tmp
    bpy.data.scenes.remove(old_scene, do_unlink=True)
    bpy.ops.outliner.orphans_purge(do_local_ids=True, do_linked_ids=True, do_recursive=True)

    #* looks like bleder automatically imports linked data
    print('Loading hair baking scene')
    try:
        with bpy.data.libraries.load(get_shell_texture_file_path()) as (data_from, data_to):  # __file__ path from were program was executed
            data_to.scenes = ["ShellTextureScn"]
            data_to.workspaces = ["Shading"]
            data_to.materials = data_from.materials # all mats

        # current_scene.user_remap(data_to.scenes['BakingHair'])
        bpy.context.window.scene = data_to.scenes[0]
        bpy.data.scenes.remove(tmp)
        bpy.context.window.workspace = data_to.workspaces[0]
        # manually hide Debug collection. Even though 'Old Method' collection is hidden correctly
        # bpy.context.view_layer.layer_collection.children['HT_Helpers'].children['Debug'].hide_viewport = True
        # bpy.ops.object.mode_set(mode='OBJECT')
        try:
            bpy.data.libraries.remove(bpy.data.libraries['shell_hair_texture_bake.blend'])
        except Exception as e:
            print(e)
        print('Loaded ok')
        if load_shell_scene in bpy.app.handlers.load_post:
            bpy.app.handlers.load_post.remove(load_shell_scene)
    except Exception as e:
        print(e)
        print('Failed to import baking scene')
        if load_shell_scene in bpy.app.handlers.load_post:
            bpy.app.handlers.load_post.remove(load_shell_scene)

