# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
from os import path
import bmesh
from .utils.general_utils import link_obj_to_same_collections, import_node_group, unlink_from_scene, get_nested_node_groups, remap_node_groups, safe_rename
import random
from .utils.helper_functions import gold_random
import numpy as np

# DONE: add convert to meshs (with uv bake)
# DONE: drav UV box - make corner dragging better
# DONE: add different modes of mesh triangulation for hair mesh  - make steps interactive...
# DONE: re-import/refresh short_hair node gr
# DONE: convert to mesh and back
# TODO: UV width to rad -finish

short_hair_node_group = 'ShortHairGN'
short_hair_mod_name = 'Short Hair'
SHORT_HAIR_BLEND_FILE = path.dirname(__file__)+"\short_hair_geo_nodes.blend"
hair_mat_name = 'ShortHairMaterial'


sampled_obj_input_name = 'Input_33'
sampled_color_input_name = 'Input_13'
sampled_uv_input_name = 'Input_34'


def import_short_hair_mat(mat_name, force_update=False):
    old_mat = bpy.data.materials.get(mat_name)
    if not old_mat or force_update:
        with bpy.data.libraries.load(SHORT_HAIR_BLEND_FILE) as (data_from, data_to):
            data_to.materials = [mat for mat in data_from.materials if mat == mat_name]

        if old_mat:
            old_mat.user_remap(data_to.materials[0])
            if old_mat.users == 0:
                # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
                # we have to remove data block that holds the name. Then use that name, on new data block...
                bpy.data.materials.remove(old_mat)
                data_to.materials[0].name = mat_name
            else:
                safe_rename(bpy.data.materials, data_to.materials[0], mat_name)

        return data_to.materials[0]
    else:
        return old_mat

def get_node_group(node_name):
    node_group = bpy.data.node_groups.get(node_name)
    if not node_group:
        import_node_group(node_name, lib_file=SHORT_HAIR_BLEND_FILE)    # XXX: does it actually works (node.name vs node.node_tree.name)
        node_group = bpy.data.node_groups.get(node_name)
    return node_group


def get_short_hair_mod(obj):
    guide_curve_mod = [m for m in obj.modifiers if m.type=='NODES' and m.node_group.name == short_hair_node_group]
    return guide_curve_mod[0] if guide_curve_mod else None

def ensure_short_hair_mat_setup_ok(obj):
    short_hair_mod = get_short_hair_mod(obj)
    short_hair_mat = short_hair_mod['Input_16']
    if not short_hair_mat:
        if obj.data.materials and obj.data.materials[0]:
            short_hair_mat = obj.data.materials[0]
        else:
            short_hair_mat = bpy.data.materials.get(hair_mat_name)
            if not short_hair_mat:
                short_hair_mat = import_short_hair_mat(hair_mat_name)
                short_hair_mat.use_fake_user = True
        short_hair_mod['Input_16'] = short_hair_mat
    # checkc if first obj material is short_hair_mat. If not set it to short_hair_mat
    if obj.data.materials and obj.data.materials[0] != short_hair_mat:
        obj.data.materials[0] = short_hair_mat
    elif not obj.data.materials:
        obj.data.materials.append(short_hair_mat)
    return short_hair_mat


def three_d_bincount_add(out ,e_k, weights):
    out[:,0] += np.bincount(e_k, weights[:,0], out.shape[0])
    out[:,1] += np.bincount(e_k, weights[:,1], out.shape[0])
    out[:,2] += np.bincount(e_k, weights[:,2], out.shape[0])

def new_bincount(out, e_k, weights = None):
    out += np.bincount(e_k, weights, out.shape[0])


class HTOOL_OT_SetupShortHair(bpy.types.Operator):
    bl_idname = "object.setup_short_hair"
    bl_label = "Setup Short Hair"
    bl_description = "Great for animal fur but not only\nTo comb generated hair use sculpting tool on generated GUIDE object (it is drawn as bounding box)"
    bl_options = {"REGISTER","UNDO"}

    smoothing: bpy.props.IntProperty(name="Initial Mesh Smoothing", description='Smooth hair base geometry - helps to even out generated hair distribution and normals', default= 1, min=0, max=6)
    apply_modifiers: bpy.props.BoolProperty(name="Apply Modifiers", description='Apply existing modifiers when generating hair base mesh', default= True)
    topology: bpy.props.EnumProperty(name='Remeshing Mode', description='Topology of generated base mesh has big influence on hair look and density',
        items=[
            ('KEEP', 'Use original', 'Keep mesh topology unchanged'),
            ('CELLS_TRIS', 'Triangulated Cells', 'Build mesth topology from Triangulated Veronoi cells. Gives nicest, even triangulation, but quite high-poly'),
            ('DIAMOND', 'Diamonds', 'Build topology from Diamond cells, lowpoly but not as even as Cells topology'),
            ('DIAMOND_TRIS', 'Triangulated Diamonds', 'Same As Diamond, gives slightly more tris, but with more even haircards density'),
        ], default='CELLS_TRIS')
    boundary: bpy.props.EnumProperty(name='Boundary', description='Pick smooth or shapr of boundary edges',
        items=[
            ('ALL', 'Smooth', 'Smooth boundary edges, when subdividing geometry'),
            ('PRESERVE_CORNERS', 'Keep Sharp', 'Keep share boundary edges, when subdividing geometry'),
        ], default='PRESERVE_CORNERS')


    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'apply_modifiers', icon='MODIFIER')
        layout.prop(self, 'topology')
        if self.topology != 'KEEP':
            layout.prop(self, 'smoothing')
            if self.topology == 'CELLS_TRIS':
                if bpy.app.version >= (3, 2, 0):
                    layout.prop(self, 'boundary')

    def numpy_smooth(self,bm, smooth=0.4, iters=5):
        bm.verts.ensure_lookup_table()
        bm.edges.ensure_lookup_table()

        o_normal = np.array([vert.normal for vert in bm.verts])
        o_co =  np.array([vert.co for vert in bm.verts])
        vert_count =  len(bm.verts)

        edge_keys = np.array([[edge.verts[0].index, edge.verts[1].index] for edge in  bm.edges])
        freeze_verts = np.array([vert.is_boundary or not vert.is_manifold for vert in bm.verts]) #manifold for pocket sewing edges

        p_co = np.copy(o_co) #current new vertex position
        b_co_sum_neighbors = np.copy(o_co) #current new vertex position

        #sum edges that have vert X  and either end of edge (Va, Vb)
        v_connections = np.zeros(vert_count, dtype=np.float32)
        new_bincount(v_connections, edge_keys[:,1] )
        new_bincount(v_connections, edge_keys[:,0])

        alpha = 0.5   # Iriginal Vert co influence
        beta = 1 - smooth # 1 - smooth strength
        for _ in range(iters):
            q_co = np.copy(p_co) #prevoius iteration vert position
            p_co.fill(0)
            b_co_sum_neighbors.fill(0)
            #calc average position weighted...
            three_d_bincount_add(p_co, edge_keys[:,1], q_co[edge_keys[:,0]])
            three_d_bincount_add(p_co, edge_keys[:,0], q_co[edge_keys[:,1]])

            p_co = p_co / v_connections[:,None]   #new aver posision
            b_co = p_co - (alpha*o_co + (1-alpha) * q_co)  # difference

            #average of neibors vert differences
            three_d_bincount_add(b_co_sum_neighbors, edge_keys[:, 1], b_co[edge_keys[:, 0]])
            three_d_bincount_add(b_co_sum_neighbors, edge_keys[:, 0], b_co[edge_keys[:, 1]])

            p_co[freeze_verts] = q_co[freeze_verts] # not selected ver - reset to original pos o_co

            #project p_normal* p_normal.dot(b_sum)  # u(u dot v)  - projection of v on u
            #TO prevent border vert slide ahead of loop (caused by sum_b being to big.., so reduce it by projecting on vert normal)
            #projected b_sum on normal  - helps also stabilize noise...
            b_sum_projected_to_normal = np.einsum('ij,i->ij', o_normal, np.einsum('ij,ij->i', o_normal, b_co_sum_neighbors/v_connections[:, None]))
            p_co = p_co - (beta*b_co + (1-beta) * b_sum_projected_to_normal)
            p_co[freeze_verts] = q_co[freeze_verts]  # not selected ver - reset to original pos o_co

        for v in bm.verts:
            v.co = p_co[v.index]

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        is_short_hair = False
        if obj and obj.type == 'MESH':
            # find short hair mods on object
            for mod in obj.modifiers:
                if mod.type == 'NODES' and mod.node_group.name == short_hair_node_group:
                    is_short_hair = True
                    break
        return not is_short_hair
        # return obj and obj.type == 'MESH' and not obj.ht_props.short_hair_props.is_short_hair


    def veronoi_cells_topo(self, bm):
        sel_verts = [v for v in bm.verts if len(v.link_edges) != 3]
        bm.select_mode = {'VERT', 'EDGE'}
        for v in bm.verts:
            v.select = False
        for v in sel_verts:
            v.select = True
        bm.select_flush_mode()
        bm.select_mode = {'EDGE'}
        del_edges = [e for e in bm.edges if e.select]
        bmesh.ops.dissolve_edges(bm, edges=del_edges, use_verts=True, use_face_split=True)
        bmesh.ops.poke(bm, faces=bm.faces, offset=0.1, center_mode='MEAN_WEIGHTED', use_relative_offset=True)
        # self.numpy_smooth(bm, smooth=0.5, iters=3, only_selected=False) # because smooth in subdivide_edges is not worksing...


    def diamond_topo(self, bm):
        bmesh.ops.poke(bm, faces=bm.faces, offset=0.1, center_mode='MEAN_WEIGHTED', use_relative_offset=True)
        # we will  dissolve edges with verts that have more than 4 edges
        sel_verts = [v for v in bm.verts if len(v.link_edges) > 4]
        bm.select_mode = {'VERT', 'EDGE'}
        for v in bm.verts:
            v.select = False
        for v in sel_verts:
            v.select = True
        bm.select_flush_mode()
        bm.select_mode = {'EDGE'}
        del_edges = [e for e in bm.edges if e.select]
        bmesh.ops.dissolve_edges(bm, edges=del_edges, use_verts=False, use_face_split=True)
        # remaining_sel_verts = [v for v in bm.verts if v.select]
        self.numpy_smooth(bm)
        # bmesh.ops.smooth_vert(bm, verts=remaining_sel_verts, factor=.8, mirror_clip_x=False, mirror_clip_y=False, mirror_clip_z=False, clip_dist=0.0, use_axis_x=True, use_axis_y=True, use_axis_z=True)
        # bmesh.ops.smooth_laplacian_vert(bm, verts=remaining_sel_verts, lambda_factor=3, lambda_border=1.5, use_x=True, use_y=False, use_z=False, preserve_volume=True)

    def vert_crease(self, obj):
        bm = bmesh.new()
        bm.from_mesh(obj.data)
        if bpy.app.version < (4, 0, 0):
            crease_lay = bm.verts.layers.crease.active # note: new in 3.2
            if not crease_lay:
                crease_lay = bm.verts.layers.crease.new()
        else: # new for blender 4
            crease_lay = bm.verts.layers.float.get('crease_vert', None)
            if not crease_lay:
                crease_lay = bm.verts.layers.float.new('crease_vert')
                # crease_edge_attr = obj.data.attributes.new('crease_vert', 'FLOAT','POINT')

        for _,v in enumerate(bm.verts):
            if v.is_boundary:
                v[crease_lay] = 1

        bm.to_mesh(obj.data)
        bm.free()



    def rebuild_topo(self, obj):
        if self.topology == 'CELLS_TRIS': #we do this manually before bm, since there is no working bmesh subdivide smoothing
            obj.modifiers.new('TriangulateSH', 'TRIANGULATE')
            sub_surf_mod = obj.modifiers.new('SubShortHair', type='SUBSURF')
            # select all boundary vertices

            # NOTE: we can access, write to bmesh vert[crease_lay].val so this hack
            # crease_data = obj.data.vertex_creases[0]
            # TODO: use new bmesh vert crease api
            # for i in range(len(obj.data.vertices)):
            #     crease_data.data[i].value # get crease data

            if self.boundary == "PRESERVE_CORNERS": # because of triangulation some corners will be smooth. So we have to
                self.vert_crease(obj)

            sub_surf_mod.boundary_smooth = self.boundary

            sub_surf_mod.use_limit_surface = False
            sub_surf_mod.use_creases = True
            bpy.ops.object.modifier_apply(modifier='TriangulateSH')
            bpy.ops.object.modifier_apply(modifier='SubShortHair')

        me = obj.data
        bm = bmesh.new()   # create an empty BMesh
        bm.from_mesh(me)   # fill it in from a Mesh

        if self.topology == 'DIAMOND':
            self.diamond_topo(bm)
        elif self.topology == 'DIAMOND_TRIS':
            self.diamond_topo(bm)
            bmesh.ops.triangulate(bm, faces=bm.faces, quad_method='SHORT_EDGE', ngon_method='BEAUTY')
        elif self.topology == 'CELLS_TRIS':
            self.veronoi_cells_topo(bm)
            # bmesh.ops.triangulate(bm, faces=bm.faces, use_beauty=True)
        bm.to_mesh(me)
        bm.free()

        # additional post process since Cells uses Subsurf and shrinkgs geo
        #  (project on base, and smooth)
        if self.topology == 'CELLS_TRIS':
            project_mod = obj.modifiers.new('ProjectToBase', type='SHRINKWRAP')
            project_mod.target = self.base_obj
            project_mod.wrap_method = 'PROJECT'
            project_mod.wrap_mode = 'OUTSIDE'
            project_mod.subsurf_levels = self.boundary != "PRESERVE_CORNERS" # smooth if nnot preserve corners
            project_mod.use_negative_direction = False
            bpy.ops.object.modifier_apply(modifier='ProjectToBase')
            me = obj.data
            bm = bmesh.new()   # create an empty BMesh
            bm.from_mesh(me)
            # self.numpy_smooth(bm, iters=1, smooth=0.4)
            bm.to_mesh(me)
            bm.free()


    def execute(self, context):
        if bpy.app.version < (3, 1, 0):
            self.report({'ERROR'}, "This operation requires Blender 3.1 or higher")
            return {'CANCELLED'}
        obj = context.active_object
        self.base_obj = obj
        obj.select_set(False)

        hair_obj = obj.copy() #  short hair mod will remove base geo, so we need to copy it
        hair_obj.name = obj.name + '_short_hair'
        hair_obj.data = hair_obj.data.copy()

        if self.smoothing and self.topology != 'KEEP':
            me = hair_obj.data
            bm = bmesh.new()   # create an empty BMesh
            bm.from_mesh(me)   # fill it in from a Mesh
            self.numpy_smooth(bm, iters=self.smoothing)
            bm.to_mesh(me)
            bm.free()  # free and prevent further access

        link_obj_to_same_collections(obj, hair_obj)
        context.view_layer.objects.active = hair_obj
        hair_obj.select_set(True)

        if self.topology != 'KEEP':
            if self.apply_modifiers:
                bpy.ops.object.convert(target="MESH")
            self.rebuild_topo(hair_obj)

        # clear old materials
        for _ in range(len(hair_obj.data.materials)):
            mat = hair_obj.data.materials[-1]
            hair_obj.data.materials.pop() #new
            if mat and mat.users == 0:
                bpy.data.materials.remove(mat)

        guide_obj = hair_obj.copy()
        guide_obj.data = guide_obj.data.copy()
        seg_guide_ng = bpy.data.node_groups.get(short_hair_node_group)
        if not seg_guide_ng:
            seg_guide_ng = import_node_group(short_hair_node_group, SHORT_HAIR_BLEND_FILE)
            seg_guide_ng.use_fake_user = True

        short_hair_mod = hair_obj.modifiers.new(name=short_hair_mod_name, type='NODES')
        short_hair_mod.node_group = seg_guide_ng

        guide_obj.name = obj.name + '_hair_guide'
        guide_obj.display_type = 'BOUNDS' # 'WIRE'

        link_obj_to_same_collections(obj, guide_obj)
        short_hair_mod['Input_6'] = guide_obj # Deformer
        short_hair_mod['Output_10_attribute_name'] = 'UVMap' # Deformer
        short_hair_mod['Output_27_attribute_name'] = 'UV_Base' # Deformer
        short_hair_mod[sampled_obj_input_name] = obj # sampled obj
        # if obj has uvs then set Input_35
        if obj.data.uv_layers.active:
            short_hair_mod[sampled_uv_input_name] = obj.data.uv_layers.active.name
        # if obj has vertex colors then set Input_36
        if obj.data.color_attributes.active:
            short_hair_mod[sampled_color_input_name] = obj.data.color_attributes.active.name

        guide_obj.shape_key_add(name='Base')
        deform_key = guide_obj.shape_key_add(name='Deformer')
        guide_obj.active_shape_key_index = 1
        deform_key.value = 1

        # short_hair_mod['Input_7'] = 0.1  # Initial Length

        short_hair_mat = bpy.data.materials.get(hair_mat_name)
        if not short_hair_mat:
            short_hair_mat = import_short_hair_mat(hair_mat_name)
            short_hair_mat.use_fake_user = True
        hair_obj.data.materials.append(short_hair_mat)
        hair_obj.ht_props.short_hair_props.is_short_hair = True
        # XXX:
        # obj.material_slots[0].link = 'OBJECT'
        # guide_obj.material_slots[0].link = 'OBJECT'

        if bpy.app.version < (4, 0, 0):
            bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_11_use_attribute\"]", modifier_name="Short Hair")
            bpy.ops.object.geometry_nodes_input_attribute_toggle(prop_path="[\"Input_12_use_attribute\"]", modifier_name="Short Hair")
        else:
            bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_11", modifier_name="Short Hair")
            bpy.ops.object.geometry_nodes_input_attribute_toggle(input_name="Input_12", modifier_name="Short Hair")

        short_hair_mod['Input_11_attribute_name'] = 'UV_Start'
        short_hair_mod['Input_12_attribute_name'] = 'UV_End'

        start_uv_attr = hair_obj.data.attributes.get('UV_Start')
        if not start_uv_attr:
            start_uv_attr = hair_obj.data.attributes.new('UV_Start', 'FLOAT2','EDGE')
            start = (0,0)
            for edge in hair_obj.data.edges:
                start_uv_attr.data[edge.index].vector = start
        end_uv_attr = hair_obj.data.attributes.get('UV_End')
        if not end_uv_attr:
            end_uv_attr = hair_obj.data.attributes.new('UV_End', 'FLOAT2','EDGE')
            end = (1,1)
            for edge in hair_obj.data.edges:
                end_uv_attr.data[edge.index].vector = end
        update_all_edges_uvs(hair_obj)
        context.view_layer.objects.active = guide_obj
        guide_obj.select_set(True)
        return {'FINISHED'}



def set_uv_edge(obj, e_idx,  uv_start, uv_end):
    start_uv_attr = obj.data.attributes.get('UV_Start')
    if not start_uv_attr:
        start_uv_attr = obj.data.attributes.new('UV_Start', 'FLOAT2','EDGE')
    start_uv_attr.data[e_idx].vector = uv_start
    end_uv_attr = obj.data.attributes.get('UV_End')
    if not end_uv_attr:
        end_uv_attr = obj.data.attributes.new('UV_End', 'FLOAT2','EDGE')
    end_uv_attr.data[e_idx].vector = uv_end

def assing_sel_uv_boxes_to_edges(obj, sel_edges_ids, boxes_ids, seed):
    random.seed(seed)
    g_seed = random.random()  #cos we need float..
    box_cnt = len(boxes_ids)
    first_mat = obj.material_slots[0].material
    # switch to Edit mode if not already
    # if obj.mode != 'EDIT':
    #     bpy.ops.object.mode_set(mode='EDIT')
    for e_idx in sel_edges_ids:
        rand_idx = int(gold_random(e_idx, g_seed)*box_cnt)
        picked_box = boxes_ids[rand_idx]
        hair_uv_data = first_mat.ht_props.hair_uv_points[picked_box]
        set_uv_edge(obj, e_idx, hair_uv_data.start_point, hair_uv_data.end_point)
    # if obj.mode != 'OBJECT':
    #     bpy.ops.object.mode_set(mode='OBJECT')

def update_all_edges_uvs(obj):
    # first_mat = obj.material_slots[0].material
    # uvBoxes = len(first_mat.ht_props.hair_uv_points)
    # boxes_ids = [int(box) for box in range(uvBoxes)]
    # all_edges = [i for i in range(len(obj.data.edges))]
    # assing_sel_uv_boxes_to_edges(obj, all_edges, boxes_ids, obj.ht_props.uv_seed)
    update_all_uv_new(obj)


class SHORTHAIR_UL_List(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        short_hair = data
        sh_vg_uvs = item
        idx = int(item.path_from_id()[-2]) #  '.short_hair_vg_uvs[x]'
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            # start your row layout by a label (icon + text), or a non-embossed text field, to make it selectable
            row = layout.row(align=True)
            row.label(text="", icon='GROUP_VERTEX')
            if idx == 0:
                row.label(text="Base UVs")
            else:
                row.prop_search(sh_vg_uvs, "vg_name", context.object, "vertex_groups", text="")
            row.prop_menu_enum(sh_vg_uvs, "uv_regions", text="UV Regions")
            row.prop(sh_vg_uvs, "mute", text="", icon='HIDE_ON' if sh_vg_uvs.mute else 'HIDE_OFF')
        # 'GRID' layout type should be as compact as possible (typically a single icon!).
        elif self.layout_type == 'GRID':
            layout.alignment = 'CENTER'
            layout.label(text="", icon_value=icon)


class HTOOL_PT_ShortHairPanel(bpy.types.Panel):
    bl_label = "Short Hair"
    bl_idname = "HTOOL_PT_ShortHairPanel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Hair Tool'
    # bl_context = 'objectmode'
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and context.mode == 'OBJECT'

    def draw(self, context):
        layout = self.layout
        obj = context.active_object
        short_hair_props = obj.data.ht_props.short_hair

        if not obj.ht_props.short_hair_props.is_short_hair:
            row = layout.row(align=True)
            row.operator("object.setup_short_hair")
            row.operator("hair.update_short_hair_node_group", icon="FILE_REFRESH", text='')
        else:
            short_hair_mod = get_short_hair_mod(obj)
            if short_hair_mod:

                def draw_attrib(layout,prop_name, text):
                    row = layout.row(align=True)
                    # row.label(text=text)
                    if short_hair_mod[f'{prop_name}_use_attribute']: # eg use vg
                        # row.prop(short_hair_mod, f'["{prop_name}_attribute_name"]', text=text)
                        row.prop_search(short_hair_mod,f'["{prop_name}_attribute_name"]', text=text, search_data=context.active_object, search_property='vertex_groups')
                    else:
                        row.prop(short_hair_mod, f'["{prop_name}"]', text=text)
                    op = row.operator('object.geometry_nodes_input_attribute_toggle', text='', icon='SPREADSHEET')
                    if bpy.app.version < (4, 0, 0):
                        op.prop_path=f'[\"{prop_name}_use_attribute\"]'
                    else:
                        op.input_name=f'{prop_name}'
                    op.modifier_name=short_hair_mod.name
                    return row

                # guide_obj = short_hair_mod['Input_6'] # Deformer
                col = layout.column(align=True)
                row = col.row(align=True)
                row.prop_search(short_hair_mod, '["Input_6"]', bpy.data, 'objects', text="Guide Object")
                row.operator('short_hair.toggle_guide', text="", icon='HIDE_OFF' if short_hair_mod["Input_22"] else 'HIDE_ON')

                draw_attrib(col,'Input_9', 'Placement Mask')
                draw_attrib(col,'Input_7', 'Initial Length')
                col.prop(short_hair_mod, '["Input_15"]', text="Resolution")
                if short_hair_mod["Input_15"]>1:
                    sub_r = col.row(align=True)
                    sub_r.alignment = 'RIGHT'
                    sub_r.prop(short_hair_mod, '["Input_25"]', text="Bend Factor")
                    sub_r = col.row(align=True)
                    sub_r.alignment = 'RIGHT'
                    sub_r.prop(short_hair_mod, '["Input_26"]', text="Bend Shape")
                draw_attrib(col,'Input_31', 'Taper')
                draw_attrib(col,'Input_23', 'Length Multiplier')
                draw_attrib(col,'Input_24', 'Length Increase')
                draw_attrib(col,'Input_21', 'Length Unify')
                draw_attrib(col,'Input_19', 'Length Rand')
                draw_attrib(col,'Input_30', 'Align To Comb Dir')
                draw_attrib(col,'Input_28', 'Randomize Rotation')
                draw_attrib(col,'Input_8', 'Add Noise')
                col.prop_search(short_hair_mod, '["Input_16"]', bpy.data, 'materials', text="Material")
                col.separator()
                box = col.box()
                box_col = box.column(align=True)
                box_col.label(text="   Sampling:")
                box_col.prop_search(short_hair_mod, '["Input_33"]', bpy.data, 'objects', text="Object")
                if short_hair_mod.get(sampled_obj_input_name) != None and short_hair_mod[sampled_obj_input_name]: # we can sample uvs and color
                    base_obj_data = getattr(short_hair_mod[sampled_obj_input_name], 'data', None)
                    box_col.prop_search(short_hair_mod, '["Input_34"]', base_obj_data, 'uv_layers', text="UVs")
                    box_col.prop_search(short_hair_mod, '["Input_13"]', base_obj_data, 'color_attributes', text="Colors")

            layout.label(text="UV Layers:")
            row = layout.row(align=True)
            row.template_list("SHORTHAIR_UL_List", "", short_hair_props, "short_hair_vg_uvs", short_hair_props, "active_index", rows=2)
            col = row.column(align=True)
            # TODO: implement these for short hair..
            col.operator('short_hair.create_vg_to_uvs', icon='ADD', text='')
            col.operator('short_hair.delete_vg_to_uvs', icon='REMOVE', text='')
            col.operator('short_hair.refresh_uv_vgroups', icon='FILE_REFRESH', text='')

# operator for toggling short_hair_mod   guide object  "Input_22" input
class HTOOL_OT_ToggleShortHairGuide(bpy.types.Operator):
    bl_idname = 'short_hair.toggle_guide'
    bl_label = 'Toggle Guide Object'
    bl_description = 'Toggle usage of Short Hair guide object'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        short_hair_mod = get_short_hair_mod(obj)
        if short_hair_mod:
            short_hair_mod['Input_22'] = 1 if short_hair_mod['Input_22'] == 0 else 0
            # force update in 3d view
            obj.modifiers.update()
            obj.update_tag()
            context.view_layer.update()
        return {'FINISHED'}

# implement create_vg_to_uvs operator for short hair
class HTOOL_OT_CreateVgToUv(bpy.types.Operator):
    bl_idname = 'short_hair.create_vg_to_uvs'
    bl_label = 'Create Vg To Uv'
    bl_description = 'Create Vg To Uv'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        short_hair_props = obj.data.ht_props.short_hair
        short_hair_props.short_hair_vg_uvs.add()
        short_hair_props.active_index = len(short_hair_props.short_hair_vg_uvs)-1
        if len(short_hair_props.short_hair_vg_uvs) == 0:
            update_all_uv_new(context.active_object)
        return {'FINISHED'}

# implement delete_vg_to_uvs operator for short hair
class HTOOL_OT_DeleteVgToUv(bpy.types.Operator):
    bl_idname = 'short_hair.delete_vg_to_uvs'
    bl_label = 'Delete Vg To Uv'
    bl_description = 'Delete Vg To Uv'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and  obj.data.ht_props.short_hair.active_index > 0

    def execute(self, context):
        obj = context.active_object
        short_hair_props = obj.data.ht_props.short_hair
        short_hair_props.short_hair_vg_uvs.remove(short_hair_props.active_index)
        cnt = len(short_hair_props.short_hair_vg_uvs)
        short_hair_props.active_index = short_hair_props.active_index if short_hair_props.active_index < cnt else cnt-1
        update_all_uv_new(context.active_object)
        return {'FINISHED'}


BOX_ITEMS  = []
def uv_box_items(self, context):
    obj = context.active_object
    first_mat = obj.material_slots[0].material
    uvBoxes = len(first_mat.ht_props.hair_uv_points)
    global BOX_ITEMS
    if len(BOX_ITEMS) != uvBoxes:
        BOX_ITEMS = [(str(i), f'UV box {i}','') for i in range(uvBoxes)]
    return BOX_ITEMS


# VG_ITEMS = []
# def vg_items(self, context):
#     obj = context.active_object
#     global VG_ITEMS
#     if len(VG_ITEMS) != len(obj.vertex_groups):
#         VG_ITEMS = [(vg.name, vg.name,'') for vg in obj.vertex_groups]
#     return VG_ITEMS

def update_all_uv_new(obj):
    short_hair_props = obj.data.ht_props.short_hair
    first_mat = obj.material_slots[0].material

    me = obj.data
    bm = bmesh.new()   # create an empty BMesh
    bm.from_mesh(me)   # fill it in from a Mesh
    bm.verts.layers.deform.verify()

    edges_to_weights = [[] for _ in range(len(bm.edges))] # edge_weights[e.index] -> (vg1_weight, vg2_weight, ...)
    vg_to_uv_boxes = {} # ordered vgroups name -> [uv_box1, uv_box2, ...]
    if len(short_hair_props.short_hair_vg_uvs) == 0: # no existing vgroup uvs? Create at least the background one
        short_hair_props.short_hair_vg_uvs.add()
        short_hair_props.active_index = 0

    for i,short_hair_vg_uv in enumerate(short_hair_props.short_hair_vg_uvs):
        uv_regions = short_hair_vg_uv.uv_regions #  enum with list of UV boxes used by vertex group vg_name
        vg_name = short_hair_vg_uv.vg_name
        if i == 0:  # background - always has some UV boxes assigned. And opacity set to 1
            if len(uv_regions):
                vg_to_uv_boxes['Background'] = [int(box_id) for box_id in uv_regions] # set to list
            else:    # no uv_regions assigned to background?
                vg_to_uv_boxes['Background'] = [i for i,_ in enumerate(first_mat.ht_props.hair_uv_points)] # assign all uv boxes to background
                short_hair_vg_uv.uv_regions = {str(i) for i in vg_to_uv_boxes['Background']}
        else:
            if len(uv_regions) > 0 and vg_name and  obj.vertex_groups.get(vg_name) and not short_hair_vg_uv.mute:
                vg_to_uv_boxes[vg_name]=[int(box_id) for box_id in uv_regions] # set to list

    # for v in bm.verts:
    #     weight = v[deform].get(vg_id)
    #     vg_verts[vg_id] = weight if weight else 0
    deform = bm.verts.layers.deform.active
    for e in bm.edges:
        for i,vg_name in enumerate(vg_to_uv_boxes.keys()):
            if i == 0: # special case for background layer - all edge_weights == 1
                edges_to_weights[e.index].append(1)
            else:
                uv_vgroup_mask = obj.vertex_groups.get(vg_name)
                real_vg_id = uv_vgroup_mask.index
                v1 = e.verts[0]
                v2 = e.verts[1]
                w1 = v1[deform].get(real_vg_id)
                w2 = v2[deform].get(real_vg_id)
                w1 = w1 if w1 else 0
                w2 = w2 if w2 else 0
                edges_to_weights[e.index].append((w1 + w2) / 2)

    # weight[e.id] = [vg1_weight, vg2_weight, ..., vgN_weight]  - threat wgN as layer opacity
    # thus weight for 'layer' below => weight[n] = (1- weight[n+1]) * weight[n]
    for e_weights in edges_to_weights:
        prev_weight = 0
        acc_weights = 1 # dim current layer by cumproduct of (1-weight) of layers above
        for i, weight_i in reversed(list(enumerate(e_weights))):
            new_weight = (1 - prev_weight) * weight_i
            e_weights[i] = new_weight * acc_weights # dim by layers above accumulated weight
            acc_weights *= (1-prev_weight)
            prev_weight = weight_i

        # DONE: make sure first 'background' layer opacity is 1 - and point it to some uv box or vgroup...
        # especially if sum(e_weights) < 1 - probability of picking any layer is below 100%

    # random.seed(short_hair_props.seed)
    # g_seed = random.random()  #cos we need float..
    np.random.seed(short_hair_props.seed)
    g_seed = short_hair_props.seed
    sorted_uv_boxes_groups = list(vg_to_uv_boxes.values())
    for e_idx,e_weights in enumerate(edges_to_weights):
        # pick random layer (boxes list id) based on e_weights weights
        layer_id = (np.cumsum(e_weights) > np.random.rand()).argmax()
        uv_boxes_ids = sorted_uv_boxes_groups[layer_id]

        box_cnt = len(uv_boxes_ids)
        rand_idx = int(gold_random(e_idx, g_seed)*box_cnt)
        picked_box_id = uv_boxes_ids[rand_idx] # select box from  group
        picked_uv = first_mat.ht_props.hair_uv_points[picked_box_id]
        set_uv_edge(obj, e_idx, picked_uv.start_point, picked_uv.end_point)

    # bm.to_mesh(me) # nothing to write, since we only update attribs
    bm.free()


class HTOOL_OT_RefreshShortHairUV(bpy.types.Operator):
    bl_idname = "short_hair.refresh_uv_vgroups"
    bl_label = "Update UV's"
    bl_description = "Update UVs on short hair cards based on current vertex groups setup"
    bl_options = {"REGISTER", "UNDO"}

    seed: bpy.props.IntProperty(name="Seed", description='Randomize how UV regions are assigned per selected strand',  default=0, min=0, max=1000)

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.short_hair_props.is_short_hair and len(obj.data.materials)

    def invoke(self, context, event):
        obj = context.active_object
        short_hair_props = obj.data.ht_props.short_hair
        self.seed = short_hair_props.seed
        return self.execute(context)

    def execute(self, context):
        obj = context.active_object
        short_hair_props = obj.data.ht_props.short_hair
        short_hair_props.seed = self.seed
        update_all_uv_new(obj)
        return {"FINISHED"}


class HTOOL_OT_SetShortHairUvRegion(bpy.types.Operator):
    bl_idname = "hair.set_short_hair_uv_region"
    bl_label = "Set UV Region"
    bl_description = "Set selected edges UVs to specific region by its ID number"
    bl_options = {"REGISTER", "UNDO"}

    uv_regions: bpy.props.EnumProperty(name='UV Region ID', description='Pick UV region that will be assigned to selected strands',
                                   items=uv_box_items,  options={'ENUM_FLAG'})
    seed: bpy.props.IntProperty(name="Seed", description='Randomize how UV regions are assigned per selected strand',  default=0, min=0, max=1000)

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.short_hair_props.is_short_hair and len(obj.data.materials)

    def draw(self, context):
        layout = self.layout
        if len(self.uv_regions) > 1:
            layout.prop(self, 'seed')
        layout.label(text='Select UVs that will be assigned to strands:')
        layout.prop(self, 'uv_regions')

    def invoke(self, context, event):
        global BOX_ITEMS
        BOX_ITEMS.clear()
        curve_obj = context.active_object
        self.uv_regions = set() #deselect by default all mats...
        self.seed = curve_obj.ht_props.uv_seed
        return self.execute(context)

    def execute(self, context):
        if not self.uv_regions:
            return {"FINISHED"}
        obj = context.active_object
        boxes_ids = [int(box) for box in self.uv_regions]
        bm = bmesh.from_edit_mesh(obj.data)
        sel_edges = [e.index for e in bm.edges if e.select]
        #switch to object mode or else attribute write will fail
        bpy.ops.object.mode_set(mode='OBJECT')
        assing_sel_uv_boxes_to_edges(context.active_object, sel_edges, boxes_ids, self.seed)
        obj.ht_props.uv_seed = self.seed
        bpy.ops.object.mode_set(mode='EDIT')

        return {"FINISHED"}

# operator that will import default short hair material
class HTOOL_OT_ImportShortHairMaterial(bpy.types.Operator):
    bl_idname = "short_hair.import_default_material"
    bl_label = "Import Short Hair Material"
    bl_description = "Import Short Hair material"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(self, context):  # make selection only false, if obj mode
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.short_hair_props.is_short_hair and len(obj.data.materials)

    def execute(self, context):
        obj = context.active_object
        short_hair_mod = get_short_hair_mod(obj)
        short_hair_mat = import_short_hair_mat(hair_mat_name, force_update=True)
        short_hair_mat.use_fake_user = True
        short_hair_mod['Input_16'] = short_hair_mat
        if obj.data.materials and obj.data.materials[0] != short_hair_mat:
            obj.data.materials[0] = short_hair_mat
        elif not obj.data.materials:
            obj.data.materials.append(short_hair_mat)
        return {"FINISHED"}



# operator of updating short_hair_node_group
class HTOOL_OT_UpdateShortHairNodeGroup(bpy.types.Operator):
    bl_idname = "hair.update_short_hair_node_group"
    bl_label = "Update Short Hair Node Group"
    bl_description = "Update (or Reset) Short Hair modifier node group,\nUsefull when you want to grab latest version of Short Hair modifier, after addon update"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        original_mats = {m.name: m for m in bpy.data.materials}
        original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}
        seg_guide_ng = import_node_group(short_hair_node_group, SHORT_HAIR_BLEND_FILE, force_update=True)
        seg_guide_ng.use_fake_user = True

        new_mat = None
        for m in bpy.data.materials:
            if m.name not in original_mats:
                new_mat = m
                print(new_mat)
                break

        hair_mat_inner_node_groups = get_nested_node_groups(new_mat.node_tree)
        remap_node_groups(hair_mat_inner_node_groups, original_node_groups)

        remap_node_groups([new_mat], original_mats, data_type='materials')
        for obj in bpy.data.objects:
            if obj.type == 'MESH':
                short_hair_mod = get_short_hair_mod(obj)
                obj.ht_props.short_hair_props.is_short_hair = True
                obj.modifiers.update()
                if short_hair_mod:
                    short_hair_mod.node_group = bpy.data.node_groups['ShortHairGN'] # forces inputs refresh (eg. add new inputs if any)
                    short_hair_mod['Output_10_attribute_name'] = 'UVMap' # Deformer
                    short_hair_mod['Output_27_attribute_name'] = 'UV_Base' # Deformer
                    # NOTE: we do not know original base obj?
                    # short_hair_mod[sampled_obj_input_name] = obj # For sampling (needed since short hair changes topo, and thus may break uvs)
        # short_hair_mat_name = 'ShortHairMaterial'
        # if new_mat:
        #     for obj in bpy.data.objects:
        #         short_hair_mod = obj.modifiers.get(short_hair_mod_name)
        #         if short_hair_mod:
        #             short_hair_mod['Input_16'] = new_mat.name
        return {"FINISHED"}


# oferator wthat will apply modifiers, and write uv attribute to uv
class HTOOL_OT_ConvertShortHairToMesh(bpy.types.Operator):
    bl_idname = "hair.convert_short_hair_to_mesh"
    bl_label = "Convert Short Hair To Mesh"
    bl_description = "Apply Short Hair modifier and write uvs"
    bl_options = {"REGISTER", "UNDO"}

    # @classmethod
    # def poll(self, context):  # make selection only false, if obj mode
    #     obj = context.active_object
    #     return obj and obj.type == 'MESH'

      # free and prevent further access

    @staticmethod
    def write_col_attrib(obj):
        pass
        # NOTE: output attribs are automatically outputted to mesh
        # sampled_col = obj.data.attributes.get('SampledCol')
        # if not sampled_col:
        #     return
        # col_array = np.empty((len(sampled_col.data)*4), dtype=np.float32)
        # sampled_col.data.foreach_get('color', col_array)
        #
        # mesh_vcol = obj.data.vertex_colors.get('SampledCol')
        # if not mesh_vcol:
        #     mesh_vcol = obj.data.vertex_colors.new(name='SampledCol')
        # mesh_vcol.data.foreach_set("color", np.power(col_array, 1/2.2))

        #for f in obj.data.polygons:
        #   for loop_idx in f.loop_indices:
        #      mesh_vcol.data[loop_idx].color = gn_col.data[loop_idx].color

        # bm = bmesh.new()   # create an empty BMesh
        # me = obj.data
        # bm.from_mesh(me)   # fill it in from a Mesh
        # color_layer = bm.loops.layers.color.get("Col")
        # if not color_layer:
        #     color_layer = bm.loops.layers.color.new("Col")
        #
        # for v in bm.verts:
        #     for loop in v.link_loops:
        #          loop[color_layer] = np.power(gn_col.data[v.index].color[:], 1/2.2)
        # bm.to_mesh(me)
        # bm.free()  # free and prevent further access
        #

    @staticmethod
    def write_3d_attrib_to_2d_attrib(obj, src_attr_name, target_attr_name):
        '''More like convert vec3 to vec2'''
        # if src_attr_name == target_attr_name:
        tmp_target_attr_name = target_attr_name + '_TMP_HT' # to avoid collision eg. on 'SampledUV' -> 'SampledUV'

        target_attr = obj.data.attributes.get(tmp_target_attr_name)
        if not target_attr: # should be always true
            target_attr = obj.data.attributes.new(name=tmp_target_attr_name, type='FLOAT2', domain='CORNER')

        attrib_uv = obj.data.attributes.get(src_attr_name)
        if not attrib_uv:
            print('Cant Read UVMapGN attribute. Make sure shorth hair modifier UVs output is set to "UVMapGN"')
            return

        # write gn_uv to numpy array
        uv_array = np.empty((len(attrib_uv.data)*3), dtype=np.float32)
        attrib_uv.data.foreach_get('vector', uv_array)

        # remove every third element form numpy array
        uv_array = uv_array.reshape((len(attrib_uv.data), 3))
        uv_array = uv_array[:, :2] # only copy x,y  (first 2 elements)

        # write uvs using foreach_set
        target_attr.data.foreach_set('vector', uv_array.ravel())
        # for f in obj.data.polygons:
        #    for loop_idx in f.loop_indices:
        #        default_uv.data[loop_idx].uv = gn_uv.data[loop_idx].vector.xy
        attrib_uv.name = attrib_uv.name + '_OLD'
        obj.data.attributes.remove(attrib_uv)
        target_attr.name = target_attr_name

    def write_uv_attrib(self, obj, short_hair_mod):
        # XXX: create uvs before accessing UVMapGN or writing wont work
        # default_uv = obj.data.uv_layers.get('UVMap')
        # if not default_uv:
        #     default_uv = obj.data.uv_layers.new()
        #
        # # XXX: has to be second after uv_lays.get or write error.
        # gn_uv = obj.data.attributes.get('UVMapGN')
        # if not gn_uv:
        #     print('Cant Read UVMapGN attribute. Make sure shorth hair modifier UVs output is set to "UVMapGN"')
        #     return
        # Same as in hair System nodes

        self.write_3d_attrib_to_2d_attrib(obj, 'UVMap', 'UVMap')
        self.write_3d_attrib_to_2d_attrib(obj, 'UV_Base', 'HairTool_UV')
        guide_obj = short_hair_mod[sampled_obj_input_name] # Deformer
        if guide_obj and short_hair_mod[sampled_uv_input_name]:
            self.write_3d_attrib_to_2d_attrib(obj, 'SampledUV', 'SampledUV')

        # # write gn_uv to numpy array
        # uv_array = np.empty((len(gn_uv.data)*3), dtype=np.float32)
        # gn_uv.data.foreach_get('vector', uv_array)
        #
        # # remove every third element form numpy array
        # uv_array = uv_array.reshape((len(gn_uv.data), 3))
        # uv_array = uv_array[:, :2] # only copy x,y  (first 2 elements)
        #
        # # write uvs using foreach_set
        # default_uv.data.foreach_set('uv', uv_array.ravel())
        # # for f in obj.data.polygons:
        # #    for loop_idx in f.loop_indices:
        # #        default_uv.data[loop_idx].uv = gn_uv.data[loop_idx].vector.xy
        #
        #
        #
        # base_uv = obj.data.uv_layers.get('HairTool_UV')
        # if not base_uv:
        #     base_uv = obj.data.uv_layers.new(name='HairTool_UV')
        #
        # gn_uv_base = obj.data.attributes.get('UV_Base')
        # if not gn_uv_base:
        #     print('Cant Read UV_Base attribute. Make sure shorth hair modifier UVs output is set to "UV_Base"')
        #     return
        #
        # # write gn_uv_base to numpy array
        # uv_array = np.empty((len(gn_uv_base.data)*3), dtype=np.float32)
        # gn_uv_base.data.foreach_get('vector', uv_array)
        #
        # # remove every third element form numpy array
        # uv_array = uv_array.reshape((len(gn_uv_base.data), 3))
        # uv_array = uv_array[:, :2] # only copy x,y  (first 2 elements)
        #
        # # write uvs using foreach_set
        # base_uv.data.foreach_set('uv', uv_array.ravel())
        #
        #
        # # sampled_uv = obj.data.uv_layers.get('UVMap')
        # # if not sampled_uv:
        # #     sampled_uv = obj.data.uv_layers.new()
        # # base_uv = obj.data.uv_layers.get('HairTool_UV')
        # # if not base_uv:
        # #     base_uv = obj.data.uv_layers.new(name='HairTool_UV')



    def short_hair_mod_apply(self, context, obj):
        short_hair_mod = obj.modifiers.get(short_hair_mod_name)
        if not short_hair_mod:
            return
        target_obj = obj.copy()
        target_obj.data = obj.data.copy()
        target_obj.name = obj.name+'_short_hair'
        link_obj_to_same_collections(obj, target_obj)
        target_obj.ht_props.short_hair_props.is_short_hair = False
        # target_obj.ht_props.short_hair_props.is_short_hair_output = True

        unlink_from_scene(obj)
        obj.use_fake_user = True

        context.view_layer.objects.active = target_obj
        target_obj.select_set(True)
        bpy.ops.object.modifier_apply(modifier=short_hair_mod_name)

        #NOTE:  GNodes mod creates additional mat slot. Remove first one (empty)
        bpy.ops.object.material_slot_remove()
        self.write_uv_attrib(target_obj, short_hair_mod)
        self.write_col_attrib(target_obj)
        target_obj.ht_props.short_hair_props.hair_base_mesh = obj.name

        if bpy.app.version < (3, 5, 0):
            keep_attribs = [] # skip ['UVMap'HairTool_UV] - since below B3.5 there should not be attributes -> they are obj.data.uv_layers
        else: # blender 3.5 and up
            keep_attribs = ['UVMap','HairTool_UV'] # keep them since from B3.5  2D FCorner attribs are == UV layers

        sample_obj = short_hair_mod[sampled_obj_input_name]
        if sample_obj:
            sampled_uv_name = short_hair_mod[sampled_uv_input_name]
            if sampled_uv_name:
                keep_attribs.append('SampledUV')
            sampled_color_name = short_hair_mod[sampled_color_input_name]
            if sampled_color_name:
                keep_attribs.append('SampledColor')

        for attr in reversed(target_obj.data.attributes):
            # temp_attrib = resutl_obj.data.attributes.get(attrib)
            if attr.name not in keep_attribs:
                try: # may error out in 3.5
                    target_obj.data.attributes.remove(attr)
                except Exception as e:
                    print(e)



        guide_obj = short_hair_mod['Input_6'] # Deformer
        if guide_obj:
            guide_obj.use_fake_user = True
            unlink_from_scene(guide_obj)


    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'MESH' and obj.ht_props.short_hair_props.is_short_hair:
                short_hair_mod = obj.modifiers.get(short_hair_mod_name)
                if not short_hair_mod:
                    self.report({'ERROR'}, f'{short_hair_mod_name} modifier not found. Skipping {obj.name}')
                    continue
                self.short_hair_mod_apply(context, obj)

            self.report({'INFO'}, f'{obj.name} converted to mesh')

        return {"FINISHED"}


# operator to restore short hair base mesh
class HTOOL_OT_RestoreShortHairBaseMesh(bpy.types.Operator):
    bl_idname = "object.restore_short_hair_base_mesh"
    bl_label = "Restore Short Hair Base Mesh"
    bl_description = "Restore short hair base mesh"

    @classmethod
    def poll(self, context):
        obj = context.active_object
        return obj and obj.type == 'MESH' and obj.ht_props.short_hair_props.hair_base_mesh

    def execute(self, context):
        short_hair_converted = context.active_object
        # bring back base mesh
        base_obj_name = short_hair_converted.ht_props.short_hair_props.hair_base_mesh
        base_mesh = bpy.data.objects.get(base_obj_name)
        if not base_mesh:
            return {"CANCELLED"}

        link_obj_to_same_collections(short_hair_converted, base_mesh)
        base_mesh.matrix_world = short_hair_converted.matrix_world
        base_mesh.use_fake_user = False
        base_mesh.select_set(True)

        context.view_layer.objects.active = base_mesh

        short_hair_mod = base_mesh.modifiers.get(short_hair_mod_name)
        if short_hair_mod:
            guide_obj = short_hair_mod['Input_6'] # Deformer
            if guide_obj:
                guide_obj.use_fake_user = True

                link_obj_to_same_collections(short_hair_converted, guide_obj)
                guide_obj.matrix_world = short_hair_converted.matrix_world
                guide_obj.use_fake_user = False
                guide_obj.select_set(True)

        bpy.data.objects.remove(short_hair_converted)

        return {'FINISHED'}

