# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from mathutils import Vector
from .. import __package__ as base_package

def get_addon_name():
    return base_package

def addon_name_lowercase():
    return get_addon_name().lower()

def get_addon_preferences():
    return bpy.context.preferences.addons[get_addon_name()].preferences

def clear_parent(obj):
    parent = obj.parent
    if parent:
        obj.matrix_basis = parent.matrix_world @ obj.matrix_parent_inverse @ obj.matrix_basis
        obj.parent = None


def set_parent(obj, new_parent):
    if obj.parent:
        old_parent_m_w = obj.parent.matrix_world.copy() #backup cos we will change obj.parent below, so backup
        backup_obj_mpi = obj.matrix_parent_inverse.copy()  # bug in blender: obj.parent = new_parent clears obj.matrix_parent_inverse so backup it
        obj.parent = new_parent
        new_m_p_inv = new_parent.matrix_world.inverted() @ old_parent_m_w @ backup_obj_mpi
        obj.matrix_parent_inverse = new_m_p_inv
    else:
        obj.parent = new_parent
        obj.matrix_parent_inverse = new_parent.matrix_world.inverted()

def assign_material(obj, material_name, clear_materials=True):
    '''Add material to obj. If clear_material is True - remove all slots except first.'''
    mat = bpy.data.materials.get(material_name)
    if not mat:
        print('Material %s doesn\'t exist!' % (material_name))
        return
    if clear_materials is True:
        while len(obj.material_slots) > 1:
            obj.data.materials.pop()
    if len(obj.material_slots) == 0:  # make sure first slot is assigned
        obj.data.materials.append(mat)
    else:
        obj.material_slots[0].material = mat


def unlink_from_scene(obj):
    for col in obj.users_collection:
        col.objects.unlink(obj)
    if obj.name in bpy.context.scene.collection.objects.keys():
        bpy.context.scene.collection.objects.unlink(obj)
    obj.use_fake_user = True


def link_child_to_collection(parent_obj, clone):
    ''' Link clone to collection where parent_obj is located '''
    if parent_obj.users_collection:
        for col in parent_obj.users_collection:
            if clone.name not in col.objects.keys():
                col.objects.link(clone)
            # else:
            #     return
    else:
        if clone.name not in bpy.context.scene.collection.objects.keys():
            bpy.context.scene.collection.objects.link(clone)
        else:
            return

def replace_material(old_mat, new_mat):
    ''' replace old_mat in objs slots '''
    for data in bpy.data.curves:
        for slot_id, mat in enumerate(data.materials):
            if mat == old_mat:
                data.materials[slot_id] = new_mat
    for data in bpy.data.meshes:
        for slot_id, mat in enumerate(data.materials):
            if mat == old_mat:
                data.materials[slot_id] = new_mat
    old_name = old_mat.name
    # if old_mat.users != 0: #wont work sice curve.ht_props.linked_mesh_mat - is not cleared...
        # print(f'Somethign went wrong, material {old_name} should have been replaced.')
    bpy.data.materials.remove(old_mat)
    new_mat.name = old_name

def get_seam_edges(bm):
    '''get edges that are seams but non border '''
    seam_edges = []
    # uv_lay = bm.loops.layers.uv['UB']
    uv_lay = bm.loops.layers.uv.active
    for e in bm.edges:
        #noo need to split this since alrady border
        # if len(e.link_faces) == 1:
        #     e.select = True
        if len(e.link_faces) == 2:
            e_loop1 = e.link_loops[0]
            e_loop2 = e.link_loops[1]
            if e_loop1[uv_lay].uv != e_loop2.link_loop_next[uv_lay].uv or e_loop2[uv_lay].uv != e_loop1.link_loop_next[uv_lay].uv:
                seam_edges.append(e)
    return seam_edges


def deleteGroup(currentGroup):
    for groupObj in currentGroup.objects:
        bpy.data.objects.remove(groupObj)
    for obj in bpy.data.objects:    # then delete all empties in scene that used that group  -- including currentEmpty (which is now linked to scene)
        # or if different empties are left empty, then delete those too
        if obj.type == 'EMPTY' and obj.instance_collection and (obj.instance_collection.name == currentGroup.name or len(obj.instance_collection.objects) == 0):
            bpy.data.objects.remove(obj)
    bpy.data.collections.remove(currentGroup, do_unlink=True)

def get_ancestors_collections(context, searched_coll):
    '''Returns [searched_col, .., context.scene.collection ]'''
    def get_child_recursive(coll, target):
        if coll == target:
            return [coll]
        for child in coll.children:
            result = get_child_recursive(child, target)
            if result:
                result.append(coll)
                return result
        return []
    get_child_recursive(context.scene.collection, searched_coll)

def link_obj_to_same_collections(source_obj, clone, force_linking = True):
    ''' Link obj to collections where source_obj is located '''
    was_linked = False
    for source_obj_coll_parent in source_obj.users_collection:
        if clone.name not in source_obj_coll_parent.objects.keys():
            source_obj_coll_parent.objects.link(clone)
            was_linked = True

    if source_obj.name in bpy.context.scene.collection.objects.keys():
        if clone.name not in bpy.context.scene.collection.objects.keys():
            bpy.context.scene.collection.objects.link(clone)
            was_linked = True
    if not was_linked and force_linking: #link to at least one collection
        bpy.context.scene.collection.objects.link(clone)


def link_coll_to_same_collections(source_obj, clone_coll):
    ''' Link clone col to collections where source_obj is located '''
    for source_obj_coll_parent in source_obj.users_collection:
        if clone_coll.name not in source_obj_coll_parent.children.keys():
            source_obj_coll_parent.children.link(clone_coll)
    if source_obj.name in bpy.context.scene.collection.objects.keys():
        if clone_coll.name not in bpy.context.scene.collection.children.keys():
            bpy.context.scene.collection.children.link(clone_coll)


def find_create_collection(parent_col, searched_col_name):
    ''' Try to find searched collection name in parent childrens. If not found create it and link it to master'''
    def find_coll(p_coll, search_col):
        if search_col == p_coll.name:
            return True
        if search_col in p_coll.children.keys():
            return True
        else:
            for child_col in p_coll.children:
                if find_coll(child_col, search_col):
                    return True
        return False

    if find_coll(parent_col, searched_col_name):
        collection = bpy.data.collections[searched_col_name]
    else:
        if searched_col_name in bpy.data.collections.keys():
            collection = bpy.data.collections[searched_col_name]
        else:
            collection = bpy.data.collections.new(searched_col_name)
        parent_col.children.link(collection)
    return collection



def unlink_obj_from_scene(context, obj, exclude_coll = None):
    ''' ulink form all collections, exept exclude_coll'''
    obj.use_fake_user = True
    for coll in obj.users_collection:
        if exclude_coll and coll == exclude_coll:
            continue
        coll.objects.unlink(obj)  # needs to be visible for matrix transform to work so we hide it now
    if obj.name in context.scene.collection.objects.keys():  # if in master collection unlinke it
        context.scene.collection.objects.unlink(obj)

def link_to_scene_coll(context, obj):
    if obj.name not in context.scene.collection.objects.keys():
        context.scene.collection.objects.link(obj)


def unlink_col_from_coll_children_nested(parent_col, remove_coll):
    '''unlink remove_coll from parent_col.children recursive '''
    for child_coll in parent_col.children:  # and its child cools
        unlink_col_from_coll_children_nested(child_coll, remove_coll)
    if remove_coll.name in parent_col.children.keys():
        parent_col.children.unlink(remove_coll)

def vec3d_to_screen2d(context, co_3d):
    r3d = context.region_data
    #r3d.view_matrix - cam mat world
    #r3d.window_matrix - only fov ?
    # perspective_matrix = window_matrix * view_matrix
    co_cam_3d = r3d.perspective_matrix @ co_3d.to_4d()  # should be (x,y,z,1) - point
    cam_co_2d_normalized = Vector((co_cam_3d.x/co_cam_3d.w, co_cam_3d.y/co_cam_3d.w))  # in <-1,1> range
    # print(cam_co_2d_normalized)
    cam_co_2d = Vector((context.region.width, context.region.height)) * (cam_co_2d_normalized + Vector((1, 1)))/2  # remap to (0, win_width)
    # print(cam_co_2d)
    return cam_co_2d

# function to get all node groups in a node tree recursively
def get_nested_node_groups(node_tree):
    node_groups = []
    for node in node_tree.nodes:
        if node.bl_idname == 'GeometryNodeGroup':
            node_groups.append(node.node_tree)
            node_groups.extend(get_nested_node_groups(node.node_tree))
    return set(node_groups) # without duplicates

def operator_is_redo(cls, context):
    op = context.active_operator
    return op and (getattr(op.bl_rna,'bl_idname',None) == cls.bl_idname)


def modifier_move_to_index(obj, mod_name, index):
    with bpy.context.temp_override(object=obj, active_object=obj, selected_objects=[obj], selected_editable_objects=[obj], editable_objects=[obj]):
        bpy.ops.object.modifier_move_to_index(modifier=mod_name, index=index)


def remap_node_groups(new_ngroups, original_ngroups, data_type='node_groups'):
    for orig_ng_name, old_node_gr in original_ngroups.items(): # old name: Twist
        for new_ng in new_ngroups: # imported name: Twist.001
            if new_ng.name.startswith(orig_ng_name) and abs(len(new_ng.name)-len(orig_ng_name))<=4:   # 'Twist.001'.startswith('Twist')
                # new_data = data_to.node_groups[all_lib_node_groups_names.index(orig_node_gr_name)] # we access by index
                print(f'remapping: {orig_ng_name} to new: {new_ng.name}')
                try:
                    old_node_gr.user_remap(new_ng)
                except Exception as e:
                    print(f'failed to remap: {orig_ng_name} to new: {new_ng.name}')
                else: # remap went ok
                    # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
                    # we have to remove data block that holds the name. Then use that name, on new data block...
                    data_handle = getattr(bpy.data, data_type)
                    data_handle.remove(old_node_gr)
                    new_ng.name = orig_ng_name



data_types = ['node_groups']  # 'materials', 'textures' etc from gpro

def import_node_group(node_name, lib_file="\hair_geo_nodes.blend", force_update=False):
    ''' import node gr. from hair_geo_nodes.blend '''
    #TODO: we should also update inner group nodes....
    old_node_gr = bpy.data.node_groups.get(node_name)

    if not old_node_gr or force_update:
        # store name now before it gets overridden
        original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}
        old_node_for_remap = bpy.data.node_groups.get(node_name)
        with bpy.data.libraries.load(lib_file) as (data_from, data_to):
            data_to.node_groups = [node_n for node_n in data_from.node_groups if node_n == node_name]
            imported_gr_names = [node_n for node_n in data_from.node_groups if node_n == node_name]  # stores names
            # all_lib_node_groups_names = data_from.node_groups[:] # part gets imported

        if data_to.node_groups:
            print(f"{imported_gr_names=}")
        else:
            print(f"node group {node_name} not found in {lib_file}")
            return None


        if old_node_for_remap: #remap main node
            old_node_for_remap.user_remap(data_to.node_groups[0])
            # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
            # we have to remove data block that holds the name. Then use freed name, on new data block...
            bpy.data.node_groups.remove(old_node_for_remap)
            data_to.node_groups[0].name = node_name

        # imported node_gr - may contain nested node groups - scan for them and remap old ngroups to new ones
        imported_node_groups = get_nested_node_groups(data_to.node_groups[0])
        remap_node_groups(imported_node_groups, original_node_groups)

        return data_to.node_groups[0]
    else:
        return old_node_gr


def safe_rename(data_blocks, id_data_to_rename, new_name: str):
    ''' renames id_data to new_name if new_name is not taken, else swaps current id_data with taken one'''
    # due to https://projects.blender.org/blender/blender/issues/119139  in blender 4.1
    # we have to free name from data block that holds the name. Then use that name, on new data block...
    old_data =  data_blocks.get(new_name)
    if old_data: # name already taken...  perform swap..
        current_name = id_data_to_rename.name
        old_data.name = 'random_available_tmp_name'  # free up 'a_temp' name (no id is using it now)
        id_data_to_rename.name = new_name
        old_data.name = current_name
    else:
        id_data_to_rename.name = new_name


def create_new_curve_obj(obj_name, parent_obj = None, set_active = True):
    # curveData = bpy.data.curves.new(obj_name+'curve', type='CURVE')
    curveData = bpy.data.hair_curves.new(obj_name+'curve')
    # curveData.surface = obj
    # curveData.surface_uv_map = ''
    # curveData.attributes.get()..
    new_curve_obj = bpy.data.objects.new(obj_name, curveData)
    if parent_obj:
        link_obj_to_same_collections(parent_obj, new_curve_obj)
    else:
        bpy.context.scene.collection.objects.link(new_curve_obj)

    if set_active:
        bpy.context.view_layer.objects.active = new_curve_obj
        new_curve_obj.select_set(True)
        parent_obj.select_set(False)

    return new_curve_obj

def set_attributes(new_id_data, attributes, ignored_attributes=()):
    """copies the list of attributes from the old to the new prop if the attribute exists"""
    # for k, v in attributes.items():
    for k, v in attributes.items(): # like loc, width etc
        if k in ignored_attributes:
            continue
        if hasattr(new_id_data, k):
            setattr( new_id_data, k, v )

def get_attributes(old_id_data, attrib_override=None):
    """copies the list of attributes from the old to the new prop if the attribute exists"""
    # for k, v in attributes.items():
    attribs = {}
    if attrib_override:
        for k in attrib_override:
            if hasattr(old_id_data, k):
                attribs[k]= getattr( old_id_data, k )
    else:
        for k, v in old_id_data.bl_rna.properties.items(): # like loc, width etc
            if not v.is_readonly:  # value may be eg. <bpy_struct, EnumProperty("type")>; use: path_resolve(k) instead
                attribs[k]= getattr( old_id_data, k )
    return attribs

def cache_old_mod_inputs(old_nodegroups, attr_over):
    ''' attr_over - list of attributes to override
        returns list of tuples: (
            old_nodes_cache[old_n.name] = get_attributes(old_n),  - mod inputs
            old_nodes_input_values[old_n.name] = []  - old mod.node_tree.nodes inputs
        )
    '''
    hair_cache = []
    for hair_ng in old_nodegroups:
        old_nodes_cache = {}
        old_nodes_input_values = {}
        for old_n in hair_ng.nodes:
            old_nodes_cache[old_n.name] = get_attributes(old_n)
            old_nodes_input_values[old_n.name] = []
            for old_input in old_n.inputs:
                # old_nodes_input_vaues[old_input.name] = old_input.default_value
                old_nodes_input_values[old_n.name].append(get_attributes(old_input, attrib_override=attr_over))
        hair_cache.append( (old_nodes_cache, old_nodes_input_values) )
    return hair_cache

def move_before_mod(moved_mod, target_mod):
    ''' moves modifier moved_mod before target_mod in obj.modifiers list'''
    # profile_idx = [i for i,m in enumerate(obj.modifiers) if m==target_mod][0]
    obj = moved_mod.id_data
    profile_idx = obj.modifiers.find(target_mod.name)
    modifier_move_to_index(obj, moved_mod.name, profile_idx)


# write function that will calculate object size in x, y, z axis based on its vertices positins
def obj_dimensions(mesh_object):
    if not mesh_object:
        return Vector((1,1,1))
    vertices = [vertex.co for vertex in mesh_object.data.vertices]
    x_min, y_min, z_min = min([vertex[0] for vertex in vertices]), min([vertex[1] for vertex in vertices]), min([vertex[2] for vertex in vertices])
    x_max, y_max, z_max = max([vertex[0] for vertex in vertices]), max([vertex[1] for vertex in vertices]), max([vertex[2] for vertex in vertices])
    return Vector((x_max - x_min, y_max - y_min, z_max - z_min))

