'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import gpu
from gpu_extras.batch import batch_for_shader
import blf
import bpy
from mathutils import Vector
import time


def draw_text_line(lines, left_margin= 30, bottom_margin = 30, width_override = False):
    h1_font_size = 17
    p_font_size = 15
    line_height_h1 = int(1.4*h1_font_size)
    line_height_p = int(1.3*p_font_size)
    font_id = 0

    if width_override:
        box_width = width_override
    else:
        max_string_len = len(max(lines, key=lambda p: len(p[0]))[0])
        box_width = max_string_len * (0.5*p_font_size)
    draw_background_box(width=box_width, height=(len(lines)+2)*line_height_p, left_margin=left_margin-line_height_p, bottom_margin=bottom_margin-line_height_h1)

    accum_y_pos = bottom_margin  # go down line by line from top
    blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
    for line in reversed(lines):
        text, line_type = line
        font_size = h1_font_size if line_type == 'H1' else p_font_size

        if bpy.app.version < (4, 0, 0):
            blf.size(font_id, font_size, 60)
        else: # for blender 4
            blf.size(font_id, font_size)

        blf.position(font_id, left_margin, accum_y_pos, 0)
        blf.draw(font_id, text)
        if line_type == 'H1':
            accum_y_pos += line_height_h1
        else:
            accum_y_pos += line_height_p



shader2d = gpu.shader.from_builtin('UNIFORM_COLOR')

def draw_background_box(width, height, left_margin, bottom_margin, bg_color = (0., 0., 0., 0.3) ):
    '''  0  - 1
         |  / |
         2  - 3
    '''
    gpu.state.blend_set("ALPHA") # if bpy.app.version >= (3, 5, 0) else bgl.glEnable(bgl.GL_BLEND)
    vertices = ((left_margin, bottom_margin+height), (left_margin+width, bottom_margin+height),
                (left_margin, bottom_margin), (left_margin+width, bottom_margin))
    indices = ((0, 1, 2), (2, 3, 1))
    batch = batch_for_shader(shader2d, 'TRIS', {"pos": vertices}, indices=indices)

    shader2d.bind()
    shader2d.uniform_float("color", bg_color)
    batch.draw(shader2d)
    gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)


def draw_background_box_abs(p1, p2, p3, p4, col):
    '''  0  - 1
         |  / |
         2  - 3
    '''
    gpu.state.blend_set("ALPHA") # if bpy.app.version >= (3, 5, 0) else bgl.glEnable(bgl.GL_BLEND)
    vertices = (p1, p2, p3, p4)
    indices = ((0, 1, 2), (2, 3, 1))
    batch = batch_for_shader(shader2d, 'TRIS', {"pos": vertices}, indices=indices)

    shader2d.bind()
    shader2d.uniform_float("color", col)
    batch.draw(shader2d)
    gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)



smooth_shader_2d = gpu.shader.from_builtin('SMOOTH_COLOR')
def draw_gradient_box(x0, y0, x1, y1, color_top=(1, 1, 1, 0), color_bottom=(1, 1, 1, 1)):
    '''  0  - 1
         |  / |
         2  - 3
    '''
    gpu.state.blend_set("ALPHA") # if bpy.app.version >= (3, 5, 0) else bgl.glEnable(bgl.GL_BLEND)

    vertices = ((x0, y0), (x1, y0),
                (x0, y1), (x1, y1))
    indices = ((0, 1, 2), (2, 3, 1))

    colors_grad_verts = [color_top, color_top, color_bottom, color_bottom] #per vert color
    batch = batch_for_shader(smooth_shader_2d, 'TRIS', {"pos": vertices, "color": colors_grad_verts}, indices=indices)
    smooth_shader_2d.bind()
    batch.draw(smooth_shader_2d)
    gpu.state.blend_set("NONE") # if bpy.app.version >= (3, 5, 0) else bgl.glDisable(bgl.GL_BLEND)


def depth_offset_matrix_method(obj_mw, dynamic_depth_off):
    # bgl.glDepthRange(0, maxRange) # XXX: breaks colors alpha in B3.5
    winmat = gpu.matrix.get_projection_matrix()
    # winmat[3][2] -= 0.1*dynamic_depth_off  # bias toward camera offset -- simple ver

    # bias toward cam - adv version
    is_persp = winmat[3][3] == 0.0
    if is_persp:
        near = winmat[2][3] / (winmat[2][2] - 1.0)
        far_ = winmat[2][3] / (winmat[2][2] + 1.0)
    else:
        near = (winmat[2][3] + 1.0) / winmat[2][2]
        far_ = (winmat[2][3] - 1.0) / winmat[2][2]

    far_ += dynamic_depth_off
    near += dynamic_depth_off
    range = far_ - near
    if is_persp:
        winmat[2][2] = -(far_ + near) / range
        winmat[2][3] = (-2 * far_ * near) / range
    else:
        winmat[2][3] = -(far_ + near) / range

    gpu.matrix.multiply_matrix(obj_mw)
    gpu.matrix.push_projection()
    gpu.matrix.load_projection_matrix(winmat)



def draw_button_box(width, height=60, pos=Vector((5, 60)), text='Button', bg_color=(0.8, 0.8, 0.8, 0.3)):
    ''' Draw from bottom left corner, to bottom + height, and left + width '''
    draw_background_box(width, height, pos.x, pos.y, bg_color)
    font_id = 0
    if bpy.app.version < (4, 0, 0):
        blf.size(font_id, 20, 65)
    else: # for blender 4
        blf.size(font_id, 20)

    blf.position(font_id, pos.x+25, pos.y+6, 0) # move font to the right and up
    blf.draw(font_id, text)



# draw simple confirmation dialog, using draw_background_box and draw_button_box.
class ConfirmDialog():
    def __init__(self, context, width, position: Vector, bg_color=(0.0, 0.0, 0.0, 0.3)):
        self.width = width

        # make sure position is not outside of context region
        if position.x < width / 2 + 20:
            position.x = width / 2 + 20
        if position.x > context.region.width - width / 2 - 20:
            position.x = context.region.width - width / 2 - 20
        if position.y < 40:
            position.y = 40
        if position.y > context.region.height - 40:
            position.y = context.region.height - 40

        self.position = position
        self.bg_color = bg_color
        self.response = {'msg': None}
        self.layout_elements = [] # list buttons as [type, pos_start, width, height, text, color]  Type: 'LABEL', 'OK', 'CANCEL', 'BG'
        self.button_color = (0.8, 0.8, 0.8, 0.3)
        self.hover_color = (0.0, 0.9, 0.0, 0.4)

    def __init_confirm_box_layout_elements(self, label):
        h1_font_size = 20
        line_height = int(1.4*h1_font_size)

        lines_count = 3 # label, ok, cancel
        total_height = lines_count*line_height

        x0 = self.position[0] - self.width / 2
        y0 = self.position[1] - total_height / 2
        draw_background_box(self.width, total_height, x0, y0, bg_color=self.bg_color)
        self.layout_elements.append(['BG', Vector((x0, y0)), self.width, 3*line_height, '', (0.0, 0.0, 0.0, 0.6)])

        label_pos = Vector((x0, y0 + 2*line_height)) # 2 lines up
        self.layout_elements.append(['LABEL', label_pos, self.width, line_height, label, (0.0, 0.0, 0.0, 0.3)])

        ok_pos = Vector((x0, y0 + line_height))
        self.ok_button_pos = (ok_pos, Vector((ok_pos.x+self.width, ok_pos.y+line_height)))
        self.layout_elements.append(['OK', ok_pos, self.width, line_height, 'OK, Quit', self.button_color])

        cancell_pos = Vector((x0, y0))
        self.cancel_button_pos = (cancell_pos, Vector((cancell_pos.x+self.width, cancell_pos.y + line_height)))
        self.layout_elements.append(['CANCEL', cancell_pos, self.width, line_height, 'Keep Editing', self.button_color])

    @staticmethod
    def draw_dialog(self):
        for element in self.layout_elements:
            type, pos, width, height, text, color = element
            if type == 'BG':
                draw_button_box(width, height, pos, text, bg_color=color)
            else:
                draw_button_box(width, height, pos, text=text, bg_color=color)

    def finish(self, msg):
        self.response['msg'] = msg
        self.restore_ops()


    def my_modal(self, context, event):
        if time.time() - self.start_time < 0.2: # give 0.2 sec delay, before handling mouse events
            return {'PASS_THROUGH'}

        if event.type == 'MOUSEMOVE':
            for element in self.layout_elements: # list of  (type, pos_start, width, height, text, color)
                if element[0] in ('OK','CANCEL'):
                    element[5] = self.hover_color if self.is_mouse_over(element, event) else self.button_color
            context.area.tag_redraw()
        elif event.type in {'ESC','RIGHTMOUSE'} and event.value == 'RELEASE':
            self.finish('OK')

        elif event.type in {'RET'} and event.value == 'RELEASE':
            self.finish('OK')

        elif event.type == 'LEFTMOUSE' and event.value == 'PRESS': # not sure why RELEASE does not work.. is not detected
            click_element = self.check_click_result(event) # handles self.restore_ops()
            if click_element in {'CANCEL', 'OK'}:
                self.finish(click_element)
                return {'PASS_THROUGH'}

            time.sleep(0.1) # again delay, to avoid double click
            # not click on OK or CANCEL, so pass through - just cancel
            self.finish('CANCEL')
            context.area.tag_redraw()
        return {'PASS_THROUGH'}


    def draw_popup(self, bpy_space, oper_cls):
        '''By adding eg. bpy.types.SpaceView3D.draw_handler_add(self.draw_popup, args, 'WINDOW', 'POST_PIXEL')A '''
        self.__init_confirm_box_layout_elements(label='Cancelling (changes will be lost!)')
        args = (self, )
        self.space = bpy_space
        self._handle = bpy_space.draw_handler_add(self.draw_dialog, args, 'WINDOW', 'POST_PIXEL')

        # Try to override other operator modal with self.modal, will it work?
        self.backup_modal = oper_cls.my_modal
        self.call_ops = oper_cls
        self.call_ops.response = self.response # pass response to operator (works as pointer)
        self.start_time = time.time()

        oper_cls.my_modal = self.my_modal


    def restore_ops(self):
        ''' Restore operator modal, and remove popup draw handler '''
        if self._handle:
            self.space.draw_handler_remove(self._handle, 'WINDOW')
        self.call_ops.my_modal = self.backup_modal


    def is_mouse_over(self, element, event):
        type, pos, width, height, text, _ = element
        if  pos.x < event.mouse_region_x < pos.x + width  and \
            pos.y < event.mouse_region_y < pos.y + height:
            return True
        return False


    def check_click_result(self, event):
        for element in self.layout_elements:
            if self.is_mouse_over(element, event):
                type, pos, width, height, text, _ = element
                if type != 'BG':
                    return type
        return 'CANCEL'







