# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# import bpy
import bmesh
from mathutils import Vector

def build_face_islands(mesh):
    bm = bmesh.new()
    bm.from_mesh(mesh)
    # bm.from_mesh(obj.data)
    bm.faces.ensure_lookup_table()
    for face in bm.faces: # not visited
        face.tag = False

    # Function to visit adjacent faces and mark them as visited
    def visit_faces(start_face):
        faces_to_visit = [start_face]
        island_faces = []
        while faces_to_visit:
            face = faces_to_visit.pop()
            if not face.tag:
                face.tag = True
                # Add the face ID to the island_faces list
                island_faces.append(face.index)

                # Visit adjacent faces
                for edge in face.edges:
                    for adjacent_face in edge.link_faces:
                        if not adjacent_face.tag:
                            faces_to_visit.append(adjacent_face)
        return island_faces

    island_face_ids = []
    for face in bm.faces:
        if not face.tag:
            island_faces = visit_faces(face)
            island_face_ids.append(island_faces)
    bm.free()
    return island_face_ids


# build verts inslands for mesh
def build_vert_islands(mesh):
    ''' If mesh is bmesh type, return list of verts, else return list of vert ids'''
    # check if mesh is bmesh type
    if isinstance(mesh, bmesh.types.BMesh):
        bm = mesh
        free_bmesh = False
    else:
        bm = bmesh.new()
        bm.from_mesh(mesh)
        free_bmesh = True

    # bm.from_mesh(obj.data)
    bm.verts.ensure_lookup_table()
    for vert in bm.verts: # not visited
        vert.tag = False

    # Function to visit adjacent verts and mark them as visited
    def visit_verts(start_vert):
        verts_to_visit = [start_vert]
        island_verts = []
        while verts_to_visit:
            vert = verts_to_visit.pop()
            if not vert.tag:
                vert.tag = True
                # Add the vert ID to the island_verts list
                island_verts.append(vert.index if free_bmesh else vert)

                # Visit adjacent verts
                for edge in vert.link_edges:
                    for adjacent_vert in edge.verts:
                        if not adjacent_vert.tag:
                            verts_to_visit.append(adjacent_vert)
        return island_verts

    island_verts_out = [] # list of vert ids (or verts)
    for vert in bm.verts:
        if not vert.tag:
            island_verts = visit_verts(vert)
            island_verts_out.append(island_verts)

    if free_bmesh:
        bm.free()
    return island_verts_out


#build edge islands for mesh
def build_edge_islands(mesh):
    ''' If mesh is bmesh type, return list of edges, else return list of edge ids '''
    # Check if mesh is bmesh type
    if isinstance(mesh, bmesh.types.BMesh):
        bm = mesh
        free_bmesh = False
    else:
        bm = bmesh.new()
        bm.from_mesh(mesh)
        free_bmesh = True

    bm.edges.ensure_lookup_table()
    for edge in bm.edges: # not visited
        edge.tag = False

    # Function to visit adjacent edges and mark them as visited
    def visit_edges(start_edge):
        edges_to_visit = [start_edge]
        island_edges = []
        while edges_to_visit:
            edge = edges_to_visit.pop()
            if not edge.tag:
                edge.tag = True
                # Add the edge ID to the island_edges list
                island_edges.append(edge.index if free_bmesh else edge)

                # Visit adjacent edges
                for vert in edge.verts:
                    for adjacent_edge in vert.link_edges:
                        if not adjacent_edge.tag:
                            edges_to_visit.append(adjacent_edge)
        return island_edges

    island_edges_out = [] # list of edge ids (or edges)
    for edge in bm.edges:
        if not edge.tag:
            island_edges = visit_edges(edge)
            island_edges_out.append(island_edges)

    if free_bmesh:
        bm.free()
    return island_edges_out


def sort_vertices_by_uv_v(bm, bm_verts, reverse=False):
    # NOTE: assume verts are not on seams...
    # Create a UV layer if it doesn't exist
    # if not bm.loops.layers.uv.active:
    #     bm.loops.layers.uv.new()

    # Create a list to hold the vertices and their UV V coordinates
    sorted_verts = []

    uv_lay = bm.loops.layers.uv.active
    # Iterate over all vertices
    for vert in bm_verts:
        link_loops_cnt = len(vert.link_loops)
        if link_loops_cnt == 0:
            continue
        # Get the UV V coordinate
        uv_v = sum(loop[uv_lay].uv.y for loop in vert.link_loops) / link_loops_cnt
        # Add the vertex and its UV V coordinate to the list
        sorted_verts.append((vert, uv_v))

    # Sort the vertices by their UV V coordinates
    # sorted_verts.sort(key=lambda x: x[1])
    sorted_verts = sorted(sorted_verts, key=lambda x: x[1], reverse=reverse)

    return sorted_verts # as list of tuples (vert, uv_v)

def sort_edges_by_uv_v(bm, bm_edges, reverse=False):
    # NOTE: assume edges are not on seams...

    # Create a list to hold the edges and their averaged UV V coordinates
    sorted_edges = []

    uv_lay = bm.loops.layers.uv.active
    if uv_lay is None:
        raise ValueError("No active UV layer found.")

    # Iterate over all edges
    for edge in bm_edges:
        uv_v = 0
        for vert in edge.verts:
            link_loops_cnt = len(vert.link_loops)
            if link_loops_cnt == 0:
                continue
            # Get the UV V coordinate
            uv_v += sum(loop[uv_lay].uv.y for loop in vert.link_loops) / link_loops_cnt
            # Add the edge and its averaged UV V coordinate to the list
        sorted_edges.append((edge, uv_v/2))

    # Sort the edges by their UV V coordinates
    sorted_edges = sorted(sorted_edges, key=lambda x: x[1], reverse=reverse)

    return sorted_edges # as list of tuples (edge, uv_v)


def verts_to_strips_by_uv(sorted_verts, uv_threshold=0.01):
    """
    Write new edges loop strip from averaging verts by uv threshold.
    If uv_v difference is greater than threshold, create new vert. Else average positions of verts below threshold.

    Args:
        sorted_verts ([type]): list of tuples (vert, uv_v)
        uv_threshold (float, optional): threshold below which group of vertices will be merged into one point. Defaults to 0.01.

    Returns:
        [type]: [description]
    """

    # Create a list to hold the new vertices. Inputs: sorted_verts : is tuple (vert, uv_v)

    # first average verts, by uv threshold. If uv_v difference is greater than threshold, create new vert. Else average positions of verts below threshold
    out_verts_pos = []
    current_uv = sorted_verts[0][1]
    current_vert_co = sorted_verts[0][0].co.copy()
    step_vert_cnt = 1
    got_first_vert = False
    min_thr = uv_threshold * 0.2

    for i,(next_v,next_uv) in enumerate(sorted_verts[1:]):
        # for last and first verts, use a smaller threshold - to get nice match at root and tips
        uv_thr = uv_threshold if got_first_vert else min_thr
        add_to_aver = abs(next_uv - current_uv) < uv_thr or step_vert_cnt < 2 # average at least 2 verts
        if add_to_aver:
            current_vert_co += next_v.co
            step_vert_cnt += 1
        else: # finish averaging verts, and init for next vert
            got_first_vert = True
            out_verts_pos.append(current_vert_co/step_vert_cnt)
            current_vert_co = next_v.co.copy()
            step_vert_cnt = 1
            current_uv = next_uv
    # out_verts_pos.append(current_vert_co/step_vert_cnt) # for last vert - dealt below

    # same but in reverse to get good matching last vert
    step_vert_cnt = 1
    current_uv = sorted_verts[-1][1]
    current_vert_co = sorted_verts[-1][0].co.copy()
    for next_v,next_uv in reversed(sorted_verts[:-1]):
        add_to_aver = abs(next_uv - current_uv) < min_thr or step_vert_cnt < 2
        if add_to_aver:
            current_vert_co += next_v.co
            step_vert_cnt += 1
        else: # finish averaging verts, and init for next vert
            out_verts_pos.append(current_vert_co/step_vert_cnt)
            break

    return out_verts_pos

def calc_edge_center(edge):
    return (edge.verts[0].co + edge.verts[1].co) / 2

def edges_to_strips_by_uv(sorted_edges, uv_threshold=0.01, smooth_factor=0.5):
    """
    Write new edges loop strip from averaging edge centers by uv threshold.
    If uv_v difference is greater than threshold, create new vertex. Else average positions of edge centers below threshold.

    Args:
        sorted_edges ([type]): list of tuples (edge, uv_v)
        uv_threshold (float, optional): threshold below which group of edges will be merged into one point. Defaults to 0.01.

    Returns:
        [type]: list of vertex positions
    """

    # NOTE: normalize edges uvs - to 0-1
    uv_min = min(sorted_edges, key=lambda x: x[1])[1]
    uv_max = max(sorted_edges, key=lambda x: x[1])[1]
    if uv_max - uv_min > 1:
        uv_max -= 1
    for i in range(len(sorted_edges)):
        sorted_edges[i] = (sorted_edges[i][0], (sorted_edges[i][1] - uv_min) / (uv_max - uv_min))

    # Create a list to hold the new vertices positions
    out_verts_pos = []
    if not sorted_edges:
        return out_verts_pos

    current_uv = sorted_edges[0][1]
    current_weight = sorted_edges[0][0].calc_length()
    weighted_pos = calc_edge_center(sorted_edges[0][0]) * current_weight
    total_weight = current_weight
    longest_edge_at_step = current_weight # output points segment should be at least as long as the longest edge that create that segment point
    got_first_edge = False
    min_thr = uv_threshold * 0.2 #   smaller for first edge  > nicer approximation at root and tip

    for next_edge, next_uv in sorted_edges[1:]:
        uv_thr = uv_threshold if got_first_edge else min_thr # smaller threshold when generating first pt - should help get closer to root
        edge_center = calc_edge_center(next_edge)
        edge_length = next_edge.calc_length()
        below_uv_threshold  = abs(next_uv - current_uv) < uv_thr #
        prev_to_current_dist = (out_verts_pos[-1] - edge_center).length if out_verts_pos else 0 # new point should be at least as long as the longest edge that create that segment point
        add_to_aver =  prev_to_current_dist < 1.5*longest_edge_at_step and got_first_edge

        if add_to_aver or below_uv_threshold: # add to average
            weighted_pos += edge_center * edge_length
            total_weight += edge_length
            if edge_length > longest_edge_at_step:
                longest_edge_at_step = edge_length
        else:  # finish averaging edges, and init next step
            got_first_edge = True
            out_verts_pos.append(weighted_pos / total_weight if total_weight > 0.0001 else edge_center)
            current_uv = next_uv
            weighted_pos = edge_center * edge_length
            total_weight = edge_length
            longest_edge_at_step = 0.3*edge_length+0.7*longest_edge_at_step # prevet bad results on short edges (it would add too many interpolated points on short edges)

    # Append the last averaged position
    out_verts_pos.append(weighted_pos / total_weight if total_weight > 0.0001 else edge_center)

    # Usually last point is not  close enough to haircard tip, so run same algo in reverse - for only one iteration
    current_uv = sorted_edges[-1][1]
    edge_center = calc_edge_center(sorted_edges[-1][0])
    current_weight = sorted_edges[-1][0].calc_length()
    weighted_pos = edge_center * current_weight
    total_weight = current_weight
    got_first_edge = False

    for next_edge, next_uv in reversed(sorted_edges[:-1]):
        uv_thr = uv_threshold if got_first_edge else min_thr # smaller threshold when generating first pt - should help get closer to root
        edge_center = calc_edge_center(next_edge)
        edge_length = next_edge.calc_length()
        below_uv_threshold  = abs(next_uv - current_uv) < uv_thr
        prev_to_current_dist = (out_verts_pos[-1] - edge_center).length if out_verts_pos else 0 # new point should be at least as long as the longest edge that create that segment point
        add_to_aver =  prev_to_current_dist < 1.5*longest_edge_at_step and got_first_edge

        if add_to_aver or below_uv_threshold: # add to average
            weighted_pos += edge_center * edge_length
            total_weight += edge_length
            if edge_length > longest_edge_at_step:
                longest_edge_at_step = edge_length
        else:  # finish averaging edges, and init next step
            got_first_edge = True
            if total_weight > 0.0001:
                out_verts_pos.append(weighted_pos / total_weight)
            else:
                out_verts_pos.append(edge_center)
            break

    smooth_out_verts_pos(out_verts_pos, smooth_factor) # smooth the out_verts_pos
    return out_verts_pos


# smooth the out_verts_pos
def smooth_out_verts_pos(out_verts_pos, smooth_factor=0.5):
    # Smooth the out_verts_pos - is list of 3d Vector(). smooth without first and last point
    pts_cp = [p.copy() for p in out_verts_pos]
    for _ in range(3):
        for i in range(1, len(out_verts_pos)-1):
            avr = (out_verts_pos[i-1] + out_verts_pos[i+1]) / 2 - out_verts_pos[i] # laplacian smooth
            out_verts_pos[i] = out_verts_pos[i] + 0.5 * avr

    for i, (orig_co, smooth_co) in enumerate(zip(pts_cp, out_verts_pos)):
        out_verts_pos[i] = orig_co.lerp(smooth_co, smooth_factor)
    return smooth_co

